unit SPCLoader;
{$I DFS.inc}{ Copyright (C) 2007 C M Western; see NLFit.pas for license}
{ Import Thermo Scientific (formerly Galactic Industries) .SPC spectrometer files }
interface

uses ExpPlotLoader, Classes;

type

  { TSPCFileLoader }

  TSPCFileLoader = class(TExperimentalPlotLoader)
  public
    function Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean; override;
  end;

implementation

uses Units, Math, SysUtils;

{ TSPCFileLoader }

const
  TSPREC: Byte =   1; { 16 bit integer Y values }
  TMULTI: Byte =   4; { Multiple files }
  TXYXYS: Byte = $40; { Individual X values }
  TXVALS: Byte = $80; { Uneven X values }

function TSPCFileLoader.Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean;
  procedure CheckRead(out v; Length: Integer);
  begin
    if Stream.Read(v, Length) <> Length then
      Connector.ReportError('Unexpected end of file', True);
    FixEndian(v, Length);
  end;
  procedure CheckPosition(NewPosition: Int64);
  begin
    Stream.Position := NewPosition;
    if Stream.Position <> NewPosition then
      Connector.ReportError('File too short', True);
  end;
  procedure AddString(Len: Integer);
  var
    buffer: string;
    P: Integer;
  begin
    SetLength(buffer, Len);
    Stream.Read(buffer[1],Len);
    P := StrLen(PChar(buffer));
    if (P < Len-1) and (buffer[P+2] <> #0)  then
      buffer[P+1] := ' '; { More than one value, separated by NUL }
    SetLength(buffer, StrLen(PChar(buffer)));
    if Length(buffer) > 0 then begin
      if Connector.Comment <> '' then
        Connector.Comment := Connector.Comment + ' ';
      Connector.Comment := Connector.Comment + buffer;
    end;
  end;
var
  ftFlags, fexp: Byte;
  nPoints: Integer;
  Xscale: Double;
  Xtype: Byte;
  procedure ReadSubHeader;
  var
    Start: Int64;
  begin
    Start := Stream.Position;
    if (ftFlags and TMULTI) <> 0 then begin
      CheckRead(fexp,1);
      CheckRead(fexp,1);
      if (ftFlags and TXYXYS) <> 0 then begin
        CheckPosition(Stream.Position + 14);
        CheckRead(nPoints,Sizeof(nPoints));
      end;
    end;
    CheckPosition(Start + 32);
  end;
  procedure ReadYdata(Index: Integer);
  begin
    if fexp = $80 then
      ReadAndCopyArray(Index,1,tySingle) { 4 byte floating point }
    else begin
      if (ftFlags and TSPREC) = 0 then
        ReadAndCopyArray(Index,1,tyI32,-1,IntPower(2, fexp - 32)) { 4 byte integer }
      else
        ReadAndCopyArray(Index,1,tyI16,-1,IntPower(2, fexp - 16)); { 2 byte integer }
    end;
  end;
  procedure HandleXtype;
  begin
    Xscale := 1;
    case Xtype of
      2: begin Connector.SetUnits(-1,nmVac); Xscale := 1/1000; end;
      3: Connector.SetUnits(-1,nmVac);
      6: begin Connector.SetUnits(-1,MHz); Xscale := 1e-6; end;
      7: begin Connector.SetUnits(-1,MHz); Xscale := 1e-3; end;
      8: Connector.SetUnits(-1,MHz);
     14: Connector.SetUnits(-1,eV);
     26: begin Connector.SetUnits(-1,MHz); Xscale := 1e3; end;
    else
      Connector.SetUnits(-1,cm1);
    end;
  end;
var
  Position: Int64;
  buffer: string[2];
  OldFormat: Boolean;
  i, nSubfiles: Integer;
  FirstX, LastX: double;
  Ytype, Ztype: Byte;
  Sng: Single;
  Wrd: Word;
begin
  Result := False;
  if (Ext = '') or (Ext = '.spc') or (Ext = '.cgm') then begin
    Position := Stream.Position;
    if (Stream.Read(Buffer[1],2) = 2)
       and (buffer[2] in [#$4B..#$4D]) then begin
      LittleEndian := buffer[2] = #$4B;
      OldFormat := buffer[2] = #$4D; { implies different header layout }
    end else begin
      { Unrecognized }
      Stream.Position := Position;
      Exit;
    end;
  end else
    Exit; { Wrong extension }
  Result := True;
  ftFlags := Ord(buffer[1]);
  if not OldFormat then begin
    CheckRead(fexp,1); { Ignore Instrument technique }
    CheckRead(fexp,1);
    CheckRead(nPoints, Sizeof(nPoints));
    CheckRead(FirstX, Sizeof(FirstX));
    CheckRead(LastX, Sizeof(LastX));
    CheckRead(nSubfiles, Sizeof(nSubfiles));
    if (ftFlags and TMULTI) = 0 then
      nSubfiles := 1;
    CheckRead(Xtype, Sizeof(Xtype));
    HandleXtype;
    CheckRead(Ytype, Sizeof(Ytype)); { TODO: Could use this }
    CheckRead(Ztype, Sizeof(Ztype)); { ignore (?) }
    CheckPosition(Stream.Position+5);
    AddString(9);
    AddString(9);
    CheckPosition(Stream.Position+34);
    AddString(130);
    AddString(30);
    CheckPosition(Position + 512); { Skip rest of header }
  end else begin
    { Old style file }
    LittleEndian := True;
    CheckRead(Wrd, SizeOf(Wrd)); fexp := Wrd;
    CheckRead(Sng, SizeOf(Sng)); nPoints := Round(Sng);
    CheckRead(Sng, SizeOf(Sng)); FirstX := Sng;
    CheckRead(Sng, SizeOf(Sng)); LastX := Sng;
    CheckRead(Xtype, Sizeof(Xtype));
    HandleXtype;
    CheckRead(Ytype, Sizeof(Ytype)); { TODO: Could use this }
    CheckPosition(Stream.Position+6);
    AddString(8);
    CheckPosition(Stream.Position+32);
    AddString(130);
    AddString(30);
    nSubfiles := 1;
    ReadSubHeader;
    SwapWords := True;
  end;
  SetupChannels(nSubfiles, nPoints, (ftFlags and TXYXYS) <> 0);
  if (ftflags and TXVALS) = 0 then
    { X values evenly spaced }
    for i := 0 to nPoints - 1 do
      SetValue(-2, 0, i, FirstX + (LastX-FirstX)*i/(nPoints-1))
  else if (ftFlags and TXYXYS) = 0 then
    { Read X values in }
    ReadAndCopyArray(-2, 0, tySingle, -1, Xscale);
  for i := 0 to nSubfiles-1 do begin
    if (i > 0) or not OldFormat then
      ReadSubHeader;
    if (ftFlags and TXYXYS) <> 0 then begin
      ReadAndCopyArray(i, 0, tySingle, -1, XScale);
      Connector.SetLineList(i,True);
    end;
    ReadYdata(i);
  end;
end;

initialization
  RegisterPlotLoader(TSPCFileLoader, 'SPC Files', '*.spc; *.cgm');
end.

