unit SVD;

interface

uses MatrixVec, Vector;

{$DEFINE LAPACK}

{$IFNDEF LAPACK}
procedure SV_Decompose(A:TMatrix; W:TColumnVector; U:TMatrix; V:TMatrix);

{ Perform the singular value decomposition of a rectangular matrix

  Converted from svd.f from SLATEC library

     This subroutine is a translation of the ALGOL procedure SVD,
     NUM. MATH. 14, 403-420(1970) by Golub and Reinsch.
     HANDBOOK FOR AUTO. COMP., VOL II-LINEAR ALGEBRA, 134-151(1971).

     This subroutine determines the singular value decomposition
          T
     A=USV  of a REAL M by N rectangular matrix.  Householder
     bidiagonalization and a variant of the QR algorithm are used.

     On Input

        A contains the rectangular input matrix to be decomposed.  A is
          a two-dimensional REAL array, dimensioned A(M,N):
          M is the number of rows of A and U.
          N is the number of columns of A and U and the order of V.


     On Output

        A is unaltered (unless overwritten by U or V).

        W contains the N (non-negative) singular values of A (the
          diagonal elements of S).  They are unordered. W is a one-dimensional
          REAL array, dimensioned W(N).

        U contains the matrix U (orthogonal column vectors) of the
          decomposition if U was not nil on entry. U may coincide with A.
          U is a two-dimensional REAL array, dimensioned U(M,N).

        V contains the matrix V (orthogonal) of the decomposition if
          V was not nil on entry. V may also coincide with A if U does not.
          V is a two-dimensional REAL array, dimensioned V(NM,N).

     Questions and comments should be directed to B. S. Garbow,
     APPLIED MATHEMATICS DIVISION, ARGONNE NATIONAL LABORATORY
     ------------------------------------------------------------------
***SEE ALSO  EISDOC
***ROUTINES CALLED  PYTHAG
***REVISION HISTORY  (YYMMDD)
   811101  DATE WRITTEN
   890531  Changed all specific intrinsics to generic.  (WRB)
   890831  Modified array declarations.  (WRB)
   891214  Prologue converted to Version 4.0 format.  (BAB)
   900402  Added TYPE section.  (WRB)
}
{$ELSE}
procedure SV_DecomposeAlt(A:TMatrix; W:TColumnVector; U:TMatrix; VT:TMatrix);
{$ENDIF}
procedure SV_Solve(U: TMatrix; W: TColumnVector; V: TMatrix; IsTransposed: Boolean; B: TColumnVector; X: TColumnVector);

implementation

uses {$IFDEF LAPACK} Lapack, {$ENDIF}
     SysUtils, Math;

{$IFDEF LAPACK}

procedure SV_DecomposeAlt(A:TMatrix; W:TColumnVector; U:TMatrix; VT:TMatrix);
{ Note that A can be the same as U or V, but not both }
{ U or V can also be nil, if not required }
var
  JOBZ: Char;
  UP, VTP: PDouble;
  n, m: Integer;
  work: array of Double;
  iwork: array of Integer;
  info, lwork, ldvt: Integer;
  t: Double;
begin
  m := A.RowCount;
  n := A.ColCount;
  if A = U then begin
    JOBZ := 'O';
    if m < n then
      raise Exception.CreateFmt('SV_Decompose:O %d %d',[m,n]);
  end else if A = VT then begin
    JOBZ := 'O';
    if n <= m then
      raise Exception.CreateFmt('SV_Decompose:O %d %d',[m,n]);
  end else
    JOBZ := 'N';
  if Assigned(U) and (A <> U) then begin
    U.RowCount := m;
    U.ColCount := m;
    UP := U.FirstItem;
  end else
    UP := nil;
  if Assigned(VT) and (VT <> A) then begin
    VT.RowCount := n;
    VT.ColCount := n;
    VTP := VT.FirstItem;
  end else
    VTP := nil;
  if Assigned(VT) then
    ldvt := Max(1,VT.RowCount)
  else
    ldvt := 1;
  W.Size := Min(m,n);
  SetLength(iwork,8*Min(m,n)+1);
  lwork := -1;
  dgesdd_(JOBZ,m,n,A.FirstItem^,m,W.FirstItem^,UP^,m,VTP^,ldvt,t,lwork,iwork[0],info,1);
  if info = 0 then begin
    lwork := Round(t);
    SetLength(work, lwork);
    dgesdd_(JOBZ,m,n,A.FirstItem^,m,W.FirstItem^,UP^,m,VTP^,ldvt,work[0],lwork,iwork[0],info,1);
  end;
  if JOBZ = 'O' then
    if m >= n then
      U.ColCount := n
    else
      VT.RowCount := n;
end;
{$ENDIF}

{$IFDEF NON_DIVIDE_AND _CONQUER}
function dgesvd_(var JOBU: Char; var JOBVT: Char; var M: Integer; var N: Integer; var A: Double {1..LDA,1..*}; var LDA: Integer; var S: Double {1..*};
                 var U: Double {1..LDU,1..*}; var LDU: Integer; var VT: Double {1..LDVT,1..*}; var LDVT: Integer; var WORK: Double {1..*};
                 var LWORK: Integer; var INFO: Integer; jobu_len, jobvt_len: Integer): Integer; cdecl; external;

function dgesvd(JOBU, JOBVT: Char; M, N: Integer; var A, S, U, VT: Double):Integer;
var
  work: array of Double;
  lwork, ldvt: Integer;
  t: Double;
begin
  case JOBVT of
    'A': ldvt := N;
    'S': ldvt := Min(M,N);
  else
    ldvt := 1;
  end;
  lwork := -1;
  dgesvd_(JOBU, JOBVT, M, N, A, M, S, U, M, VT, ldvt, t, lwork, Result, 1, 1);
  if Result = 0 then begin
    lwork := Round(t);
    SetLength(work, lwork);
    dgesvd_(JOBU, JOBVT, M, N, A, M, S, U, M, VT, ldvt, work[0], lwork, Result, 1, 1);
  end;
end;

procedure SV_DecomposeAlt1(A:TMatrix; W:TColumnVector; U:TMatrix; VT:TMatrix);
{ Note that A can be the same as U or V, but not both }
{ U or V can also be nil, if not required }
var
  JOBU, JOBVT: Char;
  UP, VTP: PDouble;
  n, m: Integer;
begin
  m := A.RowCount;
  n := A.ColCount;
  if A = U then begin
    JOBU := 'O';
  end else if U <> nil then begin
    JOBU := 'A';
    U.RowCount := m;
    U.ColCount := m;
  end else
    JOBU := 'N';
  if Assigned(U) then begin
    UP := U.FirstItem;
  end else
    UP := nil;
  if A = VT then
    JOBVT := 'O'
  else if VT <> nil then begin
    JOBVT := 'A';
    VT.RowCount := n;
    VT.ColCount := n;
  end else
    JOBVT := 'N';
  if Assigned(VT) then
    VTP := VT.FirstItem
  else
    VTP := nil;
  W.Size := Min(m,n);
  dgesvd(JOBU,JOBVT,M,N,A.FirstItem^,W.FirstItem^,UP^,VTP^);
  if JOBVT = 'O' then
    VT.RowCount := n;
end;
{$ENDIF}

{$IFNDEF LAPACK}
function Sign(a,b: Double):Double;
begin
  Result := a;
  if a < 0 then Result := -a;
  if b < 0 then Result := -Result;
end;

function Pythag(a, b: Extended):Extended;
{ Finds sqrt(A**2+B**2) without overflow or destructive underflow }
var
  Q,R,S,T: Extended;
begin
  Result := Abs(a);
  Q := Abs(b);
  if Result < Q then begin
    Result := Abs(b);
    Q := Abs(a);
  end;
  if Q <> 0 then
    repeat
      R := Sqr(Q/Result);
      T := 4 + R;
      if t = 4 then Break;
      S := R/T;
      Result := Result + 2*Result*S;
      Q := Q*S;
    until false;
end;

procedure SV_Decompose(A:TMatrix; W:TColumnVector; U:TMatrix; V:TMatrix);
var
  RV1: TColumnVector;
  I, J, K, L, M, N, MN: Longint;
  ITS: Integer;
  MatU: Boolean; { If the U matrix in the decomposition is desired }
  MatV: Boolean; { If the V matrix in the decomposition is desired }
  C, S, F, G, H, SCALE, S1, X, Y, Z: Extended;
  Cancel: Boolean;
begin
  M := A.RowCount;
  N := A.ColCount;
  RV1 := nil;
  MatU := U <> nil;
  L := 0;
  try
    RV1 := TColumnVector.Create(N);
    if not MatU then
      U := TMatrix.Create(M,N);
    MatV := V <> nil;
    U.Assign(A);
{     .......... HOUSEHOLDER REDUCTION TO BIDIAGONAL FORM .......... }
    G := 0;
    SCALE := 0;
    S1 := 0;
    for I := 1 to N do begin
       L := I + 1;
       RV1[I] := SCALE * G;
       G := 0;
       S := 0;
       SCALE := 0;
       if I <= M then begin
         for K := I to M do
           SCALE := SCALE + Abs(U[K,I]);
         if SCALE <> 0 then begin
           for K := I to M do begin
              U[K,I] := U[K,I] / SCALE;
              S := S + Sqr(U[K,I]);
           end;
           F := U[I,I];
           G := -Sign(Sqrt(S),F);
           H := F * G - S;
           U[I,I] := F - G;
           if I <> N then
             for J := L to N do begin
                S := 0;
                for K := I to M do
                  S := S + U[K,I] * U[K,J];
                F := S / H;
                for K := I to M do
                  U[K,J] := U[K,J] + F * U[K,I];
             end;
           for K := I to M do
             U[K,I] := SCALE * U[K,I];
         end;
       end;
       W[I] := SCALE * G;
       G := 0;
       S := 0;
       SCALE := 0;
       if (I <= M) and (I <> N) then begin
         for K := L to N do
           SCALE := SCALE + Abs(U[I,K]);
         if SCALE <> 0 then begin
           for K := L to N do begin
              U[I,K] := U[I,K] / SCALE;
              S := S + Sqr(U[I,K]);
           end;
           F := U[I,L];
           G := -Sign(Sqrt(S),F);
           H := F * G - S;
           U[I,L] := F - G;
           for K := L to N do
             RV1[K] := U[I,K] / H;
           if I <> M then
             for J := L to M do begin
                S := 0;
                for K := L to N do
                  S := S + U[J,K] * U[I,K];
                for K := L to N do
                   U[J,K] := U[J,K] + S * RV1[K];
             end;
           for K := L to N do
             U[I,K] := SCALE * U[I,K];
         end;
       end;
       S := Abs(W[I])+Abs(RV1[I]);
       if S > S1 then S1 := S;
    end;


{     .......... ACCUMULATION OF RIGHT-HAND TRANSFORMATIONS .......... }
    if MATV then
      for I := N downto 1 do begin
         IF I <> N then begin
           if G <> 0 then begin
             for J := L to N do
    {     .......... DOUBLE DIVISION AVOIDS POSSIBLE UNDERFLOW .......... }
               V[J,I] := (U[I,J] / U[I,L]) / G;
             for J := L to N do begin
                S := 0;
                for K := L to N do
                  S := S + U[I,K] * V[K,J];
                for K := L to N do
                  V[K,J] := V[K,J] + S * V[K,I];
             end;
           end;
           for J := L to N do begin
              V[I,J] := 0;
              V[J,I] := 0;
           end;
         end;
         V[I,I] := 1;
         G := RV1[I];
         L := I;
      end;

{     .......... ACCUMULATION OF LEFT-HAND TRANSFORMATIONS .......... }
    if MatU then begin
      MN := N;
      IF M < N then MN := M;
      for I := MN downto 1 do begin
         L := I + 1;
         G := W[I];
         if I <> N then
           for J := L to N do
             U[I,J] := 0;
         if G <> 0 then begin
           if I <> MN then
             for J := L to N do begin
                S := 0;
                for K := L to M do
                  S := S + U[K,I] * U[K,J];
{     .......... DOUBLE DIVISION AVOIDS POSSIBLE UNDERFLOW .......... }
                F := (S / U[I,I]) / G;
                for K := I to M do
                  U[K,J] := U[K,J] + F * U[K,I];
             end;
           for J := I to M do
             U[J,I] := U[J,I] / G;
         end else
           for J := I to M do
             U[J,I] := 0;
         U[I,I] := U[I,I] + 1;
      end;
    end;
{     .......... DIAGONALIZATION OF THE BIDIAGONAL FORM .......... }
    for k := N downto 1 do begin
       ITS := 0;
       repeat
         L := K;
         Cancel := False;
{     .......... TEST FOR SPLITTING. }
         repeat
{     .......... RV1[1] IS ALWAYS ZERO, SO THERE IS NO EXIT
                 THROUGH THE BOTTOM OF THE LOOP .......... }
           if S1 + Abs(RV1[L]) = S1 then Break;
           if S1 + Abs(W[L-1]) = S1 then
             Cancel := True
           else
             Dec(L);
         until Cancel;
         if Cancel then begin
{     .......... CANCELLATION OF RV1[L] IF L GREATER THAN 1 .......... }
           C := 0;
           S := 1;
           for I := L to K do begin
              F := S * RV1[I];
              RV1[I] := C * RV1[I];
              if S1 + Abs(F) = S1 then
                Break;
              G := W[I];
              H := Pythag(F,G);
              W[I] := H;
              C := G / H;
              S := -F / H;
              if MatU then
                for J := 1 to M do begin
                   Y := U[J,L-1];
                   Z := U[J,I];
                   U[J,L-1] := Y * C + Z * S;
                   U[J,I] := -Y * S + Z * C;
                end;
           end;
         end;
{     .......... TEST FOR CONVERGENCE .......... }
         Z := W[K];
         if L = K then Break;
{     .......... SHIFT FROM BOTTOM 2 BY 2 MINOR .......... }
         if ITS = 30 then
           raise Exception.CreateFmt('No convergence for %d-th singular value after 30 iterations', [K]);
         ITS := ITS + 1;
         X := W[L];
         Y := W[K-1];
         G := RV1[K-1];
         H := RV1[K];
         F := 0.5 * (((G + Z) / H) * ((G - Z) / Y) + Y / H - H / Y);
         G := Pythag(F,1);
         F := X - (Z / X) * Z + (H / X) * (Y / (F + Sign(G,F)) - H);
{     .......... NEXT QR TRANSFORMATION .......... }
         C := 1;
         S := 1;
         for I := L+1 to K do begin
            G := RV1[I];
            Y := W[I];
            H := S * G;
            G := C * G;
            Z := Pythag(F,H);
            RV1[I-1] := Z;
            C := F / Z;
            S := H / Z;
            F := X * C + G * S;
            G := -X * S + G * C;
            H := Y * S;
            Y := Y * C;
            if MatV then
              for J := 1 to N do begin
                 X := V[J,I-1];
                 Z := V[J,I];
                 V[J,I-1] := X * C + Z * S;
                 V[J,I] := -X * S + Z * C;
              end;
            Z := Pythag(F,H);
            W[I-1] := Z;
{     .......... ROTATION CAN BE ARBITRARY IF Z IS ZERO .......... }
            if Z <> 0 then begin
              C := F / Z;
              S := H / Z;
            end;
            F := C * G + S * Y;
            X := -S * G + C * Y;
            if MatU then
              for J := 1 to M do begin
                 Y := U[J,I-1];
                 Z := U[J,I];
                 U[J,I-1] := Y * C + Z * S;
                 U[J,I] := -Y * S + Z * C;
              end;
         end;
         RV1[L] := 0;
         RV1[K] := F;
         W[K] := X;
       until false;

{     .......... CONVERGENCE .......... }
       if Z < 0 then begin
{     .......... W[K] IS MADE NON-NEGATIVE .......... }
         W[K] := -Z;
         if MatV then
           for J := 1 to N do
             V[J,K] := -V[J,K];
       end;
    end;
  finally
    RV1.Free;
    if not MatU then
      U.Free;
  end;
end;

procedure SV_Solve(U : TMatrix; W : TColumnVector; V : TMatrix; B : TColumnVector;
                   X : TColumnVector);
{ Solves a system of equations by singular value decomposition, after
  the matrix has been transformed by SV_Decompose

  On Input
    U(M,N)   As outout by SV_Decompose
    W(N)
    V(N,N)
    B(M)     Vector of length N

  On Output
                                 T
    X(N)     V * Diag(1/s(i)) * U  * B
}
var
  I, J, JJ, M, N : Longint;
  Sum : Extended;
  Tmp : TColumnVector;
begin
  M := U.RowCount;
  N := U.ColCount;
  Tmp := nil;
  try
    Tmp := TColumnVector.Create(N);
    for J := 1 to N do begin
      Sum := 0.0;
      if W[J] <> 0.0 then begin
        for I := 1 to M do
          Sum := Sum + U[I,J] * B[I];
        Sum := Sum / W[J];
      end;
      Tmp[J] := Sum;
    end;
    for J := 1 to N do begin
      Sum := 0.0;
      for JJ := 1 to N do
        Sum := Sum + V[J,JJ] * Tmp[JJ];
      X[J] := Sum;
    end;
  finally
    Tmp.Free;
  end;
end;
{$ENDIF}

procedure SV_Solve(U: TMatrix; W: TColumnVector; V: TMatrix; IsTransposed: Boolean; B: TColumnVector; X: TColumnVector);
{ Solves a system of equations by singular value decomposition, after
  the matrix has been transformed by SV_Decompose

  On Input
    U(M,N)   As outout by SV_Decompose
    W(N)
    V(N,N)
    B(M)     Vector of length N

  On Output
                                 T
    X(N)     V * Diag(1/s(i)) * U  * B
}
var
  J, N : Longint;
begin
  // M := U.RowCount;
  N := U.ColCount;
  X.Assign(B);
  X.PreMultiply(U,True);
  for J := 1 to N do
    if W[J] <> 0 then
      X[J] := X[J] / W[J]
    else
      X[J] := 0;
  X.PreMultiply(V,IsTransposed);
end;

end.
