unit Units;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see NLFit.pas for license}
interface

type
  TResonanceUnits = (cm1=0, MHz, Kelvin, eV, sqrtcm1, nmVac, nmAir);
  TEnergyUnits = cm1..eV;

function NameOfUnit(Units: TResonanceUnits): string; overload;
function TypeOfUnit(Units: TResonanceUnits): string;
function StrToUnits(s: string): TResonanceUnits;
function ConversionFactor(UFrom, UTo: TEnergyUnits):Extended;
function Convert(Value:Double; UFrom, UTo: TResonanceUnits):Extended;
function ConvertWidth(Value, Centre:Double; UFrom, UTo: TResonanceUnits):Extended;

{ Is the unit proportional to energy? }
function IsEnergyUnit(Units: TResonanceUnits): Boolean; inline;

{ Is a reciprocal involved in converting to any energy unit? }
function IsWavelengthUnit(Units: TResonanceUnits): Boolean; inline;

{ Is a square root involved in converting to energy ? }
function IsSqrtUnit(Units: TResonanceUnits): Boolean; inline;

type
  TIntensityUnits = (Arbitrary=0,
                     Normalized,
                     Squared,
                     HonlLondon,   { Dimensionless line strength factor }
                     EinsteinASum, { s-1; summed over all degenerate states }
                     EinsteinA,    { s-1; summed over all lower degnerate states only, so reciprocal is radiative lifetime }
                     nm2MHzperMolecule {nm^2 MHz / molecule},
                     cm2WavenumberperMolecule { cm^2 cm-1 / molecule },
                     HzperMolecule {s-1, emission including Boltzmann factor },
                     PopDist { Population Distrubution },
                     NormPopDist { Normalised Population Distrubution } );

function NameOfUnit(Units: TIntensityUnits; Integrated: Boolean; FrequencyUnits:TResonanceUnits): string; overload;
function IntensityFactor(Frequency: Double; FrequencyUnits: TResonanceUnits; IntensityUnits: TIntensityUnits): Double; overload;
function ConvertIntensity(Value, Frequency: Double; FFrom: TResonanceUnits; IFrom: TIntensityUnits; FTo: TResonanceUnits; ITo: TIntensityUnits; Integrated: Boolean = False):Double;
function IsArbitraryStyle(Units: TIntensityUnits): Boolean;
function HasUnits(Units: TIntensityUnits):Boolean;
function IsNormalized(Units: TIntensityUnits):Boolean;
function IsPopDist(Units: TIntensityUnits):Boolean;
function IncludesPopulation(Units: TIntensityUnits):Boolean;

var
  BoltzmannConstants: array [TEnergyUnits] of Double;
{ This is computed to ensure the values are identical, even when switching units.
  The previous values were 0.6950356 ( for cm-1 ) and 20836.644 (for MHz )
  but 20836.644 / c * 1e4 =0.69503563028260037 }
  MomentOfInertiaTimesRotationalConstant: array [TEnergyUnits] of Double;

const
  { 2006 CODATA recommended values }
  ge: Double = 2.0023193043622;
  BohrMagneton: Double = 927.400915e-26; { e hcross / 2 / me }
  SpeedOfLight: Double = 299792458.0;
  Planck: Double = 6.62606896e-34;
  Avagadro: Double = 6.02214179e23;
  eVTocm1: Double = 8065.54465;
  KelvinToMHz: Double = 20836.644;
  BohrRadius: Double = 0.52917720859e-10;
  MeOverMp: Double = 5.4461702177e-4;

implementation

uses AirVac, SysUtils, Math;

function NameOfUnit(Units: TResonanceUnits): string;
begin
  case Units of
    cm1: Result := 'cm-1';
    MHz: Result := 'MHz';
    Kelvin: Result := 'Kelvin';
    eV:  Result := 'eV';
    sqrtcm1: Result := 'cm-1/2';
    nmVac:  Result := 'nm';
    nmAir:  Result := 'nm';
  end;
end;

function StrToUnits(s: string): TResonanceUnits;
begin
  Result := cm1;
  repeat
    if CompareText(s, NameOfUnit(Result)) = 0 then
      Exit;
    if Result = High(Result) then
      raise Exception.Create(s + ' is not a valid unit name');
    Inc(Result);
  until False;
end;

function TypeOfUnit(Units: TResonanceUnits): string;
begin
  case Units of
    cm1: Result := 'Wavenumber';
    MHz, eV: Result := 'Energy';
    Kelvin: Result := 'Temperature';
    sqrtcm1: Result := 'Velocity';
    nmAir:  Result := 'Air Wavelength';
    nmVac:  Result := 'Vacuum Wavelength';
  end;
end;

function IsEnergyUnit(Units: TResonanceUnits): Boolean; inline;
begin
  Result := Units <= High(TEnergyUnits);
end;

function IsWavelengthUnit(Units: TResonanceUnits): Boolean; inline;
begin
  Result := Units >= nmVac;
end;

function IsSqrtUnit(Units: TResonanceUnits): Boolean; inline;
begin
  Result := Units = sqrtcm1;
end;

const
  SpeedOfLight4: Double = 29979.2458;

function ConversionFactor(UFrom, UTo: TEnergyUnits):Extended;
begin
  { Can't use array of constants, because expressions like 1/SpeedOfLight don't count as constant expressions }
  if UTo < UFrom then
    Result := 1/ConversionFactor(UTo, UFrom)
  else if UFrom = UTo then
    Result := 1
  else
    case UFrom of
      cm1:    case UTo of
                MHz:    Result := SpeedOfLight4;
                Kelvin: Result := SpeedOfLight4/KelvinToMHz;
                ev:     Result := 1/eVTocm1;
              end;
      MHz:    case UTo of
                Kelvin: Result := 1/KelvinToMHz;
                ev:     Result := 1/eVTocm1/SpeedOfLight4;
              end;
      Kelvin: case UTo of
                ev:     Result := KelvinToMHz/SpeedOfLight4/eVTocm1;
              end;
    end;
end;

function Convert(Value:Double; UFrom, UTo: TResonanceUnits):Extended;
begin
  Result := Value;
  if UFrom = UTo then
    Exit;
  case UFrom of
    sqrtcm1: begin UFrom := cm1; Result := Sqr(Result); end;
    nmVac:   begin UFrom := cm1; Result := nmcm1(Result); end;
    nmAir:   begin UFrom := cm1; Result := vac(Result); end;
  end;
  if IsEnergyUnit(UTo) then
    Result := Result*ConversionFactor(UFrom, UTo)
  else begin
    Result := Result*ConversionFactor(UFrom, cm1);
    case UTo of
      sqrtcm1: if Result < 0 then
                 { Allow energy scale to go through zero }
                 Result := -Sqrt(Abs(Result))
               else
                 Result := Sqrt(Result);
      nmVac:   Result := nmcm1(Result);
      nmAir:   Result := air(Result);
    end
  end;
end;

{ Note that this function is not reversible, and as it depends on the centre
  frequency it can't be made usefully so. }
function ConvertWidth(Value, Centre:Double; UFrom, UTo: TResonanceUnits):Extended;
begin
  if UFrom = UTo then
    Result := Value
  else if IsSqrtUnit(UFrom) xor IsSqrtUnit(UTo) then
    Result := Abs(Convert(Centre,UFrom,UTo)-Convert(Centre+Value,UFrom,UTo))
  else if IsWavelengthUnit(UFrom) or IsWavelengthUnit(UTo) then
    { Taking inverse or air <-> nm conversion - scale according to Value/Centre }
    if Centre > 1/Max_nm_cm1 then
      Result := Value/Centre * Convert(Centre, UFrom, UTo)
    else
      Result := 0 { Invalid, but zero is probably harmless }
  else
    Result := Convert(Value, UFrom, UTo)
end;

function NameOfUnit(Units: TIntensityUnits; Integrated: Boolean; FrequencyUnits:TResonanceUnits): string;
begin
  case Units of
    Arbitrary:    Result := '';
    Squared:      Result := 'Intensity^2';
    Normalized:   Result := 'Normalized Intensity';
    HonlLondon:   Result := 'Honl-London Factor';
    EinsteinASum: Result := 'Summed Einstein A coefficient / s-1';
    EinsteinA:    Result := 'Einstein A coefficient per state / s-1';
    nm2MHzperMolecule: Result := 'nm^2 MHz / molecule';
    cm2WavenumberperMolecule: Result := 'cm^2 cm-1 / molecule';
    HzperMolecule: Result := 's-1 / molecule';
    PopDist: Result := 'Population';
    NormPopDist: Result := 'Normalized Population';
  end;
  if not Integrated and (Units <> Arbitrary) then
    Result := Result + ' / ' + NameOfUnit(FrequencyUnits);
end;

const
  FrequencyUnit: array [TIntensityUnits] of TResonanceUnits =
    (cm1 { Actually none },
     cm1 { Actually none },
     cm1 { Actually none },
     cm1 { Actually none },
     cm1 { Hz },
     cm1 { Hz },
     MHz, { nm2MHzperMolecule }
     cm1, { cm2WavenumberperMolecule }
     cm1  { HzperMolecule },
     cm1 { Actually none },
     cm1 { Actually none }
     );

  FrequencyPower: array [TIntensityUnits] of Integer =
    { Power of frequency in relationship to dipole moment }
    (0 { Arbitrary },
     0 { Normalized },
     0 { Squared },
     0 { Honl-London },
     3 { EinsteinASum },
     3 { EinsteinA },
     1, { nm2MHzperMolecule }
     1, { cm2WavenumberperMolecule }
     3, { HzperMolecule }
     0, { PopDist }
     0  { NormPopDist } );

var { Actually const, but compiler can't cope with these as expressions just involving constants }
  IntensityFactors: array [TIntensityUnits] of Double;
  { Conversion factors excluding frequency }

function IntensityFactor(Frequency: Double; FrequencyUnits: TResonanceUnits; IntensityUnits: TIntensityUnits): Double;
begin
  if HasUnits(IntensityUnits) then begin
    if FrequencyUnits <> FrequencyUnit[IntensityUnits] then
      Frequency := Convert(Abs(Frequency),FrequencyUnits,FrequencyUnit[IntensityUnits])
    else
      Frequency := Abs(Frequency);
    Result := 3 { Convert to unpolarised light }
              * IntensityFactors[IntensityUnits]
              * IntPower(Frequency,FrequencyPower[IntensityUnits])
  end else if IntensityUnits = HonlLondon then
    Result := 3
  else
    Result := 1;
end;

function ConvertIntensity(Value, Frequency: Double;
                          FFrom: TResonanceUnits; IFrom: TIntensityUnits;
                          FTo: TResonanceUnits; ITo: TIntensityUnits;
                          Integrated: Boolean):Double;
var
  s: Double;
begin
  { Attempt to convert intetsity units; some conversions are only approximate,
    and values without a specific unit (Arbitrary, Normalized, Squared) are left unchanged.
    Call with Integrated=True for integrated line intensities }
  { Note that EinsteinA and EinsteinASum are considered as equivalent in this routine } 
  { TODO: Do the conversions make sense for sqrtcm1? }
  Result := Value; { If all else fails, leave unchanged }
  case IFrom of
    Arbitrary, Normalized, PopDist, NormPopDist, HonlLondon, EinsteinASum, EinsteinA: if ITo = Squared then Result := Sqr(Result);
    Squared: if IFrom <> Squared then Result := Sqrt(Abs(Result));
  else
    case ITo of
      Arbitrary: ;
      Normalized: ;
      Squared: Result := Sqr(Result);
      HonlLondon: ;
      EinsteinASum: ;
      EinsteinA: ;
      PopDist: ;
      NormPopDist: ;
    else
      { below here both From and To have units }
      Result := Result * IntensityFactors[ITo]/IntensityFactors[IFrom];
      if (FrequencyUnit[IFrom] <> FrequencyUnit[ITo]) or (FrequencyPower[ITo] <> FrequencyPower[IFrom]) then
        if FrequencyPower[ITo] = FrequencyPower[IFrom] then
          Result := Result * IntPower(ConversionFactor(FrequencyUnit[IFrom],FrequencyUnit[ITo]),FrequencyPower[IFrom])
        else begin
          Frequency := Convert(Frequency, FFrom, FrequencyUnit[IFrom]);
          if Frequency <= 0 then
            Result := 0
          else
            Result := Result * IntPower(Convert(Frequency, FrequencyUnit[IFrom], FrequencyUnit[ITo]), FrequencyPower[ITo])
                             / IntPower(Frequency, FrequencyPower[IFrom]);
        end;
    end;
  end;
  if not Integrated and (FFrom <> FTo) and not(IFrom in [Arbitrary,Normalized]) then begin
    { This conversion fails if the zero level is incorrect in experimental spectra }
    { TODO: Review this for experimental spectra } 
    s := Convert(Frequency, FFrom, FTo);
    if s > 0 then
      Result := Result * Frequency / s
    else
      Result := 0;
  end;
end;

function IsArbitraryStyle(Units: TIntensityUnits): Boolean;
begin
  Result := Units in [Arbitrary, Squared, PopDist];
end;

function HasUnits(Units: TIntensityUnits):Boolean;
begin
  Result := Units in [EinsteinASum, EinsteinA, nm2MHzperMolecule, cm2WavenumberperMolecule, HzperMolecule];
end;

function IsNormalized(Units: TIntensityUnits):Boolean;
begin
  Result := Units in [Normalized, nm2MHzperMolecule, cm2WavenumberperMolecule, HzperMolecule, NormPopDist];
end;

function IsPopDist(Units: TIntensityUnits):Boolean;
begin
  Result := Units in [PopDist, NormPopDist];
end;

function IncludesPopulation(Units: TIntensityUnits):Boolean;
begin
  Result := not (Units in [HonlLondon, EinsteinASum, EinsteinA]);
end;

var
  I: TEnergyUnits;

initialization
  IntensityFactors[Arbitrary] := 1;
  IntensityFactors[Squared] := 1;
  IntensityFactors[Normalized] := 1;
  IntensityFactors[HonlLondon] := 1;
  IntensityFactors[nm2MHzperMolecule] := 8*Pi*Pi*Pi/(3e31*SpeedOfLight*Planck);
  IntensityFactors[cm2WavenumberperMolecule] := IntensityFactors[nm2MHzperMolecule] / 1e14;
  IntensityFactors[HzperMolecule] := 64*IntPower(pi,4)*1e-43/3/Planck;
  IntensityFactors[EinsteinASum] := IntensityFactors[HzperMolecule];
  IntensityFactors[EinsteinA] := IntensityFactors[HzperMolecule];
  IntensityFactors[PopDist] := 1;
  IntensityFactors[NormPopDist] := 1;
  BoltzmannConstants[cm1] := Convert(1,Kelvin,cm1);
  BoltzmannConstants[MHz] := Convert(1,Kelvin,MHz);
  BoltzmannConstants[Kelvin] := 1;
  BoltzmannConstants[eV] := Convert(1,Kelvin,eV);
  MomentOfInertiaTimesRotationalConstant[MHz] :=  Planck / (8000*Sqr(Pi)) * 1e20 * Avagadro;
  for I := Low(I) to High(I) do
    MomentOfInertiaTimesRotationalConstant[I] := Convert(MomentOfInertiaTimesRotationalConstant[MHz], MHz, I);
end.
