unit AsymT;

{ Closed Shell Asymmetric Top - use the open shell equivalent normally }

{$I DFS.inc}
interface

uses SysUtils, Classes, Mixture, QuantumNumber;

const
  FirstWidthPar = LSP+26;

type

  TClosedAsymmetricMolecule = class;

  T4Sym = (symA1, symA2, symB1, symB2); { Note that if the molecule is not C2v, then some of these are not used }

  TClosedAsymmetricTop = class (TState)
  private
    function GetSymmetry(const Index: Integer): T4Sym;
    procedure SetSymmetry(const Index: Integer; const AValue: T4Sym);
  public
    RoSyms: array [0..3] of Integer; { Mapping from overall sym number to rotational eo symmetry;
                                       Implicitly Contains more than one value subgroups of C2v }
    constructor Create(AOwner: TComponent); override;
    class function OwnerType: TComponentClass; override;
    function Molecule: TClosedAsymmetricMolecule;
    function OddSpin: TQuantumNumber; override;
    function AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer; override;
    function FixupProperties: Exception; override;
    function HRotA(J, Kp, K: TQuantumNumber): Double;
    function HRotS(J, Kp, K: TQuantumNumber): Double;
    function HasWidth: Boolean; override;
    function HWidth(J, Kp, K: TQuantumNumber): Double;
    function HDip(TransitionMoment: TTransitionMoment; Jp, Kp, J, K: TQuantumNumber): Double;
    function MatrixElement(ABra, AKet: TBasisState; MatElemType: TMatElemType): Double; override;
    function TransitionMatrixElement(TransitionMoment: TTransitionMoment; ABra, AKet: TBasisState): Double; override;
    function PerturbationMatrixElement(Perturbation: TPerturbation; ABra, AKet: TBasisState; MatElemType: TMatElemType): Double; override;
    function GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar; override;
  published
    property Symmetry: T4Sym index LSI read GetSymmetry write SetSymmetry;
    property A: Double index LSP   read GetParameter write SetParameter;
    property B: Double index LSP+1 read GetParameter write SetParameter;
    property C: Double index LSP+2 read GetParameter write SetParameter;
    { Quartic Centrifugal Distortion }
    property DJ:      Double index LSP+3 read GetParameter write SetParameter;
    property DJK:     Double index LSP+4 read GetParameter write SetParameter;
    property DK:      Double index LSP+5 read GetParameter write SetParameter;
    property deltaJ:  Double index LSP+6 read GetParameter write SetParameter;
    property deltaK:  Double index LSP+7 read GetParameter write SetParameter;
    { Sextic Centrifugal Distortion }
    property HJ:    Double index LSP+8 read GetParameter write SetParameter;
    property HJK:   Double index LSP+9 read GetParameter write SetParameter;
    property HKJ:   Double index LSP+10 read GetParameter write SetParameter;
    property HK:    Double index LSP+11 read GetParameter write SetParameter;
    property phiJ:  Double index LSP+12 read GetParameter write SetParameter;
    property phiJK: Double index LSP+13 read GetParameter write SetParameter;
    property phiK:  Double index LSP+14 read GetParameter write SetParameter;
    { Octic Centrifugal Distortion }
    property LJ:    Double index LSP+15 read GetParameter write SetParameter;
    property LJJK:  Double index LSP+16 read GetParameter write SetParameter;
    property LJK:   Double index LSP+17 read GetParameter write SetParameter;
    property LKKJ:  Double index LSP+18 read GetParameter write SetParameter;
    property LK:    Double index LSP+19 read GetParameter write SetParameter;
    property llJ:   Double index LSP+20 read GetParameter write SetParameter;
    property llJK:  Double index LSP+21 read GetParameter write SetParameter;
    property llKJ:  Double index LSP+22 read GetParameter write SetParameter;
    property llK:   Double index LSP+23 read GetParameter write SetParameter;
    { Alternate form for rotational constants }
    property BBar:  Double index LSP+24 read GetParameter write SetParameter;
    property BDelta:Double index LSP+25 read GetParameter write SetParameter;
    { Linewidth parameters - initial stab }
    property wKa:   Double index FirstWidthPar read GetParameter write SetParameter;
    property wKb:   Double index FirstWidthPar+1 read GetParameter write SetParameter;
    property wKc:   Double index FirstWidthPar+2 read GetParameter write SetParameter;
    property wBar:  Double index FirstWidthPar+3 read GetParameter write SetParameter;
    property wDelta:Double index FirstWidthPar+4 read GetParameter write SetParameter;
  end;

  TClosedAsymmetricTopPerturbation = class (TPerturbation)
  public
    function MatrixElement(Jp, Kp, J, K: TQuantumNumber): Double;
    function Molecule: TClosedAsymmetricMolecule;
    function PairName: string; override;
  published
    property JPower:Integer index LPI   read GetSetting write SetSetting;
    property aPower:Integer index LPI+1 read GetSetting write SetSetting;
    property bPower:Integer index LPI+2 read GetSetting write SetSetting;
    property cPower:Integer index LPI+3 read GetSetting write SetSetting;
  end;

  TClosedAsymmetricTopBasisState = class(TBasisState)
  public
    K: TQuantumNumber;
    Ka, Kc: TQuantumNumber;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel: string; override; { State name, assuming energy ordering preserved }
    function StateLabelKey: string; override; { Explanation of State Name }
  end;

  TClosedAsymmetricManifold = class(TManifold)
    class function OwnerType: TComponentClass; override;
    function FirstSym(J: TQuantumNumber):Integer; override;
    function NextSym(J: TQuantumNumber; Sym: Integer):Integer; override;
    function nSym: Integer; override; { Number of different symmetries }
  end;

  TRepresentation = (Ir=0, IIr, IIIr, Il, IIl, IIIl);
  TPointGroup = (C1=0, Ci, C2, Cs, D2, C2v, C2h, D2h);

  TClosedAsymmetricMolecule = class(TMolecule)
  private
    FOddRankSym: Integer;
    function GetPointGroup(Index: integer): TPointGroup;
    function GetRepresentation(Index: integer): TRepresentation;
    procedure SetPointGroup(Index: integer; const AValue: TPointGroup);
    procedure SetRepresentation(Index: integer; const AValue: TRepresentation);
    function GetAxis(const Index: Integer): TabcAxis;
    procedure SetAxis(const Index: Integer; const AValue: TabcAxis);
  public
    constructor Create(AOwner: TComponent); override;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; override;
    function eoSymC2v(Sym4: T4Sym): integer;
    function FixupProperties: Exception; override;
    function GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar; override;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function LastSym: Integer; override;
    property OddRankSym: Integer read FOddRankSym; { Symmetry of dipole moment }
  published
    property PointGroup:     TPointGroup index LMI read GetPointGroup write SetPointGroup;
    property Representation: TRepresentation index LMI+1 read GetRepresentation write SetRepresentation;
    property SReduction:     Boolean  index LMI+2 read GetFlag write SetFlag; { Use Watson's S reduction (as opposed to A) }
    property eeWt:           Integer  index LMI+3 read GetSetting write SetSetting;
    property eoWt:           Integer  index LMI+4 read GetSetting write SetSetting;
    property oeWt:           Integer  index LMI+5 read GetSetting write SetSetting;
    property ooWt:           Integer  index LMI+6 read GetSetting write SetSetting;
    property C2zAxis:        TabcAxis index LMI+7 read GetAxis write SetAxis;
    property C2xAxis:        TabcAxis index LMI+8 read GetAxis write SetAxis;
    property PseudoC2v:      Boolean  index LMI+9 read GetFlag write SetFlag; { Ignore any operators breaking the C2v symmetry }
    property FakeSym:        Boolean  index LMI+10 read GetFlag write SetFlag; { Turn off transition moment test }
  end;

implementation

uses Math, nj;

const
  FirstStatisticalWeight = LMI+3;

{ Symmetry is binary representation of ee=00 eo=01 oe=10 oo=11, so multiplication
  is simply by bitwise xor
  These specify the oddness/evenness of Ka and Kc.
  The axis for which K is used in the basis is specified by the representation
  Ir: z=a and so on
  To use the same ideas for the vibronic symmetry, we use the identification (for C2v):
  R2z' = C2
  R2x' = sigma(y'z')
  R2y' = sigma(x'z')
  where yet another axis system is being introduced for the point group. This is
  defined by the postion of C2 axis (z') and the out of plane axis (x'). The
  location of these wrt the inertial axes is defined by the C2zAxis and C2xAxis
  variables.

  For Cs we have:
  R2z' = sigma(x'y')
  where C2zAxis is now the out of plane axis, and C2xAxis is unused

  For C1, no simplification is possible
   }

{ TClosedAsymmetricTop }

const
  C2vCharacterTable: array [T4Sym] of array [0..3] of Integer =
   (    { E C2 s(zx) s*(yz) }
        {   Rz   Ry     Rx  }
   {A1} ( 1, 1,   1,     1),
   {A2} ( 1, 1,  -1,    -1),
   {B1} ( 1,-1,   1,    -1),
   {B2} ( 1,-1,  -1,     1));
  CsCharacterTable: array [0..1] of array [0..1] of Integer =
   (    { E s(xy) } { z out of plane }
        {    Rz   }
   {A1} ( 1,  1   ),
   {A2} ( 1, -1));


  { Antisymmetric wrt rotation about the inertial axis implies the following symmetries: }
  AntiSymWrt: array [TabcAxis] of Integer = ( 2 {a - 10}, 3 {b - 11}, 1 {c - 01} );
  { Non totally symmetric function that is symmetric wrt the given axis: }
  SymWrt: array [TabcAxis] of Integer = ( 1 {a - 01}, 3 {b - 11}, 2 {c - 10} );
  { Mask to switch to functions that are antisymmetric wrt reflection in Cs }
  CsSwitch: array [TabcAxis] of Integer = ( 2 {a - 10}, 1 {b - arbitrary}, 1 {c - 01} );


function TClosedAsymmetricMolecule.eoSymC2v(Sym4: T4Sym): Integer;
{ Convert between C2v point group labels and eo symmetry labels }
var
  Sym: Integer;
begin
  Result := 0;
  { We know directly from eo label the symmetry wrt Ra and Rc, but not Rb, so
    must use only these two only to match irreducible representation labels }
  Sym := AntiSymWrt[C2zAxis];
  if (Sym <> 3) then { if not b axis }
    if C2vCharacterTable[Sym4][1{=C2}] < 0 then
      Result := Sym;
  Sym := AntiSymWrt[C2zAxis] xor AntiSymWrt[C2xAxis]; { = AntiSymWrt[C2yAxis] }
  if (Sym <> 3) then { if not b axis }
    if C2vCharacterTable[Sym4][2{=sigma(zx)}] < 0 then
      Result := Result or Sym;
  Sym := AntiSymWrt[C2xAxis];
  if (Sym <> 3) then { if not b axis }
    if C2vCharacterTable[Sym4][3{=sigma(yz)}] < 0 then
      Result := Result or Sym;
end;

function TClosedAsymmetricTop.FixupProperties: Exception;
var
  i, j, w, w1: Integer;
begin
  case Molecule.PointGroup of
    C2v: begin
            for i := 0 to 3 do
              RoSyms[i] := Molecule.eoSymC2v(Symmetry) xor i;
         end;
    Cs:  begin
          { Pair of symmetric functions }
          RoSyms[0] := 0;
          RoSyms[2] := SymWrt[Molecule.C2zAxis];
          { Pair of antisymmetric functions }
          RoSyms[1] := RoSyms[0] xor CsSwitch[Molecule.C2zAxis];
          RoSyms[3] := RoSyms[2] xor CsSwitch[Molecule.C2zAxis];
          if Symmetry <> symA1 then
            { Antisymmetric vibronic state - switch over }
            for i := 0 to 3 do
              RoSyms[i] := RoSyms[i] xor CsSwitch[Molecule.C2zAxis];
         end;
    C1:  begin
            for i := 0 to 3 do
              RoSyms[i] := i;
         end;
  else
    raise NotYet;
  end;
  Result := inherited FixupProperties;
  if Assigned(Result) then
    Exit;
  { TODO: Derive effective symmetries from transition moments? }
  { Check for consistent statistical weights }
  for i := 0 to Manifold.nSym-1 do begin
    w := Molecule.GetSetting(FirstStatisticalWeight+(RoSyms[i] xor RoSyms[0]));
    j := i + Manifold.nSym;
    while j < 4 do begin
      w1 := Molecule.GetSetting(FirstStatisticalWeight+(RoSyms[j] xor RoSyms[0]));
      if w <> w1 then
        Result := Exception.Create('Inconsistent statistical weights for ' + Name);
      Inc(j, Manifold.nSym);
    end;
  end;
end;


const
  C2vNames: array [0..4] of PChar = ('A1', 'A2', 'B1', 'B2', nil);
  CsNames: array [0..2] of PChar = ('A''', 'A"', nil);
  C1Names: array [0..1] of PChar = ('A', nil);

function TClosedAsymmetricTop.GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar;
begin
  if AIndex = LSI then
    case Molecule.PointGroup of
      C2v: Result := @C2vNames;
      Cs: Result := @CsNames;
      C1: Result := @C1Names;
    else
      Raise Exception.CreateFmt('%s: Point Group Not implmented: %d', [Name, Ord(Molecule.PointGroup)]);
    end
  else
    Result := inherited GetSettingsNames(AIndex,LowestValue);
end;

function TClosedAsymmetricTop.GetSymmetry(const Index: Integer): T4Sym;
begin
  Result := T4Sym(GetSetting(Index));
end;

procedure TClosedAsymmetricTop.SetSymmetry(const Index: Integer; const AValue: T4Sym);
begin
  SetSetting(Index, Ord(AValue));
end;

constructor TClosedAsymmetricTop.Create(AOwner: TComponent);
begin
  inherited;
end;

class function TClosedAsymmetricTop.OwnerType: TComponentClass;
begin
  Result := TClosedAsymmetricManifold;
end;

function TClosedAsymmetricTop.Molecule: TClosedAsymmetricMolecule;
begin
  Result := Owner.Owner as TClosedAsymmetricMolecule;
end;

function TClosedAsymmetricTop.OddSpin: TQuantumNumber;
begin
  Result := 0; { Closed shell for now }
end;

{ Mapping between x/y/z and a/b/c axes }
const                                             {   Ir,   IIr,  IIIr,    Il,   IIl,  IIIl}
  ZAxisFor: array [TRepresentation] of TabcAxis = (aAxis, bAxis, cAxis, aAxis, bAxis, cAxis);
  XAxisFor: array [TRepresentation] of TabcAxis = (bAxis, cAxis, aAxis, cAxis, aAxis, bAxis);
  YAxisFor: array [TRepresentation] of TabcAxis = (cAxis, aAxis, bAxis, bAxis, cAxis, aAxis);
  { Reverse Mapping }
  aAxisFor: array [TRepresentation] of TxyzAxis = (zAxis, yAxis, xAxis, zAxis, xAxis, yAxis);
  bAxisFor: array [TRepresentation] of TxyzAxis = (xAxis, zAxis, yAxis, yAxis, zAxis, xAxis);
  cAxisFor: array [TRepresentation] of TxyzAxis = (yAxis, xAxis, zAxis, xAxis, yAxis, zAxis);

function xyzAxis(Representation: TRepresentation; abcAxis: TabcAxis): TxyzAxis;
{ Given a b or c, return corresponding x y or z axis }
begin
  if ZAxisFor[Representation] = abcAxis then
    Result := zAxis
  else if YAxisFor[Representation] = abcAxis then
    Result := yAxis
  else
    Result := xAxis;
end;

function TClosedAsymmetricTop.AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer;
var
  K, Kmax, Kinc, Ka, Kc: TQuantumNumber;
  cSym, Phase, i: Integer;
  function AntiSymmetricWrtAxis(Axis: TabcAxis): Boolean;
  { Given axis, determine from Sym whether want wavefunction to be symmetric or antisymmetric
    wrt C2 about that axis }
  begin
    with Eigenvalues do
      case Axis of
        aAxis: Result := (cSym and 2) > 0;
        bAxis: Result := (((cSym shr 1) xor cSym) and 1) > 0;
        cAxis: Result := (cSym and 1) > 0;
      end;
  end;
  procedure Add(K: TQuantumNumber);
  var
    BasisState: TClosedAsymmetricTopBasisState;
  begin
    BasisState := TClosedAsymmetricTopBasisState.Create(Eigenvalues);
    BasisState.State := Self;
    BasisState.K := K;
    BasisState.Phase := Phase;
    BasisState.Ka := Ka;
    BasisState.Kc := Kc;
    BasisState.J := J;
    BasisState.HyperfineState := HyperfineState;
  end;
begin
  i := Eigenvalues.Sym;
  repeat
    cSym := RoSyms[i];
    { Statistical Weight }
    Result := Molecule.GetSetting(FirstStatisticalWeight+(cSym xor RoSyms[0]));
    if Result = 0 then
      Exit;
    if AntiSymmetricWrtAxis(ZAxisFor[Molecule.Representation]) then
      K := 2
    else
      K := 0;
    if AntiSymmetricWrtAxis(XAxisFor[Molecule.Representation]) xor (Minus(J) < 0) then
      Phase := -1
    else
      Phase := 1;
    if Minus(J-K) < 0 then
      Kmax := J-2
    else
      Kmax := J;
    { To generate Ka Kc assignments, must generate K states in order of increasing energy
      Note that this could generate odd results for if the centrifugal distortion
      is dominant, or if B ~ (A+C)/2.
      TODO: This would be fixed by ranking by diagonal matrix elements }
    case ZAxisFor[Molecule.Representation] of
      { Prolate - definately increasing with K }
      aAxis: Kinc := 4;
      { Oblate - definately decreasing with K }
      cAxis: Kinc := -4;
      { K is intermediate axis - depends on whether B is closer to A or C }
      bAxis: if B < (A+C)/2 then
               Kinc := -4
             else
               Kinc := 4;
    end;
    if Kinc < 0 then
      K := Kmax;
    { Work out Ka and Kc for lowest state }
    { Kc is simply J or J-1 to match xo or xe }
    if (Minus(J) < 0) xor ((cSym and 1) <> 0) then
      Kc := J-2
    else
      Kc := J;
    { Ka is 1 for oe or oo }
    if (cSym and 2) <> 0 then
      Ka := 2 { oe or oo }
    else begin { ee or eo - use the requirement that Ka+Kc = J or J+1 }
      Ka := J - Kc;
      if Minus(Ka) < 0 then
        Inc(Ka, 2);
    end;
    while (K <= J) and (K >= 0) do begin
      if (K <> 0) or (Phase = 1) then begin
        { E- matrix has K = 0 missing }
        Add(K);
        Inc(Ka,4);
        Dec(Kc,4);
      end;
      Inc(K,Kinc);
    end;
    Inc(i, (Owner as TClosedAsymmetricManifold).nSym);
  until i >= 4;
end;

function TClosedAsymmetricTop.HWidth(J, Kp, K: TQuantumNumber): Double;
{ Linewidth parameters - as a first stab, go for <Ka^2>, etc. }
{ TODO: This is straight copy of HRot - can this be generalised? }
var
  Izz, Iyy, Ixx, JSq, KSq, IxxpIyy, IxxmIyy: Double;
begin
  if K = Kp then begin
    Izz := GetParameter(FirstWidthPar+Ord(ZAxisFor[Molecule.Representation]));
    Ixx := GetParameter(FirstWidthPar+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(FirstWidthPar+Ord(YAxisFor[Molecule.Representation]));
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    IxxpIyy := Ixx + Iyy + wBar*2;
    Result := Width + Izz*KSq  + IxxpIyy*(JSq - KSq)/2;
  end else if Abs(K-Kp) = 4 then begin
    Ixx := GetParameter(FirstWidthPar+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(FirstWidthPar+Ord(YAxisFor[Molecule.Representation]));
    IxxmIyy := Ixx-Iyy + wDelta;
    JSq := J*(J+2) / 4;
    Result := IxxmIyy/4 * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else
    Result := 0;
end;

function TClosedAsymmetricTop.HRotA(J, Kp, K: TQuantumNumber): Double;
{ Hamiltonian using Watson's A reduction }
var
  Izz, Iyy, Ixx, JSq, KSq, KpSq, IxxpIyy, IxxmIyy: Double;
begin
  if K = Kp then begin
    Izz := GetParameter(LSP+Ord(ZAxisFor[Molecule.Representation]));
    Ixx := GetParameter(LSP+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(LSP+Ord(YAxisFor[Molecule.Representation]));
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    IxxpIyy := Ixx + Iyy + BBar*2;
    Result := Origin + Izz*KSq  + IxxpIyy*(JSq - KSq)/2;
    Result := Result - DJ*Sqr(JSq) - DJK*JSq*KSq - DK*Sqr(KSq);
    Result := Result + HJ*Sqr(JSq)*JSq + HJK*Sqr(JSq)*KSq + HKJ*Jsq*Sqr(KSq) + HK*Sqr(KSq)*KSq;
    Result := Result + LJ*Sqr(Sqr(JSq)) + LJJK*Sqr(JSq)*JSq*KSq + LJK*Sqr(JSq)*Sqr(KSq)
                     + LKKJ*JSq*KSq*Sqr(KSq) + LK*Sqr(Sqr(KSq))
  end else if Abs(K-Kp) = 4 then begin
    Ixx := GetParameter(LSP+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(LSP+Ord(YAxisFor[Molecule.Representation]));
    IxxmIyy := Ixx-Iyy + BDelta;
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    KpSq := Kp*Kp / 4;
    Result := IxxmIyy/4 - deltaJ*JSq - deltaK*(KSq+KpSq)/2;
    Result := Result + phiJ*Sqr(JSq) + phiJK*Jsq*(KSq+KpSq)/2 + phiK*(Sqr(KSq)+Sqr(KpSq))/2;
    Result := Result + llJ*JSq*Sqr(JSq) + llJK*Sqr(JSq)*KSq + llKJ*Jsq*Sqr(KSq) + llK*Sqr(KSq)*KSq;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else
    Result := 0;
end;

function TClosedAsymmetricTop.HRotS(J, Kp, K: TQuantumNumber): Double;
{ Hamiltonian using Watson's S reduction }
var
  Iyy, Ixx, JSq, IxxmIyy: Double;
begin
  if K = Kp then
    Result := HRotA(J, KP, K)
  else if Abs(K-Kp) = 4 then begin
    Ixx := GetParameter(LSP+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(LSP+Ord(YAxisFor[Molecule.Representation]));
    IxxmIyy := Ixx-Iyy + BDelta;
    JSq := J*(J+2) / 4;
    Result := IxxmIyy/4 + (deltaJ + (phiJ + llJ*JSq)*JSq)*JSq;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else if Abs(K-Kp) = 8 then begin
    JSq := J*(J+2) / 4;
    Result := (deltaK + (phiJK + llJK*JSq)*JSq)*rl(J, K, Kp)
  end else if Abs(K-Kp) = 12 then begin
    JSq := J*(J+2) / 4;
    Result := (phiK + llKJ*JSq)*rl(J, K, Kp)
  end else if Abs(K-Kp) = 18 then
    Result := llK*rl(J, K, Kp)
  else
    Result := 0;
end;

function TClosedAsymmetricTop.HDip(TransitionMoment: TTransitionMoment; Jp, Kp, J, K: TQuantumNumber): Double;
var
  q: TQuantumNumber;
  Phase: Integer;
begin
  Result := 0;
  Phase := 0;
  if TransitionMoment is TSphericalTransitionMoment then begin
    q := TSphericalTransitionMoment(TransitionMoment).Component;
    { To match symmetries, treat + and - q as actually T(k,|q|) +- T(k,-|q|) }
    if (Kp < K) and (q < 0) then
      Phase := 2;
  end else if TransitionMoment is TCartesianTransitionMoment then begin
    case xyzAxis(Molecule.Representation, TCartesianTransitionMoment(TransitionMoment).Axis) of
      zAxis: q := 0;
      yAxis: q := 2;
      xAxis: begin
               q := 2;
               if K < Kp then
                 Phase := 2;
             end;
    end;
    { TODO: may need 1/Sqrt(2); Check for complex matrix elements }
  end else
    raise Exception.CreateFmt('%s Can''t Handle a %s', [Name, TransitionMoment.ClassName]);
  if Abs(q) = Abs(Kp-K) then
    Result := Sqrt((J+1.0)*(Jp+1))*Minus(q+Phase)*threej(Jp, Kp, TransitionMoment.Rank, K-Kp, J, -K)*TransitionMoment.Strength;
end;

function TClosedAsymmetricTop.MatrixElement(ABra, AKet: TBasisState; MatElemType: TMatElemType): Double;
{ Matrix elements in a symmetrised case (a) basis. }
var
  Bra, Ket: TClosedAsymmetricTopBasisState;
  K, Kp: TQuantumNumber;
  H: function (J, Kp, K: TQuantumNumber): Double of object;
begin
  Bra := ABra as TClosedAsymmetricTopBasisState;
  Ket := AKet as TClosedAsymmetricTopBasisState;
  Result := 0;
  if (Bra.State <> Ket.State) or (Bra.Sym <> Ket.Sym) or (Bra.J <> Ket.J) then
    Exit; { TODO : Hyperfine, Perturbations }
  case MatElemType of
    meHrot:  if Molecule.SReduction then
               H := {$IFDEF FPC}@{$ENDIF}HrotS
             else
               H := {$IFDEF FPC}@{$ENDIF}HrotA;
    meWidth: H := {$IFDEF FPC}@{$ENDIF}HWidth;
  else
    raise Exception.Create('Not Yet');
    // NotYet;
  end;
  K := Ket.K;
  Kp := Bra.K;
  if K = 0 then
    if Kp = 0 then
      Result := H(Bra.J, Kp, K)
    else
      Result := (H(Bra.J, Kp, K) + Bra.Phase*H(Bra.J, -Kp, K))/Sqrt(2.0)
  else if Kp = 0 then
    Result := (H(Bra.J, Kp, K) + Ket.Phase*H(Bra.J, Kp, -K))/Sqrt(2.0)
  else
    { All 4 here to permit C1 symmetry to work - could be speeded up with specific test }
    Result := (H(Bra.J, Kp, K) + Bra.Phase*H(Bra.J, -Kp, K)
               + Ket.Phase*H(Bra.J, Kp, -K) + Ket.Phase*Bra.Phase*H(Bra.J, -Kp, -K))/2
end;

function TClosedAsymmetricTop.TransitionMatrixElement(TransitionMoment: TTransitionMoment; ABra, AKet: TBasisState): Double;
var
  Bra, Ket: TClosedAsymmetricTopBasisState;
begin
  Bra := ABra as TClosedAsymmetricTopBasisState;
  Ket := AKet as TClosedAsymmetricTopBasisState;
  if Bra.K = 0 then
    if Ket.K = 0 then
      Result := HDip(TransitionMoment, Bra.J, Bra.K, Ket.J, Ket.K)
    else
      Result := (HDip(TransitionMoment, Bra.J, Bra.K, Ket.J, Ket.K) + Ket.Phase*HDip(TransitionMoment, Bra.J, Bra.K, Ket.J, -Ket.K))/Sqrt(2.0)
  else if Ket.K = 0 then
    Result := (HDip(TransitionMoment, Bra.J, Bra.K, Ket.J, Ket.K) + Bra.Phase*HDip(TransitionMoment, Bra.J, -Bra.K, Ket.J, Ket.K))/Sqrt(2.0)
  else
    Result := (HDip(TransitionMoment, Bra.J, Bra.K, Ket.J, Ket.K) + Bra.Phase*HDip(TransitionMoment, Bra.J, -Bra.K, Ket.J, Ket.K)
               +Ket.Phase*HDip(TransitionMoment, Bra.J, Bra.K, Ket.J, -Ket.K) + Bra.Phase*Ket.Phase*HDip(TransitionMoment, Bra.J, -Bra.K, Ket.J, -Ket.K))/2;
end;

function TClosedAsymmetricTop.PerturbationMatrixElement(Perturbation: TPerturbation; ABra, AKet: TBasisState; MatElemType: TMatElemType): Double;
var
  Bra, Ket: TClosedAsymmetricTopBasisState;
  P: TClosedAsymmetricTopPerturbation;
begin
  if MatElemType <> meHrot then
    raise NotYet;
  Bra := ABra as TClosedAsymmetricTopBasisState;
  Ket := AKet as TClosedAsymmetricTopBasisState;
  P := Perturbation as TClosedAsymmetricTopPerturbation;
  if Bra.J <> Ket.J then
    raise Exception.Create('TClosedAsymmetricTop.PerturbationMatrixElement');
  if Bra.K = 0 then
    if Ket.K = 0 then
      Result := P.MatrixElement(Bra.J, Bra.K, Ket.J, Ket.K)
    else
      Result := (P.MatrixElement(Bra.J, Bra.K, Ket.J, Ket.K) + Ket.Phase*P.MatrixElement(Bra.J, Bra.K, Ket.J, -Ket.K))/Sqrt(2.0)
  else if Ket.K = 0 then
    Result := (P.MatrixElement(Bra.J, Bra.K, Ket.J, Ket.K) + Bra.Phase*P.MatrixElement(Bra.J, -Bra.K, Ket.J, Ket.K))/Sqrt(2.0)
  else
    Result := (P.MatrixElement(Bra.J, Bra.K, Ket.J, Ket.K) + Bra.Phase*P.MatrixElement(Bra.J, -Bra.K, Ket.J, Ket.K)
               +Ket.Phase*P.MatrixElement(Bra.J, Bra.K, Ket.J, -Ket.K) + Bra.Phase*Ket.Phase*P.MatrixElement(Bra.J, -Bra.K, Ket.J, -Ket.K))/2;
end;


function TClosedAsymmetricTop.HasWidth: Boolean;
var
  i: Integer;
begin
  Result := inherited HasWidth;
  if not Result then
    for i := FirstWidthPar to nParameters do
      if Parameters[i] <> 0 then begin
        Result := True;
        Exit;
      end;
end;

{ TClosedAsymmetricManifold }

class function TClosedAsymmetricManifold.OwnerType: TComponentClass;
begin
  Result := TClosedAsymmetricMolecule;
end;

function TClosedAsymmetricManifold.FirstSym(J: TQuantumNumber):Integer;
begin
  Result := 0;
end;

function TClosedAsymmetricManifold.NextSym(J: TQuantumNumber; Sym: Integer):Integer;
begin
  Result := Sym+1;
  if Result >= nSym then
    Result := -1;
end;

function TClosedAsymmetricManifold.nSym: Integer;
begin
  Result := TClosedAsymmetricMolecule(Owner).LastSym+1;
end;

{ TClosedAsymmetricMolecule }

function TClosedAsymmetricMolecule.GetPointGroup(Index: integer): TPointGroup;
begin
  Result := TPointGroup(GetSetting(Index));
end;

function TClosedAsymmetricMolecule.GetRepresentation(Index: integer): TRepresentation;
begin
  Result := TRepresentation(GetSetting(Index));
end;

procedure TClosedAsymmetricMolecule.SetPointGroup(Index: integer; const AValue: TPointGroup);
begin
  SetSetting(Index, Ord(AValue));
end;

procedure TClosedAsymmetricMolecule.SetRepresentation(Index: integer; const AValue: TRepresentation);
begin
  SetSetting(Index, Ord(AValue));
end;

constructor TClosedAsymmetricMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',PointGroup=C1,Representation=Ir,eeWt=1,eoWt=1,oeWt=1,ooWt=1,C2xAxis=c';
  inherited Create(AOwner);
end;

function TClosedAsymmetricMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
// var
//  q: TQuantumNumber;
//  Sym: Integer; { TODO: Precompute }
//  function AntiSymmetricWrtAxis(Axis: TxyzAxis): Boolean;
//  { Given axis, determine from Rank and q whether wavefunction is symmetric or antisymmetric
//    wrt C2 about that axis }
//  begin
//    case Axis of
//      xAxis: Result := Minus(TransitionMoment.Rank) < 0;
//      yAxis: Result := Minus(TransitionMoment.Rank-q) < 0;
//      zAxis: Result := Minus(q) < 0;
//    end;
//    if (Axis <> xAxis) and (q < 0) then
//      Result := not Result; { T(k,|q|) - T(k,-|q|) }
//  end;
begin
//  if TransitionMoment is TSphericalTransitionMoment then begin
//    { Symmetry is the same as a rotational wavefunction with J=Rank, K=Component }
//    { But to get operators with the correct symmetry, must take
//      T(k,|q|) +- T(k,-|q|)
//      so treat the sign of q as the sign of this combination }
//    { z Symmetry is (-1)^q }
//    { x Symmetry is (-1)^Rank }
//    q := TSphericalTransitionMoment(TransitionMoment).Component;
//    Sym := 0;
//    if AntiSymmetricWrtAxis(aAxisFor[Representation]) then
//      Sym := Sym or 2;
//    if AntiSymmetricWrtAxis(cAxisFor[Representation]) then
//      Sym := Sym or 1;
//  end else if TransitionMoment is TCartesianTransitionMoment then begin
//    case TCartesianTransitionMoment(TransitionMoment).Axis of
//      aAxis: Sym := 1;
//      bAxis: Sym := 3;
//      cAxis: Sym := 2;
//    end;
//  end else
//    raise Exception.CreateFmt('%s can''t handle %s', [Name, TransitionMoment.ClassName]);
//  Result := (Sym xor Sym1 xor Sym2) = 0;
  { TODO: Copy to individual matrix element }
  { This is essentially testing for a parity change }
  if FakeSym or PseudoC2v then
    Result := True
  else if not TransitionMoment.Magnetic and  (Minus(TransitionMoment.Rank) < 0) then
    Result := (OddRankSym xor Sym1 xor Sym2) = 0
  else
    Result := Sym1 = Sym2;
end;

function TClosedAsymmetricMolecule.FixupProperties: Exception;
begin
  if PseudoC2v then
    FOddRankSym := eoSymC2v(symA2)
  else
    case PointGroup of
      C2v: FOddRankSym := eoSymC2v(symA2);
      Cs:  FOddRankSym := 1;
      C1:  FOddRankSym := 0;
    else
      raise NotYet;
    end;
  Result := Inherited FixupProperties;
end;

function TClosedAsymmetricMolecule.GetAxis(const Index: Integer): TabcAxis;
begin
  Result := TabcAxis(GetSetting(Index));
end;

procedure TClosedAsymmetricMolecule.SetAxis(const Index: Integer; const AValue: TabcAxis);
begin
  SetSetting(Index, Ord(AValue));
end;

const
  AxisNames: array [0..3] of PChar = ('a', 'b', 'c', nil);

function TClosedAsymmetricMolecule.GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar;
begin
  if AIndex in [LMI+7, LMI+8] then
    Result := @AxisNames
  else
    Result := inherited GetSettingsNames(AIndex,LowestValue);
end;

const
                                                   {Ir, IIr, IIIr, Il, IIl, IIIl}
  JAdjustMask: array[TRepresentation] of Integer = ( 1,   3,    2,  1,   3,    2);

function TClosedAsymmetricMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
  if Mixture.JAdjustSym and (Minus(J) < 0) then
    if PseudoC2v then
      Result := Result xor JAdjustMask[Representation]
    else
      case PointGroup of
        C2v: Result := Result xor JAdjustMask[Representation];
        { Three cases here:
          1. The J alternation flips bewteen the two C2v symmetries that are in the same matrix - don't alternate
          2. The J alternation flips to the other matrix for both symmetries - alternate
          3. The J alternation flips one symmetry but not the other to the other matrix - In this case
             neither choice is likely to make sense, so alternate, so user has choice }
        Cs:  if JAdjustMask[Representation] <> SymWrt[C2zAxis] then
               Result := Result xor 1;
        C1: { All matrices the same }
      end;
  { TODO : other point groups }
end;

function TClosedAsymmetricMolecule.LastSym: Integer;
begin
  if PseudoC2v then
    Result := 4
  else
    case PointGroup of
      C2v: Result := 4;
      Cs: Result := 2;
      C1: Result := 1;
    else
      raise NotYet;
    end;
  Dec(Result);
end;

{ TClosedAsymmetricTopBasisState }

function TClosedAsymmetricTopBasisState.BasisLabel: string;
var
  c: Char;
begin
  if Phase > 0 then
    c := '+'
  else
    c := '-';
  Result := Format('%s %2s%s', [inherited BasisLabel, State.Manifold.Mixture.QuantumNumberToStr(K), c]);
end;

function TClosedAsymmetricTopBasisState.StateLabel: string;
begin
  Result := Format('%s %2s %2s %2s', [inherited StateLabel, State.Manifold.Mixture.QuantumNumberToStr(J), State.Manifold.Mixture.QuantumNumberToStr(Ka), State.Manifold.Mixture.QuantumNumberToStr(Kc)]);
end;

function TClosedAsymmetricTopBasisState.StateLabelKey: string;
begin
  Result := inherited StateLabelKey + ' J Ka Kc';
end;

function TClosedAsymmetricTopBasisState.BasisLabelKey: string;
begin
  Result := inherited BasisLabelKey + ' K +/-';
end;

{ TClosedAsymmetricTopPerturbation }

function TClosedAsymmetricTopPerturbation.MatrixElement(Jp, Kp, J, K: TQuantumNumber): Double;
var
  xPower, yPower, zPower: Integer;
  dK: Integer;
begin
  xPower := GetSetting(LPI+1+Ord(XAxisFor[Molecule.Representation]));
  yPower := GetSetting(LPI+1+Ord(YAxisFor[Molecule.Representation]));
  zPower := GetSetting(LPI+1+Ord(ZAxisFor[Molecule.Representation]));
  dK := Abs(K-Kp) div 2;
  Result := 0;
  if (dK <= xPower + yPower) and not Odd(dK+xPower+yPower) then begin
    if (xPower = 1) and (yPower = 0) and (zPower = 1) then begin
      Result := Value * (K + Kp)*rl(J, K, Kp);
      if JPower > 0 then
        Result := Result * IntPower(J*(J+2), JPower div 2);
    end else
      raise NotYet;
    Result := Result / IntPower(2,xPower + yPower + zPower + JPower) * Minus(yPower);
  end;
end;

function TClosedAsymmetricTopPerturbation.Molecule: TClosedAsymmetricMolecule;
begin
  Result := Owner.Owner as TClosedAsymmetricMolecule;
end;

function TClosedAsymmetricTopPerturbation.PairName: string;
  function Show(Letter: Char; Power: Integer): string;
  begin
    case Power of
      0: Result := '';
      1: Result := Letter;
    else
      Result := Letter + IntToStr(Power)
    end;
  end;
begin
  Result := Show('J',JPower) + Show('a',aPower) + Show('b',bPower) + Show('c',cPower)
end;

initialization
  RegisterClasses([TClosedAsymmetricMolecule, TClosedAsymmetricManifold, TClosedAsymmetricTop, TClosedAsymmetricTopPerturbation]);

end.

