unit CalSource;
{$I DFS.inc}{Copyright (C) 2006 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} Graphs, Menus, {$ENDIF}
     GraphRange, SpectrumPlot, Overlay, Mixture, Units, Classes, SysUtils, GraphM;

type

  TCalibrationSource = class(TStretchablePlot)
  public
    procedure AddToSpectrum(Spectrum: TSpectrum; Fmin, Fmax: Double); virtual; abstract;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    function MakeSpectrum(Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits):TSpectrumPlot;
    function LimitsInRange(Simulation: TSimulation; var AIMin, AImax: Double; PlotIntensityUnits: TIntensityUnits):Boolean; override;
    function Show(G: TGraphM; Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits;
                  PlotRange: TPlotRange; var PlotRanges: TPlotRanges; OwnAxis: Boolean; ShowName: Boolean; ForceAxis: Boolean = False):Boolean; override;
  end;

  TLineListSource = class(TCalibrationSource)
  protected
    FLineList: TLineList;
  public
    procedure AddToSpectrum(Spectrum: TSpectrum; Fmin, Fmax: Double); override;
    function FindPeak(AFmin, AFmax: Double; out Position: Double; PlotUnits: TResonanceUnits): Boolean; override;
    function FixupProperties:Exception; override;
    procedure SetName(const Value: TComponentName); override;
    property LineList: TLineList read FLineList write FLineList;
  published
    property IgnoreIntensity: Boolean index 7 read GetFlag write SetFlag;
  end;

  TCalculatedSource = class(TCalibrationSource)
    function FindPeak(AFmin, AFmax: Double; out Position: Double; PlotUnits: TResonanceUnits): Boolean; override;
  end;

function MakeCalibrationSource(Name: string; Owner: TComponent):TCalibrationSource;

{$IFNDEF CONSOLE}
procedure StuffCalibrationMenu(ParentMenu: TMenuItem; OnClick: TNotifyEvent);
{$ENDIF}

implementation

uses Profile, Global, MathSup, I2, Atomic, MNode;

{ TCalculatedSource }

function TCalculatedSource.FindPeak(AFmin, AFmax: Double; out Position: Double;
  PlotUnits: TResonanceUnits): Boolean;
var
  LineList: TLineList;
begin
  LineList := TLineList.Create(Units,False);
  try
    AddToSpectrum(LineList, AFmin, AFmax);
    Result := LineList.FindPeak(AFmin,AFmax,Position,PlotUnits,Colour);
  finally
    LineList.Free;
  end;
end;

{ TLineListSource }

procedure TLineListSource.AddToSpectrum(Spectrum: TSpectrum; Fmin,
  Fmax: Double);
var
  i: Integer;
  Int: Double;
begin
  Int := 1;
  for i := 0 to LineList.nLines - 1 do
    with LineList.Line[i] do begin
      if not IgnoreIntensity then
        Int := Strength;
      Spectrum.AddLine(AdjustFrequency(Frequency),Int,Width,Colour,LineList.Units);
    end;
end;

function TLineListSource.FindPeak(AFmin, AFmax: Double; out Position: Double;
  PlotUnits: TResonanceUnits): Boolean;
begin
  AFmin := Convert(DeAdjustFrequency(Convert(AFmin, PlotUnits, Units)),Units,PlotUnits);
  AFmax := Convert(DeAdjustFrequency(Convert(AFmax, PlotUnits, Units)),Units,PlotUnits);
  Result := LineList.FindPeak(AFmin,AFmax,Position,PlotUnits,Colour);
  if Result then begin
    Position := AdjustFrequency(Position);
    PeakMeasured(Convert(Position,Units,PlotUnits),PlotUnits,Name,nil);
  end;
end;

function TLineListSource.FixupProperties: Exception;
begin
  Result := inherited FixupProperties;
  FHMin := LineList.FMin;
  FHMax := LineList.FMax;
end;

procedure TLineListSource.SetName(const Value: TComponentName);
var
  i: Integer;
begin
  for i := 0 to High(AtomicLineLists) do
    if (CompareText(Value,AtomicLineLists[i].Name) = 0)
       or (CompareText(Value,AtomicLineLists[i].Name+' Calibration') = 0) then begin
      ModifySetting(1, Ord(nmVac), False);
      IgnoreIntensity := AtomicLineLists[i].IgnoreIntensities;
      LineList := AtomicLineLists[i].LineList;
      inherited SetName(AtomicLineLists[i].Name + ' Calibration');
      Exit;
    end;
  if (Value <> '') and (Value <> 'LineListSource') then
    raise EReportException.Create('Unknown line list type:'+Value)
  else
    inherited;
end;

{ TCalibrationSource }

var
 CalibrationSpectrumTime: string = 'Calibration Plot';

function TCalibrationSource.Show(G: TGraphM; Simulation: TSimulation;
  PlotIntensityUnits: TIntensityUnits; PlotRange: TPlotRange;
  var PlotRanges: TPlotRanges; OwnAxis, ShowName: Boolean; ForceAxis: Boolean): Boolean;
var
  Spectrum: TSpectrumPlot;
begin
  Spectrum := MakeSpectrum(Simulation, PlotIntensityUnits);
  try
    if Assigned(Spectrum) then begin
      PlotRange.yx := Spectrum.IMax;
      Spectrum.Show(G, PlotRange, False, False, PlotRanges);
      Result := True;
    end else
      Result := False;
    Result := inherited Show(G,Simulation,PlotIntensityUnits,PlotRange,PlotRanges,OwnAxis,False,ForceAxis)
              or Result;
  finally
    Spectrum.Free;
  end;
end;

function TCalibrationSource.LimitsInRange(Simulation: TSimulation;
  var AIMin, AImax: Double;
  PlotIntensityUnits: TIntensityUnits): Boolean;
var
  Spectrum: TSpectrumPlot;
begin
  Spectrum := MakeSpectrum(Simulation, PlotIntensityUnits);
  if Assigned(Spectrum) then begin
    AIMin := Spectrum.IMin;
    AImax := Spectrum.IMax;
    Spectrum.Free;
    Result := True;
  end else
    Result := False;
end;

function TCalibrationSource.MakeSpectrum(Simulation: TSimulation;
  PlotIntensityUnits: TIntensityUnits): TSpectrumPlot;
var
  Xi, Xx: Double;
begin
  Result := TSpectrumPlot.Create(Simulation.PlotUnits,False);
  try
    Result.InitializeFromSimulation(Simulation);
    Result.IntensityUnits := PlotIntensityUnits;
    Result.Colour := Colour;
    Result.NoBaseLine := True;
    Result.nDF := Result.nDF; { Force recalc of convolution stuff }
    { Avoid potential divide by zero in conversion }
    Xi := Result.FMin;
    Xx := Result.FMax;
    InOrder(Xi,Xx);
    if Xi <= 0 then begin
      Xx := Abs(Xx);
      if Xx = 0 then
        Xx := 1;
      Xi := Xx/100;
    end;
    Xi := Convert(Xi, Result.Units, Units);
    Xx := Convert(Xx, Result.Units, Units);
    InOrder(Xi,Xx);
    TStart(CalibrationSpectrumTime);
    AddToSpectrum(Result,Xi,Xx);
    TEnd(CalibrationSpectrumTime);
    if Result.nLinesInRange > 0 then begin
      Result.Convolute;
      Result.FindLimits;
    end else
      FreeAndNil(Result);
  except
    Result.Free;
    raise;
  end;
end;

procedure TCalibrationSource.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  if AIndex <> 1 then
    inherited; { Can't change units }
end;

function MakeCalibrationSource(Name: string; Owner: TComponent):TCalibrationSource;
var
  i: Integer;
begin
  if Name = 'I2' then
    Result := TI2.Create(Owner)
  else begin
    for i := 0 to High(AtomicLineLists) do
      if AtomicLineLists[i].Name = Name then begin
        Result := TLineListSource.Create(Owner);
        Result.Name := Name + ' Calibration';
        Result.ModifySetting(1, Ord(nmVac), False);
        TLineListSource(Result).IgnoreIntensity := AtomicLineLists[i].IgnoreIntensities;
        TLineListSource(Result).LineList := AtomicLineLists[i].LineList;
      end;
  end;
end;

{$IFNDEF CONSOLE}
procedure StuffCalibrationMenu(ParentMenu: TMenuItem; OnClick: TNotifyEvent);
var
  i: Integer;
  MenuItem: TMenuItem;
begin
  for i := 0 to High(AtomicLineLists) do begin
    MenuItem := TMenuItem.Create(ParentMenu.Owner);
    MenuItem.Caption := AtomicLineLists[i].Name;
    MenuItem.Tag := i+1;
    MenuItem.OnClick := OnClick;
    MenuItem.Hint := 'Add overlay for transitions of atomic ' + AtomicLineLists[i].Name;
    ParentMenu.Insert(i+1,MenuItem);
  end;
end;
{$ENDIF}

initialization
  TLineListSource.RegisterClass(TOverlays);
end.
