unit CheckJB95;
{ Read a JB95 output (.asn) file, and compare to a calculation of the same thing }

{$I DFS.inc}{Copyright (C) 2007 C M Western; see commandline.pas for license}

interface

uses Mixture, TextStrm;

function CheckJB95ViewFile(Mixture: TMixture; S: TTextStream;
                           var nTried: Integer; Tolerance: Double; var Failures: Integer
                          ):Double;

implementation

uses SysUtils, Worker, QuantumNumber, MathSup, MatrixVec, MNode;

function CheckJB95ViewFile(Mixture: TMixture; S: TTextStream;
                           var nTried: Integer; Tolerance: Double; var Failures: Integer
                          ):Double;
var
  Failed: Boolean;
  MaxDifference: Double;
  Simulation: TSimulation;
  Line: TLine;
  Problem: Exception;
  Jp, J, Kap, Kcp, Ka, Kc: TQuantumNumber;
  np: Smallint; { Upper State }
  symp: Shortint;
  n: SmallInt; { Lower State }
  sym: Shortint;
  StrU, StrL: string;
  ManifoldL, ManifoldU: TManifold;
  Molecule: TMolecule;
  Frequency, Intensity, cstr: Double;
  i: Integer;

  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
      WriteLn(Log, Jp:3, J:5, Frequency:12:4, Intensity:8:4, ' ', StrU, ' ', StrL);
      if Assigned(Line.EigenvaluesU) then
        Mixture.LineList.ShowLine(Line, Log, plLineList, 0, False, Problem);
    end;
    Failed := True;
  end;

  function CheckNo(n1, n2, tol: Double; Name: string; Defer: Boolean = False):Boolean;
  var
    Msg: string;
  begin
    Result := True; { Assume fail }
    if isNanOrInf(n2) then
      Msg := Format('%s fails: %g Nan', [Name, n1])
    else if Abs(n1-n2) > tol then
      Msg := Format('%s out of tolerance: %g %g %g (%g)', [Name, n1, n2, n1-n2, n1/n2])
    else
      Result := False; { OK }
    if Abs(n1-n2) > MaxDifference then
      MaxDifference := Abs(n1-n2);
    if Result then
      Fail(Msg);
  end;

  procedure NextQuantumVar(var Q: TQuantumNumber);
  var
    str: String;
  begin
    str := S.NextString('?');
    if not TryStrToQuantumNumber(str, Q, qftJ) then
      Fail('Bad quantum number:'+str);
  end;

var
  Dipole: string;
  TmAxis: TabcAxis;
  D: TMatrix;
  TransitionMoments: TTransitionMoments;
  TransitionMoment: TTransitionMoment;
begin
  S.Comments := False;
  MaxDifference := 0;
  Simulation := Mixture.Simulations[0];
  Problem := nil;
  Simulation.MakeLineList(Problem);
  if Assigned(Problem) then
    raise Problem;
  Molecule := Mixture.Species[0].Isotopomer[0];
  ManifoldL := Molecule.Manifold[0];
  if Molecule.nManifolds > 1 then
    ManifoldU := Molecule.Manifold[1]
  else
    ManifoldU := ManifoldL;
  while not S.AtEndOfFile do begin
    { Line format is <mucomp> J' Ka' Kc' J" Ka" Kc" Intensity Obs Calc }
    Failed := False;
    Line.EigenvaluesU := nil;
    Dipole := S.NextString('');
    if (Length(Dipole) <> 1) or not (Dipole[1] in ['A'..'C','a'..'c']) then
      if Dipole <> '' then
        Fail('Invalid Dipole: ' + Dipole);
    if Failed or (Dipole = '') then begin
      S.NextLine;
      Continue;
    end;
    TmAxis := TabcAxis(Ord(LowerCase(Dipole)[1])-Ord('a'));
    NextQuantumVar(Jp); NextQuantumVar(Kap); NextQuantumVar(Kcp);
    NextQuantumVar(J); NextQuantumVar(Ka); NextQuantumVar(Kc);
    S.NextDoubleVar(Intensity);
    S.NextDouble(0); { Observed }
    S.NextDoubleVar(Frequency);
    StrU := ManifoldU.FindIndex(Line.EigenvaluesU, np, symp, Jp, [Kap, Kcp], qftJ, Simulation.ExField);
    if np < 0 then
      Fail(StrU)
    else begin
      StrL := ManifoldL.FindIndex(Line.EigenvaluesL, n, sym, J, [Ka, Kc], qftJ, Simulation.ExField);
      if n < 0 then
        Fail(StrL);
    end;
    if not Failed then begin
      Line.Width := 0;
      Line.IndexU := np;
      Line.IndexL := n;
      Line.Frequency := Line.EigenvaluesU.E[Line.IndexU]-Line.EigenvaluesL.E[Line.IndexL];
      Line.CachedName := '';
      TransitionMoments := Molecule.FindTransitionMoments(ManifoldU,ManifoldL);
      if Assigned(TransitionMoments) then begin
        TransitionMoment := nil;
        for i := 0 to TransitionMoments.Count - 1 do
          with TransitionMoments.TransitionMoment[i] as TCartesianTransitionMoment do
            if Axis = TmAxis then begin
              TransitionMoment := TransitionMoments.TransitionMoment[i];
              Break;
            end;
        if not Assigned(TransitionMoment) then
          Fail('Dipole not found:'+Dipole);
        { Do this here so line printout has correct intensity }
        D := TransitionMoments.TransitionMatrix(Line.EigenValuesL, Line.EigenValuesU, True, True, TransitionMoment, Simulation);
        Line.Strength := D[Line.IndexL, Line.IndexU]*Line.EigenvaluesL.StatisticalWeight(Line.IndexL)
                         *Line.EigenvaluesL.Manifold.Molecule.Abundance
                         *Line.EigenvaluesL.Manifold.Molecule.Species.Concentration;
      end;
      CheckNo(Frequency, Line.Frequency, Tolerance, 'Frequency');
      if not Assigned(TransitionMoments) then
        Fail('Can''t find transition moment')
      else begin
        Problem := nil;
        cstr := Simulation.LineIntensity(Line, Mixture.Units, Problem);
        if Assigned(Problem) then begin
            Fail(Problem.Message);
            Problem.Free;
          end;
        { Intensity is only printed to 0.001 }
        CheckNo(Intensity, cstr, 0.001, 'Intensity', True);
      end;
    end;
    S.NextLine;
  end;
  Result := MaxDifference;
end;

end.
