unit CheckPgopher;
{ Read a pgopher output file, and compare to a calculation of the same thing }

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

interface

uses
  Classes, SysUtils, Mixture, TextStrm;

function CheckPgopherFile(Generated: Boolean; Mixture: TMixture; S: TTextStream; var nTried: Integer;
                          Tolerance: Double; var Failures: Integer;
                          InputPgopherVersion: Integer; PseudoSym: Boolean; LineListMode: Integer):Double;

function CheckEigenvalueFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; InputPgopherVersion: Integer): Double;

implementation

uses Math, MathSup, MatrixVec, Worker, QuantumNumber, MNode, Obs, OAsymT,
     ExternalList, Abelian, ShellSort, Progress, NextTok;

function VarAndTol(const S: string; out Value, Tolerance: Double; var MaxVal: Double):Boolean;
var
  t: string;
  i, LastDigit: Integer;
begin
  if not TryStrToFloat(S,Value) then begin
    Tolerance := 0;
    if (CompareText(s,'Nan') = 0) or (CompareText(s,'Inf') = 0) then
      SetNan(Value)
    else
      raise Exception.Create('Bad number:'+S);
    Exit;
  end;
  MaxVal := Max(MaxVal, Abs(Value));
  { Figure out tolerance by replacing all digits except the last by 0 }
  t := S;
  LastDigit := -1;
  for i := 1 to Length(t) do
    case t[i] of
      '-': t[i] := ' ';
      '0'..'9': begin
                  if LastDigit > 0 then
                    t[LastDigit] := '0';
                  LastDigit := i;
                  t[i] := '1';
                end;
      'e', 'E': Break;
      'd', 'D': begin
                  t[i] := 'e';
                  Break;
                end;
      '.': { Leave as is } ;
    else
      raise Exception.Create('Bad number:'+S);
    end;
  Tolerance := StrToFloat(t)*1.6;
  { be more slack on very small values }
  Tolerance := Max(Tolerance,1e-14*MaxVal);
end;

function NextVarAndTol(S: TTextStream; out Value, Tolerance: Double; var MaxVal: Double):Boolean;
begin
  if not S.GetFreeItem then
    raise Exception.Create('No Number');
  Result := VarAndTol(S.Token, Value, Tolerance, MaxVal);
end;

{ For comparing linelists, prepare list of transitions,
  sorted by J', J", Sym', Sym" and frequency }

type

  TSLine = record
    JU, JL: TQuantumNumber;
    SymU, SymL: Integer;
    Frequency: Double;
    Intensity: Double;
    Strength: Double;
    ITol, STol: Double;
  end;

  TSLineList = array of TSLine;

  TSLineListSorter = class(TShellSorter)
    LineList: TSLineList;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Swap(I,J: Integer); override;
  end;

var
  FreqDiffThreshold: Double;

function CompareQnos(const A, B: TSLine):Integer;
begin
  Result := A.SymU - B.SymU;
  if Result = 0 then begin
    Result := A.SymL - B.SymL;
      if Result = 0 then begin
      Result := A.JU - B.JU;
      if Result = 0 then
        Result := A.JL - B.JL;
    end;
  end;
end;

function Compare(const A, B: TSLine; out Close: Boolean; AlternateTest: Boolean = False):Integer;
begin
  Close := False;
  Result := CompareQnos(A,B);
  if Result = 0 then begin
    Result := Sign(A.Frequency - B.Frequency);
    if Result = 0 then
      Close := True
    else begin
      Close := Abs(A.Frequency-B.Frequency) <= FreqDiffThreshold;
      if Close then { Actually want string representation to match }
        if AlternateTest then
          Close := Abs(A.Frequency-B.Frequency) <= FreqDiffThreshold/2
        else
          Close := RoundToIncrement(A.Frequency,FreqDiffThreshold) = RoundToIncrement(B.Frequency,FreqDiffThreshold);
    end;
  end;
end;

function TSLineListSorter.InOrder(I, J: Integer): Boolean;
var
  Close: Boolean;
begin
  Result := Compare(LineList[J],LineList[I], Close) >= 0;
end;

procedure TSLineListSorter.Swap(I, J: Integer);
var
  L: TSLine;
begin
  L := LineList[I];
  LineList[I] := LineList[J];
  LineList[J] := L;
end;

function MakeTSLineList(Simulation: TSimulation):TSLineList;
var
  Line: TLine;
  i, w: Integer;
  Problem: Exception;
  LineList: TLineList;
begin
  LineList := Simulation.Mixture.LineList;
  SetLength(Result,LineList.nLines);
  w := 0;
  for i := 0 to LineList.nLines - 1 do begin
    Line := LineList.Line[i];
    with Result[w] do begin
      JL := Line.EigenvaluesL.J;
      JU := Line.EigenvaluesU.J;
      SymL := Line.EigenvaluesL.Sym;
      SymU := Line.EigenvaluesU.Sym;
      Frequency := Line.Frequency;
      Strength := Line.Strength;
      Problem := nil;
      Intensity := Simulation.LineIntensity(Line, Simulation.Mixture.Units, Problem);
      if Assigned(Problem) then
        raise Problem;
      ITol := 0;
      STol := 0;
      if Intensity >= Simulation.OThreshold then
        Inc(w);
    end;
  end;
  SetLength(Result,w);
end;

procedure ShowSLine(var F: Text; const Line: TSLine; ShowStr: Boolean);
begin
  Write(Log,QuantumNumberToStr(Line.JU,qftJ):5, Line.SymU:3,
            QuantumNumberToStr(Line.JL,qftJ):5, Line.SymL:3,
            Line.Frequency);
  if ShowStr then
    WriteLn(Log, ' ', Line.Strength, ' ', Line.STol:6, ' ', Line.Intensity, Line.ITol:6)
   else
     WriteLn(Log);
end;

function CompareLists(CalcList, ReadList: TSLineList; MaxLineStrength: Double; IgnoreWeak: Boolean):Integer;
  function SumLines(out SSum, ISum, STol, ITol:Double; p: Integer; List: TSLineList; var Template: TSLine; ALternateTest: Boolean):Integer;
  var
    Close: Boolean;
  begin
    if Template.JU < 0 then begin
      SSum := List[p].Strength;
      ISum := List[p].Intensity;
      STol := List[p].STol;
      ITol := List[p].ITol;
      Result := p + 1;
      Template := List[p];
    end else begin
      Result := p;
      SSum := 0;
      ISum := 0;
      STol := 0;
      ITol := 0;
    end;
    while Result < Length(List) do begin
      Compare(Template,List[Result], Close, AlternateTest);
      if not Close then
        Break;
      SSum := SSum + List[Result].Strength;
      ISum := ISum + List[Result].Intensity;
      STol := STol + List[Result].STol;
      ITol := ITol + List[Result].ITol;
      Inc(Result);
    end;
  end;
var
  Sorter: TSLineListSorter;
  CP, RP, d, nC, nR, i: Integer;
  SSumC, ISumC, SSumR, ISumR, STol, ITol, Junk: Double;
  Close, AlternateTest: Boolean;
  Current: TSLine;
begin
  Sorter := TSLineListSorter.Create;
  Sorter.LineList := CalcList;
  Sorter.Sort(Length(CalcList));
  Sorter.LineList := ReadList;
  Sorter.Sort(Length(ReadList));
  Sorter.Free;
  CP := 0;
  RP := 0;
  Result := 0;

  while (Cp < Length(CalcList)) and (Rp < Length(ReadList)) do begin
    AlternateTest := False;
    if Cp >= Length(CalcList) then
      d := 1
    else if Rp >= Length(ReadList) then
      d := -1
    else begin
      d := Compare(CalcList[Cp],ReadList[Rp],Close);
      if Close then
        d := 0;
      if (d <> 0) and (CompareQnos(CalcList[Cp],ReadList[Rp]) = 0)
         and (Abs(CalcList[Cp].Frequency-ReadList[Rp].Frequency) < FreqDiffThreshold / 2 ) then begin
        { Allow for shift to other side of rounding }
        AlternateTest := True;
        d := 0;
      end;
    end;
    if d = 0 then begin
      { Sum lines with the same frequency }
      Current.JU := -1;
      nR := SumLines(SSumR, ISumR, STol, ITol, Rp, ReadList, Current, False);
      nC := SumLines(SSumC, ISumC, Junk, Junk, Cp, CalcList, Current, AlternateTest); { These will have STol = ITol = 0 }
      if ((Abs(SSumC-SSumR)>STol) or (Abs(ISumC-ISumR)>ITol))
         and (CalcList[Cp].Frequency > FreqDiffThreshold) { ignore very low frequency lines, as these can be intermittently discarded }
       then begin
        Inc(Result);
        Write(Log, 'Line intensity mismatch ', SSumC, ' ', SSumR, ' ', Abs(SSumC-SSumR)/STol:5:1, ' ', ISumC, ' ', ISumR, ' ', Abs(ISumC-ISumR)/ITol:5:1);
        ShowSLine(Log,ReadList[Rp],False);
        Write(Log, 'Calc Strength: ');
        for i := Cp to nC - 1 do
          Write(Log, CalcList[i].Strength);
        WriteLn(Log);
        Write(Log, 'Read Strength: ');
        for i := Rp to nR - 1 do
          Write(Log, ReadList[i].Strength, ' ', ReadList[i].STol:6);
        WriteLn(Log);
        Write(Log, 'Calc Intensity: ');
        for i := Cp to nC - 1 do
          Write(Log, CalcList[i].Intensity);
        WriteLn(Log);
        Write(Log, 'Read Intensity: ');
        for i := Rp to nR - 1 do
          Write(Log, ReadList[i].Intensity, ' ', ReadList[i].ITol:6);
        WriteLn(Log);
      end;
      Cp := nC;
      Rp := nR;
    end else if d < 0 then begin
      Current.JU := -1;
      nC := SumLines(SSumC, ISumC, STol, ITol, Cp, CalcList, Current, False);
      if (not IgnoreWeak or (ISumC > MaxLineStrength*1e-15))
         and (CalcList[Cp].Frequency > FreqDiffThreshold) then begin
        Inc(Result);
        Write(Log, 'Discarding ', nC-Cp, ' calculated lines (', Cp, ' ', Rp, ') :', SSumC:10, ' ', ISumC:10);
        ShowSLine(Log, CalcList[Cp], True);
        WriteLn(Log,CalcList[Cp].frequency,' ',CalcList[Cp-1].Frequency);
      end;
      Cp := nC;
    end else begin
      Current.JU := -1;
      nR := SumLines(SSumR, ISumR, STol, ITol, Rp, ReadList, Current, False);
      if (not IgnoreWeak or (SSumR > STol) or (ISumR > ITol)) { Ignore bonus low intensity lines }
         and (ReadList[Rp].Frequency > FreqDiffThreshold) then begin
        Inc(Result);
        Write(Log, 'Discarding ', nR-Rp, ' read lines (', Cp, ' ', Rp, ') :', SSumR:10, ' ', ISumR:10);
        ShowSLine(Log, ReadList[Rp], True);
        WriteLn(Log,CalcList[Cp].frequency,' ',ReadList[Rp].Frequency);
      end;
      Rp := nR;
    end;
  end;
end;

function CheckEigenvalueFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; InputPgopherVersion: Integer): Double;
var
  J, sym, n: Integer; { State Details }
  E, ETol, EMax: Double;
  Manifold: TManifold;
  Molecule: TMolecule;
  Eigenvalues: TEigenvalues;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
      WriteLn(Log, J:5, sym:2, n:2, E:12:4);
    end;
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string);
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else if Abs(n1-n2) > tol then
      Fail(Format('%s out of tolerance: Read=%g Calc=%g %g', [Name, n1, n2, n1-n2]));
    if Abs(n1-n2) > Result then
      Result := Abs(n1-n2);
  end;
  function FindSym(Manifold:TManifold; J: TQuantumNumber; SymName:String):Integer;
  begin
    if TryStrToInt(SymName, Result) then
      Result := Manifold.Molecule.FixSym(J, Result, InputPgopherVersion)
    else if not Manifold.Molecule.TryStrToAnySymName(J, SymName, Result) then begin
      Fail('Bad symmetry name:'+SymName);
      Result := -1;
    end;
  end;
var
 SymS: string;
begin
  nTried := 0;
  Molecule := Mixture.Species[0].Isotopomer[0];
  Manifold := Molecule.Manifold[0];
  Result := 0;
  EMax := 0;
  while not S.AtEndOfFile do begin
    { Skip lines not starting with an integer or manifold }
    if not S.GetFreeItem then begin
      S.Nextline;
      Continue;
    end;
    Manifold := Molecule.ManifoldWithName(S.Token);
    if Assigned(Manifold) then
      S.GetFreeItem
    else
      Manifold := Molecule.Manifold[0];
    if not TryStrToQuantumNumber(S.Token,J,Mixture.QuantumNumberFormat) then begin
      S.Nextline;
      Continue;
    end;
    S.NextStringVar(SymS);  S.NextIntegerVar(n);
    NextVarAndTol(S, E, ETol, EMax);
    Sym := FindSym(Manifold,J,SymS);
    Inc(nTried);
    if Sym >= 0 then begin
      Eigenvalues := Manifold.Eigenvalues(J, sym, [meHrot, meWidth], Mixture.Simulations[0].ExField);
      if n > Eigenvalues.Size then
        Fail(Format('Eigenvalue (%d) > %d', [n, Eigenvalues.Size]))
      else
        CheckNo(E, Eigenvalues.E[n], ETol, 'E');
     end;
    S.NextLine;
  end;
end;

function CheckPgopherFile(Generated: Boolean; Mixture: TMixture; S: TTextStream;
                          var nTried: Integer; Tolerance: Double; var Failures: Integer;
                          InputPgopherVersion: Integer; PseudoSym: Boolean; LineListMode: Integer):Double;
var
  MoleculeName: string;
  Manp, sympStr: string;
  Jp, np: Integer; { Upper State }
  symp: Shortint;
  Man, symStr: string;
  J,  n : Integer; { Lower State }
  sym: Shortint;
  f, cstr, w, Ep, E, Sum: Double;
  linestrength, linestrengthTol, linestrengthMax,
  Str, StrTol, StrMax,
  EstUnc, EstUncMax, EstUncTol,
  EinsteinA, EinsteinATol, EinsteinAMax: Double;
  ManifoldL, ManifoldU: TManifold;
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  D: TMatrix;
  Line: TLine;
  Failed: Boolean;
  CurrentFailure: string;
  MaxDifference, CurrentDifference: Double;
  Simulation: TSimulation;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
      WriteLn(Log, MoleculeName, ' ', Manp, ' ', Jp:3, symp:2, np:2, J:5, sym:2, n:2, f:12:4, str:8:4, w:8:3, Ep:12:4, E:12:4);
    end;
    Failed := True;
  end;
  function CheckNo(n1, n2, tol: Double; Name: string; Relative: Boolean = False; Defer: Boolean = False):Boolean;
  var
    Msg: string;
  begin
    Result := True; { Assume fail }
    if isNanOrInf(n2) then
      Msg := Format('%s fails: %g Nan', [Name, n1])
    else if (Relative and (Abs(n1-n2)/Abs(n1+n2) > tol)) or (not Relative and (Abs(n1-n2) > tol)) then
      Msg := Format('%s out of tolerance: %g %g %g (%g)', [Name, n1, n2, n1-n2, n1/n2])
    else
      Result := False; { OK }
    if Defer then begin
      if Abs(n1-n2) > CurrentDifference then
        CurrentDifference := Abs(n1-n2);
      if Result then begin
        if CurrentFailure <> '' then
          CurrentFailure := CurrentFailure + sLineBreak;
        CurrentFailure := CurrentFailure + Msg;
      end;
    end else begin
      if Abs(n1-n2) > MaxDifference then
        MaxDifference := Abs(n1-n2);
      if Result then
        Fail(Msg);
    end;
  end;
  function FindSym(Manifold:TManifold; J: TQuantumNumber; SymName:String):Integer;
  begin
    if Simulation.HasFields then
      Result := -1
    else if TryStrToInt(SymName, Result) then
      Result := Manifold.Molecule.FixSym(J, Result, InputPgopherVersion)
    else if not Manifold.Molecule.TryStrToAnySymName(J, SymName, Result) then
      if PseudoSym then
        Result := -1 { Symmetry and eigenvalue number must come from symmetry name }
      else
        Fail('Bad symmetry name:'+SymName);
    if PseudoSym and (Manifold.Molecule is TAsymmetricMolecule)
       and (TAsymmetricMolecule(Manifold.Molecule).PointGroup in [C2h, Ci]) then
      Result := -1; { For these point groups, Ag and Au are in both the high and low symmetry versions
                      of the point group, so force search mode }
  end;
  procedure NextQuantumVar(var Q: TQuantumNumber);
  var
    str: String;
  begin
    str := S.NextString('?');
    if not TryStrToQuantumNumber(str, Q, Mixture.QuantumNumberFormat) then
      Fail('Bad quantum number:'+str);
  end;
  procedure CheckDerivatives(Derivs: string);
    procedure CheckDerivative(const Derivative: TDerivative; Value: Double);
    var
      s, e: string;
      v, Tol, Mx: Double;
    begin
      s := NextToken(Derivs);
      e := Derivative.MNode.FullParameterName(Derivative.Index);
      if s <> e then
        CurrentFailure := 'Derivative Names: '+s+' <> '+e
      else begin
        s := NextToken(Derivs);
        VarAndTol(s, v, Tol, Mx);
        CheckNo(v, Value, Tol, e, False, True);
      end;
    end;
  var
    i: Integer;
  begin
    if Line.EigenvaluesU.Manifold = Line.EigenvaluesL.Manifold then
      for i := 0 to High(Line.EigenvaluesL.Derivatives) do
        CheckDerivative(Line.EigenvaluesU.Derivatives[i], Line.EigenvaluesU.Derivatives[i].Values[Line.IndexU]-Line.EigenvaluesL.Derivatives[i].Values[Line.IndexL])
    else begin
      for i := 0 to High(Line.EigenvaluesU.Derivatives) do
        CheckDerivative(Line.EigenvaluesU.Derivatives[i], Line.EigenvaluesU.Derivatives[i].Values[Line.IndexU]);
      for i := 0 to High(Line.EigenvaluesL.Derivatives) do
        CheckDerivative(Line.EigenvaluesL.Derivatives[i], -Line.EigenvaluesL.Derivatives[i].Values[Line.IndexL]);
    end;
  end;
var
  Problem: Exception;
  nCounted, nZero, nNotCounted, nIntensityFailures: Integer;
  TxDetails, TxOnly, ReadTx, CalcTx, AdjTx, Buffer, Derivs: string;
  StepIndex, OriginalL, OriginalU: Integer;
  i, P, Q, R: Integer;
  MaxIntensity: Double;
  HasA, HasEstUnc: Boolean;
  CalcList, ReadList: TSLineList;
  nReadList: Integer;
  LR: TSLine;
  FreqTol, FreqMax, EpTol, EpMax, ETol, EMax: Double;
begin
  { LineListMode is a mask to turn off various checks:
    1: TxDetails
    2: Line Count
    4: Compare Linelist
    8: Line Strength/Intensity/A/Width
   16: Ignore linelist differences involving very weak lines
   32: Skip linelist generation completely
  }

  nTried := 0;
  nZero := 0;
  MaxDifference := 0;
  Problem := nil;
  FreqDiffThreshold := 0;
  FreqMax := 0; { Not used }
  EpMax := 0;
  EMax := 0;
  StrMax := 0;
  linestrengthMax := 0;
  EinsteinAMax := 0;
  nIntensityFailures := 0;
  S.AllowDefaults := False;
  Simulation := Mixture.Simulations[0];
  CalcList := nil;
  ReadList := nil;
  nReadList := 0;
  Mixture.PrintLevel := plNone;
  if (LineListMode and 32) = 0 then begin
    Simulation.MakeLineList(Problem);
    if Assigned(Problem) then
      raise Problem;
    if not Generated and ((LineListMode and 4) = 0) then begin
      CalcList := MakeTSLineList(Simulation);
      SetLength(ReadList, Length(CalcList));
    end;
  end;
  i := -1;
  while not S.AtEndOfFile do begin
    Buffer := S.RestOfLine;
    i := Max(Pos('Frequency', Buffer),Pos('Position', Buffer));
    if i > 0 then
      Break;
    S.NextLine;
  end;
  { Check for A as column heading }
  HasA := (i > 0) and (Pos(' A ', Copy(Buffer,i,Length(Buffer))) > 0);
  HasEstUnc := (i > 0) and (Pos(' Est Unc ', Copy(Buffer,i,Length(Buffer))) > 0);
  S.NextLine;
  i := -1;
  Sum := 0;
  while not S.AtEndOfFile do begin
    if (S.Line mod 100) = 0 then
      SetProgress(S.Position);
    Failed := False;
    MoleculeName := S.NextString('');
    if CompareText(MoleculeName, 'stop') = 0 then
      Exit;
    if (MoleculeName = '') then begin
      S.NextLine;
      Continue;
    end;
    Molecule := Mixture.FindMolecule(MoleculeName) as TMolecule;
    if not Assigned(Molecule) and TryStrToInt(MoleculeName,i) then
      Break;
    S.NextStringVar(Manp); NextQuantumVar(Jp); S.NextStringVar(sympStr); S.NextIntegerVar(np);
    S.NextStringVar(Man);  NextQuantumVar(J);  S.NextStringVar(symStr);  S.NextIntegerVar(n);
    NextVarAndTol(S,f,FreqTol,FreqMax);
    FreqDiffThreshold := Max(FreqDiffThreshold,FreqTol);
    if HasEstUnc then
      NextVarAndTol(S,EstUnc,EstUncTol,EstUncMax);
    if Generated then begin
      S.NextDoubleVar(w); { Actually weight }
      NextVarAndTol(S,Str,StrTol,StrMax);
      if w <= 0 then begin
        S.NextLine;
        Continue; { Skip }
      end;
    end else begin
      NextVarAndTol(S,Str,StrTol,StrMax);
      NextVarAndTol(S,Ep,EpTol,EpMax);
      NextVarAndTol(S,E,ETol,EMax);
      NextVarAndTol(S,linestrength,linestrengthTol,linestrengthMax);
      if HasA then
        NextVarAndTol(S,EinsteinA,EinsteinATol,EinsteinAMax);
      S.NextDoubleVar(w);
      TxDetails := Trim(S.RestOfLine);
      if (Length(TxDetails) >0) and (TxDetails[1] = ':') then
        TxDetails := Trim(Copy(TxDetails,2,Length(TxDetails)));
      { May need to do more trimming if format changes in future }
    end;
    Inc(nTried);
    if not Assigned(Molecule) then Fail('Can''t find Molecule:'+MoleculeName);
    if not Failed then begin
      Line.Width := 0;
      ManifoldU := Molecule.ManifoldWithName(Manp);
      if not Assigned(ManifoldU) then
        Fail('Can''t find Upper Manifold:'+Manp)
      else begin
        symp := FindSym(ManifoldU, Jp, sympStr);
        if symp < 0 then
          Line.IndexU := FindEigenvaluesAndIndex(Line.EigenvaluesU, ManifoldU, Jp, symp, -1, TxDetails, False, S.Line, '', 0, Mixture.QuantumNumberFormat, Simulation.ExField)
        else begin
          Line.EigenvaluesU := ManifoldU.Eigenvalues(Jp, symp, [meHrot, meWidth], Simulation.ExField, True, plNone, nil, True, 0);
          Line.IndexU := np;
        end;
        if (Line.IndexU <= 0) or (Line.IndexU > Line.EigenvaluesU.Size) then
          Fail(Format('Upper eigenvalue (%d) > %d', [np, Line.EigenvaluesU.Size]))
        else begin
          if not Generated then
            CheckNo(Ep, Line.EigenvaluesU.Energy[Line.IndexU], EpTol, 'E''');
          if Assigned(Line.EigenValuesU.Widths) then
            Line.Width := Line.Width + Line.EigenValuesU.Widths[np];
        end;
      end;
      ManifoldL := Molecule.ManifoldWithName(Man);
      if not Assigned(ManifoldL) then
        Fail('Can''t find Lower Manifold:'+Man)
      else begin
        sym := FindSym(ManifoldL, J, symStr);
        if sym < 0 then
          Line.IndexL := FindEigenvaluesAndIndex(Line.EigenvaluesL, ManifoldL, J, sym, -1, TxDetails, True, S.Line, '', 0, Mixture.QuantumNumberFormat, Simulation.ExField)
        else begin
          Line.EigenvaluesL := ManifoldL.Eigenvalues(J, sym, [meHrot, meWidth], Simulation.ExField, True, plNone, nil, True, 0);
          Line.IndexL := n;
        end;
        if (Line.IndexL <= 0) or (Line.IndexL > Line.EigenvaluesL.Size) then
          Fail(Format('Lower eigenvalue (%d) > %d', [n, Line.EigenvaluesL.Size]))
        else begin
          if Assigned(Line.EigenValuesL.Widths) then
            Line.Width := Line.Width + Line.EigenValuesL.Widths[n];
          if not Generated then
            CheckNo(E, Line.EigenvaluesL.Energy[Line.IndexL], ETol, 'E"');
        end;
      end;
      if not failed then begin
        Line.Frequency := Line.EigenvaluesU.Energy[Line.IndexU]-Line.EigenvaluesL.Energy[Line.IndexL];
        if not (Molecule is TExternalMolecule) then
          CheckNo(f, Line.Frequency, FreqTol, 'Frequency');
        if HasEstUnc then
          CheckNo(EstUnc, EstErr(Line.EigenvaluesU, Line.EigenvaluesL, Line.IndexL, Line.IndexU), EstUncTol, 'Est Unc');
        TransitionMoments := Molecule.FindTransitionMoments(ManifoldU,ManifoldL);
        if (LineListMode and 8) <> 0 then
          { Skip intensity checks }
          Sum := Sum + Str
        else if not Assigned(TransitionMoments) then
          Fail('Can''t find transition moment')
        else if Molecule is TExternalMolecule then begin
          { Simple check of line strength }
          { First find strength }
          if TExternalMolecule(Molecule).FindLine(Line) < 0 then
            Fail('Line not found');
          CheckNo(f, Line.Frequency, Tolerance, 'Frequency');
          if not Generated then begin
            { TODO: Should be textual compare, so that precision matches }
            if not CheckNo(linestrength, Line.Strength, linestrengthTol, 'Line Strength') then begin
              if HasA then
                CheckNo(EinsteinA, Molecule.EinsteinA(Line, Simulation.PlotUnits), EinsteinATol, 'A');
              CheckNo(w, Line.Width, Tolerance, 'Line Width');
            end;
          end;
          Problem := nil;
          cstr := Simulation.LineIntensity(Line, Mixture.Units, Problem);
          if Assigned(Problem) then begin
            Fail(Problem.Message);
            Problem.Free;
          end;
          if not CheckNo(Str, cstr, StrTol, 'Intensity') then
            if (Line.CachedName <> TxDetails) and ((LineListMode and 1) = 0) then
              Fail('TxDetails mismatch|'+Trim(Line.CachedName)+'|'+TxDetails+'| ');
        end else begin
          if Line.Frequency = 0 then
            Inc(nZero); { Zero frequency lines are discarded in more recent versions }
          D := TransitionMoments.TransitionMatrix(Line.EigenValuesL, Line.EigenValuesU, True, True, nil, Simulation);
          { Accept eigenvalue number changes, if the other eigenvalue has the
            same, or very similar energy. }
          OriginalL := Line.IndexL;
          OriginalU := Line.IndexU;
          cstr := 0;
          { Cycle through all pairs of eigenvalues; the +1 ensures the original is used to give the error message }
          for StepIndex := 0 to Line.EigenvaluesL.Size * Line.EigenvaluesU.Size + 1 do begin
            Line.IndexL := (OriginalL + StepIndex - 1) mod Line.EigenvaluesL.Size + 1;
            if not SameValue(Line.EigenvaluesL.E[OriginalL],Line.EigenvaluesL.E[Line.IndexL]) then
              Continue;
            Line.IndexU := (OriginalU + StepIndex div Line.EigenvaluesL.Size - 1) mod Line.EigenvaluesU.Size + 1;
            if not SameValue(Line.EigenvaluesU.E[OriginalU],Line.EigenvaluesU.E[Line.IndexU]) then
              Continue;
//            if StepIndex <> 0 then
//              WriteLn(Log, StepIndex:3, OriginalL:3, Line.IndexL:3, OriginalU:3, Line.IndexU:3,
//                           ' ', Line.EigenvaluesL.E[OriginalL]-Line.EigenvaluesL.E[Line.IndexL],
//                           ' ', Line.EigenvaluesU.E[OriginalU]-Line.EigenvaluesU.E[Line.IndexU],
//                           CurrentFailure);
            CurrentFailure := '';
            CurrentDifference := 0;
            Line.Strength := D[Line.IndexL, Line.IndexU]*Line.EigenvaluesL.StatisticalWeight(Line.IndexL)
                             *Line.EigenvaluesL.Manifold.Molecule.Abundance
                             *Line.EigenvaluesL.Manifold.Molecule.Species.Concentration;
            if not Generated then begin
              if CheckNo(linestrength, Line.Strength, linestrengthTol, 'Line Strength', False, True) then
                Continue;
              if HasA then
                CheckNo(EinsteinA, Molecule.EinsteinA(Line, Simulation.PlotUnits), EinsteinATol, 'A', False, True);
            end;
            Problem := nil;
            cstr := Simulation.LineIntensity(Line, Mixture.Units, Problem);
            if Assigned(Problem) then begin
              Fail(Problem.Message);
              Problem.Free;
            end;
            if CheckNo(Str, cstr, StrTol, 'Intensity', False, True) then
              Continue;
            if not Generated then begin
              if CheckNo(w, Line.Width, Tolerance, 'Width', False, True) then
                Continue;
              Line.CachedName := '';
              if (TxDetails <> '') and (Trim(MakeTxName(Line)) <> TxDetails) then begin
                { Format Adjust }
                P := Pos(':',Line.CachedName);
                Q := Pos(':',TxDetails);
                if (P > 0) and (Q > 0) then begin
                  { Compare branch details }
                  ReadTx := Trim(Copy(TxDetails,1,Q-1));
                  CalcTx := Trim(Copy(Line.CachedName,1,P-1));
                  if ReadTx <> CalcTx then begin
                    { Recheck, removing spaces }
                    AdjTx := ReadTx;
                    R := 2;
                    while R < Length(AdjTx) do begin
                      if AdjTx[R] = ' ' then
                        if (AdjTx[R-1] in ['0'..'9']) and (AdjTx[R+1] in ['0'..'9']) then
                          AdjTx[R] := ','
                        else begin
                          Delete(AdjTx,R,1);
                          Dec(R);
                        end;
                      Inc(R);
                    end;
                    if AdjTx <> CalcTx then
                      CurrentFailure := 'Branch Details mismatch|'+AdjTx+'|'+CalcTx+'| ';
                  end;
                end;
                if P > 0 then
                  Line.CachedName := Trim(Copy(Line.CachedName, P+1, Length(Line.CachedName)));
                Derivs := '';
                if Q > 0 then begin
                  TxOnly := Trim(Copy(TxDetails, Q+1, Length(TxDetails)));
                  Q := Pos(':',TxOnly);
                  if Q > 0 then begin
                    Derivs := Trim(Copy(TxOnly,Q+1,Length(TxOnly)));
                    TxOnly := Trim(Copy(TxOnly,1,Q-1));
                  end;
                end else
                  TxOnly := TxDetails;
                P := Pos('- ',Line.CachedName);
                if (P > 0) and (Pos('- ',TxOnly) <= 0) then
                  Delete(Line.CachedName, P, 2);
                if (Line.CachedName <> TxOnly) and ((LineListMode and 1) = 0) then begin
                  CurrentFailure := 'TxDetails mismatch|'+Trim(Line.CachedName)+'|'+TxOnly+'| ';
                end;
                if (CurrentFailure = '') and (Derivs <> '') then
                  CheckDerivatives(Derivs);
              end;
            end;
            if CurrentFailure = '' then
              Break; { Success }
          end;
          Sum := Sum + cstr;
          if CurrentDifference > MaxDifference then
            MaxDifference := CurrentDifference;
          if CurrentFailure <> '' then begin
            Inc(nIntensityFailures);
            Fail(CurrentFailure);
          end;
          // WriteLn(Log, Line.Strength:8:4, LineIntensity(Line, Mixture.Temperature, Mixture.Tvib)/5:8:4, str:8:4);
        end;
      end;
    end;
    if Length(CalcList) > 0 then begin
      if Length(ReadList) <= nReadList then
        SetLength(ReadList, nReadList+10000);
      LR.Frequency := f;
      LR.JU := Jp;
      LR.JL := J;
      LR.SymU := symp;
      LR.SymL := sym;
      LR.Strength := linestrength;
      LR.Intensity := Str;
      LR.STol := linestrengthTol;
      LR.ITol := StrTol;
      ReadList[nReadList] := LR;
      Inc(nReadList);
    end;
    S.NextLine;
  end;
  if (i <> -1) and ((LineListMode and 32) = 0) then begin
    { Calculate expected values }
    with Simulation.Mixture.LineList do begin
      ITotal := 0;
      nCounted := 0;
      MaxIntensity := 0;
      for j := 0 to nLines-1 do begin
        Problem := nil;
        Str := Simulation.LineIntensity(Line[j], Simulation.Mixture.Units, Problem);
        if Assigned(Problem) then
          raise Problem;
        ITotal := ITotal + Str;
        if Str >= Simulation.OThreshold then
          Inc(nCounted);
      end;
      nNotCounted := nLines - nCounted;
    end;
    CheckNo(i, nTried, Tolerance, 'Line Count');
    if ((LineListMode and 2) = 0) or (Abs(i-nZero-nCounted) > 20) then
      CheckNo(i-nZero, nCounted, Tolerance, 'Simulated Line Count');
    while not S.AtEndOfFile and (S.NextString('') <> '=') do
      if S.Token = '' then Fail('Format Error');
    Str := S.NextDouble(Sum);
    if nIntensityFailures = 0 then
      CheckNo(Str, Sum, 1e-4, 'Line intensity sum(a)'); { Note that this will happen if there are any intensity errors }
    CheckNo(Str, Simulation.Mixture.LineList.ITotal, 1e-4, 'Line intensity sum(b)');
  end;
  if (Length(CalcList) > 0) and (Failures < 50) then begin
    SetLength(ReadList,nReadList);
    Inc(Failures, CompareLists(CalcList, ReadList, linestrengthMax, (LineListMode and 16) <> 0));
  end;
  Result := MaxDifference;
end;

end.

