unit CommandLine;
{$I DFS.inc}
{
PGOPHER - a program for simulating rotational structure
Copyright (C) 2005 C M Western
School of Chemistry, University of Bristol, Bristol, UK
For further information please see http://pgopher.chm.bris.ac.uk/
or contact help-pgopher@bristol.ac.uk

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}
interface

uses
  Classes, SysUtils;

function HandleCommandLine: Boolean;
procedure DisconnectLogs;

implementation

{.$DEFINE SKELETON}

uses Profile,
     {$IFNDEF CONSOLE}
       {$IFDEF MSWINDOWS} Windows, {$ENDIF}
       {$IFDEF LCL_} LCLType, {$ENDIF}
       Main, LogUnit, Forms, OvrlyFrm, SyncObjs,
     {$ENDIF}
     {$IFDEF VCL_}
       {$IFNDEF CONSOLE }
         StackDump,
       {$ENDIF}
     {$ENDIF}
     {$IFDEF SKELETON}
     Skeleton,
     {$ENDIF}
     Global, Mixture, D3hImport, PgopherImport, CheckPgopher, CheckOverlay,
     TextStrm, Overlay, MNode, JPLCat, MatrixElements, Obs, Fitter, Expr,
     TextAssign, Worker, Sysinfo, PGWork, Math, CheckSpecView, CheckJB95,
     CheckImport, Hitran, DataSource, UCompat, Plotter, ExportG, ImportLmatrix,
     GraphM, LevelPlotter, ExpCan, Progress, Units, NextTok, Clock, MathSup;

{$IFNDEF CONSOLE}
function WaitGui(var Event: TEvent; Timeout: Cardinal): TWaitResult;
var
  Handle: {$IFDEF FPC}TEventHandle{$ELSE}THandle{$ENDIF};
begin
  if GetCurrentThreadID <> MainThreadID then
    Result := SimpleWait(Event, Timeout)
  else begin
    Result := wrAbandoned;
    {$IFDEF MSWINDOWS}
    repeat
      if Aborted or Application.Terminated then
        Abort;
      if not Assigned(Event) then
        Exit; { Event no longer set up }
      Handle := Event.Handle;
      case MsgWaitForMultipleObjects(1, Handle, False, Timeout, QS_ALLINPUT) of
        WAIT_OBJECT_0: Result := wrSignaled;
        WAIT_ABANDONED_0: Result := wrAbandoned;
        WAIT_OBJECT_0+1: begin; Application.ProcessMessages; Continue; end;
        WAIT_FAILED: RaiseLastOSError;
        WAIT_TIMEOUT: Result := wrTimeout;
      end;
    until True;
    {$ELSE}
    repeat
      Application.ProcessMessages;
      if Aborted or Application.Terminated then
        Abort;
      if not Assigned(Event) then
        Exit; { Event no longer set up }
      Result := Event.WaitFor(Timeout); { Ideally should wait on message available as well }
      case Result of
        wrError: RaiseLastOSError;
      end;
    until True;
    {$ENDIF}
  end;
end;
{$ENDIF}

var
  LogConnected: Boolean = False;

procedure ConnectLog(OutFile: string);
{$IFNDEF CONSOLE}
var
  FileName: string;
{$ENDIF}
begin
  {$IFNDEF CONSOLE}
  if not Assigned(LogForm) then
    Application.CreateForm(TLogForm, LogForm);
  if not Assigned(OverlayForm) then
    Application.CreateForm(TOverlayForm, OverlayForm);
  with Application do begin
    FileName := HelpFile;
    if not FileExists(FileName) then begin
      FileName := ExtractFilePath(Application.ExeName) + FileName;
      if not FileExists(FileName) then
        FileName := ExtractFilePath(Application.ExeName) + 'Help' + PathDelim + ExtractFileName(FileName);
    end;
    HelpFile := ExpandFileName(FileName);
  end;
  if OutFile = '' then begin
    { Output connection is just to catch any Write()s to Output,
      though they should be all to Log }
    //AssignStrings(Output, LogForm.Log.Lines);
    //Rewrite(Output);
    AssignStrings(Log, LogForm.Log.Lines);
    Rewrite(Log);
  end else
  {$ENDIF}
  begin
    if OutFile <> '' then
      AssignFile(Output, OutFile);
    Rewrite(Output);
    AssignTextFile(Log,Output);
    Rewrite(Log);
  end;
  LogConnected := True;
end;

procedure DisconnectLogs;
begin
  { Clean shutdown }
  StartWorkers(0,nil);
  CloseFile(Log);
  {$IFDEF LINUX}
  AssignFile(Log,'');
  Append(Log);
  {$ENDIF}
end;

type
  TNewPgoOpt = (npgNone, npgOver, npgNew);

{$IFDEF CONSOLE}
function DoFit(const ObsFile: string; ObsType:TObsType; Check: string; NewPgoOpt: TNewPgoOpt; TakeWeight: TTakeWeight = tkObs):string;
var
  Prec, i: Integer;
  function CheckResult(Expected, Actual: string):Boolean;
  var
    E, A: string;
    EV, AV: Double;
    First: Boolean;
  begin
    { Compare tokens in the string, allowing for small differences in the numbers b}
    Result := True;
    First := True;
    repeat
      E := NextToken(Expected);
      A := NextToken(Actual);
      if E <> A then begin
        Result := False;
        if E[1] in ['0'..'9','+','-','.'] then begin
          if E[Length(E)] = ')' then
            Delete(E, Length(E), 1);
          if A[Length(A)] = ')' then
            Delete(A, Length(A), 1);
          EV := StrToFloat(E);
          AV := StrToFloat(A);
          Result := SameValue(EV,AV,Min(Abs(AV),Abs(EV))*IfThen(First,2e-8,2e-5));
          First := False;
        end;
        if not Result then
          Exit;
      end;
    until (Expected = '') and (Actual = '');
  end;
var
  ObsStream: TStream;
  OtherFileName: string;
begin
  GlobalMixture.CheckProperties;
  try
    if Check <> '' then begin
      AssignFile(Log,ObsFile+'.out.new');
      Rewrite(Log);
    end;
    if ObsType = obContour then begin
      if not Assigned(GlobalOverlays) then
        raise EReportException.Create('Overlays required');
    end else begin
      ObsStream := TFileStream.Create(ObsFile, fmOpenRead or fmShareDenyWrite);
      try
        LoadObservations(GlobalMixture, ObsStream, ObsFile, ObsType, TakeWeight);
      finally
        ObsStream.Free;
      end;
    end;
    for i := 1 to GlobalMixture.FitCycles do begin
      if ObsType = obContour then begin
        GlobalMixture.PrepareForFit(ftContour,GlobalOverlays);
        Result := FitContour(GlobalMixture, GlobalOverlays, nil, GlobalMixture.FittingParameters, False);
      end else begin
        if ObsType = obIntensity then
          Prec := -1
        else
          Prec := GlobalMixture.Precision;
        GlobalMixture.PrepareForFit(ftLine,nil);
        Result := FitLines(GlobalMixture.FittingParameters, Observations, Prec, GlobalMixture);
      end;
    end;
    if (NewPgoOpt <> npgNone) and (Length(GlobalMixture.FittingParameters) > 0) then begin
      { Write out new parameters }
      { Find unused "other" file name }
      i := 0;
      repeat
        Inc(i);
        OtherFileName := ChangeFileExt(GlobalMixture.Name, '.'+IntToStr(i)+'.pgo');
      until not FileExists(OtherFileName);
      { TODO: Consider saving overlay parameters also }
      if NewPgoOpt = npgOver then
        { Overwrite input pgo file after backing it up }
        if not RenameFile(GlobalMixture.Name, OtherFileName) then
          raise Exception.Create('Can''t rename '+GlobalMixture.Name+' to '+OtherFileName)
        else begin
          WriteLn('Writing new parameters to ', GlobalMixture.Name, ' and saving old parameters to ', OtherFileName);
          GlobalMixture.SaveToFile(GlobalMixture.Name);
        end
      else begin
        { Write to new file }
        WriteLn('Writing new parameters to ', OtherFileName);
        GlobalMixture.SaveToFile(OtherFileName);
      end;
    end;
  except
    on E: Exception do
      Result := E.Message;
  end;
  if Check <> '' then begin
    CloseFile(Log);
    AssignFile(Log,''); { Reconnect to standard output }
    {$IFNDEF FPC}
    Rewrite(Log); { Delphi Bug - Append tries to open STDIN, not STDOUT }
    {$ELSE}
    Append(Log);
    {$ENDIF}
    if CheckResult(Check,Result) then
      Write(Log, 'OK (', Result, ')')
    else
      Write(Log, '**** failure: Fit mismatch :', Result, '<>', Check);
    WriteLn(Log, ' ', GlobalMixture.Name, ' ', ObsFile);
  end else
    WriteLn(Log, Result);
end;
{$ENDIF}

procedure MakePlot(FileName: string; IsEPlot: Boolean);
var
  G: TGraphM;
  Stream: TFileStream;
  CClass: TStreamCanvasClass;
  Canvas: TExportableCanvas;
begin
  Stream := TFileStream.Create(FileName,fmCreate);
  CClass := GetStreamCanvasClassFromFilename(FileName);
  if CClass = TTextCanvas then
    CClass := TNullCanvas;
  Canvas := CClass.Create(Stream,250,179);
  if IsEPlot then
    try
      G := TGraphM.Create(nil);
      if CClass = TNullCanvas then
        G.TextStream := Stream;
      G.CurrentCanvas := Canvas;
      ReadLevelPlotter(GlobalMixture).PlotItem(G,GlobalMixture,True)
    finally
      G.CurrentCanvas := nil;
      G.TextStream := nil;
      G.Free;
    end
  else
    Plotter.MakePlot(Canvas);
  Stream.Free;
end;

function HandleCommandLine: Boolean;
var
  Stream: TTextStream;
  MStream: TMemoryStream;
  i, d, nTried, np, LineListMode, nParams: Integer;
  Params, MixtureChanges: array of string;
  Tolerance, LargestError, Offset: Double;
  nFailures: Integer;
  SortBoth, Interpolate, CvtToPert, SkipELowerCheck, DoubleABlevels: Boolean;
  NewPgoOpt: TNewPgoOpt;
  InputPgopherVersion, PgopherVersion: Integer;
  Check, OutFile: string;
  procedure ApplyMixtureChanges;
  var
    i: Integer;
  begin
    if GlobalMixture <> nil then
      for i := 0 to High(MixtureChanges) do
        GlobalMixture.HandleAssignment(MixtureChanges[i]);
  end;
  procedure FitHelper(ObsType: TObsType);
  begin
    {$IFNDEF CONSOLE}
      LogForm.FitType.ItemIndex := Ord(ObsType);
      LogForm.FitSpeedClick(LogForm.FitSpeed);
    {$ELSE}
      DoFit(Params[2],ObsType,Check,NewPgoOpt);
    {$ENDIF}
  end;
var
  Problem: Exception;
  Parsums: TParsums;
  LoadedMixture: TMixture;
  CDir: string;
  StartTime: Int64;
begin
  SetLength(MixtureChanges,0);
  Stream := nil;
  MStream := nil;
  Parsums := nil;
  {$IFDEF CONSOLE}
    Global.VersionInfo := {$I commandlinever.inc} ;
    {$IFDEF FPC}
      Global.VersionInfo := Global.VersionInfo + ' ' + {$INCLUDE %DATE%} + ' (fpc ' + {$INCLUDE %FPCVERSION%} + ')';
    {$ENDIF}
    ShowTimingAtEnd := False;
  {$ELSE}
    DoWait := @WaitGui;
    with MainForm.Project do
      Global.VersionInfo := 'Pgopher ' + Version + '.' + IntToStr(Build) + ' ' + BuildDate;
  {$ENDIF}
  PgopherVersion := DecodePgopherVersion(Global.VersionInfo);
  AddExpectedException(EReportException);
  AddExpectedException(EExprFormatError);
  AddExpectedException(EFitError);
  UCompat.VersionInfo := Global.VersionInfo;
  {$IFDEF VCL_}
  DecimalSeparator := '.'; { Ignore system setting - there are internal strings that are not localised
                             and data files should not be localised. Text input could be localised, though
                             the langauge of the program is English }
  {$ENDIF}
  np := 1; { Default is 1 as issues like Heap management slow things down for
             smaller calculations }
  // SetTrackMemory(True);
//  np := nCPUs; 2; { TODO: Testing only }
  try
    Result := False;
    Tolerance := 1e-4;
    nFailures := 0;
    SortBoth := False;
    Interpolate := False;
    CvtToPert := False;
    SkipELowerCheck := False;
    DoubleABlevels := False;
    Check := '';
    OutFile := '';
    Offset := 0;
    NewPgoOpt := npgNone;
    LineListMode := 0;
    InputPgopherVersion := PgopherVersion;
    nParams := ParamCount;
    {$IFDEF DARWIN}{$IFNDEF CONSOLE}
    nParams := 0; { Parameters to gui files come via drop files, not command line }
    {$ENDIF}{$ENDIF}
    SetLength(Params, nParams);
    d := 0;
    i := 1;
    while i <= nParams do begin
      Params[d] := ParamStr(i);
      if Copy(Paramstr(i), 1, 8) = 'Mixture.' then begin
        SetLength(MixtureChanges, Length(MixtureChanges)+1);
        MixtureChanges[High(MixtureChanges)] := Copy(Paramstr(i), 9, Length(Paramstr(i)));
      end else if ParamStr(i) = '--tolerance' then begin
        Tolerance := StrToFloatEx(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--fail' then begin
        nFailures := -StrToInt(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--linelistmode' then begin
        LineListMode := StrToInt(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--sort' then
        SortBoth := True
      else if ParamStr(i) = '--offset' then begin
        Offset := StrToFloatEx(ParamStr(i+1));
        Inc(i);
      end else if (ParamStr(i) = '-o') or (ParamStr(i) = '--o') then begin
        OutFile := ParamStr(i+1);
        Inc(i);
      end else if ParamStr(i) = '--interpolate' then
        Interpolate := True
      else if ParamStr(i) = '--check' then begin
        Check := ParamStr(i+1);
        Inc(i);
      end else if ParamStr(i) = '--np' then begin
        np := StrToInt(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--pgopherversion' then begin
        InputPgopherVersion := DecodePgopherVersion(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--cvttopert' then
        CvtToPert := True
      else if ParamStr(i) = '--skipelowercheck' then
        SkipELowerCheck := True
      else if ParamStr(i) = '--doubleablevels' then
        DoubleABlevels := True
      else if ParamStr(i) = '--exprcheck' then
        ExprChecking := True
      else if ParamStr(i) = '--overwrite' then
        NewPgoOpt := npgOver
      else if ParamStr(i) = '--newpgo' then
        NewPgoOpt := npgNew
      else if ParamStr(i) = '--trackmemory' then begin
        SetTrackMemory(True);
        {$IFDEF CONSOLE}
          ShowTimingAtEnd := True;
        {$ENDIF}
      end else if ParamStr(i) = '--times' then
        {$IFDEF CONSOLE}
          ShowTimingAtEnd := True
        {$ENDIF}
      else
        Inc(d);
      Inc(i);
    end;
    if ExprChecking then
      np := 1; { TODO: Remove when MustBeSerial removed from CheckMatrixElementsAvailable }
    SetLength(Params, d);
    ConnectLog(OutFile);
    if (d = 3) or ((d > 1) and (Params[0] = '--plot')) then begin
      { Automated testing or specific command line actions }
      StartWorkers(np,TPgopherWorker);
      if Params[0] <> '--import' then begin
        {$IFNDEF CONSOLE}
          CDir := GetCurrentDir;
          LogForm.Show;
          MainForm.EmptyMixtureItemClick(nil);
          LoadGeneralFile(Params[1],False,True,cm1);
          DoDirSeparators(Params[2]);
          LogForm.ObsFileName.Text := Params[2];
          SetCurrentDir(CDir); { As Mixture load can change current directory }
        {$ELSE}
          if LowerCase(ExtractFileExt(Params[1])) = '.pgo' then begin
            GlobalMixture := TMixture.Create(nil);
            GlobalMixture.LoadFromFile(Params[1]);
          end else
            AddToOverlay(GlobalOverlays, Params[1], True, cm1, False, Arbitrary);
        {$ENDIF}
      end;
      ApplyMixtureChanges;
      if GlobalMixture <> nil then
        GlobalMixture.CheckProperties;
      if CvtToPert then
        GlobalMixture.ConvertToPerturbations;
      try
        if Params[0] = '--fit' then
          FitHelper(obLine)
        else if Params[0] = '--intensityfit' then
          FitHelper(obIntensity)
        else if Params[0] = '--fitlower' then
          FitHelper(obELower)
        else if Params[0] = '--fitupper' then
          FitHelper(obEUpper)
        else if Params[0] = '--contourfit' then begin
          AddOverlay(Params[2], True, cm1);
          FitHelper(obContour);
        end else if Params[0] = '--fitcdlower' then
          FitHelper(obCDLower)
        else if Params[0] = '--fitcdupper' then
          FitHelper(obCDUpper)
        else if Params[0] = '--plot' then begin
          for i := 2 to d-2 do
            AddOverlay(Params[i], True, cm1);
          MakePlot(Params[d-1],False);
        end else if Params[0] = '--eplot' then
          MakePlot(Params[2],True)
        else if Params[0] = '--import' then begin
          MStream := TMemoryStream.Create;
          MStream.LoadFromFile(Params[2]);
          Stream := TFreeFormatStream.Create(MStream);
          if LowerCase(Params[1]) = 'lmatrix' then
            LoadedMixture := ImportLmatrixFile(Stream as TFreeFormatStream)
          else
            raise EReportException.Create('Unknown import:'+Params[1]);
          {$IFDEF CONSOLE}
          GlobalMixture :=
          {$ELSE}
          MainForm.Mixture :=
          {$ENDIF}
            LoadedMixture;
        end else if Params[0] = '--HITRAN' then begin
          MStream := TMemoryStream.Create;
          MStream.LoadFromFile(Params[2]);
          Stream := TFreeFormatStream.Create(MStream);
          Parsums := LoadHITRANParsumFile(Stream as TFreeFormatStream);
          FreeAndNil(Stream);
          MStream.LoadFromFile(Params[1]);
          MStream.Position := 0;
          Stream := TFreeFormatStream.Create(MStream);
          Stream.Comments := True;
          {$IFDEF CONSOLE}
          GlobalMixture :=
          {$ELSE}
          MainForm.Mixture :=
          {$ENDIF}
            ImportHITRANFile(Stream as TFreeFormatStream,Parsums);
        end
        {$IFDEF SKELETON}
        else if Params[0] = '--misc' then
          SkeletonFunction(Params[2])
        {$ENDIF}
        else begin
          MStream := TMemoryStream.Create;
          MStream.LoadFromFile(Params[2]);
          Stream := TTextStream.Create(MStream);
          Stream.Comments := True;
          nTried := 0;
          StartTime := Clocks();
          PushProgress('Checking',MStream.Size);
          try
            if Params[0] = '--testd3h' then
              LargestError := CheckD3hFile(GlobalMixture, Stream, nTried, Tolerance, nFailures)
            else if Params[0] = '--testold' then
              LargestError := CheckOldPgopherFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, False, False)
            else if Params[0] = '--testoldnoom' then
              LargestError := CheckOldPgopherFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, True, False)
            else if Params[0] = '--testoldnof1' then
              LargestError := CheckOldPgopherFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, False, True)
            else if Params[0] = '--testpgo' then
              LargestError := CheckPgopherFile(False, GlobalMixture, Stream, nTried, Tolerance, nFailures, InputPgopherVersion, False, LineListMode)
            else if Params[0] = '--testpseudo' then
              LargestError := CheckPgopherFile(False, GlobalMixture, Stream, nTried, Tolerance, nFailures, InputPgopherVersion, True, LineListMode)
            else if Params[0] = '--testgen' then
              LargestError := CheckPgopherFile(True, GlobalMixture, Stream, nTried, Tolerance, nFailures, InputPgopherVersion, False, LineListMode)
            else if Params[0] = '--testeig' then
              LargestError := CheckEigenvalueFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, InputPgopherVersion)
            else if Params[0] = '--testovr' then
              LargestError := CheckOverlays(GlobalOverlays, Stream, nTried, Tolerance, nFailures, SortBoth, Interpolate)
            else if Params[0] = '--testjplcat' then begin
              if ExtractFilePath(Params[2]) <> '' then
                SetCurrentDir(ExtractFilePath(Params[2]));
              LargestError := CheckJPLCatFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, Offset, SkipELowerCheck, DoubleABlevels)
            end else if Params[0] = '--testspecview' then
              LargestError := CheckSpecViewFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, False)
            else if Params[0] = '--testspecviewghz' then
              LargestError := CheckSpecViewFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, True)
            else if Params[0] = '--testjb95' then
              LargestError := CheckJB95ViewFile(GlobalMixture, Stream, nTried, Tolerance, nFailures)
            else if Params[0] = '--testlimport' then
              LargestError := CheckLImport(GlobalMixture, Stream, nTried, Tolerance, nFailures)
            else
              raise Exception.Create('Bad test:'+Params[0]);
          finally
            PopProgress;
          end;
          if nFailures <> 0 then begin
            Write(Log, nFailures, ' failure(s) out of ', nTried);
            ExitCode := 1;
          end else
            Write(Log, 'OK (', nTried:5, ')');
          Write(Log, ' Largest Error = ', LargestError:20, (Clocks()-StartTime)/ClockFrequency:10:3, ' s  :');
          for i := 1 to ParamCount do
            Write(Log, ' ', ParamStr(i));
          WriteLn(Log);
        end;
      finally
        FreeAndNil(Stream);
        FreeAndNil(MStream);
      end;
      {$IFDEF CONSOLE}
      FreeAndNil(GlobalMixture);
      FreeAndNil(GlobalOverlays);
      DisconnectLogs;
      {$ENDIF}
    end else begin
      { Simple Calculation }
      StartWorkers(np,TPgopherWorker);
      {$IFNDEF CONSOLE}
        if GlobalMixture = nil then
          MainForm.EmptyMixtureItemClick(nil);
          for i := 0 to High(Params) do
            try
              LoadGeneralFile(Params[i],i>0,False,MainForm.Simulation.PlotUnits);
            except
              on E:Exception do begin
                Application.MessageBox(PChar('Error on loading '+Params[i]+':'+sLineBreak+E.Message), 'Error', MB_OK+MB_ICONERROR);
                Break;
              end;
            end;
        MainForm.ConsiderReplot(GlobalMixture.nSpecies = 0);
      {$ELSE}
        if d = 0 then begin
          WriteLn(Log, Global.VersionInfo);
          WriteLn(Log, DetermineCpu);
        end else if d = 1 then begin
          GlobalMixture := TMixture.Create(nil);
          GlobalMixture.LoadFromFile(Params[0]);
        end else
          raise EReportException.Create('Arguments not recognized');
        if GlobalMixture <> nil then begin
          ApplyMixtureChanges;
          if not LineListWanted(GlobalMixture.PrintLevel) then
            GlobalMixture.PrintLevel := plLineList;
          if GlobalMixture.PrintLevel >= plDetail then
            GlobalMixture.PrettyPrint(Log, GlobalMixture.PrintLevel);
          Problem := nil;
          GlobalMixture.Simulations[0].MakeLineList(Problem);
          if Assigned(Problem) then begin
            WriteLn(Log, Problem.Message);
            Problem.Free;
          end;
          Flush(Log);
        end;
        FreeAndNil(GlobalMixture);
      {$ENDIF}
    end;
  except
    on E:Exception do begin
      if not LogConnected then
        ConnectLog('');
      WriteLn(Log, '**** failure Exception:', E.message, ' ', CmdLine);
    end;
  end;
end;

initialization
finalization
  Global.VersionInfo := '';
end.
