unit D3hImport;
{ This unit provides backward compatibility with the d3h.exe d3h32,exe program.
  It allows opending of the data files, and comparing the output of the two programs.
}

interface

uses Classes, SysUtils, Mixture, TextStrm;

function ImportD3hFile(Z: TTextStream): TMixture;
function CheckD3hFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;

implementation

uses MathSup, MatrixVec, Life, SymT, Units, IntVec, Worker, MNode, StaticField;

function CheckD3hFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;
var
  Jp, Kp, gp, symp, np: Integer; { Upper State }
  J,  K,  g,  sym,  n : Integer; { Lower State }
  f, str, w, Ep, E: Double;
  ManifoldL, ManifoldU: TSymmetricManifold;
  Molecule: TSymmetricMolecule;
  TransitionMoments: TTransitionMoments;
  i: Integer;
  D: TMatrix;
  Line: TLine;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
      WriteLn(Log, Jp:3, Kp:3, gp:2, symp:2, np:2, J:5, K:3, g:2, sym:2, n:2, f:12:4, str:8:4, w:8:3, Ep:12:4, E:12:4);
    end;
  end;
  function evNo(const EigenValues: TEigenvalues; const K, g, n: Integer): Integer;
  var
    i: Integer;
    BS: TSymmetricTopBasisState;
  begin
    i := 1; { Writen as a for loop -> FPC register allocation error }
    while i <= EigenValues.Size do begin
      BS := EigenValues.DominantBasis[i] as TSymmetricTopBasisState;
      if (Abs(BS.K) = Abs(2*K))
         and not ((BS.l <> 0) and (BS.K*BS.l <> 4*K)) then begin
          Result := i;
          Exit;
      end;
      Inc(i);
    end;
    Result := 0;
    Fail(Format('Not found %d %d %d', [K, g, n]));
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string);
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else if Abs(n1-n2) > tol then
      Fail(Format('%s out of tolerance: %g %g %g', [Name, n1, n2, n1-n2]));
    if Abs(n1-n2) > Result then
      Result := Abs(n1-n2);
  end;
var
  Problem: Exception;
  Simulation: TSimulation;
  wCalc: Double;
begin
  nTried := 0;
  Result := 0;
  Molecule := Mixture.Species[0].Isotopomer[0] as TSymmetricMolecule;
  i := 0;
  if not Molecule.Manifold[0].Initial then Inc(i);
  ManifoldL := Molecule.Manifold[i] as TSymmetricManifold;
  ManifoldU := Molecule.Manifold[1-i] as TSymmetricManifold;
  if Molecule.nTransitionMoments <> 1 then Fail(Format('%d transition moments', [Molecule.nTransitionMoments]));
  TransitionMoments := Molecule.TransitionMoments[0,1];
  Simulation := Mixture.Simulations[0];
  while not S.AtEndOfFile and (Pos('frequency', S.RestOfLine) <= 0) do
    S.NextLine;
  S.NextLine;
  while not S.AtEndOfFile do begin
    Jp := S.NextInteger(-99);
    if Jp < 0 then
      Break;
    S.NextIntegerVar(Kp); S.NextIntegerVar(gp); S.NextIntegerVar(symp); S.NextIntegerVar(np);
    S.NextIntegerVar(J);  S.NextIntegerVar(K);  S.NextIntegerVar(g);  S.NextIntegerVar(sym);  S.NextIntegerVar(n);
    S.NextDoubleVar(f);   S.NextDoubleVar(str); S.NextDoubleVar(w);   S.NextDoubleVar(Ep);    S.NextDoubleVar(E);
    Inc(nTried);
    Line.EigenvaluesL := ManifoldL.Eigenvalues(J*2, sym-1, [meHrot, meWidth], ZeroField);
    Line.EigenvaluesU := ManifoldU.Eigenvalues(Jp*2, symp-1, [meHrot, meWidth], ZeroField);
    Line.IndexL := evNo(Line.EigenvaluesL, K, g, n);
    if Line.IndexL > 0 then begin
      CheckNo(E, Line.EigenvaluesL.E[Line.IndexL], Tolerance, 'E"');
      Line.IndexU := evNo(Line.EigenvaluesU, Kp, gp, np);
      if Line.IndexU > 0 then begin
        CheckNo(Ep, Line.EigenvaluesU.E[Line.IndexU], Tolerance, 'E''');
        D := TransitionMoments.TransitionMatrix(Line.EigenValuesL, Line.EigenValuesU, True, True, nil, Simulation);
        Line.Strength := D[Line.IndexL, Line.IndexU]*Line.EigenvaluesL.StatisticalWeight(Line.IndexL);
        Problem := nil;
        CheckNo(str, Simulation.LineIntensity(Line, Mixture.Units, Problem), 3.1*Tolerance, 'Intensity');
        if Assigned(Problem) then begin
          Fail(Problem.Message);
          Problem.Free;
        end;
        wCalc := 0;
        if Assigned(Line.EigenvaluesL.Widths) then
          wCalc := Line.EigenvaluesL.Widths[Line.IndexL];
        if Assigned(Line.EigenvaluesU.Widths) then
          wCalc := wCalc + Line.EigenvaluesU.Widths[Line.IndexU];
        CheckNo(w, wCalc, Tolerance, 'Width');
      end;
    end;
    S.NextLine;
  end;
end;

{$IFDEF VER170}
{$OPTIMIZATION OFF}
{$ENDIF}

function ImportD3hFile(Z: TTextStream): TMixture;
var
  nManifolds, i: Integer;
  t: string;
  Species: TSpecies;
  Molecule: TSymmetricMolecule;
  ManifoldL, ManifoldU: TSymmetricManifold;
  procedure Record1;
  begin
    ManifoldL := TSymmetricManifold.Create(Molecule);
    ManifoldL.Name := 'Lower';
    ManifoldL.Initial := True;
    if nManifolds = 2 then begin
      ManifoldU := TSymmetricManifold.Create(Molecule);
      ManifoldU.Name := 'Upper';
    end else if nManifolds = 1 then
      ManifoldU := ManifoldL
    else
      raise Exception.CreateFmt('nstates: %d', [nManifolds]);
    Species.SetSetting(2, Z.NextInteger(0)*2);
    Z.NextInteger(0);
    Result.Simulations[0].Temperature := Z.NextDouble(300);
    Z.NextDouble(0); { STRMIN }
    Result.Simulations[0].Fmin := Z.NextDouble(0);
    Result.Simulations[0].Fmax := Z.NextDouble(0);
    Result.Simulations[0].Foffset := Z.NextDouble(0);
    Result.Simulations[0].nDF := Z.NextInteger(Result.Simulations[0].nDF);
    Result.PrintLevel := TPrintLevel(Ord(plNone) + Z.NextInteger(-3));
    Z.NextInteger(0); { Itype }
    Species.SetSetting(1, Z.NextInteger(Species.Jmin div 2)*2);
    Z.NextInteger(0); { gmin }
    Z.NextInteger(0); { gmax }
    i := Z.NextInteger(6);
    case i of
      3: Molecule.PointGroup := C3v;
      6: Molecule.PointGroup := D3h;
    else
      raise Exception.CreateFmt('Bad point group: %d', [i]);
    end;
    Z.NextInteger(0); { dres }
    Result.Simulations[0].Lorentzian := Z.NextDouble(0);
    Result.Simulations[0].Gaussian := Z.NextDouble(0);
    if Z.NextInteger(0) <> 0 then
      Result.Units := cm1;
    Result.Simulations[0].LifeModel := TLifeModel(Z.NextInteger(0));
    Result.Simulations[0].RefWidth := Z.NextDouble(0);
    Z.NextDouble(0); { overlay_baseline }
    Z.NextInteger(0); { overlay_baseline_fix }
    Z.NextDouble(0); { overlay_scale }
    Z.NextInteger(0); { overlay_scale_fix }
    Z.NextInteger(0); { add_spectra }
    Z.NextInteger(0); { nerrdigits }
    Z.NextInteger(0); { multiple }
    Z.NextInteger(0); { smargin }
    Z.NextRecord;
  end;
  procedure ReadPert(Manifold: TSymmetricManifold; Old: Boolean);
  var
    i, nStates, State1, State2: Integer;
    State: TSymmetricTop;
    Value: Double;
    function ReadLabel(ReadWts: Boolean):Integer;
    begin
      Result := Z.NextInteger(0);
      if ReadWts then begin
        Z.NextInteger(0); Z.NextInteger(0); Z.NextInteger(0);
        Molecule.Wt0 := Z.NextInteger(1);
        Molecule.Wt1 := Z.NextInteger(1);
        Molecule.Wt2 := Z.NextInteger(1);
        if Molecule.PointGroup = C3v then begin
          { Wt1 is discarded - should be the same as wt0 }
          Molecule.Wt1 := Molecule.Wt2;
          Molecule.Wt2 := 1;
        end;
      end;
      if Result >= 0 then begin
        State := TSymmetricTop.Create(Manifold);
        State.Symmetry := Result-1;
        State.Comment := Z.NextString('');
        Z.NextRecord;
        if State.Comment = '' then
          State.Comment := Z.Token;
      end else
        Z.NextRecord;
    end;
  begin
    nStates := ReadLabel(True);
    if nStates > 20 then
      raise Exception.CreateFmt('nstates: %d', [nStates]);
    if nStates > 0 then
      nStates := 1;
    if nStates < 0 then
      for i := 1 to -nStates do
        if ReadLabel((i = 1) and Old) < 0 then
          raise Exception.Create('Bad State Symmetry')
    else
      nStates := 1;
    for i := 0 to Abs(nStates)-1 do
      with Manifold.State[i] as TSymmetricTop do begin
        Name := 'State' + IntToStr(i+1);
        Origin := Z.NextDouble(0);
        B := Z.NextDouble(0);
        C := Z.NextDouble(0);
        DJ := Z.NextDouble(0);
        DJK := Z.NextDouble(0);
        DK := Z.NextDouble(0);
        zeta := Z.NextDouble(0);
        qplus := -Z.NextDouble(0);
        etaJ := Z.NextDouble(0);
        etaK := Z.NextDouble(0);
        DqJ := -Z.NextDouble(0);
        DqK := -Z.NextDouble(0);
        hJ := Z.NextDouble(0);
        hJK := Z.NextDouble(0);
        hKJ := Z.NextDouble(0);
        hK := Z.NextDouble(0);
        Z.NextRecord;
      end;
    if nStates < 0 then
      { Perturbations }
      repeat
        i := Z.NextInteger(0);
        if i > 0 then begin
          State1 := Z.NextInteger(0); { State index }
          State2 := Z.NextInteger(0); { State index }
          Value := Z.NextDouble(0);
          raise NotYet;
        end;
        Z.NextRecord;
      until i = 0;
  end;
  procedure ReadTrans;
  var
    Rank, Component, StateG, StateU: Integer;
    Value: Double;
    M: TSphericalTransitionMoment;
    MS: TTransitionMoments;
  begin
    MS := nil;
    repeat
      Rank := Z.NextInteger(-99);
      Component := Z.NextInteger(-99);
      if Component <> -99 then begin
        StateG := Z.NextInteger(0);
        StateU := Z.NextInteger(0);
        Value := Z.NextDouble(0);
        if not Assigned(MS) then
          MS := TTransitionMoments.Create(Molecule, ManifoldU.Name, ManifoldL.Name);
        M := TSphericalTransitionMoment.Create(MS,ManifoldU.State[StateU].Name, ManifoldL.State[StateG].Name);
        M.Strength := Value;
        M.Rank := Rank*2;
        M.Component := Component*2;
      end;
      Z.NextRecord;
    until Component = -99;
  end;
  procedure ReadWidth;
  var
    State: TSymmetricTop;
    Value: Double;
  begin
    State := ManifoldU.State[0] as TSymmetricTop;
    State.Width := Z.NextDouble(0);
    State.wJ := Z.NextDouble(0)*State.Width;
    { Note data file is multiplier of J(J+1) }
    State.wK := State.wJ*(Z.NextDouble(0)+1);
    Z.NextRecord;
  end;
begin
  Result := TMixture.Create(nil);
  try
    Result.Name := 'Mixture';
    t := Z.NextString('2');
    Species := TSpecies.Create(Result);
    Species.Name := 'S';
    Molecule := TSymmetricMolecule.Create(Species);
    Molecule.Name := 'M';
    if TryStrToInt(t, nManifolds) then begin
      { Old format }
      Record1;
      ReadPert(ManifoldL, True);
      if nManifolds > 1 then
        ReadPert(ManifoldU, True);
    end else begin
      { New format }
      Z.Nextline;
      nManifolds := Z.NextInteger(2);
      Record1;
      ReadPert(ManifoldL, False);
      if nManifolds > 1 then
        ReadPert(ManifoldU, False);
    end;
    ReadTrans;
    ReadWidth;
  except
    Result.Free;
    Raise;
  end;
end;

end.

