unit FitLin;

interface

uses Mixture, Obs;

procedure FitLines(Parameters: array of TParameter; Obs: array of TObservation);
procedure FitContour(Parameters: array of TParameter; Spectra: array of TExperimentalPlot);

implementation

uses SysUtils, Matrix, Vector, Nlfit;

procedure FitContour(Parameters: array of TParameter; Spectra: array of TExperimentalPlot);
begin
end;

procedure FitLines(Parameters: array of TParameter; Obs: array of TObservation);
var
  nObs, nPar: Integer;
  a: TMatrix;   { d(obs)/d(par) }
  omc: TColumnVector; { normally obs - calc }
  par: TColumnVector; { Parameters - changed by this routine }
  weights: TColumnVector; { nil or The relative standard deviation of each value in obs }
  { Output: }
  OldSD, NewSD: Double; { Previous and estimated next standard deviations }
  b: TColumnVector;  { Standard Deviations of the parameters}
  ada: TMatrix; { Correlation Matrix }
  SVD: TMatrix; SVDw: TColumnVector; { Details of Singular Value Decomposition }
  i, j: Integer;
  Eigenvalues: TEigenvalues;
  vo, vn: Double;
  procedure CalculateObs(v: TColumnVector);
  var
    i: Integer;
  begin
    for i := 0 to nObs-1 do begin
      try
        with Observations[i] do begin
          Eigenvalues := ManifoldU.Eigenvalues(JU, SymU);
          if (IndexU <= 0) or (IndexU > Eigenvalues.Size) then
            raise Exception.CreateFmt('Upper eigenvalue not found (%d)', [Eigenvalues.Size]);
          v[i+1] := Eigenvalues.E[IndexU];
        end;
        with Observations[i] do begin
          Eigenvalues := ManifoldL.Eigenvalues(JL, SymL);
          if (IndexL <= 0) or (IndexL > Eigenvalues.Size) then
            raise Exception.CreateFmt('Lower eigenvalue not found (%d)', [Eigenvalues.Size]);
          v[i+1] := v[i+1] - Eigenvalues.E[IndexL];
        end;
      except
        on E:Exception do begin
          WriteLn(E.Message, ' ', Observations[i].LineNumber);
          raise;
        end;
      end;
    end;
  end;
begin
  { Allocate Arrays }
  nPar := Length(Parameters);
  nObs := Length(Obs);
  a := TMatrix.Create(nObs, nPar);
  omc := nil;
  weights := nil;
  par := nil;
  b := nil; ada := nil; SVD := nil; SVDw := nil;
  try
    omc := TColumnVector.Create(nObs);
    weights := TColumnVector.Create(nObs);
    { Determine obs - calc }
    CalculateObs(omc);
    {  Calculate Derivatives wrt each parameter }
    for i := 0 to High(Parameters) do begin
      vo := Parameters[i].Value;
      if Parameters[i].Increment = 0 then begin
        vn := vo * 1.0001;
        if vn = 0 then
          vn := 0.001; { Random Guess }
      end else
        vn := vo + Parameters[i].Increment;
      Parameters[i].MNode.SetParameter(Parameters[i].Index, vn);
      try
        CalculateObs(weights);
      finally
        Parameters[i].MNode.SetParameter(Parameters[i].Index, vo);
      end;
      for j := 1 to nObs do
        a[j,i+1] := (omc[j] - weights[j])/(vn-vo)
    end;
    { Work out obs-calc, weights }
    for i := 0 to nObs-1 do begin
      with Observations[i] do
        WriteLn(ManifoldU.Molecule.Name, ' ',
                ManifoldU.Name, ' ', JU:3, SymU:2, IndexU:3,
                ManifoldL.Name, ' ', JL:3, SymL:2, IndexL:3,
                Frequency:12:Precision, omc[i+1]:12:Precision, omc[i+1] - Frequency:12:Precision,
                Weight:8:4, ' ', Name);
      omc[i+1] := omc[i+1] - Observations[i].Frequency;
      weights[i+1] := Observations[i].Weight;
    end;
    { Fill in parameter info }
    par := TColumnVector.Create(nPar);
    for i := 0 to nPar-1 do
      par[i+1] := Parameters[i].Value;
    b := TColumnVector.Create(nPar);
    ada := TMatrix.Create(nPar, nPar);
    SVD := TMatrix.Create(nPar, nPar); { TODO: Make use of these }
    SVDw := TColumnVector.Create(nPar);
    LeastSQ(a, omc, par, weights, OldSD, NewSD, b, ada, SVD, SVDw, 0, 1);
    WriteLn;
    WriteLn(nObs, ' Observations,  ', nPar, ' Parameters');
    WriteLn('Initial Average Error: ', OldSD:10:Precision);
    if nPar > 0 then begin
      WriteLn('Predicted New Error:   ', NewSD:10:Precision);
      WriteLn('Parameters:');
      WriteLn('  # Old                       New                       Std Dev         Name');
      for i := 0 to nPar-1 do
        WriteLn(Format('%3d %-25g %-25g %-15.6g %s', [i, Parameters[i].Value, par[i+1], b[i+1], Parameters[i].Name]));
      if nPar > 1 then begin
        WriteLn;
        WriteLn('Correlation Matrix');
        for i := 1 to nPar do
          Write(i:8);
        WriteLn;
        for i := 1 to nPar do begin
          Write(i:3);
          for j := 1 to i do
            Write(ada[i,j]:8:3);
          WriteLn;
        end;
      end;
      { Copy Back Parameters }
      for i := 0 to nPar-1 do
        Parameters[i].MNode.SetParameter(Parameters[i].Index, par[i+1]);
    end;
  finally
    a.Free;
    omc.Free;
    par.Free;
    b.Free;
    ada.Free;
    weights.Free;
    SVD.Free;
    SVDw.Free;
  end;
end;

end.
