unit Hitran;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Mixture, ExpPlotLoader, Classes, Obs, TextStrm, ExternalList, MathSup;

type

  THITRANReader = class(TObservationsReader)
    ForceGlobalClass: Integer;
    constructor Create; override;
    function DoReadOneLine: Boolean; override;
    function HandleDirective(var Handled: Boolean):Boolean; override;
  end;

  TBaseHITRANLoader = class(TExperimentalPlotLoader)
  private
    EolLength: Integer;
    EolChar: Char;
  protected
    Buffer: string;
    LineLength: Integer;
    nPoints: Integer;
    P: Integer;
    procedure Setup; virtual; abstract;
    procedure AddLine; virtual; abstract;
    procedure Done; virtual; abstract;
    function GF(Width: Integer): Double;
  public
    function Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean; override;
  end;

  TParsumEntry = record
    Name: string;
    MoleculeNumber: Integer;
    Isotopologue: Integer;
    T: TDoubleArray;
    Q: TDoubleArray;
  end;

  TParsums = array of TParsumEntry;

  THITRANLoader = class(TBaseHITRANLoader)
  protected
    procedure Setup; override;
    procedure AddLine; override;
    procedure Done; override;
  end;

  TExternalHITRANLoader = class(TBaseHITRANLoader)
  protected
    Mixture: TMixture;
    ExternalMolecule: TExternalMolecule;
    FullImport: Boolean;
    Convert: Double;
    Parsums: TParsums;
    procedure Setup; override;
    procedure AddLine; override;
    procedure Done; override;
  end;

function IsHITRANfile(FileName: string):Boolean;

function ImportHITRANFile(S: TTextStream): TMixture; overload;
function ImportHITRANFile(S: TTextStream; AParsums: TParsums ): TMixture; overload;
function LoadHITRANParsumFile(S: TFreeFormatStream):TParsums;

implementation

uses {$IFNDEF CONSOLE} IPlotLoader, {$ENDIF}
     MNode, HundA, SymT, QuantumNumber, StrUtils, SysUtils, Math, Graphics,
     Units, Worker, PartFunc, QuantumMath, StaticField;

function IsHitranBuffer(const buffer: string):Boolean;
var
  s, e: Integer;
  LineLength, nOK: Integer;
begin
  LineLength:= -1;
  nOK := 0;
  { Count number of 100 or 160 character lines }
  s := 1;
  while s <= Length(buffer) do begin
    e := s;
    while (e <= Length(buffer)) and not (buffer[e] in [#10,#13]) do
      Inc(e);
    if (e-s = 100) or (e-s = 160) then begin
      if LineLength < 0 then
        LineLength := e-s;
      if e-s = LineLength then
        Inc(nOK);
    end;
    while (e <= Length(buffer)) and (buffer[e] in [#10,#13]) do
      Inc(e);
    s := e;
  end;
  Result := nOK > 5;
end;

function IsHitranStream(Stream: TStream):Boolean;
var
  buffer: string;
  Position: Int64;
  nb: Integer;
begin
  Result := False;
  SetLength(Buffer,1024);
  Position := Stream.Position;
  nb := Stream.Read(buffer[1], Length(Buffer));
  Stream.Position := Position;
  if nb > 0 then begin
    SetLength(buffer,nb);
    Result := IsHitranBuffer(buffer);
  end;
end;

function IsHITRANfile(FileName: string):Boolean;
var
  Stream: TStream;
begin
  Result := False;
  { Quick test to see if the file might be a HITRAN .par file }
  try
    Stream := TFileStream.Create(FileName,fmOpenRead or fmShareDenyWrite);
    try
      Result := IsHitranStream(Stream);
    finally
      Stream.Free;
    end;
  except
  end;
end;

{ THITRANLoader }

function TBaseHITRANLoader.GF(Width: Integer): Double;
begin
  if not TryStrToFloatEx(Copy(Buffer,P,Width),Result) then
    FileErr('Bad Floating point value:'+Copy(Buffer,P,Width));
  Inc(P,Width);
end;

function TBaseHITRANLoader.Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean;
var
  nl, cr, e, nb: Integer;
  s: string;
begin
  Result := False;
  if ((Ext = '') and not IsHitranStream(Stream)) or (Ext <> '.par') then
    Exit;
  LineLength := -1;
  LineNumber := 1;
  nPoints := 0;
  { Determine line length, end of line style }
  SetLength(Buffer, 162);
  nb := Stream.Read(Buffer[1],162);
  SetLength(Buffer, nb);
  if nb < 100 then begin
    FileErr('File too short');
    Exit;
  end;
  nl := Pos(#10,Buffer);
  cr := Pos(#13,Buffer);
  EolLength := 1;
  if nl > 0 then
    if cr > 0 then begin
      EolLength := 2;
      EolChar := Buffer[Min(nl,cr)];
    end else
      EolChar := #10
  else if cr > 0 then
    EolChar := #13
  else
    FileErr('Line too long');
  repeat
    { Parse line }
    e := Pos(EolChar, Buffer)-1;
    if (LineLength < 0) and ((e=100) or (e=160)) then begin
      LineLength := e;
      Stream.Position := Stream.Position - nb + LineLength + EolLength;
      SetLength(Buffer, LineLength + EolLength);
      Setup;
    end;
    if e <> LineLength then begin
      s := LowerCase(Trim(Copy(Buffer,1,e)));
      if (s = 'end') or (s = 'stop') then
        Break;
      if e >= 0 then begin
        FileErr('Bad length to line');
        { Skip This line }
        Stream.Position := Stream.Position - (nb-e) + EolLength;
      end; { else Discard and skip until eol found }
    end else begin
      P := 1;
      AddLine;
      Inc(nPoints);
    end;
    Inc(LineNumber);
    nb := Stream.Read(Buffer[1],160+EolLength);
  until nb = 0;
  Result := nPoints > 0;
  Done;
end;

{ THITRANLoader }

procedure THITRANLoader.AddLine;
begin
  Inc(P,3);
  Connector.SetValue(0,0,nPoints,GF(12));
  Connector.SetValue(0,1,nPoints,GF(10));
end;

procedure THITRANLoader.Done;
begin
  if nPoints > 0 then begin
    Connector.SetLineList(0,True);
    Connector.SetLength(0,nPoints,False,True,True);
    Connector.SetIntensityUnits (cm2WavenumberperMolecule);
    Connector.SetScale(1);
  end;
end;

procedure THITRANLoader.Setup;
begin
  CheckAvailableChannels(2, Stream.Size div (LineLength + EolLength), False);
end;

{ THITRANReader }

constructor THITRANReader.Create;
begin
  inherited;
  Units := cm1;
  ForceGlobalClass := 0;
end;

function THITRANReader.DoReadOneLine: Boolean;
var
  p, nErrorsBefore: Integer;
  Buffer, Branch, LName, UName: string;
  Molecule: TMolecule;
  Fp, Fpp: TQuantumNumber;
  GlobalClass: Integer;
  ExtraU, ExtraL: string;
  AStateU, AStateL: TState;
  ss: string;
  OtherObs: TObservation; { For case where, for example, both parities are assigned to one line }
  UseOtherObs: Boolean;
  Str, ELow: Double;
  function GetField(Width: Integer): string;
  begin
    Result := Copy(buffer, p, Width);
    Result := Trim(Result);
    Inc(p, Width);
  end;
  function GetQno(Width: Integer): TQuantumNumber;
  begin
    Result := Round(StrToFloatEx(GetField(Width))*2);
  end;
  function GetOptionalQno(Width: Integer): TQuantumNumber;
  var
    s: string;
  begin
    s := Trim(GetField(Width));
    if s = '' then
      Result := Low(TQuantumNumber)
    else
      Result := Round(StrToFloatEx(s)*2);
  end;
  function DetermineGlobalClass(Field: string): Integer;
  begin
    Result := ForceGlobalClass;
    if Result > 0 then
      Exit;
    Field := Trim(Field);
    case Length(Field) of
      1, 2: Result := 1;
      3:    Result := 2; { Note 2004 documentaion seems to be in error, at leest for 2008 db - O2 is A1I7, so 8 used below }
      5,6:  Result := 6;
      7:    if Molecule.Manifold[0].State[0] is TLinear then
              Result := 4
            else
              Result := 8;
      8:    if (Molecule.Manifold[0].State[0] is TLinear)
               and (TLinear(Molecule.Manifold[0].State[0]).S > 0) then
              if TLinear(Molecule.Manifold[0].State[0]).S > 1 then
                Result := 2 { O2 }
              else
                Result := 3
            else begin
              {Could be 4 or 5 - only difference is position of possible spaces }
              Result := Pos(' ', Field);
              if (Result > 0) and not Odd(Length(Field)-Result) then
                Result := 5
              else
                Result := 4;
            end;
      9:  if Molecule.Manifold[0] is TSymmetricManifold then
            Result := 8
           else
            Result := 5;
      10:   Result := 8;
      14,15:Result := 7;
      11,12: if Molecule.Manifold[0] is TLinearManifold then
               Result := 7
             else
               Result := 9; { or 10, but do not differentiate }
    else
      ReportError('Bad global quantum numbers format:'+Field);
    end;
  end;
  function ManifoldName(Field: string; out Extra: string): string;
    procedure ReadnCh(nCh, n:Integer);
    var
      i: string;
    begin
      repeat
        i := Trim(Copy(Field, Length(Field)-nCh+1, nCh));
        if Length(Field) > nCh then
          Delete(Field, Length(Field)-nCh+1, nCh)
        else
          Field := '';
        if Result = '' then
          Result := i
        else
          Result := i + '_' + Result;
        Dec(n);
      until (n <= 0) or (Length(Field) = 0);
    end;
  begin
    Result := '';
    Extra := '';
    case GlobalClass of
      1: ReadnCh(2,1);
      2: begin ReadnCh(2,1); ReadnCh(1,1); end;
      3: begin
           ReadnCh(2,1);
           Delete(Field, Length(Field)-1, 2);
           Extra := Copy(Field, Length(Field)-2, 3);
           Delete(Field, Length(Field)-2, 3);
           ReadnCh(1,1);
         end;
      4: ReadnCh(2,4);
      5: ReadnCh(2,5);
      6: ReadnCh(2,3);
      7: begin ReadnCh(1,3); ReadnCh(2,6); end;
      8: ReadnCh(2,5);
      9: ReadnCh(2,6);
     10: Result := Trim(Field);
    else
      ReportError('Bad global class:' + IntToStr(GlobalClass));
    end;
  end;
  { Local quantum number subroutines }
  procedure Local2_5_6(Multiplicity: Integer);
  var
    NL, NU: TQuantumNumber;
    NBranch: string;
    SwitchParity: Integer;
    function FindState(SName: string; Manifold: TManifold; State: TState; out IndexF:Smallint; out SymF: Shortint; FW: TQuantumNumber; SymW: Shortint; JW, NW: TQuantumNumber):string;
    begin
      Result := Manifold.FindStateLabel(IndexF, SymF, FW, [], obs.QuantumNumberFormat, ExField, SymW, JW, NW, State, obs.RemoveSpins);
      if (SymW < 0) and (IndexF = -3) then begin
        { Parity not specified, so add both parities if there is one match of each }
        Result := Manifold.FindStateLabel(IndexF, SymF, FW, [], obs.QuantumNumberFormat, ExField, 0, JW, NW, State, obs.RemoveSpins);
        UseOtherObs := IndexF > 0;
      end;
      if IndexF < 0 then
        ReportError(SName+Result);
      IndexF := 0;
    end;
    function DetermineN(Manifold: TManifold; J: TQuantumNumber; const Extra: string): TQuantumNumber;
    var
      Linear: TLinear;
      Ninc: TQuantumNumber;
    begin
      Linear := Manifold.State[0] as TLinear;
      if Linear.A > 0 then
        Ninc := -1
      else
        Ninc := +1;
      if Extra[1] = '1' then
        Result := J + Ninc
      else
        Result := J - Ninc;
      if Result < Linear.Lambda then
        Result := Linear.Lambda;
    end;
  begin { Normal linear }
    { Group 2 Local Quantum numbers }
    GetField(10);
    Fp := GetOptionalQno(5);
    NL := -1;
    NU := -1;
    case Multiplicity of
      1: begin
           GetField(5);
           Branch := GetField(1);
           obs.JL := GetQno(3);
         end;
      2: begin
           GetField(3);
           Branch := GetField(1);
           obs.JL := GetQno(5);
           NL := DetermineN(obs.ManifoldL, obs.JL, ExtraL);
           NU := DetermineN(obs.ManifoldU, obs.JU, ExtraU);
         end;
      3: begin
           GetField(1);
           NBranch := GetField(1);
           NL := GetQno(3);
           NU := (Ord(NBranch[1])-Ord('Q'))*2+NL;
           Branch := GetField(1);
           obs.JL := GetQno(3);
         end;
    end;
    obs.JU := (Ord(Branch[1])-Ord('Q'))*2 + obs.JL;
    if Multiplicity = 3 then begin
      Fpp := GetOptionalQno(5);
      s := GetField(1)+' '; { Sym }
    end else begin
      s := GetField(1)+' '; { Sym }
      Fpp := GetOptionalQno(5);
    end;
    SwitchParity := 1;
    case s[1] of
      '+': obs.SymL := 0;
      '-': obs.SymL := 1;
      'e', 'E': obs.SymL := (obs.JL div 2) and 1;
      'f', 'F': obs.SymL := 1 - ((obs.JL div 2) and 1);
      'd', 'q': begin obs.SymL := -1; SwitchParity := 0; end;
      ' ': obs.SymL := (obs.JL div 2) and 1;
    else
      ReportError('Bad symmetry:'+s);
      Exit;
    end;
    if (Molecule.nNuclei > 0) and (Fp >= 0) then begin
      if Molecule.nNuclei = 0 then begin
        ReportError('Wrong overall spin');
        Exit;
      end;
      Obs.RemoveSpins := Obs.RemoveSpins + Max(Molecule.nNuclei-1,0); { Single hyperfine provided, so discard any others }
    end else begin
      Obs.RemoveSpins := Obs.RemoveSpins + Molecule.nNuclei; { No hyperfine provided, so discard }
      Fp := Obs.JU;
      Fpp := Obs.JL;
    end;
    LName := FindState('Lower:', obs.ManifoldL, AStateL, obs.IndexL, obs.SymL, Fpp, obs.SymL, obs.JL, NL);
    if UseOtherObs then begin
      OtherObs := Obs;
      OtherObs.Name := FindState('Other Lower:', obs.ManifoldL, AStateU, OtherObs.IndexL, OtherObs.SymL, Fpp, 1-OtherObs.SymL, obs.JL, NL)
                       + ':' + obs.Name;
      OtherObs.JL := Fpp;
    end;
    obs.JL := Fpp;
    obs.SymU := SwitchParity xor obs.SymL;
    UName := FindState('Upper:', obs.ManifoldU, AStateU, obs.IndexU, obs.SymU, Fp, obs.SymU, obs.JU, NU);
    if UseOtherObs then begin
      OtherObs.Name := FindState('Other Upper:', obs.ManifoldU, AStateU, OtherObs.IndexU, OtherObs.SymU, Fp, SwitchParity xor obs.SymU, obs.JU, NU)
                       + ' - ' + OtherObs.Name;
      OtherObs.JU := Fp;
    end;
    obs.JU := Fp;
    obs.Name := UName + ' - ' + LName + ':' + obs.Name;
  end;
  procedure Local4; { Symmetric Tops }
  var
    LastSymW: Integer;
    US: string;
    function FindState(Lab: string; Manifold: TManifold; State: TState; out F: TQuantumNumber; out Sym: Shortint; out Index: Smallint): string;
    var
      Extra: array [0..1] of TQuantumNumber;
      SymS: string;
      J: TQuantumNumber;
      SymW, SymA: Integer;
    begin
      Index := -1;
      J := GetQno(3);
      Extra[0] := GetQno(3); { K }
      Extra[1] := GetOptionalQno(2); { l }
      SymS := Trim(GetField(3));
      if SymS = 'E' then
        SymW := 2
      else if SymS = 'A+' then
        SymW := 0
      else if SymS = 'A-' then
        SymW := 1
      else if SymS = 'A' then
        if (LastSymW=0) or (LastSymW=1) then
          SymW := LastSymW { Guess that symmetry does not change - could be refined }
        else
          ReportError(Lab+'Ambiguous symmetry:A - specify A+ or A-')
      else if (SymS = '') or (SymS = 's') or (SymS = 'a') then
        if LastSymW < 0 then
          SymW := -1 { Accept any symmetry }
        else begin
          { Look for dipole allowed transition }
          if Odd(BranchRank div 2) then begin
            for SymA := 0 to Molecule.LastSym do
              if (Molecule as TSymmetricMolecule).DipoleAllowed(SymA, LastSymW) then begin
                SymW := SymA;
                Break;
              end;
          end else
            SymW := LastSymW;
        end
      else begin
        ReportError(Lab+'Unknown sym:'+SymS);
        Sym := -1;
        Exit;
      end;
      if (SymW in [0,1]) and Odd(J div 2) and (Length(SymS) = 2) then
        SymW := 1 - SymW; { May need further adjustment }
      F := GetOptionalQno(4);
      if (F < 0) or (Molecule.nNuclei-Obs.RemoveSpins <= 0) then begin
        F := J;
        Obs.RemoveSpins := Max(Obs.RemoveSpins,Molecule.nNuclei);
      end else
        Obs.RemoveSpins := Max(Molecule.nNuclei-1,0); { Single hyperfine provided, so discard any others }
      Result := Manifold.FindStateLabel(Index, Sym, F, Extra, obs.QuantumNumberFormat, ExField, SymW, J, -1, State, obs.RemoveSpins);
      if (Index = -3) and (SymW = -1) then begin
        { Two matches - take first of presumably degenerate pair }
        if ObsType = obIntensity then
          { Only half the intensity in one symmetry }
          { TODO: Should use statistical weights for this }
          Obs.Frequency := Obs.Frequency / 2;

        SymW := -1;
        while (Index < 0) and (SymW < Molecule.LastSym) do begin
          Inc(SymW);
          Result := Manifold.FindStateLabel(Index, Sym, F, Extra, obs.QuantumNumberFormat, ExField, SymW, J, -1, State, obs.RemoveSpins);
        end;
        LastSymW := SymW;
      end else if (Index < 0) and (Extra[1] <> 0) then begin
        { Try with other sign of l }
        Extra[1] := -Extra[1];
        Manifold.FindStateLabel(Index, Sym, F, Extra, obs.QuantumNumberFormat, ExField, SymW, J, -1, State, obs.RemoveSpins);
      end;
      LastSymW := SymW;
      if Index < 0 then begin
        ReportError(Lab+Result);
        Result := '';
      end else
        Index := 0;
    end;
  begin
    LastSymW := -1;
    US := FindState('Upper:',obs.ManifoldU,AStateU,obs.JU,obs.SymU,obs.IndexU);
    obs.Name := US + ' - ' + FindState('Lower:',obs.ManifoldL,AStateL,obs.JL,obs.SymL,obs.IndexL) + ':' + obs.Name;
  end;
  procedure Local1; { Asymmetric Tops }
  var
    NU, NL, JU, JL: TQuantumNumber;
    Nam: string;
    ExtraU, ExtraL: array [0..1] of TQuantumNumber;
    procedure HandleSym(out N: TQuantumNumber; var J, F: TQuantumNumber);
    var
      Sym: string;
    begin
      Sym := GetField(1);
      if Sym = '+' then begin
        N := J;
        J := N + 1;
      end else if Sym = '-' then begin
        N := J;
        J := N - 1;
      end else
        N := -1;
      if (Obs.RemoveSpins > Molecule.nNuclei) and (N >= 0) then
        J := N;
      if (F < 0) or (Molecule.nNuclei-Obs.RemoveSpins <= 0) then begin
        F := J;
        Obs.RemoveSpins := Max(Obs.RemoveSpins,Molecule.nNuclei);
      end else
        Obs.RemoveSpins := Max(Molecule.nNuclei-1,0); { Single hyperfine provided, so discard any others }
    end;
  begin
    { Asymmetric Tops }
    JU := GetQno(3);
    ExtraU[0] := GetQno(3);
    ExtraU[1] := GetQno(3);
    obs.JU := GetOptionalQno(5);
    HandleSym(NU, JU, obs.JU);
    JL := GetQno(3);
    ExtraL[0] := GetQno(3);
    ExtraL[1] := GetQno(3);
    obs.JL := GetOptionalQno(5);
    HandleSym(NL, JL, obs.JL);
    Nam := obs.ManifoldL.FindStateLabel(obs.IndexL, obs.SymL, obs.JL, ExtraL, obs.QuantumNumberFormat, ExField, -1, JL, NL, AStateL, obs.RemoveSpins);
    if Obs.IndexL < 0 then
        ReportError('Lower:'+Nam)
    else begin
      Obs.IndexL := 0;
      Obs.Name := Nam + ':' + Obs.Name;
    end;
    Nam := obs.ManifoldU.FindStateLabel(obs.IndexU, obs.SymU, obs.JU, ExtraU, obs.QuantumNumberFormat, ExField, -1, JU, NU, AStateU, obs.RemoveSpins);
    if Obs.IndexU < 0 then
        ReportError('Upper:'+Nam)
    else begin
      Obs.IndexU := 0;
      Obs.Name := Nam + ' - ' + Obs.Name;
    end;
  end;
  function FindManifold(Name: string; DefaultManifold: TManifold; out State: TState):TManifold;
  var
    i: Integer;
  begin
    State := nil;
    Result := Molecule.ManifoldWithName(Name);
    if not Assigned(Result) and Assigned(DefaultManifold) then begin
      State := DefaultManifold.StateWithName(Name);
      if Assigned(State) then
        Result := DefaultManifold;
    end;
    if not Assigned(Result) then begin
      for i := 0 to Molecule.nManifolds - 1 do begin
        State := Molecule.Manifold[i].StateWithName(Name);
        if Assigned(State) then begin
          Result := Molecule.Manifold[i];
          Break;
        end;
      end;
    end;
    if not Assigned(Result) then
      Result := DefaultManifold;
  end;
begin
  p := 1;
  Result := False;
  Buffer := StringOfChar(' ',Max(f.Column-Length(f.Token),0)) + f.Token;
  Buffer := Buffer + f.RestOfLine;
  UseOtherObs := False;
  if not ((Length(Buffer)=100) or (Length(Buffer)=160)) then
    ReportError('Bad line length')
  else begin
    s := GetField(3);
    if Mixture.nMolecules = 1 then
      Molecule := DefaultMolecule
    else
      Molecule := Mixture.FindMolecule(s);
    if not Assigned(Molecule) then
      ReportError('Molecule not found:'+s)
    else begin
      Obs.Name := Buffer; { Overriden if hyperfine present }
      Obs.Frequency := StrToFloatEx(GetField(12));
      Str := StrToFloatEx(GetField(10));
      Inc(p,10+5+5);
      ELow := StrToFloatEx(GetField(10));
      Inc(p,4+8);
      case ObsType of
        obLine, obCDLower, obCDUpper: ;
        obIntensity: Obs.Frequency := Str;
        obELower: Obs.Frequency := ELow;
        obEUpper: Obs.Frequency := Obs.Frequency + ELow;
      end;
      if Length(Buffer) = 100 then begin
        raise NotYet;
      end else begin
        { Determine Class - see Table 3 of HITRAN 2004 paper }
        s := GetField(15);
        GlobalClass := DetermineGlobalClass(s);
        ss := ManifoldName(s,ExtraU);
        obs.ManifoldU := FindManifold(ss, DefaultManifoldU, AStateU);
        if not Assigned(AStateU) then
          AStateU := StateU;
        if not Assigned(obs.ManifoldU) then begin
          ReportError('Upper Manifold/State not found:'+ss);
          Exit;
        end;
        s := GetField(15);
        ss := ManifoldName(s,ExtraL);
        obs.ManifoldL := FindManifold(ss, DefaultManifoldL, AStateL);
        if not Assigned(AStateL) then
          AStateL := StateL;
        if not Assigned(obs.ManifoldL) then begin
          ReportError('Lower Manifold/State not found:'+ss);
          Exit;
        end;
        nErrorsBefore := nErrors;
        case GlobalClass of
          2: Local2_5_6(3);
          3: Local2_5_6(2);
          1,4,5,7: Local2_5_6(1);
        else
          if Molecule.Manifold[0] is TSymmetricManifold then
            Local4
          else
            Local1;
          { Note that spherical tops are not yet implemented }
        end;
        if nErrorsBefore <> nErrors then
          Exit;
        s := GetField(1)+' ';
        if ObsType = obLine then
          case s[1] of
            '1': obs.Weight := 1;
            '2': obs.Weight := 0.1;
            '3': obs.Weight := 0.01;
            '4': obs.Weight := 0.001;
            '5': obs.Weight := 0.0001;
            '6': obs.Weight := 0.00001;
          else
            obs.Weight := 10;
          end
        else
          obs.Weight := 1; { TODO: Use intensity accuracy information }
        if UseOtherObs then
          AddObservation(OtherObs);
        Result := True;
      end;
    end;
  end;
end;

function THITRANReader.HandleDirective(var Handled: Boolean): Boolean;
begin
  Result := inherited HandleDirective(Handled);
  if not Handled and (LowerCase(s) = 'globalclass') then begin
    if f.GetFreeItem then
      ForceGlobalClass := f.IntegerToken
    else
      ForceGlobalClass := 1;
    Handled := True;
  end;
end;

procedure TExternalHITRANLoader.AddLine;
var
  Entry: TExternalEntry;
  Lower, Upper: String;
  function F(Width:Integer):string;
  begin
    Result := Copy(Buffer,P,Width);
    Inc(P,Width);
  end;
  function FirstQuantum(s: string): TQuantumNumber;
  var
    i, L: Integer;
  begin
    { Determine quantum number format from first non-space character }
    i := 1;
    while (i < 5) and (s[i] = ' ') do
      Inc(i);
    case i of
      5: L := 1; // ____n.... -> Group 3, J up to 9
      4: L := 2; // ___nn.... -> Group 3, J 10 - 99
      3: if s[4] = ' ' then
           L := 1  // __n__n... => Group 1, 4, J up to 9
         else
           L := 3; // __nnn.... -> Group 3, J 99+
      2: L := 2; // _nn_?n... => Group 1, 4, J 10 - 99
      1: L := 3; // nnn??n... => Group 1, 4, J 99+
    end;
    if not TryStrToQuantumNumber(Copy(S,i,L),Result,qftJ) then
      Result := -1;
  end;
var
  MoleculeNumber: string;
  Isotopologue: string;
  Species: TSpecies;
  Molecule: TExternalMolecule;
  nGlobal, nLocal, nSkip: Integer;
  s: string;
  DeltaJ: TQuantumNumber;
  JStart, Jlength: Integer;
begin
  with Entry do begin
    MoleculeNumber := F(2);
    Isotopologue := F(1);
    if FullImport then begin
      Species := Mixture.FindOrCreateNode(MoleculeNumber, TSpecies) as TSpecies;
      Molecule := Species.FindOrCreateNode(MoleculeNumber+Isotopologue, TExternalMolecule) as TExternalMolecule;
    end else begin
      Molecule := ExternalMolecule;
      // QuantumNumbers := MoleculeNumber+Isotopologue;
    end;
    Frequency := GF(12);
    if FullImport and (LineLength > 100) then begin
      Inc(P,10); { Ignore line strength }
      Intensity := GF(10); { Read A }
    end else begin
      Intensity := GF(10);
      Inc(P,10); { Ignore transition moment }
    end;
    Width := GF(5); { Air broadened width, gamma(air) }
    Inc(P,5); { Skip Self broadening, gamma(self) }
    ELower := GF(10);
    Inc(P,4+8); { Skip T dep gamma(air), air induced shift }
    if LineLength = 100 then begin
      nGlobal := 3;
      nLocal := 9;
      nSkip := 3+6;
    end else begin
      nGlobal := 15;
      nLocal := 15;
      nSkip := 6+12+1;
    end;
    Upper := F(nGlobal);
    Lower := F(nGlobal);
    if FullImport then begin
      ManifoldU := Molecule.FindOrCreateNode(Trim(Upper), TExternalManifold) as TExternalManifold;
      ManifoldL := Molecule.FindOrCreateNode(Trim(Lower), TExternalManifold) as TExternalManifold;
      ManifoldU.Initial := True;
      ManifoldL.Initial := True;
      UpperQuanta := '';
      LowerQuanta := '';
    end else begin
      UpperQuanta := Trim(Upper) + ' ';
      LowerQuanta := Trim(Lower) + ' ';
      ManifoldL := nil;
      ManifoldU := nil;
    end;
    Upper := F(nLocal);
    Lower := F(nLocal);
    UpperQuanta := UpperQuanta + Trim(Upper);
    LowerQuanta := LowerQuanta + Trim(Lower);
    gL := 0; { Flag unset }
    gU := 0;
    JL := -1;
    JU := -1;
    Inc(P, nSkip);
    if LineLength > 100 then begin
      gU := Round(GF(7));
      gL := Round(GF(7));
      Intensity := Intensity/(Convert*IntPower(Frequency,3))*gU;
      s := Trim(Lower);
      DeltaJ := Low(TQuantumNumber);
      JLength := 3;
      if (s <> '') and IsDeltaJName(s[1],DeltaJ) then begin
        { Diatomic, so lower state contains branch }
        JStart := Pos(s[1],Lower);
        case JStart of
          6: Inc(JStart);   { Local quanta style 2 }
          2: Inc(Jstart,5); { Local quanta style 5 }
        else
          JStart := 5;      { Local quanta style 6 }
          JLength := 5;
          if (Length(s) > 2) then
            IsDeltaJName(s[2],DeltaJ);
        end;
        if TryStrToQuantumNumber(Copy(Lower,JStart,Jlength),JL,qftJ) then
          JU := JL + DeltaJ;
      end else begin
        JL := FirstQuantum(Lower);
        JU := FirstQuantum(Upper);
      end;

    end;
  end;
  Entry.LineNumber := LineNumber;
  Molecule.AddEntry(Entry);
end;

procedure TExternalHITRANLoader.Done;
var
  i, j, k, mn: Integer;
  Q: TInterpolatedPartitionFunction;
  E: Exception;
begin
  E := Mixture.FixupProperties;
  E.Free; { Discard errors at this point }
  for i := 0 to Mixture.nSpecies - 1 do
    for j := 0 to Mixture.Species[i].Count - 1 do
      with Mixture.Species[i].Isotopomer[j] do begin
        Q := ActivePartitionFunction as TInterpolatedPartitionFunction;
        mn := StrToInt(Name);
        for k := 0 to High(Parsums) do
          if Parsums[k].MoleculeNumber*10 + Parsums[k].Isotopologue = mn then begin
            Q.Comment := Parsums[k].Name;
            Q.Plot.Frequency := Parsums[k].T;
            Q.Plot.Intensity := Parsums[k].Q;
            Break;
          end;
      end;
end;

procedure TExternalHITRANLoader.Setup;
begin
  { JQSRT 48 519 (1992) eq 42 with frequency in cm-1, dipole moment in Debye }
  Convert := 64*IntPower(Pi,4)/({3*}Planck*1e43);
  Mixture.Simulations[0].IntensityUnits := cm2WavenumberperMolecule;
end;

function ImportHITRANFile(S: TTextStream; AParsums: TParsums): TMixture;
var
  Dummy: TExperimentalPlotLoaderConnector;
begin
  Result := TMixture.Create(nil);
  Dummy := {$IFNDEF CONSOLE}TInteractivePlotLoaderConnector{$ELSE}TExperimentalPlotLoaderConnector{$ENDIF}.Create(nil,'');
  with TExternalHITRANLoader.Create(Dummy,S,'.par',cm1,cm2WavenumberperMolecule) do
    try
      Mixture := Result;
      FullImport := True;
      Parsums := AParsums;
      Load(Log, False);
    finally
      Free;
    end;
end;

function ImportHITRANFile(S: TTextStream): TMixture;
begin
  Result := ImportHITRANFile(S, nil);
end;

function LoadHITRANParsumFile(S: TFreeFormatStream):TParsums;
var
  nCol, n: Integer;
  Value: Double;
  T: TDoubleArray;
  i: Integer;
  LastMolecule, ThisMolecule: string;
  nIso, nMol: Integer;
begin
  { Load HITRAN pasrsum.dat from global data }
  { Read header line }
  nCol := 0;
  LastMolecule := '';
  nMol := 0;
  S.FlowToNextLine := False;
  while not S.GetFreeItem do
    S.NextLine; { Skip Temperature and blank lines }
  while not S.AtEndOfLine do
    if S.GetFreeItem then begin
      SetLength(Result, Length(Result)+1);
      Result[nCol].Name := s.Token;
      i := Pos('_',s.Token);
      if i > 0 then
        ThisMolecule := Copy(s.Token, 1{$IFDEF MSWINDOWS}  {$ENDIF} {$IFDEF LINUX}  {$ENDIF}, i-1)
      else
        raise Exception.Create('Unexpected molecule name :'+s.Token);
      if ThisMolecule <> LastMolecule then begin
        nIso := 1;
        Inc(nMol);
        LastMolecule := ThisMolecule;
      end else
        Inc(nIso);
      Result[nCol].MoleculeNumber := nMol;
      Result[nCol].Isotopologue := nIso;
      Inc(nCol);
    end;
  S.NextLine;
  n := 0;
  while not S.AtEndOfFile do begin
    while not S.GetFreeItem and not S.AtEndOfFile do
      S.NextLine; { Skip blank lines }
    if not S.IsRealToken(Value) then
      Break;
    if Length(T) <= n  then
      SetLength(T, n+3000);
    T[n] := Value;
    nCol := 0;
    while not S.AtEndOfLine do
      if S.GetFreeItem then begin
        if not S.IsRealToken(Value) then
          Break;
        if Length(Result[nCol].Q) <= n  then
          SetLength(Result[nCol].Q, n+3000);
        Result[nCol].Q[n] := Value;
        Inc(nCol);
      end;
    if nCol > 0 then
      Inc(n)
    else
      Dec(n);
    S.NextLine;
  end;
  SetLength(T,n);
  for i := 0 to High(Result) do begin
    Result[i].T := T;
    SetLength(Result[i].Q,n);
  end;
end;

initialization
  RegisterPlotLoader(THITRANLoader, 'HITRAN Line lists', '*.par');
end.

