unit Interlock;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF} Classes, SysUtils, Worker, SyncObjs;

type

  TInterlockedWorker = class;

  { Component that records state while being assembled, so that
    threads can co-operate on its use }

  { TInterlockedComponent }

  TInterlockedComponent = class(TComponent)
  private
    FErrorMessage: string;
    FCurrentWorker: TInterlockedWorker;
    FExceptionClass: ExceptClass;
    function GetReady: Boolean;
  protected
    FInValid: Boolean;
    FDoneEvent: TEvent;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Make(Wait:Boolean; RequiredAttributes: Pointer = nil);
    procedure DoMake(RequiredAttributes: Pointer); virtual; abstract;
    procedure RunDoMake(RequiredAttributes: Pointer);
    procedure WaitFor(Timeout: Cardinal = INFINITE);
    function HasCorrectAttributes(Attributes: Pointer):Boolean; virtual; abstract;
    property InValid: Boolean read FInValid;
    property Ready: Boolean read GetReady;
    property ErrorMessage: string read FErrorMessage;
    property ExceptionClass: ExceptClass read FExceptionClass;
    property CurrentWorker: TInterlockedWorker read FCurrentWorker;
  end;

  TInterlockedWorker = class(TWorkerThread)
  private
    FRequiredAttributes: Pointer;
    FInterlockedComponent: TInterlockedComponent;
  public
    procedure DoWork; override;
  end;

implementation

uses UCompat, Contnrs, Profile;

threadvar
  ThreadDoneEvent: TEvent;

var
  DoneEvents: TObjectList; { Used to track allocations so that they can be cleaned up }

{ TInterlockedComponent }

function TInterlockedComponent.GetReady: Boolean;
begin
  Result := not Assigned(FDoneEvent);
end;

constructor TInterlockedComponent.Create(AOwner: TComponent);
begin
  inherited;

end;

destructor TInterlockedComponent.Destroy;
begin

  inherited;
end;

var
  { One per object might be fastest, but might end up with a lot of unused objects }
  InterlockCriticalSection: {$IFDEF FPC}System.{$ENDIF}TRTLCriticalSection;

procedure TInterlockedComponent.Make(Wait: Boolean; RequiredAttributes: Pointer);
{ Ensure component is being processed so it has the RequiredAttributes;
  Wait if required }
var
  StartWork: Boolean;
begin
  if InValid then
    raise Exception.CreateFmt('Calculating invalid %s:%s', [ClassName, ErrorMessage]);
  {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(InterlockCriticalSection);
  try
    if Ready and HasCorrectAttributes(RequiredAttributes) then
      Exit; { Easy - done }
    StartWork := not Assigned(FDoneEvent);
    if StartWork then
      if nWorkers > 0 then begin
        if not Assigned(ThreadDoneEvent) then begin
          if not Assigned(DoneEvents) then
            DoneEvents := TObjectList.Create;
          ThreadDoneEvent := TEvent.Create(nil, True, False, '');
          DoneEvents.Add(ThreadDoneEvent);
        end;
        FDoneEvent := ThreadDoneEvent;
        FDoneEvent.ResetEvent;
      end else
        StartWork := Wait;
  finally
    {$IFDEF FPC}System.{$ENDIF}LeaveCriticalSection(InterlockCriticalSection);
  end;
  if StartWork then begin
    if Wait or (GetCurrentThreadID <> MainThreadID) then begin
      RunDoMake(RequiredAttributes); { This thread can do the work }
      if nWorkers > 0 then
        ThreadDoneEvent.SetEvent;
      if InValid then
        raise ExceptionClass.Create(ErrorMessage);
    end else begin
      FCurrentWorker := FindWorkerWait as TInterlockedWorker;
      if not Assigned(FCurrentWorker) then
        Abort;
      FCurrentWorker.FRequiredAttributes := RequiredAttributes;
      FCurrentWorker.FInterlockedComponent := Self;
      {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(InterlockCriticalSection);
      { Transfer done event to new worker }
      FDoneEvent := FCurrentWorker.Done;
      ThreadDoneEvent.SetEvent;
      {$IFDEF FPC}System.{$ENDIF}LeaveCriticalsection(InterlockCriticalSection);
      FCurrentWorker.StartWork;
    end;
  end else if Wait then begin
    { another thread doing it - wait for it }
    WaitFor;
    {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(InterlockCriticalSection);
    try
      if Ready and HasCorrectAttributes(RequiredAttributes) then
        Exit;
    finally
      {$IFDEF FPC}System.{$ENDIF}LeaveCriticalsection(InterlockCriticalSection);
    end;
    Make(Wait, RequiredAttributes); { Need to restart calculation }
  end;
end;

procedure TInterlockedComponent.RunDoMake(RequiredAttributes: Pointer);
begin
  try
    DoMake(RequiredAttributes);
  except
    on E:Exception do begin
      FInValid := True;
      FErrorMessage := E.Message;
      FExceptionClass := ExceptClass(E.ClassType);
    end;
  end;
  {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(InterlockCriticalSection);
  try
    FDoneEvent := nil;
  finally
    {$IFDEF FPC}System.{$ENDIF}LeaveCriticalSection(InterlockCriticalSection);
  end;
end;

procedure TInterlockedComponent.WaitFor(Timeout: Cardinal);
begin
  while Assigned(FDoneEvent) do begin
    { The calcultion could finish before the wait starts }
    DoWait(FDoneEvent, 1000); { Even if event signaled, may not be complete as DoneEvent might have been switched }
    if Aborted then Abort;
  end;
end;

{ TInterlockedWorker }

procedure TInterlockedWorker.DoWork;
begin
  if Assigned(FInterlockedComponent) then
    FInterlockedComponent.RunDoMake(FRequiredAttributes);
  FInterlockedComponent := nil;
end;

initialization
  {$IFDEF FPC}System.{$ENDIF}InitCriticalSection(InterlockCriticalSection);
finalization
  {$IFDEF FPC}System.{$ENDIF}DoneCriticalSection(InterlockCriticalSection);
  DoneEvents.Free;
end.

