unit MNode;
{ Generic node type - handles loading/saving of parameters }

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Classes, Contnrs,  Graphics, TypInfo, {$IFNDEF CONSOLE} Menus, {$ENDIF}
     pgDOM, SysUtils, Units, MathSup;

type
  TPrintLevel = (plNone=0, plDefault, plMinimal, plCSV, plText, plBinary, plDetail, plLineList, plEigenVectors, plHamiltonian, plMatrices, plAll);
const
  PrintLevelNames: array [0..Ord(plAll)+1] of PChar =
  ('None', 'Default', 'Minimal', 'CSV', 'Text', 'Binary', 'Detail', 'LineList', 'EigenVectors', 'Hamiltonian', 'Matrices', 'All', nil);

type

  TMNode = class;
  TMNodeClass = class of TMNode;
  TMNodeArray = array of TMNode;

  { TMNode }

  TMNode = class(TComponent)
  private
    FnParameters: Integer;
    FnSettings: Integer;
    FSettings: array of Integer;
    FSettingNames: TStringList;
    FDefaultSettings: array of string;
    FModified: Boolean;
    FComment: string;
    FFirstDynamicSetting, FFirstDynamicParameter: Integer;
    FSavedParameters: TDoubleArray;
    function GetItem(AIndex: Integer): TMNode; {$IFDEF INLINE} inline; {$ENDIF}
    function GetFloating(AIndex: Integer): Boolean;
    function GetIncrement(AIndex: Integer): Double;
    procedure SetComment(const Value: string);
    function GetComponentIndex: Integer;
    function GetStdDev(AIndex: Integer): Double;
    procedure SetStdDev(AIndex: Integer; const Value: Double);
  protected
    { The parameter arrays are protected rather than private as TManifold
      adjusts the size of these dynamically }
    FParameters: TDoubleArray;
    FDefaultParameters: array of Double;
    FIncrements: array of Double;
    FStdDev: array of Double;
    FFloating: array of Boolean;
    FParameterNames: TStringList;
    FName: TComponentName;
    FNodeIndex: Integer;
    {$IFNDEF CONSOLE}
    CurrentMenu: TMenuItem;
    {$ENDIF}
    function GetnParameters: Integer; virtual;
    procedure SetnParameters(AValue: Integer); virtual; { Not directly associated with property, as rarely used }
    procedure SetnSettings(AValue: Integer); virtual;
    function GetParameterNames(AIndex: Integer): string; virtual;
    function GetSettingNames(AIndex: Integer): string; virtual;
    function GetName: TComponentName; virtual;
    procedure SetName(const Value: TComponentName); override;
    procedure SetModified(const Value: Boolean); virtual;
    function GetDefaultSettings(AIndex: Integer): string; virtual;
    function GetDefaultParameters(AIndex: Integer): Double; virtual;
    procedure SetFloating(AIndex: Integer; const Value: Boolean); virtual;
    procedure SetIncrement(AIndex: Integer; const Value: Double); virtual;
    procedure SetComponentIndex(const Value: Integer); virtual;
  public
    SavedProperties: string;
    DimensionlessParameters: string;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function GetChildTypes(AIndex: Integer): TMNodeClass; virtual;
    class function OwnerType: TMNodeClass;
    class function DisplayClassName: string; virtual;
    class procedure RegisterClass(AOwnerType:TMNodeClass);
    class function nChildTypes: Integer; virtual;
    class function ChildClassWithType(MNodeClass: TMNodeClass):TMNodeClass;
    function IsValidChildClass(MNodeClass: TMNodeClass):Boolean; virtual;
    function PageName: string; virtual;
    function Count:Integer;
    function CanShift(Shift:Integer):Boolean; virtual;

    function TransferProperties(DOMNode: TDOMNode; Writing, AllParameters: Boolean; WithChildren: Boolean = True):Integer;
    function ReadChildProperties(DOMNode: TDOMNode):Integer;
    function ReadMNode(DOMNode: TDOMNode; out nUnused: Integer): TMNode; virtual;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; virtual;
    function ReadFloatableDOMNode(Index: Integer; Deflt: string; ParNode: TDOMElement):Integer;
    function ReadFloatableValue(DOMNode: TDOMNode; Index: Integer; const Prop, Deflt: string):Integer;
    function nUnreadProperties(DOMNode: TDOMNode): Integer;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; virtual;
    procedure WriteFloatableValue(Index: Integer; Deflt: string; DOMNode: TDOMNode; DefaultCheck: Boolean); virtual;
    procedure SaveToStream(Stream: TStream; WithLinkedMNodes: Boolean = False; MNodeList: TObjectList = nil);
    function SimpleSaveToStream(Stream: TStream; WithLinkedMNodes: Boolean = False): Boolean; virtual;
    procedure GetDefaults(MNode: TMNode);
    procedure GetAll(MNode: TMNode);
    procedure AdjustDefaults; virtual;
    procedure ClearStdDev;

    procedure ClearCache; virtual; { Call after significant property value has changed }
    procedure CheckProperties;
    function FixupProperties: Exception; virtual; { Call after change to set up links other than parent/child }
    procedure FixAll(Recurse: Boolean);
    procedure HandleAssignment(NameVal: string);
    function SubPropInfo(const Token: string; out MNode: TMNode; out ErrMsg: string): PPropInfo; virtual;
    function SubPropValue(const Token: string): string;
    procedure SetSubPropValue(const Token, NewValue: string);

    function GetParameter(AIndex: Integer): Double; overload; {$IFDEF INLINE} inline; {$ENDIF}
    function GetParameter(const AName: string): Double; overload;
    procedure GetParameters(MNode: TMNode);
    procedure ModifyParameter(AIndex: Integer; Value: Double; ChangeModify: Boolean; Notify: Boolean = True);
    function AddParameter(AName: string; Default: Double = 0; Position: Integer = -1):Integer;
    function AddSetting(AName: string; Default: string = ''):Integer;
    procedure SetParameterName(AIndex: Integer; const Value: string);
    procedure SetSettingName(AIndex: Integer; const Value: string);
    procedure DeleteParameter(AIndex: Integer);
    procedure DeleteSetting(AIndex: Integer);
    procedure SetParameter(AIndex: Integer; const Value: Double); virtual;
    function HelpOnParameter(AIndex: Integer): string; virtual;
    function IndexOfParameter(const AName: string):Integer;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); virtual;
    procedure StartDerivatives;
    procedure StopDerivatives;
    function DerivativeRequired(Index: Integer):Boolean;

    function GetSetting(AIndex: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}
    procedure GetSettings(MNode: TMNode);
    procedure ModifySetting(AIndex: Integer; const AValue: Integer; ChangeModify: Boolean);
    procedure SetSetting(AIndex: Integer; const AValue: Integer); virtual;
    function IndexOfSetting(const AName: string):Integer;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; virtual;
    function HelpOnSetting(AIndex: Integer): string; virtual;
    function SettingToStr(AIndex: Integer): string; virtual;
    procedure SetSettingFromStr(AIndex: Integer; const s: string); overload; virtual;
    procedure SetSettingFromStr(const AName: string; const s: string); overload;
    function IsSettingEnum(AIndex: Integer):Boolean; virtual;
    function IsSettingQuantum(AIndex: Integer):Boolean; virtual;
    procedure GetEnumNames(AIndex: Integer; List: TStrings); virtual;
    function GetFlag(AIndex: Integer): Boolean;
    procedure SetFlag(AIndex: Integer; const AValue: Boolean);
    function GetColour(AIndex: Integer): TColor;
    procedure SetColour(AIndex: Integer; const AValue: TColor);
    function IsSettingColour(AIndex: Integer): Boolean; virtual;

    function SettingToPrettyStr(AIndex: Integer): string; virtual;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); virtual;
    procedure PrettyPrintHeader(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); virtual;
    procedure PrettyPrintSettings(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); virtual;
    procedure PrettyPrintParameters(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); virtual;
    procedure PrettyPrintComponents(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); virtual;

    procedure AddHandler(const ACaption, AHint: String; OnClick: TNotifyEvent);
    procedure AddHandlers; virtual;
    {$IFNDEF CONSOLE}
    procedure AddHandlersToMenu(Menu:TMenuItem);
    {$ENDIF}

    function IsAncestorOf(MNode: TMnode):Boolean;
    function IsValidChildOf(MNode: TComponent):Boolean;
    function CommonAncestor(MNode: TMNode): TMNode;
    function RootNode: TMNode;
    function LinkedNodesList(List: TObjectList = nil):Integer;
    function ChildWithClass(MNodeClass: TMNodeClass; Index: Integer = -1):TMNode;
    function AncestorWithClass(MNodeClass: TMNodeClass):TMNode;
    procedure CycleUnits; virtual;
    function AdjustNodeIndexes(NodeClass: TMNodeClass; var NodeArray: TMNodeArray; AComponent: TComponent = nil; Operation: TOperation = opInsert; ExpectedNumber: Integer = -1):Integer;

    function UniqueName: string; virtual;
    function UniquePrefix: string; virtual;
    function FullHeirarchicName(Quote: Boolean = False; RemoveRoot: Boolean = False): string;
    function FullParameterName(Index: Integer; Quote: Boolean = False): string;
    function ComponentWithName(const AName: string; AClass: TClass; Exclude: TMNode; SearchChildren: Boolean = False): TMNode;
    function nComponentsWithName(const AName: string; out First: TMNode):Integer;
    function FindOrCreateNode(AName:string; NodeClass: TMNodeClass):TMNode;

    property nParameters: Integer read GetnParameters;
    property Parameters[AIndex: Integer]: Double read GetParameter write SetParameter;
    property Increment[AIndex: Integer]: Double read GetIncrement write SetIncrement;
    property StdDev[AIndex: Integer]: Double read GetStdDev write SetStdDev;
    property nSettings: Integer read FnSettings;
    property Settings[AIndex: Integer]: Integer read GetSetting write SetSetting;
    property Items[AIndex: Integer]: TMNode read GetItem;
    property Floating[AIndex: Integer]: Boolean read GetFloating write SetFloating;
    property ParameterNames[AIndex: Integer]: string read GetParameterNames;
    property SettingNames[AIndex: Integer]: string read GetSettingNames;
    property Modified: Boolean read FModified write SetModified;
    property DefaultSettings[AIndex: Integer]: string read GetDefaultSettings;
    property DefaultParameters[AIndex: Integer]: Double read GetDefaultParameters;
    property FirstDynamicSetting: Integer read FFirstDynamicSetting;
    property FirstDynamicParameter: Integer read FFirstDynamicParameter;
    property NodeIndex: Integer read FNodeIndex;
    { This hides the property in TComponent, so may require care is use.
      The default is just to use the inherited method }
    property ComponentIndex: Integer read GetComponentIndex write SetComponentIndex;

  published
    property Name: TComponentName read GetName write SetName;
    property Comment: string read FComment write SetComment;
    {$IFNDEF CONSOLE}
    procedure CopyToClipboard(WithLinkedMNodes: Boolean = False; MNodeList: TObjectList = nil); virtual;
    {$ENDIF}
  end;

  TMNodeChange = (mnCreate, mnDestroy, mnParameterChange, mnSettingChange, mnNameChange, mnNewBaseNode, mnPropNameChange, mnGraphicChanged, mnCacheCleared, mnObsCleared);
  OnMNodeChange = procedure(Sender: TMNode; Change: TMNodeChange; Index: Integer) of object;

  { TMNodePair }

  TMNodePair = class(TMNode)
  protected
    FBra, FKet: TMNode;
    procedure DoMNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
    function GetName: TComponentName; override;
    procedure SetName(const Value: TComponentName); override;
  public
    FBraName, FKetName: string; { Only used during load/save, hence use of F...Name }
    class function BraKetClass: TMNodeClass; virtual; abstract;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function BraOwner: TMNode; virtual;
    function KetOwner: TMNode; virtual;
    function FixupProperties: Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    procedure SetBraKetName(const ABraName, AKetName: string);
    function SetBraKetNameE(const ABraName, AKetName: string; ChangingMNode: TMNode = nil): Exception;
    property Bra: TMNode read FBra;
    property Ket: TMNode read FKet;
    function PairName: string; virtual; abstract; { For use in display }
  end;

  EReportException = class(Exception);
  EWarningException = class(EReportException);

  ECantFindState = class(Exception)
  private
    FMissingState: string;
    FMNodePair: TMNodePair;
  public
    property MNodePair: TMNodePair read FMNodePair;
    property MissingState: string read FMissingState;
    constructor Create(AMNodePair: TMNodePair; const AMissingState: string);
  end;

procedure AddMNodeNotification(MNodeChange: OnMnodeChange);
procedure RemoveMNodeNotification(MNodeChange: OnMnodeChange);
procedure PerformMNodeNotification(Sender: TMNode; Change: TMNodeChange; Index: Integer);

function ReadValue(DOMNode: TDOMNode; const PropName, Dflt: string; const AltPropName: string = ''; Delete: Boolean = True): string;
procedure WriteValue(DOMNode: TDOMNode; PropName, Value, Dflt: string);
function nUnreadProperties(Name: string; DOMNode: TDOMNode): Integer;

function ReadMNodeClass(DOMNode: TDOMNode; ParentClass: TClass): TMNodeClass;

procedure RemoveAttributeNode(Element: TDOMElement; Attr: TDOMAttr);

implementation

uses
     {$IFNDEF CONSOLE}
       {$IFDEF MSWINDOWS} Windows, {$ENDIF}
       MNodeClip, Clipbrd, Huge, HugeClip, Dialogs,
     {$ENDIF}
     MNodeColour, Math, pgXMLWrite, PropHelp,
     StrUtils, Worker, QuantumNumber;

{ TMNode }

function TMNode.GetParameter(const AName: string): Double;
var
  Index: Integer;
begin
  Index := IndexOfParameter(AName);
  if Index >= 0 then
    Result := GetParameter(Index)
  else
    raise Exception.Create('Parameter ' + AName + ' does not exist in ' + ClassName);
end;

function TMNode.GetParameterNames(AIndex: Integer): string;
begin
  Result := FParameterNames[AIndex];
end;

function TMNode.GetSettingNames(AIndex: Integer): string;
begin
  Result := FSettingNames[AIndex];
end;

function TMNode.GetItem(AIndex: Integer): TMNode; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  {$IFOPT R+}
  Result := Components[AIndex] as TMNode;
  {$ELSE}
  Result := TMNode(Components[AIndex]);
  {$ENDIF}
end;

constructor TMNode.Create(AOwner: TComponent);
var
  XMLDocument: TXMLDocument;
  PropList: PPropList;
  i, nProps, p, NewNparameters, NewNSettings: Integer;
  Prop: string;
  Propl, spl: string;
begin
  inherited Create(AOwner);
  FNodeIndex := -1; { Flag creating }
  NewNparameters := 0;
  NewNSettings := 0;
  { After this, SavedProperties contains a comma delimited list list of all properies to be saved }
  SavedProperties := ',' + SavedProperties + ',';
  {$IFDEF FPC}
  nProps := GetTypeData(PTypeInfo(Self.ClassInfo))^.Propcount;
  GetMem(PropList,nProps*SizeOf(Pointer));
  GetPropInfos(PTypeInfo(Self.ClassInfo), PropList);
  {$ELSE}
  nProps := GetPropList(Self, PropList);
  {$ENDIF}
  for i := nProps-1 downto 0 do begin
    Prop := PropList^[i]^.Name;
    Propl := LowerCase(Prop);
    spl := LowerCase(SavedProperties);
    if (Pos(','+Propl+',',spl) = 0) and (Pos(','+Propl+'=',spl) = 0) then
      SavedProperties := ',' + Prop + SavedProperties;
    if PropList^[i]^.Index > 0 then begin
      case PropList^[i]^.PropType^.Kind of
        tkInteger, tkChar, tkEnumeration {$IFDEF FPC}, tkBool {$ENDIF}:
          NewNSettings := Max(NewNSettings, PropList^[i]^.Index);
        tkFloat: NewNparameters := Max(NewNparameters, PropList^[i]^.Index);
      else
        raise Exception.CreateFmt('Unknown Type: %d', [Ord(PropList^[i]^.PropType^.Kind)]);
      end;
    end;
  end;
  FFirstDynamicSetting := NewNSettings+1;
  FFirstDynamicParameter := NewNparameters+1;
  SetnParameters(NewNparameters);
  FnSettings := -1;
  SetnSettings(NewNSettings);

  { ReScan properties, assigning names and recording default values }
  for i := nProps-1 downto 0 do begin
    if PropList^[i]^.Index >= 0 then begin
      { Extract Default value }
      Prop := ','+LowerCase(PropList^[i]^.Name)+'=';
      p := Pos(Prop,spl);
      if p > 0 then begin
        Prop := Copy(spl,p+Length(PropList^[i]^.Name)+2,Length(spl));
        p := Pos(',',Prop);
        Prop := Copy(Prop,1,p-1);
      end else
        Prop := '';
      case PropList^[i]^.PropType^.Kind of
        tkInteger, tkChar, tkEnumeration{$IFDEF FPC}, tkBool {$ENDIF}:
          if FSettingNames[PropList^[i]^.Index] <> '' then
            raise Exception.CreateFmt('Duplicate property index (%d) for %s: %s and %s',
                                      [PropList^[i]^.Index, ClassName, FSettingNames[PropList^[i]^.Index], PropList^[i]^.Name])
          else begin
            FSettingNames[PropList^[i]^.Index] := PropList^[i]^.Name;
            if Prop <> '' then
              SetSettingFromStr(PropList^[i]^.Index,Prop)
            else begin
              SetSetting(PropList^[i]^.Index, 0);
              Prop := SettingToStr(PropList^[i]^.Index);
            end;
            FDefaultSettings[PropList^[i]^.Index] := Prop;
          end;
        tkFloat:
          if FParameterNames[PropList^[i]^.Index] <> '' then
            raise Exception.CreateFmt('Duplicate property index (%d) for %s: %s and %s',
                                      [PropList^[i]^.Index, ClassName, FParameterNames[PropList^[i]^.Index], PropList^[i]^.Name])
          else begin
            FParameterNames[PropList^[i]^.Index] := PropList^[i]^.Name;
            if Prop <> '' then
              FDefaultParameters[PropList^[i]^.Index] := StrToFloat(Prop)
            else
              FDefaultParameters[PropList^[i]^.Index] := 0;
          end;
      end;
    end;
  end;
  FreeMem(PropList);
  XMLDocument := TXMLDocument.Create;
  TransferProperties(XMLDocument, False, False); { Set to defaults by reading from a blank file }
  XMLDocument.Free;
  Modified := False;
  FNodeIndex := 0;
  if Assigned(AOwner) then
    for i := 0 to AOwner.ComponentCount - 1 do
      if AOwner.Components[i] = Self then
        Break
      else if AOwner.Components[i].ClassType = ClassType then
        Inc(FNodeIndex);
  PerformMNodeNotification(Self, mnCreate, -1);
end;

function TMNode.Count: Integer;
begin
  Result := ComponentCount;
end;

function EncodePropName(const Name: string):string;
  function ReplaceChar(s: string; f: Char; t: string):string;
  var
    p: Integer;
  begin
    repeat
      p := Pos(f, s);
      if p > 0 then begin
        Delete(s, p, Length(f));
        Insert(s, t, p);
      end;
    until p <= 0;
    Result := s;
  end;
begin
  Result := AnsiReplaceStr(Name, '''', 'prime');
  Result := AnsiReplaceStr(Result, '"', 'dprime');
  Result := AnsiReplaceStr(Result, '(', '_');
  Result := AnsiReplaceStr(Result, ')', '_');
  Result := AnsiReplaceStr(Result, ',', '_');
  Result := AnsiReplaceStr(Result, ' ', '_');
  if Result[1] in ['0'..'9'] then
    Result := '_' + Result; { Property names must not begin with numbers }
end;

procedure WriteValue(DOMNode: TDOMNode; PropName, Value, Dflt: string);
var
  Attr: TDOMAttr;
  IsDefault: Boolean;
begin
  if PropName[1] = '(' then
    Value := Dflt; { Do not write bracketed properties }
  IsDefault := CompareText(Value, Dflt) = 0;
  PropName := EncodePropName(PropName);
  Attr := (DOMNode as TDOMElement).GetAttributeNode(PropName);
  if Assigned(Attr) then begin
    if IsDefault then
      RemoveAttributeNode(DOMNode as TDOMElement, Attr)
    else if Attr.Value <> Value then
      Attr.Value := Value;
  end else if not IsDefault then
    { Create new node }
    (DOMNode as TDOMElement)[PropName] := Value;
end;

function ReadValue(DOMNode: TDOMNode; const PropName, Dflt: string; const AltPropName: string; Delete: Boolean): string;
var
  Attr: TDOMAttr;
begin
  Result := Dflt;
  if DOMNode is TDOMElement then begin
    Attr := TDOMElement(DOMNode).GetAttributeNode(EncodePropName(PropName));
    if Assigned(Attr) then
      Result := Attr.NodeValue
    else if (AltPropName <> '') and (AltPropName[1] <> '(') then begin { ignore properties with brackets in name }
      Attr := TDOMElement(DOMNode).GetAttributeNode(EncodePropName(AltPropName));
      if Assigned(Attr) then
        Result := Attr.NodeValue
    end;
    if Assigned(Attr) and Delete then { Delete all read attributes to allow check for unread ones }
      RemoveAttributeNode(TDOMElement(DOMNode), Attr);
  end;
end;

function TMNode.IsSettingEnum(AIndex: Integer): Boolean;
var
  PropInfo: PPropInfo;
  LowestValue: Integer;
begin
  if Assigned(GetSettingsNames(AIndex,LowestValue)) then
    Result := True
  else if AIndex >= FirstDynamicSetting then
    Result := False
  else begin
    PropInfo := GetPropInfo(Self, FSettingNames[AIndex]);
    Result := PropInfo^.PropType^.Kind = tkEnumeration;
  end;
end;

function TMNode.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := False;
end;

function TMNode.IsValidChildClass(MNodeClass: TMNodeClass): Boolean;
begin
  if MNodeClass = nil then
    Result := False
  else if Assigned(Owner) then
    Result := (Owner as TMNode).IsValidChildClass(MNodeClass)
  else
    Result := True;
end;

function TMNode.IsValidChildOf(MNode: TComponent): Boolean;
var
  i: Integer;
begin
  if not Assigned(MNode) then begin
    Result := False;
    Exit;
  end;
  Result := True;
  { Similar to IsAncestorOf(), but does not require self to be valid }
  if Self = MNode then
    Exit;
  for i := 0 to MNode.ComponentCount - 1 do
    if Self = MNode.Components[i] then
      Exit;
  for i := 0 to MNode.ComponentCount - 1 do
    if IsValidChildOf(MNode.Components[i]) then
      Exit;
  Result := False;
end;

procedure TMNode.GetEnumNames(AIndex: Integer; List: TStrings);
var
  PropInfo: PPropInfo;
  TypeData: PTypeData;
  i: Longint;
  SettingsNames: PPChar;
begin
  List.Clear;
  SettingsNames := GetSettingsNames(AIndex,i);
  if Assigned(SettingsNames) then
    while SettingsNames^ <> nil do begin
      if SettingsNames^^ <> #0 then
        List.Add(SettingsNames^);
      Inc(SettingsNames);
    end
  else begin
    PropInfo := GetPropInfo(Self, FSettingNames[AIndex]);
    if not Assigned(PropInfo) or (PropInfo^.PropType^.Kind <> tkEnumeration) then
      raise Exception.Create('Not Enum');
    TypeData := GetTypeData(PropInfo^.PropType{$IFNDEF FPC}^{$ENDIF});
    for i := TypeData^.MinValue to TypeData^.MaxValue do
      List.Add(GetEnumName(PropInfo^.PropType{$IFNDEF FPC}^{$ENDIF},i));
  end;
end;

function TMNode.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  if IsSettingColour(AIndex) then
    if GetColour(AIndex) = clNone then
      Exit;
  Result := SettingToStr(AIndex);
  if IsSettingQuantum(AIndex) and (Result[1] in ['-', '0'..'9']) then
    Result := QuantumNumberToStr(Settings[AIndex],qftJ);
end;

function TMNode.SettingToStr(AIndex: Integer): string;
var
  SettingsNames: PPChar;
  LowestValue, V: Integer;
begin
  if IsSettingColour(AIndex) then
    Result := ColourToStr(GetColour(AIndex))
  else begin
    SettingsNames := GetSettingsNames(AIndex,LowestValue);
    if Assigned(SettingsNames) then begin
      V := GetSetting(AIndex)-LowestValue;
      while (V > 0) and (SettingsNames^ <> nil) do begin
        Dec(V);
        Inc(SettingsNames);
      end;
      if (V <> 0) or (SettingsNames^ = nil) then
        raise Exception.CreateFmt('Bad value for %s.%s: %d',[FullHeirarchicName,SettingNames[AIndex],GetSetting(AIndex)]);
      Result := SettingsNames^;
    end else if AIndex >= FirstDynamicSetting then
      Result := IntToStr(GetSetting(AIndex))
    else
      Result := PropertyToStr(Self, FSettingNames[AIndex]); { Use real name here }
  end;
end;

procedure TMNode.SetSettingFromStr(AIndex: Integer; const s: string);
var
  SettingsNames: PPChar;
  i, LowestValue: Integer;
begin
  if IsSettingColour(AIndex) then
    SetColour(AIndex, StrToColour(s,False))
  else begin
    SettingsNames := GetSettingsNames(AIndex,LowestValue);
    if Assigned(SettingsNames) then begin
      i := 0;
      while SettingsNames^ <> nil do
        if CompareText(SettingsNames^, s) = 0 then begin
          SetSetting(AIndex, i+LowestValue);
          Exit;
        end else begin
          Inc(i);
          Inc(SettingsNames);
        end;
    end;
    if AIndex >= FFirstDynamicSetting then
      SetSetting(AIndex,StrToInt(s))
    else
      SetPropertyFromStr(Self, FSettingNames[AIndex], s); { Use real name here }
  end;
end;

procedure TMNode.SetSettingFromStr(const AName, s: string);
var
  Index: Integer;
begin
  Index := IndexOfSetting(AName);
  if Index <= 0 then
    raise Exception.Create(AName+' not found in '+Name)
  else
    SetSettingFromStr(Index, s);
end;

function ReadMNodeClass(DOMNode: TDOMNode; ParentClass: TClass): TMNodeClass;
var
  NewChildClass: TPersistentClass;
  ErrMsg: string;
begin
  NewChildClass := GetClass('T' + DOMNode.NodeName);
  if (NewChildClass = nil) then
    if Copy(DOMNode.NodeName,1,14) = 'OpenAsymmetric' then
      NewChildClass := GetClass('T' + Copy(DOMNode.NodeName,5,255)) { Compatibility - Delete Open }
    else if Copy(DOMNode.NodeName,1,5) = 'HundA' then
      NewChildClass := GetClass('TLinear' + Copy(DOMNode.NodeName,6,255)); { Compatibility - TLinear was THundA }
  if NewChildClass = nil then
    ErrMsg := 'Can''t find ' + DOMNode.NodeName
  else if not NewChildClass.InheritsFrom(TMNode) then
    ErrMsg := 'Class is not a subclass of TMNode: ' + NewChildClass.ClassName
  else if Assigned(ParentClass) and not ParentClass.InheritsFrom(TMNodeClass(NewChildClass).OwnerType())  then
    ErrMsg := DOMNode.NodeName + ' OwnerClass does not match :'
              + TMNodeClass(NewChildClass).OwnerType().ClassName
              +  ': ' + ParentClass.ClassName
  else begin
    Result := TMNodeClass(NewChildClass);
    Exit; { OK }
  end;
  { Error exit }
  Result := nil;
  {$IFNDEF CONSOLE}
    ShowMessage(ErrMsg);
  {$ELSE}
    raise EReportException.Create(ErrMsg);
  {$ENDIF}
end;

function nUnreadProperties(Name: string; DOMNode: TDOMNode): Integer;
var
  Child: TDOMNode;
  Element: TDOMElement;
  Attr: TDOMAttr;
  n: string;
begin
  Result := 0;
  Child := DOMNode.LastChild;
  while Child <> nil do begin
    Inc(Result);
    WriteLn(Log, 'Unread Item in ', Name, ':', Child.NodeName, ' ', ReadValue(Child, 'Name', ''));
    DOMNode.RemoveChild(Child);
    Child := DOMNode.LastChild;
  end;
  { Check for unread attributes also }
  if not (DOMNode is TDOMElement) then
    Exit;
  { Discard name, if not already read }
  n := ReadValue(DOMNode, 'Name', '');
  Element := DOMNode as TDOMElement;
  while Element.LastAttribute <> nil do begin
    Attr := Element.LastAttribute;
    WriteLn(Log, 'Unknown attribute in ', Name, ' for ', Element.NodeName, ' ', n, ':', Attr.NodeName);
    RemoveAttributeNode(Element,Attr);
    Inc(Result);
  end;
end;

function TMNode.ReadMNode(DOMNode: TDOMNode; out nUnused: Integer): TMnode;
var
  NewChildClass: TMNodeClass;
begin
  NewChildClass := ReadMNodeClass(DOMNode, ClassType);
  if NewChildClass = nil then begin
    Result := nil;
    nUnused := 0;
  end else begin
    if not IsValidChildClass(NewChildClass) then
      raise Exception.Create(NewChildClass.ClassName+' is not a valid child of '+FullHeirarchicName);
    Result := NewChildClass.Create(Self) as TMNode;
    Result.Name := DOMNode.NodeName;
    nUnused := Result.ReadProperties(DOMNode);
  end;
end;

function TMNode.ReadChildProperties(DOMNode: TDOMNode):Integer;
var
  Child, NextChild: TDOMNode;
  nUnused: Integer;
begin
  Result := 0;
  Child := DOMNode.FirstChild;
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if (Child.NodeName <> 'Parameter') and (Child.NodeName <> 'Calibration')
       and not (Child is TDOMCDATASection) then begin
      ReadMNode(Child, nUnused);
      Inc(Result,nUnused);
      DOMNode.RemoveChild(Child);
    end;
    Child := NextChild;
  end;
end;

function TMNode.ReadFloatableDOMNode(Index: Integer; Deflt: string; ParNode: TDOMElement):Integer;
var
  s: string;
begin
  Floating[Index] := False;
  Increment[Index] := 0;
  StdDev[Index] := 0;
  s := Deflt;
  Result := 0;
  if Assigned(ParNode) then begin
    s := ReadValue(ParNode, 'Float', 'false');
    if StrToBoolean(s) then
      Floating[Index] := True;
    s := ReadValue(ParNode, 'Increment', '');
    if s <> ''then
      Increment[Index] := StrToFloatEx(s);
    s := ReadValue(ParNode, 'StdDev', '');
    if s <> ''then
      StdDev[Index] := StrToFloatEx(s);
    s := ReadValue(ParNode, 'Value', Deflt);
    Result := Result + nUnreadProperties(ParNode);
  end;
  Parameters[Index] := StrToFloatEx(s);
end;

procedure TMNode.WriteFloatableValue(Index: Integer; Deflt: string; DOMNode: TDOMNode; DefaultCheck: Boolean);
var
  ParNode: TDOMNode;
  D: Double;
  NonDefault:Boolean;
  s: string;
begin
  NonDefault := True;
  if DefaultCheck and TryStrToFloatEx(Deflt, D) then
    if not IsNanOrInf(GetParameter(Index)) then begin
      NonDefault := Abs(GetParameter(Index)-D)>Abs(D)*1e-16;
      s := FloatToStr(GetParameter(Index));
    end else
      s := 'Nan'
  else
    if IsNanOrInf(GetParameter(Index)) then
      s := Deflt
    else
      s := FloatToStr(GetParameter(Index));
  if NonDefault or Floating[Index] or (Increment[Index] <> 0) or (StdDev[Index] > 0) then begin
    ParNode := DomNode.OwnerDocument.CreateElement('Parameter');
    DomNode.AppendChild(ParNode);
    WriteValue(ParNode, 'Name', ParameterNames[Index], '');
    WriteValue(ParNode, 'Value', s, Deflt);
    if Floating[Index] then
      WriteValue(ParNode, 'Float', 'true', '');
    WriteValue(ParNode, 'Increment', FloatToStr(Increment[Index]), '0');
    if StdDev[Index] > 0 then
      WriteValue(ParNode, 'StdDev', FloatToStr(StdDev[Index]), '0');
  end;
end;

function TMNode.ReadFloatableValue(DOMNode: TDOMNode; Index: Integer; const Prop, Deflt: string):Integer;
var
  Attr: TDOMAttr;
  ParNode: TDOMNode;
begin
  ParNode := DomNode.FirstChild;
  while Assigned(ParNode) do begin
    if CompareText(ParNode.NodeName, 'Parameter') = 0 then begin
      Attr := (ParNode as TDOMElement).GetAttributeNode('Name');
      if Assigned(Attr) and (CompareText(Attr.NodeValue,Prop) = 0) then
        Break; { Found it }
    end;
    ParNode := ParNode.NextSibling;
  end;
  Result := ReadFloatableDOMNode(Index, Deflt, ParNode as TDOMElement);
  if Assigned(ParNode) then
    DOMNode.RemoveChild(ParNode);
end;

function TMNode.TransferProperties(DOMNode: TDOMNode; Writing, AllParameters: Boolean; WithChildren: Boolean):Integer;
var
  i, Index: Integer;
  Props, Prop, Deflt, AltPropName, AlreadySet: string;
  PropInfo: PPropInfo;
  TypeData: PTypeData;
  procedure SetValue(Value: string);
  begin
    WriteValue(DOMNode, AltPropName, Value, Deflt);
  end;
  function GetValue: string;
  begin
    Result := ReadValue(DOMNode, Prop, Deflt, AltPropName);
  end;
begin
  { Note that order in SavedProperties is important for overriden property names;
    property that controls a name must be before the property itself }
  Result := 0;
  Props := SavedProperties;
  AlreadySet := ',';
  while Props <> '' do begin
    i := Pos(',', Props);
    if i > 0 then begin
      Prop := Copy(Props, 1, i-1);
      System.Delete(Props, 1, i);
    end else begin
      Prop := Props;
      Props := '';
    end;
    if Prop = '' then Continue;
    Deflt := '';
    i := Pos('=', Prop);
    if i > 0 then begin
      Deflt := Copy(Prop, i+1, Length(Prop));
      System.Delete(Prop, i, Length(Prop));
    end;
    if Pos(','+Prop+',',AlreadySet) > 0 then
      Continue; { Only take first setting of property }
    AlreadySet := AlreadySet + Prop + ',';
    PropInfo := GetPropInfo(Self, Prop);
    if (PropInfo = nil) {$IFNDEF FPC} or (PropInfo^.Index = 0) {$ENDIF} then
      raise Exception.Create('No such property: '+ClassName+':'+Prop);
    { Check for overridden name }
    AltPropName := Prop;
    Index := PropInfo^.Index;
    if (Index > 0) and (Index < FSettingNames.Count)
       and (PropInfo^.Name = FSettingNames[Index]) then
      AltPropName := SettingNames[Index];
    case PropInfo^.PropType^.Kind of
      tkInteger:
               begin
                 if Deflt = '' then Deflt := '0';
                 if Index >= 0 then begin
                   Deflt := DefaultSettings[Index];
                   if Writing then
                     SetValue(SettingToStr(Index))
                   else
                     SetSettingFromStr(Index, GetValue())
                 end else if Writing then
                   SetValue(IntToStr(GetOrdProp(Self, PropInfo)))
                 else
                   SetOrdProp(Self, PropInfo, StrToInt(GetValue()));
               end;
      tkEnumeration:
               begin
                 if Deflt = '' then begin
                   TypeData := GetTypeData(PropInfo^.PropType{$IFNDEF FPC}^{$ENDIF});
                   Deflt := GetEnumName(PropInfo^.PropType{$IFNDEF FPC}^{$ENDIF}, TypeData^.MinValue);
                 end;
                 if PropInfo^.Index >= 0 then begin
                   Deflt := DefaultSettings[Index];
                   if Writing then
                     SetValue(SettingToStr(Index))
                   else
                     SetSettingFromStr(Index, GetValue());
                 end else
                   if Writing then
                     SetValue(GetEnumProp(Self, PropInfo))
                   else
                     SetEnumProp(Self, PropInfo, GetValue());
               end;
{$IFDEF FPC}
      tkBool:  begin
                 if Deflt = '' then Deflt := 'False';
                 if Writing then begin
                   Deflt := DefaultSettings[Index];
                   if GetOrdProp(Self, PropInfo) <> 0 then
                     SetValue('True')
                   else
                     SetValue('False')
                 end else begin
                   if StrToBoolean(LowerCase(GetValue())) then
                     SetOrdProp(Self, PropInfo, 1)
                   else
                     SetOrdProp(Self, PropInfo, 0);
                  end;
               end;
{$ENDIF}
      tkFloat: begin
                 if Deflt = '' then Deflt := '0';
                 if Index >= 0 then begin
                   Deflt := FloatToStr(DefaultParameters[Index]);
                   if Writing then
                     WriteFloatableValue(Index, Deflt, DOMNode, True)
                   else
                     Result := Result + ReadFloatableValue(DOMNode,Index,Prop,Deflt);
                 end else if Writing then
                   SetValue(FloatToStr(GetFloatProp(Self, PropInfo)))
                 else
                   SetFloatProp(Self, PropInfo, StrToFloatEx(GetValue()));
               end;
      tkString, tkLString {$IFDEF FPC}, tkAString {$ENDIF}:
                   if Writing then
                     SetValue(GetStrProp(Self, PropInfo))
                   else
                     SetStrProp(Self, PropInfo, GetValue());
      tkClass: { Handled as a child MNode or by a TransferProperties descendant }
    else
      raise Exception.Create('Bad property type (' + IntToStr(Ord(PropInfo^.PropType^.Kind)) + ') :' +ClassName+':'+Prop);
    end;
  end;
  if Writing then begin
    { Note that dynamic parameters are written automatically, but not read }
    for i := FFirstDynamicParameter to nParameters do
      WriteFloatableValue(i,FloatToStr(DefaultParameters[i]),DOMNode,not AllParameters);
    for i := FFirstDynamicSetting to nSettings do
      WriteValue(DOMNode, SettingNames[i], SettingToStr(i), DefaultSettings[i]);
    if WithChildren then
      for i := 0 to Count-1 do
        Items[i].WriteProperties(DOMNode)
  end else begin
    { Special handling for dynamic settings }
    for i := FirstDynamicSetting to nSettings do
      SetSettingFromStr(i, ReadValue(DOMNode, SettingNames[i], SettingToStr(i)));
    if WithChildren then
      Result := Result + ReadChildProperties(DOMNode);
  end;
end;

{ Note difference between the two procedures - read takes the current xml node, write takes the parent xml node and returns the current node }
function TMNode.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
begin
  if 'T' + Node.NodeName <> ClassName then begin
    { Compatibility fixes }
    if not ( ((ClassName = 'TLinear') and (Node.NodeName = 'HundA'))
             or ((ClassName = 'TLinearNucleus') and (Node.NodeName = 'HundANucleus'))
             or ((ClassName = 'TLinearPerturbation') and (Node.NodeName = 'HundAPerturbation'))
             or ((ClassName = 'TAsymmetricManifold') and (Node.NodeName = 'OpenAsymmetricManifold'))
             or ((ClassName = 'TAsymmetricTop') and (Node.NodeName = 'OpenAsymmetricTop'))
             or ((ClassName = 'TAsymmetricMolecule') and (Node.NodeName = 'OpenAsymmetricMolecule'))
             or ((ClassName = 'TLevelPlotter') and (Node.NodeName = 'FormSettings'))
             or ((ClassName = 'TSimulation') and (Node.NodeName = 'Mixture')) { Some TSimulation settings in TMixture }
           )
      then
        raise EReportException.Create('Expected '+ClassName+' read T'+Node.NodeName);
  end;
  Result := TransferProperties(Node, False, False);
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TMNode.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
begin
  Result := Parent.OwnerDocument.CreateElement(Copy(ClassName,2,Length(ClassName)));
  Parent.AppendChild(Result);
  TransferProperties(Result, True, AllParameters, WithChildren);
end;

procedure TMNode.ClearCache;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    Items[i].ClearCache;
end;

procedure TMNode.ClearStdDev;
var
  i: Integer;
begin
  for i := 1 to nParameters do
    StdDev[i] := 0;
  for i := 0 to Count-1 do
    (Components[i] as TMNode).ClearStdDev;
end;

function TMNode.FixupProperties: Exception;
var
  i: Integer;
begin
  Result := nil;
  if csDestroying in ComponentState then Exit;
  for i := 0 to Count-1 do begin
    if not (Components[i] is TMNode) then
      raise Exception.CreateFmt('TMNode.FixupProperties:%s', [Components[i].Name]);
    if not (csDestroying in Items[i].ComponentState) then begin
      { Default is just to make sure it is called for all children }
      Result := Items[i].FixupProperties;
      if Assigned(Result) then
        Exit;
    end;
  end;
end;

function TMNode.GetnParameters: Integer;
begin
  Result := FnParameters;
end;

procedure TMNode.SetnParameters(AValue: Integer);
var
  i: Integer;
begin
  { Note that index 0 is never used }
  if FnParameters = AValue then Exit;
  SetLength(FParameters, AValue+1);
  SetLength(FDefaultParameters, AValue+1);
  SetLength(FFloating, AValue+1);
  SetLength(FIncrements, AValue+1);
  SetLength(FStdDev, AValue+1);
  if not Assigned(FParameterNames) then begin
    FParameterNames := TStringList.Create;
    FParameterNames.Add(''); { Unused entry }
  end;
  FParameterNames.Capacity := Max(AValue+1,FParameterNames.Capacity);
  for i := FnParameters+1 to AValue do begin
    FParameters[i] := 0; { Default is 0, unless otherwise specified }
    FParameterNames.Add('');
    FFloating[i] := False;
    FIncrements[i] := 0;
    FStdDev[i] := 0;
  end;
  for i := FnParameters downto AValue+1 do
    if i < FParameterNames.Count then
      FParameterNames.Delete(i);
  FnParameters := AValue;
end;

procedure TMNode.SetnSettings(AValue: Integer);
var
  i: Integer;
begin
  { Note that index 0 is never used }
  if FnSettings = AValue then
    Exit;
  SetLength(FSettings, AValue+1);
  SetLength(FDefaultSettings, AValue+1);
  if not Assigned(FSettingNames) then
    FSettingNames := TStringList.Create;
  FSettingNames.Capacity := Max(AValue+1,FSettingNames.Capacity);
  for i := FnSettings+1 to AValue do begin
    FSettings[i] := 0; { Default is 0, unless otherwise specified }
    FDefaultSettings[i] := '0';
    FSettingNames.Add('')
  end;
  for i := FnSettings downto AValue + 1 do
    if i < FSettingNames.Count then
      FSettingNames.Delete(i);
  FnSettings := AValue;
end;

procedure TMNode.AddHandler(const ACaption, AHint: String; OnClick: TNotifyEvent);
{$IFNDEF CONSOLE}
var
  Item: TMenuItem;
{$ENDIF}
begin
  {$IFNDEF CONSOLE}
  Item := TMenuItem.Create(CurrentMenu);
  Item.Caption := ACaption;
  Item.Hint := AHint;
  Item.OnClick := OnClick;
  CurrentMenu.Add(Item);
{$ENDIF}
end;

{$IFNDEF CONSOLE}
procedure TMNode.AddHandlersToMenu(Menu:TMenuItem);
begin
  CurrentMenu := Menu;
  AddHandlers;
end;
{$ENDIF}

procedure TMNode.AddHandlers;
begin
  { By default, add no handlers }
end;

{ Should not be used on built in parameters, only on those dynamically added }
function TMNode.AddParameter(AName: string; Default: Double; Position: Integer): Integer;
begin
  if IndexOfParameter(AName) >= 0 then
    raise Exception.Create('Duplicate parameter for ' + ClassName + ' ' + Name + ':' + AName);
  SetnParameters(nParameters+1);
  if (Position <= 0) or (Position >= nParameters) then
    Result := nParameters
  else begin
    Result := Position;
    { Inserting a parameter; copy up remaining }
    Move(FDefaultParameters[Result], FDefaultParameters[Result+1], (nParameters-Result)*SizeOf(FDefaultParameters[Result]));
    Move(FParameters[Result], FParameters[Result+1], (nParameters-Result)*SizeOf(FParameters[Result]));
    { Special case for TStringList }
    FParameterNames.Delete(nParameters);
    FParameterNames.Insert(Result,'');
  end;
  FParameterNames[Result] := AName;
  FDefaultParameters[Result] := Default;
  FParameters[Result] := Default;
  FFirstDynamicParameter := Min(FFirstDynamicParameter, Result);
end;

procedure TMNode.SetParameterName(AIndex: Integer; const Value: string);
begin
  FParameterNames[AIndex] := Value;
end;

procedure TMNode.DeleteParameter(AIndex: Integer);
var
  i: Integer;
begin
  for i := AIndex to nParameters-1 do begin
    FParameters[i] := FParameters[i+1];
    FDefaultParameters[i] := FDefaultParameters[i+1];
    FFloating[i] := FFloating[i+1];
    FIncrements[i] := FIncrements[i+1];
    FStdDev[i] := FStdDev[i+1];
  end;
  FParameterNames.Delete(AIndex);
  SetnParameters(nParameters-1);
end;

function TMNode.AddSetting(AName: string; Default: string): Integer;
begin
  if IndexOfSetting(AName) >= 0 then
    raise Exception.Create('Duplicate setting for ' + ClassName + ' ' + Name + ':' + AName);
  SetnSettings(nSettings+1);
  Result := nSettings;
  FSettingNames[Result] := AName;
  FDefaultSettings[Result] := Default;
  SetSettingFromStr(Result, Default);
  FFirstDynamicSetting := Min(FFirstDynamicSetting, nSettings);
end;

procedure TMNode.SetSettingName(AIndex: Integer; const Value: string);
begin
  FSettingNames[AIndex] := Value;
end;

procedure TMNode.DeleteSetting(AIndex: Integer);
var
  i: Integer;
begin
  for i := AIndex to nSettings-1 do begin
    FSettings[i] := FSettings[i+1];
    FDefaultSettings[i] := FDefaultSettings[i+1];
  end;
  FSettingNames.Delete(AIndex);
  SetnSettings(nSettings-1);
end;

procedure TMNode.AdjustDefaults;
begin
  { Adjust parameters or settings if absolutely necessary to be consistent
    with nearby nodes; }
end;

function TMNode.AdjustNodeIndexes(NodeClass: TMNodeClass; var NodeArray: TMNodeArray; AComponent: TComponent; Operation: TOperation; ExpectedNumber: Integer):Integer;
{ Fixup NodeIndex, and optionally handle nodes that are added or removed when a
  setting changes, such as the number of nuclei.
  Fixup Node Array;
  Optionally Add or remove Nodes to agree with ExpectedNumber, if ExpectedNumber >= 0 }
var
  i, j: Integer;
begin
  Result := 0;
  for i := 0 to ComponentCount - 1 do
    if not ((Components[i] = AComponent) and (Operation = opRemove))
       and Components[i].ClassType.InheritsFrom(NodeClass) then begin
      (Components[i] as TMNode).FNodeIndex := Result;
      Inc(Result);
    end;
  if ExpectedNumber >= 0 then begin
    while Result < ExpectedNumber do begin
      ChildClassWithType(NodeClass).Create(Self);
      Inc(Result);
    end;
    if Result > ExpectedNumber then
      for j := ComponentCount-1 downto 0 do
        if Components[j].ClassType.InheritsFrom(NodeClass) then begin
          Components[j].Free;
          Dec(Result);
          if Result <= ExpectedNumber then
            Break;
        end;
  end;
  SetLength(NodeArray, Result);
  j := 0;
  for i := 0 to ComponentCount - 1 do
    if not ((Components[i] = AComponent) and (Operation = opRemove))
       and Components[i].ClassType.InheritsFrom(NodeClass) then begin
      NodeArray[j] := Components[i] as TMNode;
      Inc(j);
    end;
end;

function TMNode.GetParameter(AIndex: Integer): Double; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
{ if AIndex > nParameters then
    Result := 0
  else }
    Result := FParameters[AIndex];
end;

procedure TMNode.SetParameter(AIndex: Integer; const Value: Double);
begin
  ModifyParameter(AIndex, Value, True);
end;

function TMNode.IndexOfParameter(const AName: string): Integer;
begin
  for Result := 1 to nParameters do
    if CompareText(ParameterNames[Result],AName) = 0 then
      Exit;
  Result := -1;
end;

procedure TMNode.ModifyParameter(AIndex: Integer; Value: Double; ChangeModify: Boolean; Notify: Boolean);
begin
  if IsNanOrInf(Value) then begin
    Value := DefaultParameters[AIndex];
    if IsNanOrInf(FParameters[AIndex]) then
      Exit; { Avoid problems with Nan <> Nan }
  end;
  if (FParameters[AIndex] <> Value) or IsNanOrInf(FParameters[AIndex]) then begin
    FParameters[AIndex] := Value;
    if ChangeModify then
      Modified := True;
    if Notify then
      PerformMNodeNotification(Self, mnParameterChange, AIndex);
  end;
end;

function TMNode.GetSetting(AIndex: Integer): Integer; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
{ if AIndex >= nSettings then
    Result := 0
  else }
    Result := FSettings[AIndex];
end;

procedure TMNode.ModifySetting(AIndex: Integer; const AValue: Integer; ChangeModify: Boolean);
begin
  if FSettings[AIndex] <> AValue then begin
    FSettings[AIndex] := AValue;
    if ChangeModify then
      Modified := True;
    PerformMNodeNotification(Self, mnSettingChange, AIndex);
  end;
end;

procedure TMNode.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  ModifySetting(AIndex, AValue, True);
end;

function TMNode.IndexOfSetting(const AName: string): Integer;
begin
  for Result := 1 to FSettingNames.Count-1 do
    if CompareText(SettingNames[Result], AName) = 0 then
      Exit;
  Result := -1;
end;

function TMNode.GetFlag(AIndex: Integer): Boolean;
begin
  Result := GetSetting(AIndex) <> 0;
end;

procedure TMNode.SetFlag(AIndex: Integer; const AValue: Boolean);
begin
  SetSetting(AIndex, Ord(AValue));
end;

{ Should be a class var, but this is essentially equivalent as FPC does not suport them }

var 
  FChildTypes: array of array of TMNodeClass = nil;

class procedure TMNode.RegisterClass(AOwnerType:TMNodeClass);
var
  i, j: Integer;
begin
  Classes.RegisterClass(Self);
  j := -1;
  for i := 0 to High(FChildTypes) do
    if AOwnerType = FChildTypes[i,0] then begin
      j := i;
      Break;
    end;
  if j < 0 then begin
    j := Length(FChildTypes);
    SetLength(FChildTypes, j+1);
    SetLength(FChildTypes[j], 1);
    FChildTypes[j,0] := AOwnerType;
  end;
  for i := 1 to High(FChildTypes[j]) do
    if Self = FChildTypes[j,i] then
      Exit; { Already registered }
  SetLength(FChildTypes[j], Length(FChildTypes[j])+1);
  FChildTypes[j,High(FChildTypes[j])] := Self;
end;

procedure TMNode.GetAll(MNode: TMNode);
begin
  GetDefaults(MNode);
  GetParameters(MNode);
  GetSettings(MNode);
end;

class function TMNode.GetChildTypes(AIndex: Integer): TMNodeClass;
var
  i: Integer;
begin
  for i := 0 to High(FChildTypes) do
    if Self = FChildTypes[i,0] then begin
      Result := FChildTypes[i,AIndex+1];
      Exit;
    end;
  Result := nil;
end;

class function TMNode.nChildTypes: Integer;
var
  i: Integer;
begin
  for i := 0 to High(FChildTypes) do
    if Self = FChildTypes[i,0] then begin
      Result := High(FChildTypes[i]);
      Exit;
    end;
  Result := 0;
end;

function TMNode.nUnreadProperties(DOMNode: TDOMNode): Integer;
begin
  Result := MNode.nUnreadProperties(UniqueName, DOMNode);
end;

class function TMNode.OwnerType: TMNodeClass;
var
  i, j: Integer;
begin
  for i := 0 to High(FChildTypes) do
    for j := 1 to High(FChildTypes[i]) do
      if Self = FChildTypes[i,j] then begin
        Result := FChildTypes[i,0];
        Exit;
      end;
  Result := nil;
end;

class function TMNode.ChildClassWithType(MNodeClass: TMNodeClass): TMNodeClass;
var
  i: Integer;
begin
  for i := 0 to nChildTypes - 1 do
    if GetChildTypes(i).InheritsFrom(MNodeClass) then begin
      Result := GetChildTypes(i);
      Exit;
    end;
  Result := nil;
end;


{ Special case handling for colour, to avoid lots of overrides }

function TMNode.GetColour(AIndex: Integer): TColor;
begin
  Result := GetSetting(AIndex);
end;

function TMNode.GetComponentIndex: Integer;
begin
  Result := inherited ComponentIndex;
end;

procedure TMNode.SetComponentIndex(const Value: Integer);
begin
  inherited ComponentIndex := Value;
end;

function TMNode.GetDefaultParameters(AIndex: Integer): Double;
begin
  Result := FDefaultParameters[AIndex];
end;

procedure TMNode.GetDefaults(MNode: TMNode);
var
  i: Integer;
begin
  for i := 1 to Min(nParameters, MNode.nParameters) do
    FDefaultParameters[i] := MNode.DefaultParameters[i];
  for i := 1 to Min(nSettings, MNode.nSettings) do
    FDefaultSettings[i] := MNode.DefaultSettings[i];
end;

function TMNode.GetDefaultSettings(AIndex: Integer): string;
begin
  Result := FDefaultSettings[AIndex];
end;

procedure TMNode.SetColour(AIndex: Integer; const AValue: TColor);
begin
  SetSetting(AIndex, AValue);
end;

function TMNode.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := False;
end;

function TMNode.GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar;
begin
  if IsSettingColour(AIndex) then
    Result := GetColourNames(LowestValue)
  else begin
    Result := nil; { Default is to use Delphi Mechanism }
    LowestValue := 0;
  end;
end;

procedure TMNode.SetModified(const Value: Boolean);
var
  i: Integer;
begin
  FModified := Value;
  if Modified then begin
    { Propagate Setting Modified upwards }
    if Owner <> nil then
      (Owner as TMNode).Modified := True;
  end else begin
    { Propagate Clearing Modified downwards }
    for i := 0 to ComponentCount-1 do
      if Items[i] is TMNode then
        TMNode(Items[i]).Modified := False;
  end;
end;

destructor TMNode.Destroy;
begin
  if Owner <> nil then
    (Owner as TMNode).Modified := True;
  PerformMNodeNotification(Self, mnDestroy, -1);
  FreeAndNil(FParameterNames);
  FreeAndNil(FSettingNames);
  inherited Destroy;
end;

function TMNode.GetFloating(AIndex: Integer): Boolean;
begin
  if AIndex > High(FFloating) then
    Result := False
  else
    Result := FFloating[Aindex];
end;

procedure TMNode.SetFloating(AIndex: Integer; const Value: Boolean);
begin
  if FFloating[Aindex] <> Value then begin
    FFloating[Aindex] := Value;
    PerformMNodeNotification(Self, mnParameterChange, AIndex);
  end;
end;

function TMNode.GetIncrement(AIndex: Integer): Double;
begin
  if AIndex > High(FIncrements) then
    Result := 0
  else
    Result := FIncrements[AIndex];
end;

procedure TMNode.SetIncrement(AIndex: Integer; const Value: Double);
begin
  if FIncrements[AIndex] <> Value then begin
    FIncrements[AIndex] := Value;
    PerformMNodeNotification(Self, mnParameterChange, AIndex);
  end;
end;

function TMNode.GetStdDev(AIndex: Integer): Double;
begin
  if AIndex > High(FStdDev) then
    Result := 0
  else
    Result := FStdDev[AIndex];
end;

procedure TMNode.SetStdDev(AIndex: Integer; const Value: Double);
begin
  if FStdDev[AIndex] <> Value then begin
    FStdDev[AIndex] := Value;
    PerformMNodeNotification(Self, mnParameterChange, AIndex);
  end;
end;

function TMNode.FullHeirarchicName(Quote: Boolean; RemoveRoot: Boolean): string;
begin
  Result := Name;
  if Quote and (Pos('.', Result) > 0) then
    Result := '"' + Name + '"';
  if Owner<> nil then
    Result := (Owner as TMNode).FullHeirarchicName(Quote) + '.' + Result;
  if RemoveRoot then
   Delete(Result,1,Length(RootNode.FullHeirarchicName(Quote))+1);
end;

function TMNode.FullParameterName(Index: Integer; Quote: Boolean): string;
begin
  Result := FullHeirarchicName(Quote) + '.' + ParameterNames[Index];
  Delete(Result,1,Length(RootNode.FullHeirarchicName(Quote))+1);
end;

function TMNode.UniqueName: string;
begin
  Result := Name;
  if Owner<> nil then
    Result := (Owner as TMNode).UniquePrefix + Result;
end;

function TMNode.UniquePrefix: string;
var
  i, nSameClass: Integer;
begin
  Result := '';
  if not Assigned(Owner) then Exit;
  nSameClass := 0;
  for i := 0 to Owner.ComponentCount-1 do
    if Self.ClassType = Owner.Components[i].ClassType then
      Inc(nSameClass);
  if nSameClass > 1 then
    Result := UniqueName + ' ' + Result
  else
    Result := (Owner as TMNode).UniquePrefix + Result;
end;

function TMNode.RootNode: TMNode;
begin
  Result := Self;
  while Assigned(Result.Owner) do
    Result := Result.Owner as TMNode;
end;

function TMNode.IsAncestorOf(MNode: TMnode):Boolean;
{ True if self is the given MNode or an ancestor of MNode }
begin
  Result := True;
  while MNode <> nil do begin
    if MNode = Self then
      Exit;
    MNode := MNode.Owner as TMNode;
  end;
  Result := False;
end;

function TMNode.CommonAncestor(MNode: TMNode): TMNode;
begin
  Result := Self;
  if not Assigned(MNode) then
    Exit;
  while Assigned(Result) and not Result.IsAncestorOf(MNode) do
    Result := Result.Owner as TMNode;
end;

procedure TMNode.SetName(const Value: TComponentName);
var
  i:Integer;
begin
  if FName <> Value then begin
    FName := Trim(Value);
    { Delete any control characters }
    for i := Length(FName) downto 1 do
      if Ord(FName[i]) < Ord(' ') then
        FName[i] := ' ';
    while Pos('  ',Fname) > 0 do
      FName := AnsiReplaceStr(FName,'  ',' ');
    Modified := True;
    PerformMNodeNotification(Self, mnNameChange, -1);
  end;
end;

function TMNode.FindOrCreateNode(AName: string; NodeClass: TMNodeClass): TMNode;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Count - 1 do
    if Items[i] is NodeClass then begin
      if CompareText(Items[i].Name,AName)=0 then begin
        Result := Items[i]; { Match found }
        Exit;
      end else if Items[i].Name = '' then
        Result := Items[i]; { Found unused/not yet named node }
    end;
  if Result = nil then { No possible matches found }
    Result := NodeClass.Create(Self);
  Result.Name := AName;
end;

procedure TMNode.FixAll(Recurse: Boolean);
var
  i: Integer;
begin
  for i := 1 to nParameters do
    Floating[i] := False;
  if Recurse then
    for i := 0 to Count-1 do
      Items[i].FixAll(Recurse);
end;

procedure TMNode.ScaleParameters(UFrom, UTo: TResonanceUnits);
var
  i: Integer;
  s: string;
begin
  s := ','+LowerCase(DimensionlessParameters)+',';
  for i := 1 to nParameters do
    if Pos(','+LowerCase(ParameterNames[i])+',', s) = 0 then begin
      Parameters[i] := Convert(Parameters[i], UFrom, UTo);
      StdDev[i] := Convert(StdDev[i], UFrom, UTo);
    end;
  for i := 0 to Count-1 do
    Items[i].ScaleParameters(UFrom, UTo);
end;

procedure TMNode.SaveToStream(Stream: TStream; WithLinkedMNodes: Boolean; MNodeList: TObjectList);
{ Save either current MNode, or the MNodeList to the given stream;
  if the MNodeList is given the current MNode will only be saved to the stream if present on the list }
var
  XMLDocument: TXMLDocument;
  ListNode: TDOMNode;
  i: Integer;
  LocalMNodeList: Boolean;
begin
  XMLDocument := TXMLDocument.Create;
  LocalMNodeList := not Assigned(MNodeList);
  try
    if WithLinkedMNodes or Assigned(MNodeList) then begin
      ListNode := XMLDocument.OwnerDocument.CreateElement('List');
      XMLDocument.AppendChild(ListNode);
      if LocalMNodeList then begin
        MNodeList := TObjectList.Create(False);
        LinkedNodesList(MNodeList);
      end;
      for i := 0 to MNodeList.Count - 1 do
        (MNodeList[i] as TMNode).WriteProperties(ListNode);
    end else
      WriteProperties(XMLDocument);
    WriteXMLFile(XMLDocument, Stream);
  finally
    XMLDocument.Free;
    if LocalMNodeList then
      MNodeList.Free;
  end;
end;

{$IFNDEF CONSOLE}
procedure TMNode.CopyToClipboard(WithLinkedMNodes: Boolean; MNodeList: TObjectList);
var
  Stream, Stream1: THugeMemoryStream;
begin
  CheckProperties;
  Stream1 := nil;
  Stream := THugeMemoryStream.Create;
  try
    SaveToStream(Stream, WithLinkedMNodes, MNodeList);
    Stream1 := THugeMemoryStream.Create;
    if Assigned(MNodeList) or not SimpleSaveToStream(Stream1, WithLinkedMNodes) then begin
      Stream.Position := 0;
      Stream1.CopyFrom(Stream, Stream.Size);
    end;
    Clipboard.Open;
    try
      Stream1.CopyToClipboard(CF_TEXT);
      Stream.CopyToClipboard(MNodeClipboardFormat);
    finally
      Clipboard.Close;
    end;
  finally
    Stream.Free;
    Stream1.Free;
  end;
end;
{$ENDIF}

var
  MNodeNotifications: array of OnMnodeChange = nil;

procedure AddMNodeNotification(MNodeChange: OnMnodeChange);
begin
  SetLength(MNodeNotifications, Length(MNodeNotifications)+1);
  MNodeNotifications[High(MNodeNotifications)] := MNodeChange;
end;

procedure RemoveMNodeNotification(MNodeChange: OnMnodeChange);
var
  i: Integer;
begin
  i := 0;
  if Length(MNodeNotifications) = 0 then Exit; { Probably already finalised unit }
//  while (i < Length(MNodeNotifications)) and (@MNodeNotifications[i] <> @MNodeChange) do
  while (i < Length(MNodeNotifications))
        and not ((TMethod(MNodeNotifications[i]).Data = TMethod(MNodeChange).Data)
                 and (TMethod(MNodeNotifications[i]).Code = TMethod(MNodeChange).Code)) do
    Inc(i);
  if i >= Length(MNodeNotifications) then begin
    { Used to use raise here, but don't as if a form fails to load under the LCL
      because of a component problem this is the exception that is raised }
    WriteLn(Output, 'Warning:Removal of non-existant MNodeNotification');
    Exit;
  end;
  while i < High(MNodeNotifications) do begin
    MNodeNotifications[i] := MNodeNotifications[i+1];
    Inc(i);
  end;
  SetLength(MNodeNotifications, Length(MNodeNotifications)-1);
end;

procedure PerformMNodeNotification(Sender: TMnode; Change: TMNodeChange; Index: Integer);
var
  i: Integer;
begin
  i := 0;
  while i <= High(MNodeNotifications) do begin
    { Loop this way to allow array to change during loop }
    MNodeNotifications[i](Sender, Change, Index);
    Inc(i);
  end;
end;

function TMNode.ComponentWithName(const AName: string; AClass: TClass; Exclude: TMNode; SearchChildren: Boolean): TMNode;
var
  i: Integer;
begin
  for i := 0 to ComponentCount-1 do begin
    Result := TMNode(Components[i]);
    if (Result is AClass) and (Result <> Exclude) and ((AName = '*') or (CompareText(Result.GetName,AName)=0))
       and not (csDestroying in Result.ComponentState) then
      Exit;
  end;
  Result := nil;
  if SearchChildren then
    for i := 0 to ComponentCount-1 do begin
      Result := TMNode(Components[i]).ComponentWithName(AName, AClass, Exclude, True);
      if Assigned(Result) then
        Exit;
    end;
end;

function TMNode.nComponentsWithName(const AName: string; out First: TMNode):Integer;
var
  Child, GrandChild: TMNode;
  i: Integer;
begin
  First := nil;
  Result := 0;
  for i := 0 to ComponentCount-1 do begin
    Child := TMNode(Components[i]);
    if CompareText(Child.GetName,AName) = 0 then
      Inc(Result);
    if (Result = 1) and not Assigned(First) then
      First := Child;
    Result := Result + Child.nComponentsWithName(AName,GrandChild);
    if not Assigned(First) then
      First := GrandChild;
  end;
end;

function TMNode.AncestorWithClass(MNodeClass: TMNodeClass): TMNode;
begin
  Result := Self;
  while Assigned(Result) and not (Result is MNodeClass) do
    Result := Result.Owner as TMNode;
end;

function TMNode.CanShift(Shift: Integer): Boolean;
begin
  { If the MNode can be shifted by the given amount }
  Result := Assigned(Owner)
            and (ComponentIndex+Shift >=0)
            and (ComponentIndex+Shift < Owner.ComponentCount);
end;

procedure TMNode.CheckProperties;
var
  E: Exception;
begin
  E := FixupProperties;
  if Assigned(E) then
    raise E;
end;

procedure TMNode.SetComment(const Value: string);
begin
  if FComment <> Value then begin
    FComment := Value;
    Modified := True;
  end;
end;

function TMNode.GetName: TComponentName;
begin
  Result := Fname;
end;

function TMNode.LinkedNodesList(List: TObjectList): Integer;
  { Calculate number of nodes linked to the current one }
  { Optionally add the nodes to the List }
  procedure AddNode(MNode: TMNode);
  begin
    if Assigned(List) and (List.IndexOf(MNode) < 0) then
      List.Add(MNode);
  end;
  function LinkedCount(MNode: TMNode):Integer;
  var
    i: Integer;
    LCount: Integer;
    function CheckLink(BraOrKet: TMNode): Boolean;
    begin
      Result := True;
      if BraOrKet = Self then begin
        AddNode(MNode);
        Inc(LCount);
        Result := False; { Don't count a node twice }
      end;
    end;
  begin
    LCount := 0;
    if MNode is TMNodePair then begin
      if CheckLink(TMNodePair(MNode).Bra) then
        CheckLink(TMNodePair(MNode).Ket);
    end;
    Result := LCount;
    if not Assigned(MNode) or (MNode = Self) then Exit;
    for i := MNode.ComponentCount-1 downto 0 do
      if MNode.Components[i] is TMNode then
        Inc(Result,LinkedCount(TMNode(MNode.Components[i])));
  end;
begin
  AddNode(Self);
  Result := LinkedCount(RootNode);
end;

function TMNode.ChildWithClass(MNodeClass: TMNodeClass; Index: Integer = -1): TMNode;
{ Default is to return the only child with the right class, providing there is only one }
var
  i, n: Integer;
begin
  Result := nil;
  n := 0;
  for i := 0 to Count-1 do
    if Items[i] is MNodeClass then begin
      if Assigned(Result) then begin
        Result := nil;
        Exit; { More than one match }
      end else if n = Index then begin
        Result := Items[i];
        Exit;
      end else if Index < 0 then
        Result := Items[i];
      Inc(n);
    end;
end;

function TMNode.HelpOnSetting(AIndex: Integer): string;
begin
  Result := '';
end;

function TMNode.HelpOnParameter(AIndex: Integer): string;
begin
 Result := '';
end;

class function TMNode.DisplayClassName: string;
var
  i: Integer;
begin
  Result := Copy(ClassName, 2, 255);
  i := 2;
  while i < Length(Result) do begin
    if (Result[i] in ['A'..'Z']) and not (Result[i-1] in ['A'..'Z']) then begin
      Insert(' ', Result, i);
      Inc(i);
    end;
    Inc(i);
  end;
end;

function TMNode.SubPropInfo(const Token: string; out MNode: TMNode;  out ErrMsg: string): PPropInfo;
{ Get PropInfo for Child[.Child].property type strings }
var
  dot, i: Integer;
  Child, Info: string;
begin
  ErrMsg := '';
  Result := nil;
  if (Length(Token) > 2) and (Token[1] in ['''', '"']) then begin
    dot := 2;
    while (Token[dot] <> Token[1]) and (dot < Length(Token)) do
      Inc(Dot);
    Child := Copy(Token,2,dot-2);
    if Dot = Length(Token) then
      Dot := 0 { end of string - no dot }
    else begin
      Inc(Dot);
      if (Length(Token) = Dot) or (Token[Dot] <> '.') then begin
        ErrMsg := 'Bad string in '+Token;
        Exit;
      end;
    end;
  end else begin
    dot := Pos('.',Token);
    if dot > 1 then
      Child := Copy(Token,1,dot-1)
    else
      Child := Token
  end;
  if dot > 1 then begin
    { Child.property - find child }
    MNode := ComponentWithName(Child, TMNode, nil);
    if (MNode = nil) and (nComponentsWithName(Child, MNode) <> 1) then
      MNode := nil; { If not present at current level, look for match at any level, but require only one match }
    if MNode = nil then begin
      ErrMsg := ', expecting one of ';
      for i := 0 to ComponentCount-1 do begin
        if Length(Info) > 50 then begin
          ErrMsg := ErrMsg + ' ...';
          Break;
        end else
        ErrMsg := ErrMsg + ' ' + (Components[i] as TMNode).Name;
      end;
      ErrMsg := Child + ' not found in '+ Token + ErrMsg;
    end else
      Result := MNode.SubPropInfo(Copy(Token,dot+1,Length(Token)), MNode, ErrMsg);
  end else begin
    MNode := Self;
    Result := GetPropInfo(Self, Child);
    if Result = nil then begin
      i := IndexOfParameter(Child);
      if i > 0 then
        Result := Pointer(i) { TODO: Generate proper PropInfo entry }
      else
        ErrMsg := Child  + ' not found in '+ UniqueName;
    end;
  end;
end;

function TMNode.SubPropValue(const Token: string): string;
var
  PropInfo: PPropInfo;
  MNode: TMNode;
  ErrMsg: string;
begin
  PropInfo := SubPropInfo(Token, MNode, ErrMsg);
  if not Assigned(PropInfo) then
    raise Exception.Create('Property not found in '+UniqueName+':' + Token + ' - '+ErrMsg);
  if (PropInfo^.Index > 0) and (PropInfo^.PropType^.Kind <> tkFloat) then
    Result := MNode.SettingToStr(PropInfo^.Index)
  else
    Result := PropertyToStr(MNode, PropInfo);
end;

procedure TMNode.SetSubPropValue(const Token, NewValue: string);
var
  PropInfo: PPropInfo;
  MNode: TMNode;
  Qno: TQuantumNumber;
  ErrMsg: string;
begin
  PropInfo := SubPropInfo(Token, MNode, ErrMsg);
  if not Assigned(PropInfo) then
    raise Exception.Create('Property not found in '+UniqueName+':' + Token + ' - '+ErrMsg);
  if (PropInfo^.Index <= 0) or (PropInfo^.PropType^.Kind = tkFloat) then
    SetPropertyFromStr(MNode, PropInfo, NewValue)
  else if MNode.IsSettingQuantum(PropInfo^.Index) then
    if TryStrToQuantumNumber(NewValue, Qno, qftJ) then
      MNode.SetSetting(PropInfo^.Index, Qno)
    else
      raise EReportException.Create('Bad quantum number for '+Token+':'+NewValue)
  else
    MNode.SetSettingFromStr(PropInfo^.Index, NewValue)
end;

procedure TMNode.HandleAssignment(NameVal: string);
{ Handle Child.Child.property=value type strings }
{ TODO: Does not handle MNodePairs }
var
  Value: string;
  P: Integer;
begin
  { Look for last equal sign, to allow names to contain an equals sign }
  P := Length(NameVal);
  while (P > 0) and (NameVal[P] <> '=') do
    Dec(P);
  Value := Copy(NameVal,P+1,Length(NameVal));
  if (P > 1) and (NameVal[P-1] = ':') then
    Dec(P);
  if P <= 1 then
    raise EReportException.Create('No name to set in '+NameVal);
  SetSubPropValue(Copy(NameVal,1,P-1), Value);
end;

procedure TMNode.CycleUnits;
begin
  RootNode.CycleUnits;
end;

function TMNode.SimpleSaveToStream(Stream: TStream; WithLinkedMNodes: Boolean): Boolean;
begin
  Result := False;
end;

procedure TMNode.StartDerivatives;
var
  I: Integer;
begin
  { Provide a simple mechanism for evaluating a Hamiltonina matrix with all
    parameters zero except one; not thread safe }
  FSavedParameters := FParameters;
  FParameters := nil; { Zero current parameter set }
  SetLength(FParameters, Length(FSavedParameters));
  for I := 0 to Count - 1 do
    (Components[I] as TMNode).StartDerivatives;
end;

procedure TMNode.StopDerivatives;
var
  I: Integer;
begin
  FParameters := FSavedParameters;
  FSavedParameters := nil;
  for I := 0 to Count - 1 do
    (Components[I] as TMNode).StopDerivatives;
end;

function TMNode.DerivativeRequired(Index: Integer): Boolean;
begin
  Result := Floating[Index] or (FSavedParameters[Index] <> 0);
end;

function TMNode.PageName: string;
begin
  { Help on this class is at PageName.htm }
  { Default is basically <classname>.htm (excluding leading T) }
  Result := Copy(ClassType.ClassName,2,255);
end;

procedure TMNode.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
begin
  PrettyPrintHeader(F, PrintLevel, Indent);
  Indent := Indent + '  ';
  PrettyPrintSettings(F, PrintLevel, Indent);
  PrettyPrintParameters(F, PrintLevel, Indent);
  PrettyPrintComponents(F, PrintLevel, Indent);
end;

procedure TMNode.PrettyPrintHeader(var F: Text; PrintLevel: TPrintLevel; Indent: string);
begin
  WriteLn(F, Indent, Copy(ClassName,2,255), ': ', UniqueName);
  if Comment <> '' then
    WriteLn(F, Indent, '  ', Comment);
end;

procedure TMNode.PrettyPrintSettings(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  i, l: Integer;
begin
  l := 0;
  for i := 1 to nSettings do
    if SettingToPrettyStr(i) <> '' then
      l := Max(l,Length(SettingNames[i]));
  for i := 1 to nSettings do
    if SettingToPrettyStr(i) <> '' then
      WriteLn(F, Indent, SettingNames[i], StringOfChar(' ',l+1-Length(SettingNames[i])), SettingToPrettyStr(i));
end;

procedure TMNode.PrettyPrintParameters(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  i, l: Integer;
  s: string;
begin
  l := 0;
  for i := 1 to nParameters do
    if Parameters[i] <> 0 then
      l := Max(l,Length(ParameterNames[i]));
  for i := 1 to nParameters do
    if Parameters[i] <> 0 then begin
      s := ParameterNames[i] + StringOfChar(' ',l+1-Length(ParameterNames[i])) + FloatToStr(Parameters[i]);
      if Floating[i] then begin
        s := s + StringOfChar(' ', l+28-Length(s)) + 'Float';
        if Increment[i] > 0 then
          s := s + ', Increment = ' + FloatToStr(Increment[i]);
        if StdDev[i] > 0 then
          s := s + ', Std Dev = ' + FloatToStr(StdDev[i]);
      end;
      WriteLn(F, Indent, s);
    end;
end;

procedure TMNode.PrettyPrintComponents(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    (Components[i] as TMNode).PrettyPrint(F, PrintLevel, Indent);
end;

{ TMNodePair }

function TMNodePair.BraOwner: TMNode;
begin
  Result := Owner as TMNode;
end;

constructor TMNodePair.Create(AOwner: TComponent);
var
  p: Integer;
begin
  inherited;
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}DoMNodeChange);
  p := Pos(',Name,',SavedProperties);
  if p > 0 then
    Delete(SavedProperties,p+1,5);
end;

destructor TMNodePair.Destroy;
begin
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}DoMNodeChange);
  inherited;
end;

procedure TMNodePair.DoMNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
  procedure Process(var State: TMNode; var StateName: string);
  begin
    if Sender = State then
      case Change of
        mnNameChange: StateName := Sender.Name;
        mnDestroy:    State := nil;
      end;
  end;
begin
  Process(FBra, FBraName);
  Process(FKet, FKetName);
end;

function TMNodePair.FixupProperties: Exception;
var
  BO, KO: TMNode;
  ErrMsg: string;
  procedure FixUpStateS(N1, N2: string);
  begin
    if {not Assigned(FBra) and } Assigned(BO) then
      FBra := BO.ComponentWithName(N1, BraKetClass, nil)
    else
      FBra := nil;
    if { not Assigned(FKet) and } Assigned(KO) then
      FKet := KO.ComponentWithName(N2, BraKetClass, nil)
    else
      FKet := nil;
  end;
begin
  BO := BraOwner;
  KO := KetOwner;
  FBraName := Trim(FBraName);
  FKetName := Trim(FKetName);
  FixUpStates(FBraName, FKetName);
  if not Assigned(FBra) and not Assigned(FKet) and (KO <> BO) then begin
    { Try other way round }
    FixUpStates(FKetName, FBraName);
  end;
  ErrMsg := '';
  if not Assigned(FBra) then
    ErrMsg := FBraName;
  if not Assigned(FKet) then
    ErrMsg := FKetName;
  if not Assigned(FKet) or not Assigned(FBra) then
    Result := ECantFindState.Create(Self, ErrMsg)
  else begin
    { This will swap names over if required }
    FBraName := FBra.Name;
    FKetName := FKet.Name;
    Result := nil;
  end;
  if not Assigned(Result) then
    Result := inherited FixupProperties;
end;

function TMNodePair.KetOwner: TMNode;
begin
  Result := BraOwner;
end;

function TMNodePair.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
begin
  Result := inherited ReadProperties(Node, False);
  FBraName := ReadValue(Node, 'Bra', '');
  FKetName := ReadValue(Node, 'Ket', '');
  { Below for compatibility only }
  if FBraName = '' then
    FBraName := ReadValue(Node, 'Manifold1', '');
  if FBraName = '' then
    FBraName := ReadValue(Node, 'State1', '');
  if FKetName = '' then
    FKetName := ReadValue(Node, 'Manifold2', '');
  if FKetName = '' then
    FKetName := ReadValue(Node, 'State2', '');
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TMNodePair.SetBraKetNameE(const ABraName, AKetName: string; ChangingMNode: TMNode): Exception;
begin
  Result := nil;
  if FBraName <> ABraName then begin
    FBraName := ABraName;
    FBra := nil;
  end;
  if FKetName <> AKetName then begin
    FKetName := AKetName;
    FKet := nil;
  end;
  if not Assigned(FKet) or not Assigned(FBra) then begin
    if Assigned(ChangingMNode) then
      Result := ChangingMNode.FixupProperties
    else
      Result := FixupProperties;
    if not Assigned(Result) then begin
      Modified := True;
      PerformMNodeNotification(Self, mnNameChange, -1);
    end;
  end;
end;

procedure TMNodePair.SetBraKetName(const ABraName, AKetName: string);
var
  E: Exception;
begin
  E := SetBraKetNameE(ABraName, AKetName);
  if Assigned(E) then
    raise E;
end;

function TMNodePair.GetName: TComponentName;
begin
  Result := '<';
  if Assigned(Bra) then
    Result := Result + Bra.Name
  else if FBraName <> '' then
    Result := Result + FBraName
  else
    Result := Result + '?';
  Result := Result + '|' + PairName + '|';
  if Assigned(Ket) then
    Result := Result + Ket.Name
  else if FKetName <> '' then
    Result := Result + FKetName
  else
    Result := Result + '?';
  Result := Result + '>';
end;

procedure TMNodePair.SetName(const Value: TComponentName);
begin
  { Ignore }
end;

function TMNodePair.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
begin
  Result:=inherited WriteProperties(Parent, AllParameters, WithChildren);
  if Assigned(FBra) then
    FBraName := FBra.Name;
  if Assigned(FKet) then
    FKetName := FKet.Name;
  WriteValue(Result, 'Bra', FBraName, '');
  WriteValue(Result, 'Ket', FKetName, '');
end;

procedure TMNode.GetSettings(MNode: TMNode);
var
  i: Integer;
begin
  for i := 1 to Min(nSettings, MNode.nSettings) do
    SetSettingFromStr(i, MNode.SettingToStr(i));
end;

procedure TMNode.GetParameters(MNode: TMNode);
var
  i: Integer;
begin
  for i := 1 to Min(nParameters, MNode.nParameters) do begin
    Parameters[i] := MNode.Parameters[i];
    Floating[i] := MNode.Floating[i];
    Increment[i] := MNode.Increment[i];
    StdDev[i] := MNode.StdDev[i];
  end;
end;

{ ECantFindState }

constructor ECantFindState.Create(AMNodePair: TMNodePair; const AMissingState: string);
begin
  inherited Create('Can''t find State:' + AMissingState);
  FMissingState := AMissingState;
  FMNodePair := AMNodePair;
end;

{ Helper function }

procedure RemoveAttributeNode(Element: TDOMElement; Attr: TDOMAttr);
begin
  Element.RemoveAttributeNode(Attr);
  Attr.Free;
end;

end.
