unit OscFields;
{$I DFS.inc}{Copyright (C) 2009 C M Western; see commandline.pas for license}

{ Objects under TSimulation to describe polarization of applied field }

interface

uses Classes, MNode, QuantumNumber, Mixture;

type

  TOscillatingField = class(TMNode)
  public
    constructor Create(AOwner: TComponent); override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure SetModified(const Value: Boolean); override;
  published
    property Rank: TQuantumNumber index 1 read GetSetting write SetSetting;
    property Strength:  Double index 1 read GetParameter write SetParameter;
  end;

  TDirection = (poRandom, poParallel, poPerpendicular, poCircular);

  TPolarization = class(TOscillatingField)
  private
    function GetDirection(const Index: Integer): TDirection;
    procedure SetDirection(const Index: Integer; const Value: TDirection);
  public
    function GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar; override;
  published
    property Direction: TDirection index 2 read GetDirection write SetDirection;
  end;

function Mfactor(Jp, Mp, J, M: TQuantumNumber; Simulation:TSimulation):Double;

implementation

uses Math, MathSup, Global, nj;

{ TOscillatingField }

constructor TOscillatingField.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Rank=2,Strength=1';
  DimensionlessParameters := DimensionlessParameters + ',Strength';
  inherited;
end;

function TOscillatingField.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Relative strength of field - note that the relative intensity is proportional to the square of this value.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TOscillatingField.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Rank of field - ' + GlobalMixture.QuantumNumberToStr(2) + ' for a normal single photon transition.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TOscillatingField.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex = 1;
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

procedure TOscillatingField.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    RootNode.ClearCache;
end;

{ TPolarization }

const
  DirectionNames: array [0..4] of PChar = ('Random', 'Parallel', 'Perpendicular', 'Circular', nil);

function TPolarization.GetDirection(const Index: Integer): TDirection;
begin
  Result := TDirection(GetSetting(Index));
end;

function TPolarization.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if SettingNames[AIndex] = 'Direction' then
    Result := @DirectionNames[0];
end;

procedure TPolarization.SetDirection(const Index: Integer; const Value: TDirection);
begin
  SetSetting(Index, Ord(Value));
end;

{ M dependence to transtion moment }
function Mfactor(Jp, Mp, J, M: TQuantumNumber; Simulation:TSimulation):Double;
  function Tkq(Rank, p: TQuantumNumber):Double;
  begin
    Result := Minus(Jp-Mp) * threej(Jp, -Mp, Rank, p, J, M);
  end;
  function PolFactor(Direction: TDirection; Rank: TQuantumNumber):Double;
  begin
    { Choose right values of p; scale factors to give same total
      absorption independent of polarization }
    case Direction of
      poRandom:
        begin
          Result := Tkq(Rank,Min(Max(Mp-M,-2),2));
          if Mp = M then
            Result := Result/Sqrt(Two)
          else
            Result := Result/Two;
        end;
      poParallel: Result := Tkq(Rank,0);
      poPerpendicular:
        if Mp > M then
          Result := Tkq(Rank,2)/Sqrt(Two)
        else
          Result := Tkq(Rank,-2)/Sqrt(Two);
      poCircular: Result := Tkq(Rank,2);
    end;
  end;
var
  i: Integer;
begin
  if Assigned(Simulation) then
    if Simulation.ComponentCount > 0 then begin
      Result := 0;
      for i := 0 to Simulation.ComponentCount - 1 do
        if Simulation.Components[i] is TPolarization then
          with TPolarization(Simulation.Components[i]) do
            Result := Result + Strength * PolFactor(Direction, Rank)
    end else
      Result := PolFactor(poRandom,2) { Default if polarisation not specified }
  else
    Result := TKq(2,0) { Normally simple static field }
end;

initialization
  TPolarization.RegisterClass(TSimulation);
end.
