unit Plate;

interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Graphics, MathSup, Classes, MNode;

function PixelSum(Rect: TRect; Bitmap: TBitmap; out nPixels: Integer): Double;
function DeterminePlateAngle(Rect: TRect; Bitmap: TBitmap; Log: Boolean; var Positions: TDoubleArray):Double;
procedure ScanPlateSpectrum(D: TBitmap; Angle: Double; Bounds: array of TPoint;
                            Threshold: Double; var Frequency, Intensity: TDoubleArray);

implementation

uses Math, Types, Graphs;

{$IFNDEF MSWINDOWS}
function GetRValue(Color: TColor): Integer;
begin
  Result := Color and $ff;
end;

function GetGValue(Color: TColor): Integer;
begin
  Result := (Color shr 8) and $ff;
end;

function GetBValue(Color: TColor): Integer;
begin
  Result := (Color shr 16) and $ff;
end;

{$ENDIF}

function PixelSum(Rect: TRect; Bitmap: TBitmap; out nPixels: Integer): Double; overload;
var
  SweepY: Boolean;
  x, y, x1, y1, Xstep, Ystep, DeltaX, DeltaErr, Error: Integer;
  procedure HandlePoint;
  var
    Pixel: TColor;
    xp, yp: Integer;
  begin
    if SweepY then begin
      yp := x;
      xp := y;
    end else begin
      yp := y;
      xp := x;
    end;
    Pixel := Bitmap.Canvas.Pixels[xp,yp];
    if (Pixel = -1) or (Pixel = clAqua) then
      Exit; { Out of range }
    Result := Result + GetRValue(Pixel) + GetGValue(Pixel) + GetBValue(Pixel);
    Inc(nPixels);
    Bitmap.Canvas.Pixels[xp,yp] := clAqua;
  end;
begin
  Result := 0;
  nPixels := 0;
  { Step though line with Bresenham's algorithm }
  with Rect do begin
    SweepY := Abs(Top-Bottom) > Abs(Left-Right);
    if SweepY then begin
      x := Bitmap.Height - Top - 1;
      y := Left;
      x1 := Bitmap.Height - Bottom - 1;
      y1 := Right;
    end else begin
      y := Bitmap.Height - Top - 1;
      x := Left;
      y1 := Bitmap.Height - Bottom - 1;
      x1 := Right;
    end;
  end;
  DeltaX := abs(x1 - x);
  DeltaErr := abs(y1 - y);
  Error := 0;
  if x < x1 then
    Xstep := 1
  else
    Xstep := -1;
  if y < y1 then
    Ystep := 1
  else
    Ystep := -1;
  HandlePoint;
  while x <> x1 do begin
    Inc(x, Xstep);
    Inc(Error, DeltaErr);
    if 2*Error >= DeltaX then begin
      Inc(y, Ystep);
      Dec(Error, DeltaX);
    end;
    HandlePoint;
  end;
end;

function NetChanges(Values: array of Integer; var Mi, Mx: Integer):Integer;
var
  i, h: Integer;
begin
  Result := 0;
  for i := 0 to High(Values) do begin
    if Values[i] > Mx then Mx := Values[i];
    if Values[i] < Mi then Mi := Values[i];
  end;
  h := Length(Values) div 2;
  for i := 0 to h-1 do
    Result := Result + Values[i+1]-Values[i];
  for i := h to High(Values)-1 do
    Result := Result - Values[i+1]+Values[i];
end;

function PeakPosition(Values: array of Integer; Inverted: Boolean; var Position: Double; Threshold: Integer):Boolean;
var
  i, V, n: Integer;
  SumI, SumIX : Double;
begin
  SumI := 0;
  SumIX := 0;
  n := 0;
  for i := 0 to High(Values) do begin
    V := Values[i]-Threshold;
    if Inverted then
      V := -V;
    if V > 0 then begin
      Inc(n);
      SumI := SumI + V;
      SumIX := SumIX + V * i;
    end;
  end;
  Result := n > 1;
  if Result then
    Position := SumIX / SumI
  else
    Position := -1;
end;

function Angle(Positions: array of Double):Double;
var
  r, NPoints: Longint;
  x, y: Double;
  sx, sxx, sy, sxy, x0, y0, bDelta, Delta: Extended;
begin
  NPoints := 0;
  sx := 0;
  sxx := 0;
  sy := 0;
  sxy := 0;
  x0 := High(Positions) div 2;
  y0 := 0;
  for r := 0 to High(Positions) do begin
    if Positions[r] < 0 then
      Continue;
    y := Positions[r];
    if NPoints = 0 then
      y0:= y;  { Assumed mean, to reduce rounding errors }
    NPoints := NPoints + 1;
    { Least Squares Sum }
    x := r - x0;
    y := y - y0;
    sx := sx + x;
    sy := sy + y;
    sxx := sxx + x*x;
    sxy := sxy + x*y;
  end;
  if NPoints >= 2 then begin
    Delta := NPoints*sxx - sx*sx;
    bDelta := NPoints*sxy - sx*sy;
    Result := ArcTan2(bDelta,Delta);
  end else
    SetNan(Result);
end;

function DeterminePlateAngle(Rect: TRect; Bitmap: TBitmap; Log: Boolean; var Positions: TDoubleArray):Double;
var
  SweepY: Boolean;
  Mx, Mi, x, y, cycle, Changes: Integer;
  Pixel: TColor;
  Values: array of Integer;
begin
  with Rect do begin
    SweepY := Abs(Top-Bottom) > Abs(Left-Right);
    if SweepY then begin
      SetLength(Positions, Bottom-Top+1);
      SetLength(Values, Right-Left+1);
    end else begin
      SetLength(Values, Bottom-Top+1);
      SetLength(Positions, Right-Left+1);
    end;
    Mx := -MaxInt;
    Mi := MaxInt;
    Changes := 0;
    for cycle := 0 to 1 do begin
      for y := 0 to High(Positions) do begin
        for x := 0 to High(Values) do begin
          Pixel := Bitmap.Canvas.Pixels[Rect.Left,Rect.Top];
          Values[x] := GetRValue(Pixel) + GetGValue(Pixel) + GetBValue(Pixel);
          if Log and (Cycle = 0) then
            WriteLn(Log, Rect.Left, ' ', Rect.Top, ' ', Values[x]);
          if SweepY then
            Inc(Rect.Left)
          else
            Inc(Rect.Top);
        end;
        if Cycle=0 then
          Changes := Changes + NetChanges(Values,Mi,Mx)
        else begin
//          if Log then
//            Writeln(Log, y, ' ', NetChanges(Values),Mi,Mx);
          PeakPosition(Values, Changes<0, Positions[y], (Mx+Mi) div 2);
          if Log then
            WriteLn(Log, y, ' ', Positions[y]);
        end;
        if SweepY then begin
          Inc(Rect.Top);
          Dec(Rect.Left,Length(Values));
        end else begin
          Inc(Rect.Left);
          Dec(Rect.Top,Length(Values))
        end;
      end;
      if SweepY then
        Dec(Rect.Top,Length(Positions))
      else
        Dec(Rect.Left,Length(Positions));
    end;
  end;
  Result := Angle(Positions);
  if SweepY and not IsNan(Result) then
    Result := Pi/2 - Result;
end;

type
  { Parametric description of line passing through Point.
    General equation is Point.X + Grad.X*t, Point.Y + Grad.Y*t }
  TParamLine = record
    Point: TPoint;
    Grad:  TDPoint;
  end;

function Point(Line: TParamLine; T: Double):TDPoint; overload;
begin
  with Line do begin
    Result.X := Point.X + T * Grad.X;
    Result.Y := Point.Y + T * Grad.Y;
  end;
end;

function ParamLine(A, B: TPoint):TParamLine; overload;
{ Line though two points }
begin
  Result.Point := A;
  with Result.Grad do begin
    x := B.X - A.X;
    y := B.Y - A.Y;
  end;
end;

function ParamLine(A: TPoint; Angle: Double):TParamLine; overload;
{ Line though point at Angle A to x axis }
begin
  Result.Point := A;
  Angle := DegToRad(Angle);
  with Result.Grad do begin
    x := Cos(Angle);
    y := Sin(Angle);
  end;
end;

function Intersect(A, B: TParamLine):TDPoint;
var
  Num, Den: Double;
begin
  Den := A.Grad.X*B.Grad.Y - A.Grad.Y*B.Grad.X;
  Num := B.Grad.Y*(B.Point.X-A.Point.X) - B.Grad.X*(B.Point.Y-A.Point.Y);
  Result := Point(A,Num/Den);
end;

function Point(P: TDPoint):TPoint; overload;
begin
  Result.X := Round(P.X);
  Result.Y := Round(P.Y);
end;

function Xvalue(Y:Double; Line: TParamLine):Double;
{ Given Y, find X }
begin
  Y := (Y - Line.Point.Y)/Line.Grad.Y;
  Result := Line.Point.X + Y * Line.Grad.X;
end;

function Yvalue(X:Double; Line: TParamLine):Double;
{ Given X, find Y }
begin
  X := (X - Line.Point.X)/Line.Grad.X;
  Result := Line.Point.Y + X * Line.Grad.Y;
end;

function PixelSum(Scan: TParamLine; nScan:Integer; D: TBitmap; ScanY: Boolean; out nPixels: Integer): Double; overload;
var
  Pixel: TColor;
  Error, DError: Double;
  Xstep, Ystep: Integer;
  I: Integer;
  Other: Boolean;
begin
  Result := 0;
  nPixels := 0;
  if Scan.Grad.X < 0 then
    Xstep := -1
  else
    Xstep := 1;
  if Scan.Grad.Y < 0 then
    Ystep := -1
  else
    Ystep := 1;
  Error := 0;
  if ScanY then
    DError := Abs(Scan.Grad.X/Scan.Grad.Y)
  else
    DError := Abs(Scan.Grad.Y/Scan.Grad.X);
  for i := 0 to nScan - 1 do begin
    Pixel := D.Canvas.Pixels[Scan.Point.X,Scan.Point.Y];
    if (Pixel <> -1) and (Pixel <> clAqua) then begin
      Result := Result + GetRValue(Pixel) + GetGValue(Pixel) + GetBValue(Pixel);
      Inc(nPixels);
      D.Canvas.Pixels[Scan.Point.X,Scan.Point.Y] := clAqua;
    end;
    Error := Error + DError;
    Other := Error > 0.5;
    if Other then
      Error := Error - 1;
    if ScanY then begin
      Inc(Scan.Point.Y,YStep);
      if Other then
        Inc(Scan.Point.X,Xstep);
    end else begin
      Inc(Scan.Point.X,XStep);
      if Other then
        Inc(Scan.Point.Y,Ystep);
    end;
  end;
end;

procedure ScanPlateSpectrum(D: TBitmap; Angle: Double; Bounds: array of TPoint; Threshold: Double; var Frequency, Intensity: TDoubleArray);
var
  i, j, nDFp, nScan, nPixels: Integer;
  v: Double;
  Mid, IThresh: Integer;
  ScanY: Boolean;
  Base, Scan, Top: TParamLine;
  E: TPoint;
  f: Double;
  procedure ScanLine(i: Integer);
  var
    ChangeSign: Boolean;
  begin
    Scan.Point := Point(Point(Base,i/nDFp));
    E := Point(Intersect(Scan, Top));
    if ScanY then begin
      nScan := E.Y-Scan.Point.Y;
      ChangeSign := nScan*Scan.Grad.Y < 0;
      f := Xvalue(Mid,Scan); { Pixel position in middle of plate }
    end else begin
      nScan := E.X-Scan.Point.X;
      f := Yvalue(Mid,Scan); { Pixel position in middle of plate }
      ChangeSign := nScan*Scan.Grad.X < 0;
    end;
    nScan := Abs(nScan)+1;
    if ChangeSign then begin
      Scan.Grad.X := -Scan.Grad.X;
      Scan.Grad.Y := -Scan.Grad.Y;
    end;
  end;
begin
  Base := ParamLine(Bounds[0], Bounds[1]);
  Top := ParamLine(Bounds[2],Bounds[3]);
  ScanY := Abs(Base.Grad.Y) < Abs(Base.Grad.X);
  if ScanY then begin
    nDFp := Bounds[1].X-Bounds[0].X;
    Mid := D.Height div 2;
  end else begin
    nDFp := Bounds[1].Y-Bounds[0].Y;
    Mid := D.Width div 2;
  end;
  nDFp := ABs(nDFp);
  SetLength(Frequency, nDFp+1);
  SetLength(Intensity, nDFp+1);
  Scan := ParamLine(Bounds[0],Angle); { Start point is overwritten below }
  ScanLine(0);
  IThresh := nScan;
  ScanLine(nDFp);
  IThresh := Round(Min(IThresh, nScan)*Threshold);
  j := 0;
  for i := 0 to nDFp do begin
    ScanLine(i);
    v := PixelSum(Scan, nScan, D, ScanY, nPixels);
    if nPixels > IThresh then begin
      Frequency[j] := f;
      Intensity[j] := v/nPixels;
      Inc(j);
    end;
  end;
  SetLength(Frequency, j);
  SetLength(Intensity, j);
end;

end.
