unit Plotter;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses MNode, Mixture, Global, GraphM, GraphRange, Overlay, SpectrumPlot, ExpCan;

procedure MakePlot(Canvas: TExportableCanvas);
function MakeUpperPlot(G: TGraphM; Simulation: TSimulation; out LastFortrat: TFortratPlot; var FortratPlotRanges: TPlotRanges): Boolean;
function MakeLowerPlot(G: TGraphM; Mixture: TMixture; Simulation: TSimulation; Overlays: TOverlays;
                       var SpectrumPlotRanges: TPlotRanges; var PlotColours: TColours):string;

implementation

uses Classes, SysUtils, Math, Units, Obs, Fitter, MathSup, Worker, ExportG, Profile;

function XLabel(Simulation: TSimulation):string;
begin
  if Simulation.Foffset = 0 then
    XLabel := TypeOfUnit(Simulation.PlotUnits) + '/' + NameOfUnit(Simulation.PlotUnits)
  else
    XLabel := Format('(%s-%g)/%s', [TypeOfUnit(Simulation.PlotUnits), Simulation.Foffset, NameOfUnit(Simulation.PlotUnits)]);
end;

function MakeUpperPlot(G: TGraphM; Simulation: TSimulation; out LastFortrat: TFortratPlot; var FortratPlotRanges: TPlotRanges): Boolean;
var
  Problem: Exception;
  xl, xu: Double;
begin
  Result := True;
  G.Start;
  xl := Simulation.Fmin-Simulation.Foffset;
  xu := Simulation.Fmax-Simulation.Foffset;
  SetLength(FortratPlotRanges,0);
  LastFortrat := TFortratPlot.Create(Simulation.PlotUnits);
  with LastFortrat do
    try
      MakeFromSimulation(Simulation, not Simulation.Mixture.NoLineList, Problem);
      if Assigned(Problem) then
        G.Title1(Problem.Message);
      G.Map(xl, xu, IMin-1, IMax+1);
      G.Axis(xl, xu, IMin-1, true, XLabel(Simulation));
      Show(G, NullPlotRange, Simulation.GraphStyle in [gsOffsetAll, gsInvertedAll], False, FortratPlotRanges);
      Result := not Assigned(Problem);
    finally
      Problem.Free;
    end;
  G.Finish;
end;

var
  MakeLowerPlotTime: string = 'MakeLowerPlot';

function MakeLowerPlot(G: TGraphM; Mixture: TMixture; Simulation: TSimulation; Overlays: TOverlays;
                       var SpectrumPlotRanges: TPlotRanges; var PlotColours: TColours):string;
var
  xl, xu: Double;
  procedure ShowObservedLines;
  var
    i, nPlotted: Integer;
    v, vcalc: Double;
    s: string;
  begin
    nPlotted := 0;
    for i := 0 to High(Observations) do begin
      if Observations[i].ObsType = obParameter then
        Continue;
      if IsPopDist(Mixture.Simulations[0].IntensityUnits) then begin
        if Observations[i].ObsType <> obELower then
          Continue;
      end else if Observations[i].ObsType <> obLine then
        Continue;
      v := Observations[i].Frequency;
      try
        if not CalculateObs(Observations[i],vcalc,s,False,Observations[i].ObsType) then
          vcalc := v;
      except
        vcalc := v;
      end;
      v := Convert(v,Mixture.Units,Simulation.PlotUnits);
      vcalc := Convert(vcalc,Mixture.Units,Simulation.PlotUnits) - Mixture.Simulations[0].Foffset;
      if Within(v,xl,xu) or Within(vcalc,xl,xu) then begin
          G.StyleP(Observations[i].Colour, G.LineStyle.Width, G.LineStyle.Pattern);
          G.Point(v, 1.05);
          G.Join(vcalc, 1.02);
          Inc(nPlotted);
        end;
    end;
    if nPlotted > 0 then
      AddPlotRange(SpectrumPlotRanges, PlotRange(G, 1.02, 1.05, 1.02, 1.05, -2));
  end;
var
  yi, yx: Double;
  AnySpectrum, UseOverlayAxis, AllArbitrary: Boolean;
  SpectrumPlot: TSpectrumPlot;
  Problem: Exception;
  IU: TIntensityUnits;
  PR: TPlotRange;
begin
  Result := '';
  TStart(MakeLowerPlotTime);
  SetLength(SpectrumPlotRanges,0);
  xl := Simulation.Fmin-Simulation.Foffset;
  xu := Simulation.Fmax-Simulation.Foffset;
  AnySpectrum := Simulation.ShowSum or Simulation.ShowParts;
  if Simulation.ShowParts then
    SpectrumPlot := TMultiColouredSpectrumPlot.Create(Simulation)
  else if AnySpectrum then
    SpectrumPlot := TSpectrumPlot.Create(Simulation.PlotUnits,IsPopDist(Simulation.IntensityUnits))
  else
    SpectrumPlot := nil;
  if Assigned(SpectrumPlot) then
  try
    SpectrumPlot.MakeFromSimulation(Simulation, not Simulation.Mixture.NoLineList, Problem);
    SpectrumPlot.GetColoursUsed(PlotColours);
    G.Start;
    if Assigned(Problem) then begin
      G.Title1(Problem.Message);
      Problem.Free;
      AnySpectrum := True;
    end else
      AnySpectrum := SpectrumPlot.nLinesInRange > 0;
    G.Map(xl, xu, -0.05, 1);
    if Assigned(Overlays) then begin
      Overlays.FixupProperties;
      G.Style(1);
      UseOverlayAxis := Simulation.GraphStyle <> gsNoOffset; { Overlays separate? }
      IU := Simulation.IntensityUnits;
      if not UseOverlayAxis then begin
        { Overlays on simulation, so normally scale on simulation }
        AnySpectrum := AnySpectrum and SpectrumPlot.FindLimits;
        UseOverlayAxis := not AnySpectrum; { no simulation, so scale on overlay }
      end else if AnySpectrum then
        G.Map(xl, xu, -1.1, 1); { Overlay will be in 0..1 range, simulation in -1..0 }
      if UseOverlayAxis then begin
        IU := Overlays.IntensityUnits;
        PR := PlotRange(G, 0, 1, 0, 1, 1);
      end else
        if IsArbitraryStyle(IU)
           or IsArbitraryStyle(Overlays.IntensityUnits)
           or not Overlays.LimitsInRange(Simulation,Simulation.PlotUnits,yi,yx,IU,AllArbitrary)
           or AllArbitrary
        then
          PR := PlotRange(G, 0, 1, SpectrumPlot.IMin, SpectrumPlot.IMax)
        else
          PR := PlotRange(G, 0, 1, Min(SpectrumPlot.IMin,yi), Max(SpectrumPlot.IMax,yx));
      Overlays.Show(G, Simulation, Simulation.PlotUnits, IU, PR, SpectrumPlotRanges, UseOverlayAxis, Simulation.GraphStyle in [gsOffsetAll, gsInvertedAll]);
    end else
      UseOverlayAxis := True; { Actually not, but force recompute }
    if UseOverlayAxis then begin
      SpectrumPlot.FindLimits;
      PR := PlotRange(G, 0, 1, SpectrumPlot.IMin, SpectrumPlot.IMax);
    end;

    ShowObservedLines;
    if Assigned(Overlays) then
      case Simulation.GraphStyle of
        gsOffset, gsOffsetAll:     G.Map(xl, xu, -0.1, 2);
        gsInverted, gsInvertedAll: G.Map(xl, xu, 1.1, -1);
                                   { Default is -0.05, 1}
      end
    else
      case Simulation.GraphStyle of
        gsInverted, gsInvertedAll: G.Map(xl, xu, 1.05, 0);
                                   { Default is -0.05, 1}
      end;
    G.Style(0);
    G.Axis(xl, xu, G.Ymin, true, XLabel(Simulation));
    if AnySpectrum then begin
      if Simulation.GraphStyle <> gsNoOffset { or Simulation.ShowParts } then
        SpectrumPlot.Show(G, NullPlotRange, Simulation.GraphStyle in [gsOffsetAll, gsInvertedAll], True, SpectrumPlotRanges)
      else if Simulation.ShowParts then
        SpectrumPlot.Show(G, PR, False, True, SpectrumPlotRanges)
      else begin
        PR.Tag := -9;
        SpectrumPlot.Show(G, PR, False, True, SpectrumPlotRanges);
      end;
      Result := Format('%d lines in range with total intensity %s',
                              [SpectrumPlot.nLinesInRange, DtosM(SpectrumPlot.IntensityInRange,12,-1)]);
      if HasUnits(Simulation.IntensityUnits) then
        Result := Result + ' ' + NameOfUnit(Simulation.IntensityUnits, True, Simulation.PlotUnits);
    end;
    G.Finish;
  finally
    SpectrumPlot.Free;
  end;
  TEnd(MakeLowerPlotTime);
end;

procedure MakePlot(Canvas: TExportableCanvas);
var
  Simulation: TSimulation;
  AnySpectrum: Boolean;
  G: TGraphM;
  xl, xu, yl, yu: Double;
  SpectrumPlotRanges: TPlotRanges;
  PlotColours: TColours;
  LastFortrat: TFortratPlot;
begin
  Simulation := GlobalMixture.Simulations[0];
  AnySpectrum := Simulation.ShowSum or Simulation.ShowParts;
  if not (AnySpectrum or Simulation.ShowFortrat) then
    Exit;
  if GlobalMixture.PrintLevel >= plDetail then
    GlobalMixture.PrettyPrint(Log, GlobalMixture.PrintLevel);
  G := TGraphM.Create(nil);
  try
    if Canvas is TNullCanvas then
      G.TextStream := TNullCanvas(Canvas).Stream;
    G.CurrentCanvas := Canvas;
    G.PspaceArg(xl,xu,yl,yu);
    if AnySpectrum then begin
      if Simulation.ShowFortrat then
        G.Pspace(xl,xu,yl,0.5-(1-yu));
      MakeLowerPlot(G,GlobalMixture,Simulation,GlobalOverlays,SpectrumPlotRanges,PlotColours);
    end;
    if Simulation.ShowFortrat then begin
      if AnySpectrum then
        G.Pspace(xl,xu,0.5+yl,yu);
      MakeUpperPlot(G, Simulation, LastFortrat, SpectrumPlotRanges);
      LastFortrat.Free; { Not used }
    end;
  finally
    G.Free;
  end;
end;

end.
