unit SampleMols;
{ Simple sample molecules }
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Mixture, MNode;

type
  TSampleProc = procedure (Mixture:TMixture);

procedure AddSampleLinearMolecule(Mixture: TMixture);
procedure AddSampleAsymmetricTop(Mixture: TMixture);
procedure AddSampleSymmetricTop(Mixture: TMixture);
procedure AddSampleVibratingMolecule(Mixture: TMixture);

implementation

uses HundA, OAsymT, SymT, Abelian, Vib;

procedure AddSampleLinearMolecule(Mixture: TMixture);
var
  Mol, MNode: TMNode;
  Manifold: TManifold;
  Species: TSpecies;
begin
  Mixture.Name := 'Sample';
  Species := TSpecies.Create(Mixture);
  with Species do begin
    Name := 'Species';
    JMax := 100;
  end;
  Mol := TLinearMolecule.Create(Species);
  Mol.Name := 'LinearMolecule';
  Manifold := TLinearManifold.Create(Mol);
  with Manifold do begin
    Name := 'Ground';
    Initial := True;
    with TLinear.Create(Manifold) do begin
      Name := 'v=0';
      B := 1;
    end;
  end;
  Manifold := TLinearManifold.Create(Mol);
  with Manifold do begin
    Name := 'Excited';
    with TLinear.Create(Manifold) do begin
      Name := 'v=1';
      Origin := 1000;
      B := 1;
    end;
  end;
  MNode := TTransitionMoments.Create(Mol, 'Excited', 'Ground');
  with TSphericalTransitionMoment.Create(MNode, 'v=1', 'v=0') do begin
    Strength := 1;
    SetSettingFromStr(IndexOfSetting('Component'),'auto');
  end;
end;

procedure AddSampleAsymmetricTop(Mixture: TMixture);
var
  Mol: TAsymmetricMolecule;
  MNode: TMNode;
  Manifold: TManifold;
  Species: TSpecies;
begin
  Mixture.Name := 'Sample';
  Species := TSpecies.Create(Mixture);
  with Species do begin
    Name := 'Species';
    JMax := 60;
  end;
  with Mixture.Simulations[0] do begin
    Fmin := 1900;
    Fmax := 2100;
  end;
  Mol := TAsymmetricMolecule.Create(Species);
  with Mol do begin
    Name := 'AsymmetricTop';
    PointGroup := Abelian.C2v;
    C2zAxis := aAxis;
  end;
  Manifold := TAsymmetricManifold.Create(Mol);
  with Manifold do begin
    Name := 'Ground';
    Initial := True;
    with TAsymmetricTop.Create(Manifold) do begin
      Name := 'v=0';
      A := 9;
      B := 1.2;
      C := 1;
    end;
  end;
  Manifold := TAsymmetricManifold.Create(Mol);
  with Manifold do begin
    Name := 'Excited';
    with TAsymmetricTop.Create(Manifold) do begin
      Name := 'v=1';
      Origin := 2000;
      A := 9;
      B := 1.2;
      C := 1;
    end;
  end;
  MNode := TTransitionMoments.Create(Mol, 'Excited', 'Ground');
  with TCartesianTransitionMoment.Create(MNode, 'v=1', 'v=0') do begin
    Strength := 1;
  end;
end;

procedure AddSampleSymmetricTop(Mixture: TMixture);
var
  Mol: TSymmetricMolecule;
  MNode: TMNode;
  Manifold: TManifold;
  Species: TSpecies;
begin
  Mixture.Name := 'Sample';
  Species := TSpecies.Create(Mixture);
  with Species do begin
    Name := 'Species';
    JMax := 120;
  end;
  with Mixture.Simulations[0] do begin
    Fmin := 1440;
    Fmax := 1560;
  end;
  Mol := TSymmetricMolecule.Create(Species);
  with Mol do begin
    Name := 'SymmetricTop';
  end;
  Manifold := TSymmetricManifold.Create(Mol);
  with Manifold do begin
    Name := 'Ground';
    Initial := True;
    with TSymmetricTop.Create(Manifold) do begin
      Name := 'v=0';
      B := 0.5;
      C := 5;
    end;
  end;
  Manifold := TSymmetricManifold.Create(Mol);
  with Manifold do begin
    Name := 'Excited';
    with TSymmetricTop.Create(Manifold) do begin
      Name := 'v=1';
      Origin := 1500;
      B := 0.5;
      C := 5;
    end;
  end;
  MNode := TTransitionMoments.Create(Mol, 'Excited', 'Ground');
  with TSphericalTransitionMoment.Create(MNode, 'v=1', 'v=0') do begin
    Strength := 1;
    SetSettingFromStr(IndexOfSetting('Component'),'auto');
  end;
end;

procedure AddSampleVibratingMolecule(Mixture: TMixture);
var
  Mol: TVibratingMolecule;
  MNode: TMNode;
  Manifold: TManifold;
  Species: TSpecies;
begin
  Mixture.Name := 'Sample';
  Species := TSpecies.Create(Mixture);
  with Species do begin
    Name := 'Species';
    JMax := 2;
  end;
  with Mixture.Simulations[0] do begin
    Fmin := 47000;
    Fmax := 64000;
  end;
  Mol := TVibratingMolecule.Create(Species);
  with Mol do begin
    Name := 'VibratingMolecule';
    nModes := 1;
  end;
  Manifold := TVibrationalManifold.Create(Mol);
  with Manifold do begin
    Name := 'Ground';
    Initial := True;
    with TElectronicState.Create(Manifold) do begin
      Name := 'X';
      Mode(0).Omega := 1500;
      Mode(0).vMax := 3;
    end;
  end;
  Manifold := TVibrationalManifold.Create(Mol);
  with Manifold do begin
    Name := 'Excited';
    with TElectronicState.Create(Manifold) do begin
      Name := 'A';
      Origin := 50000;
      Mode(0).Omega := 1000;
      Mode(0).vMax := 15;
    end;
  end;
  MNode := TTransitionMoments.Create(Mol, 'Excited', 'Ground');
  with TMultidimensionalFCF.Create(MNode, 'A', 'X') do begin
    Strength := 1;
    Mode(0).Displacement := 0.5;
  end;
end;

end.

