unit SimpleQnos;
{$I DFS.inc}{Copyright (C) 2010 C M Western; see commandline.pas for license}
{ Simple quantum number and spfit quantum number helper routines }
interface

uses QuantumNumber, StaticField, Mixture, MathSup;

type
  TSimpleQnos = record
    QNP, QN: array [0..9] of Integer;
  end;

function ComputeQ(AStateU, AStateL: TState; NQN, RemoveSpins: Integer): Integer;
function ConvertQuantumNumbers(var ManifoldU: TManifold; var StateU: TState; var JU:TQuantumNumber; var SymU: Shortint; var IndexU: Smallint;
                               var ManifoldL: TManifold; var StateL: TState; var JL:TQuantumNumber; var SymL: Shortint; var IndexL: Smallint;
                               var Name: string;
                               Q, NQN: Integer; Qnos: TSimpleQnos;
                               const ExField: TStaticFields;
                               const MapUpper, MapLower: TIntegerArray;
                               const QuantumNumberFormat: TQuantumNumberFormat;
                               const RemoveSpins: Integer;
                               { Last two only needed/used for symmetric tops }
                               BranchRank: TQuantumNumber;
                               out NotKZero: Boolean):Boolean;
procedure AddToMap(var Map: TIntegerArray; n: Integer; ms: string);

type
  TnnTable = array of TQuantumNumberArray;
  { First index is NN; second is quantum number }
function MakennTable(ElectronSpin:TQuantumNumber; Molecule: TMolecule; RemoveSpins: Integer): TnnTable;
procedure ClearnnTableCache;

implementation

uses SysUtils, Math, OAsymT, SymT, HundA, Worker;

var { Cache results of last call to MakennTable for convenience } { TODO: not thread safe }
  LastElectronSpin: TQuantumNumber;
  LastMolecule: TMolecule;
  nnTable: TnnTable;
  LastRemoveSpins: Integer;

procedure ClearnnTableCache;
begin
  LastMolecule := nil;
end;
  
{ To limit the number of quantum numbers for multiple spins,
  a composite quantum number, nn, is used, which is essentially
  an index into a table containing offsets from the total quantum number F.
  For example, for the case S=1/2, I1=1, I2=I3=1/2 with I2 and I3 equivalent
  and coupled to give Itot = 0 and 1, the table is:
    nn    N     J     F1   Itot
     0  -2.5  -2.0  -1.0   1.0
     1  -1.5  -1.0   0.0   0.0
     2  -1.5  -1.0   0.0   1.0
     3  -1.5  -1.0  -1.0   1.0
     4  -1.5  -2.0  -1.0   1.0
     5  -0.5   0.0   0.0   0.0
     6  -0.5  -1.0   0.0   0.0
     7  -0.5   0.0   1.0   1.0
     8  -0.5   0.0   0.0   1.0
     9  -0.5  -1.0   0.0   1.0
    10  -0.5   0.0  -1.0   1.0
    11  -0.5  -1.0  -1.0   1.0
    12   0.5   1.0   0.0   0.0
    13   0.5   0.0   0.0   0.0
    14   0.5   1.0   1.0   1.0
    15   0.5   0.0   1.0   1.0
    16   0.5   1.0   0.0   1.0
    17   0.5   0.0   0.0   1.0
    18   0.5   0.0  -1.0   1.0
    19   1.5   1.0   0.0   0.0
    20   1.5   2.0   1.0   1.0
    21   1.5   1.0   1.0   1.0
    22   1.5   1.0   0.0   1.0
    23   2.5   2.0   1.0   1.0
  The order of the table is:
    N = min .. max
      Itot = min .. max
        and all other spins max .. min
}

function MakennTable(ElectronSpin:TQuantumNumber; Molecule: TMolecule; RemoveSpins: Integer): TnnTable;
var
  SpinTotal, N: TQuantumNumber;
  i, nPerm, Perm, nItot, nSpins, NucOffset: Integer;
  Itots, Itot: TQuantumNumberArray;
  Spins, Offsets, BaseValue, HighValue: TQuantumNumberArray;
  ItotIndex: array of Integer;
  OK: Boolean;
  function OffsetOK(i:Integer):Boolean;
  { Is Offset i OK ? }
  begin
    if ItotIndex[i-1] < 0 then begin
      { Check spin before }
      Result := Abs(Offsets[i]-Offsets[i-1]) <= Spins[i-1];
      if Result and (i<nSpins-1) then
        { Check spin after }
        if ItotIndex[i] < 0 then
          Result := Abs(Offsets[i+1]-Offsets[i]) <= Spins[i];
    end else
      Result := True; { Enforced elsewhere }
  end;
begin
  { TODO: Needs testing for RemoveSpins > 0 }
  nSpins := Molecule.nNuclei - RemoveSpins;
  NucOffset := 0;
  if nSpins < 0 then
    nSpins := 0
  else if ElectronSpin > 0 then begin
    Inc(nSpins);
    Inc(NucOffset);
  end;
  SetLength(Spins, nSpins); { All spins in one place for simplicity }
  SetLength(Offsets, nSpins); { Offsets from base spin }
  SetLength(BaseValue, nSpins); { Lowest possible value }
  SetLength(HighValue, nSpins); { Lowest possible value }
  Setlength(ItotIndex, nSpins); { Index into Itot(s) arrays, or -1 }
  for i := 0 to nSpins - 1 do
    ItotIndex[i] := -1; { Initially flag no spins as equivalent }
  Spins[0] := ElectronSpin;
  SpinTotal := ElectronSpin;
  nPerm := ElectronSpin + 1; { Number of combinations of spin additions/projections }
  nItot := 0; { Number of sets of equivalent spins }
  for i := 0 to Molecule.nNuclei - 1  - RemoveSpins do begin
    SpinTotal := SpinTotal + Molecule.Nuclei[i].Spin;
    nPerm := nPerm*(Molecule.Nuclei[i].Spin+1);
    Spins[NucOffset+i] := Molecule.Nuclei[i].Spin;
    if Molecule.Nuclei[i].AsNext then begin
      if (i = 0) or not Molecule.Nuclei[i-1].AsNext then begin
        Inc(nItot); { First nucleus of equivalent set }
        SetLength(Itots, nItot);
        Itots[nItot-1] := Molecule.Nuclei[i].Spin;
      end;
      Inc(Itots[nItot-1],Molecule.Nuclei[i].Spin);
      ItotIndex[NucOffset+i] := nItot-1;
      ItotIndex[NucOffset+i+1] := nItot-1;
    end;
  end;
  SetLength(Result, nPerm);
  Perm := 0;
  SetLength(Offsets, nSpins);
  SetLength(Itot, nItot);
  N := -SpinTotal-2; { Acutally offset of N from F }
  while N <= SpinTotal do begin
    Inc(N,2);
    { Cycle through possible Itot values }
    for i := 0 to nItot - 1 do { Initialize to all at smallest values }
      Itot[i] := SmallestProjection(Itots[i]);
    repeat
      Offsets[0] := N;
      BaseValue[0] := N;
      HighValue[0] := N;
      { Intitialise offsets to highest possible values }
      { and record highest and lowest possible values }
      for i := 0 to nSpins - 2 do begin
        { This logic requires assumes equivalent spins at the end }
        if ItotIndex[i] >= 0 then begin
          Offsets[i+1] := Itot[ItotIndex[i]];
          BaseValue[i+1] := Offsets[i+1];
        end else begin
          Offsets[i+1] := Offsets[i] + Spins[i];
          BaseValue[i+1] := BaseValue[i] - Spins[i];
        end;
        HighValue[i] := Offsets[i];
      end;
      repeat
        { Do we have a valid set of offsets? }
        OK := True;
        for i := 1 to nSpins - 1 do begin
          OK := OffsetOK(i);
          if not OK then
            Break;
        end;
        if OK then      
          { check on last spin }
          if ItotIndex[nSpins-1] >= 0 then
            { TODO: This test assumes only 2 equivalent spins }
            OK := Itot[ItotIndex[nSpins-1]] >= Abs(Offsets[nSpins-2])
          else
            OK := Spins[nSpins-1] >= Abs(Offsets[nSpins-1]);
//        if OK then begin
//          Write(Log, OK:5, Perm:5);
//          for i := 0 to High(Offsets) do
//            Write(Log, QuantumNumberToStr(Offsets[i], qftJ):6);
//          WriteLn(Log);
//        end;
        if OK then begin
          { Add to target list }
          Result[Perm] := Copy(Offsets, 0, nSpins);
          Inc(Perm);
        end;
        { Bump up to next combinations of offsets }
        i := 1;
        repeat
          if ItotIndex[i-1] < 0 then begin
            Dec(Offsets[i],2);
            if Offsets[i] >= BaseValue[i] then
              Break;
            { Need to step next spin - reset this one }
            Offsets[i] := HighValue[i];  
          end;
          Inc(i);
        until i = nSpins;
      until i = nSpins;
      { Bump up to next combination of Itot values }
      i := 0;
      repeat
        Inc(Itot[i],2);
        if Itot[i] > Itots[i] then begin
          Itot[i] := SmallestProjection(Itots[i]);
          Inc(i);
        end else
          Break;
      until i = nItot; { Stop if last Itot reached maximum }
    until i = nItot;
  end;
  if Perm <> nPerm then { Should not happen }
    raise Exception.CreateFmt('MakennTable: %d %d', [Perm, nPerm]);
end;

procedure AddToMap(var Map: TIntegerArray; n: Integer; ms: string);
var
  v, c, j: Integer;
begin
  c := Pos('.',ms);
  if c > 0 then
    { Manifold.State format }
    v := (StrToInt(Copy(ms,1,c-1))+1)*2000 + StrToInt(Copy(ms,c+1,Length(ms)))
  else
    v := StrToInt(ms);
  c := Length(Map);
  if c <= n then begin
    SetLength(Map,n+1);
    for j := c to n - 1 do
      Map[j] := j;
  end;
  Map[n] := v;
end;

function ComputeQ(AStateU, AStateL: TState; NQN, RemoveSpins: Integer): Integer;
var
  nSpins, nBase: Integer;
begin
  { Work out spfit quantum number format index }
  nSpins := AStateL.Molecule.nNuclei - RemoveSpins;
  if (AStateL.S > 0) or (AStateU.S > 0) then
    Inc(nSpins);
  { Work out Q }
  { TODO: Does not yet handle equivalent spins }
  if AStateL is TAsymmetricTop then
    nBase := 3
  else if AStateL is TSymmetricTop then
    nBase := 2
  else if AStateL is TLinear then
    nBase := 1
  else
    raise NotYet;
  Result := nBase;
  if (AStateL.Molecule.nStates > 1) and (NQN > nBase) then begin
    Inc(Result,11); { vibrational quantum number present }
    Inc(nBase);
  end;
  if nSpins + nBase > 6 then
    Inc(Result,40);
end;

function ConvertQuantumNumbers(var ManifoldU: TManifold; var StateU: TState; var JU:TQuantumNumber; var SymU: Shortint; var IndexU: Smallint;
                               var ManifoldL: TManifold; var StateL: TState; var JL:TQuantumNumber; var SymL: Shortint; var IndexL: Smallint;
                               var Name: string;
                               Q, NQN: Integer; Qnos: TSimpleQnos;
                               const ExField: TStaticFields;
                               const MapUpper, MapLower: TIntegerArray;
                               const QuantumNumberFormat: TQuantumNumberFormat;
                               const RemoveSpins: Integer;
                               { Last two only needed/used for symmetric tops }
                               BranchRank: TQuantumNumber;
                               out NotKZero: Boolean):Boolean;

  function MatchHyperfine(BasisState: TBasisState; const Q: array of Integer):Boolean;
  var
    i, ioff: Integer;
    FF, Fa : TQuantumNumber;
  begin
    Result := False;
    ioff := NQN-1;
    FF := Q[ioff];
    for i := StateU.Molecule.nNuclei-1-RemoveSpins downto 0 do begin
      if StateU.Molecule.Nuclei[i].AsNext then
        Fa := BasisState.HyperfineState[i]
      else begin
        Fa := FF;
        Dec(FF,BasisState.HyperfineState[i]);
      end;
      if Q[ioff] <> Fa then
        Exit;
      Dec(ioff);
    end;
    Result := True;
  end;

  function GetQnos(
             vflag: Integer;        { 1 if v quantum number present, 0 otherwise }
             AggregateQnos:Boolean; { If nn aggregate hyperfine quantum number present }
             nBase:Integer          { 1 for linear, 2 for symmetric top, 3 for asymmetric top }
                                    { The number of non J, v and hyperfine quantum numbers }
            ): Boolean;
    function GetSym(var F: TQuantumNumber; const Q: array of Integer; var Manifold: TManifold; var State: TState;
                    var ASym: Shortint; var Number: Smallint; var AName: string; const Map: TIntegerArray;
                    OtherSym: Integer):Boolean;
    var
      OtherQno: array [0..1] of Integer;
      Hyperfine: THyperfineBasis;
      N, J, FF: TQuantumNumber;
      nn, SpinOffset, i, Sindex: Integer;
      klToggle: Boolean;
    begin
      Result := False;
      { Find state }
      if Assigned(State) then
        Sindex := State.NodeIndex
      else
        Sindex := 0;
      klToggle := False; { Symmetric tops only }
      if vflag > 0 then begin
        Sindex := Q[nBase] div 2;
        if Sindex <= High(Map) then begin
          Sindex := Map[Sindex];
          if Sindex >= 1000 then begin
            i := Sindex div 1000;
            klToggle := Odd(i); { Symmetric tops only - flag for other half of state }
            if i > 1 then
              Manifold := Manifold.Molecule.Manifold[(i div 2)-1];
          end;
          Sindex := Sindex mod 1000;
        end;
      end;
      if Sindex >= Manifold.nStates then begin
        WriteLn(Log, 'Requested State ', Sindex, ' but only ', Manifold.nStates, ' available');
        Exit;
      end;
      State := Manifold.State[Sindex];

      N := Q[0];
      F := Q[NQN-1];
      SetLength(Hyperfine, Max(State.Molecule.nNuclei-RemoveSpins,0));
      if AggregateQnos then begin
        if (State.Molecule <> LastMolecule) or (State.S <> LastElectronSpin) or (RemoveSpins <> LastRemoveSpins) then begin
          nnTable := MakennTable(State.S, State.Molecule, RemoveSpins);
          LastElectronSpin := State.S;
          LastMolecule := State.Molecule;
          LastRemoveSpins := RemoveSpins;
        end;
        nn := Q[nBase+vflag] div 2; { Was assumed to be a quantum number and doubled }
        if N <> F+nnTable[nn][0] then begin
          WriteLn(Log,'Inconsistent nn quantum number: ', nn, ' ', QuantumNumberToStr(N,qftJ), ' ', QuantumNumberToStr(F,qftJ));
          Exit;
        end;
        if State.S > 0 then
          SpinOffset := 1
        else
          SpinOffset := 0;
        J := F + nnTable[nn][SpinOffset];
        for i := 0 to State.Molecule.nNuclei - 2 - RemoveSpins do
          if State.Molecule.Nuclei[i].AsNext then
            Hyperfine[i] := nnTable[nn][i+SpinOffset+1]
          else
            Hyperfine[i] := F + nnTable[nn][i+SpinOffset+1];
      end else begin
        if NQN = nBase+vflag+Max(State.Molecule.nNuclei-RemoveSpins,0) then begin
          J := Q[0]; { J Not given, so same as N }
          if State.Molecule.nNuclei-RemoveSpins <= 0 then
            F := J; { F not given }
        end else
          J := Q[nBase+vflag];
        { Copy remaining hyperfine quantum numbers }
        for i := 0 to State.Molecule.nNuclei - 2 - RemoveSpins do
          Hyperfine[i] := Q[NQN+i-State.Molecule.nNuclei];
      end;
      { Convert from quantum numbers to differences }
      FF := J;
      if State.Molecule.nNuclei-RemoveSpins > 0 then
        Hyperfine[High(Hyperfine)] := F;
      for i := 0 to State.Molecule.nNuclei - 1 - RemoveSpins do
        if State.Nucleus(i).AsNext then
          { Leave quantum number as is }
        else begin
          Hyperfine[i] := Hyperfine[i]-FF;
          FF := Hyperfine[i]+FF; { Value of qunantum number }
        end;
      Asym := -1; { Search all symmetries by default }
      OtherQno[0] := Low(TQuantumNumber);
      OtherQno[1] := Low(TQuantumNumber);
      for i := 0 to nBase-2 do
        OtherQno[i] := Q[i+1];
      if State is TSymmetricTop then begin
        OtherQno[0] := Abs(Q[1]); { |K| }
        NotKZero := NotKZero and (OtherQno[0] <> 0);
        if (OtherQno[0] > 0) and (State.Molecule.SymName(TSymmetricTop(State).Symmetry)[1] = 'E') then begin
          if klToggle then
            OtherQno[1] := 2
          else
            OtherQno[1] := -2
        end;
      end else if State is TLinear then begin
        ASym := 0;
        if nBase = 3 then begin
          { J, Omega, Lambda given; Lambda<0 for negative parity }
          if Q[2] < 0 then
            Asym := 1;
          OtherQno[1] := Low(TQuantumNumber);
          N := -1;
        end else if nBase = 2 then begin
          { N, Lambda given; <0 for negative parity }
          if Q[1] < 0 then
            Asym := 1;
          OtherQno[0] := Low(TQuantumNumber);
        end else if nBase = 1 then begin
          { Sigma state}
          if Odd(N div 2) xor (TLinear(State).Lambda < 0) then
            ASym := 1 { - parity }
        end;
      end;
      AName := Manifold.FindStateLabel(Number,ASym,F,OtherQno,QuantumNumberFormat,ExField,Asym,J,N,State,RemoveSpins,Hyperfine);
      Result := ASym >= 0;
      if not Result then begin
        if (Number = -3) and (State is TSymmetricTop) then begin
          { Two matches - first check for allowed transition if information available }
          { TODO: May need another pass if choice is only available for one of the states }
          if (BranchRank >= 0) and (OtherSym >= 0) then
            if TSymmetricMolecule(Manifold.Molecule).DipoleAllowed(OtherSym,-ASym-1) then
              ASym := -ASym-1
            else { Assume the alternate choice is allowed }
              ASym := ASym-1
          else
            { Two matches - assume sign of K is given in file, and take second if K < 0 }
            if (Q[1] >= 0) xor (TSymmetricMolecule(Manifold.Molecule).CharacterTable.IReps[TSymmetricTop(State).Symmetry].rho > 0) then
              ASym := -ASym-1 { Take lowest symmetry number if K >= 0 }
            else
              ASym := ASym-1; { Otherwise take higher symmetry number }
          AName := Manifold.FindStateLabel(Number,ASym,F,OtherQno,QuantumNumberFormat,ExField,Asym,J,N,State,RemoveSpins,Hyperfine);
          Result := ASym >= 0;
        end;
        if not Result then begin
          WriteLn(Log,AName);
          Manifold.FindStateLabel(Number,ASym,F,OtherQno,QuantumNumberFormat,ExField,-1,J,N,State,RemoveSpins,Hyperfine,True);
        end;
      end;
      if Result and ((Manifold.nStates > 1) or (NQN > nBase+vflag)) then
        Number := 0; { Force eigenvalue search for hyperfine }
    end;
    var
      NameU, NameL: string;
  begin
    NotKZero := True;
    Result := GetSym(JU, Qnos.QNP, ManifoldU, StateU, SymU, IndexU, NameU, MapUpper, -1);
    Result := Result and GetSym(JL, Qnos.QN, ManifoldL, StateL, SymL, IndexL, NameL, MapLower, SymU);
//      if Result and ((IndexU = 0) or (IndexL = 0)) then
    Name := ':' + NameU + ' - ' + NameL + ':' + Name;
  end;

begin
  if Assigned(StateU) and not Assigned(ManifoldU) then
    ManifoldU := StateU.Manifold;
  if Assigned(StateL) and not Assigned(ManifoldL) then
    ManifoldL := StateL.Manifold;
  { This will need adjustment }
  case Q of
       { Symmetric tops }
    13,33: Result := not GetQnos(1,False,2);
     2,22: Result := not GetQnos(0,False,2);
// ns=??  62,73: Result := not GetQnos(0,False,3);
       53: Result := not GetQnos(1,True,2);
       42: Result := not GetQnos(0,True,2);

       { Linear Sigma States }
    12,32: Result := not GetQnos(1,False,1);
  1,72,21: Result := not GetQnos(0,False,1);
// ns=??  61: Result := not GetQnos(0,False,1);
       52: Result := not GetQnos(1,True,1);
       41: Result := not GetQnos(0,True,1);

       { Linear Case (a) }
     8,19: with Qnos do begin
             QN[0] := QN[0]-1;
             QN[1] := QN[1]-1;
             QNP[0] := QNP[0]-1;
             QNP[1] := QNP[1]-1;
             if Q = 19 then
               Result := not GetQnos(1,False,3)
             else
               Result := not GetQnos(0,False,3)
           end;

//  1,72,21: Result := not GetLinearQnos(0);
//    12,32: Result := not GetLinearQnos(1);
//        2: if StateL is TLinear then
//             Result := not GetLinearQnos(0, 1)
//           else
//             Result := not GetSymmetricTopQnos(0);
//       13: if StateL is TLinear then
//             Result := not GetLinearQnos(1, 1)
//           else
//             Result := not GetSymmetricTopQnos(1);
//    3,23: if StateL is TAsymmetricTop then
//             Result := not GetAsymmetricTopQnos(0)
//           else
//             raise NotYet;
//       14: Result := not GetAsymmetricTopQnos(1);
       { Asymmetric tops }
       14: Result := not GetQnos(1,False,3);
        3: if StateL is TAsymmetricTop then
             Result := not GetQnos(0,False,3)
           else
             raise NotYet; { Hund's case (a) }
       23: Result := not GetQnos(0,False,3);
// ns=??  63: Result := not GetQnos(0,False,3);
       54: Result := not GetQnos(1,True,3);
       43: Result := not GetQnos(0,True,3);
  else
    WriteLn(Log, 'Quantum number format ', Q, ' not implemented');
    Result := True; { Not implemented }
  end;
end;

end.
