unit EigLA;
{$I DFS.inc}
interface

uses MatrixVec, Vector;

procedure Eigen(Eigenvals: TColumnVector; Matrix: TMatrix; EvReqd: Boolean; Offset: Integer = 0);
{ Return eigenvalues and optionally eigenvectors of Matrix;
  Eigenvalue matrix (E) overwrites original matrix (M);
  Each column of E contains an eigenvector and Transpose(E) * M * E is a diagonal matrix
}

type
  TRangeWanted = (rwAll, rwInterval, rwIndex);

function dsyevr(
  n: Integer;
  var a, w, z: Double;
  WantEigenVectors: Boolean = True;
  UpperTriangle: Boolean = False;
  Range: TRangeWanted = rwAll;
  VL: Double = 0;  VU:Double = 0;
  IL: Integer = 0; IU: Integer = 0;
  AbsTol:Double = 0
): Integer;

implementation

{$IFNDEF DESIGNPACKAGE}
{$DEFINE USE_LAPACK}
{$ENDIF}
{ Note that a quick test has ATLAS slower than a simple lapack + unoptimised blas }

{$IFDEF USE_LAPACK}

uses Clock, Huge, SysUtils, f2c, Worker, {$IFNDEF DESIGNPACKAGE} Lapack, {$ENDIF} Profile;

function dsyevr(
  n: Integer;
  var a, w, z: Double;
  WantEigenVectors: Boolean = True;
  UpperTriangle: Boolean = False;
  Range: TRangeWanted = rwAll;
  VL: Double = 0;  VU:Double = 0;
  IL: Integer = 0; IU: Integer = 0;
  AbsTol:Double = 0
): Integer;
var
  jobz, rnge, uplo, ss: Char;
  isuppz, iwork: array of Integer;
  work: array of Double;
  twork: Double;
  info, tiwork, lwork, liwork: Integer;
begin
  Result := 0;
  if WantEigenVectors then
    jobz := 'V'
  else
    jobz := 'N';
  case Range of
    rwAll: rnge := 'A';
    rwInterval: rnge := 'V';
    rwIndex: rnge := 'I'
  end;
  if UpperTriangle then
    uplo := 'U'
  else
    uplo := 'L';
  if AbsTol = 0 then begin
    ss := 'S';
    AbsTol := dlamch_(ss, 1);
  end;
  if Range <> rwAll then
    SetLength(isuppz, n);
  lwork := -1;
  liwork := -1;
  dsyevr_(jobz, rnge, uplo, n, a, n, VL, VU, IL, IU, AbsTol, Result, w, z, n, tiwork, twork, lwork, liwork, liwork, info, 1, 1, 1);
  if info = 0 then begin
    SetLength(work, Round(twork));
    SetLength(iwork, liwork);
    lwork := Length(work);
    if (Range = rwAll) or ((Range = rwIndex) and (IU-IL = n-1)) then begin
      SetLength(isuppz, 2*n);
      dsyevr_(jobz, rnge, uplo, n, a, n, VL, VU, IL, IU, AbsTol, Result, w, z, n, isuppz[0], work[0], lwork, iwork[0], liwork, info, 1, 1, 1)
    end else
      dsyevr_(jobz, rnge, uplo, n, a, n, VL, VU, IL, IU, AbsTol, Result, w, z, n, tiwork, work[0], lwork, iwork[0], liwork, info, 1, 1, 1);
  end;
  if info < 0 then
    xerbla_('dsyevr', info, 6)
  else if info > 0 then
    Result := -Info;
end;

var
 EigenValuesTime: string = 'Matrix Diagonalization';
 
procedure Eigen(Eigenvals: TColumnVector; Matrix: TMatrix; EvReqd: Boolean; Offset: Integer);
var
  info: Integer;
  EigenVectors: TVector;
  PEigenVectors, PEigenVals: PDouble;
begin
  if Matrix.ColCount = 0 then exit;
  TStart(EigenValuesTime);
  if EvReqd then begin
    EigenVectors := TMatrix.Create(Matrix.RowCount, Matrix.ColCount);
    PEigenVectors := EigenVectors.FirstItem;
  end else begin
    EigenVectors := nil;
    PEigenVectors := nil; { Separate variable to avoid using EigenVectors below }
  end;
  PEigenVals := Eigenvals.FirstItem;
  Inc(PEigenVals, Offset);
  try
    info := dsyevr(Matrix.ColCount, Matrix.FirstItem^, PEigenVals^, PEigenVectors^, EvReqd);
    if info < 0 then
      raise Exception.CreateFmt('Eigen: internal error in dsyevr %d', [-info])
    else if info <> Matrix.RowCount then
      raise Exception.CreateFmt('Eigen: Only %d of %d eigenvalues found', [info, Matrix.RowCount]);
  finally
    if Assigned(EigenVectors) then
      Matrix.AssignAndNil(EigenVectors);
  end;
  TEnd(EigenValuesTime);
end;

{$ELSE}

uses Math, SysUtils, Profile;

procedure Tred2(A: TMatrix; D, E: TColumnVector; EVREQD: Boolean);
var
  N: Integer;
  I, J, K, L: Integer;
  H, SCALE, F, G, HH: Double;
begin
  N := A.RowCount;
  IF N > 1 then begin
    for I := N downto 2 do begin
      L := I-1;
      H := 0;
      SCALE := 0;
      if  L > 1 then begin
        for K := 1 to L do
          SCALE := SCALE+ABS(A[I,K]);
        if SCALE = 0 then
          E[I] := A[I,L]
        else begin
          for K := 1 to L do begin
            A[I,K] := A[I,K]/SCALE;
            H := H+Sqr(A[I,K]);
          end;
          F := A[I,L];
          G := SQRT(H);
          if F > 0 then
            G := -G;
          E[I] := SCALE*G;
          H := H-F*G;
          A[I,L] := F-G;
          F := 0;
          for J := 1 to L do begin
            if EVREQD then
              A[J,I] := A[I,J]/H;
            G := 0;
            for K := 1 to J do
              G := G+A[J,K]*A[I,K];
            if L > J then
              for K := J+1 to L do
                G := G+A[K,J]*A[I,K];
            E[J] := G/H;
            F := F+E[J]*A[I,J];
          end;
          HH := F/(H+H);
          for J := 1 to L do begin
            F := A[I,J];
            G := E[J]-HH*F;
            E[J] := G;
            for K := 1 to J do
              A[J,K] := A[J,K]-F*E[K]-G*A[I,K];
          end;
        end;
      end else
        E[I] := A[I,L];
      D[I] := H;
    end;
  end;
  IF EVREQD then
    D[1] := 0;
  E[1] := 0;
  for I := 1 to N do begin
    IF EVREQD then begin
      IF D[I] <> 0 then
        for J := 1 to I-1 do begin
          G := 0;
          for K := 1 to I-1 do
            G := G+A[I,K]*A[K,J];
          for K := 1 to I-1 do
            A[K,J] := A[K,J]-G*A[K,I];
        end;
    end;
    D[I] := A[I,I];
    if EVREQD then begin
      A[I,I] := 1;
      IF I >= 2 then
        for J := 1 to I-1 do begin
          A[I,J] := 0;
          A[J,I] := 0;
        end;
    end;
  end;
end;

procedure TQLI(D, E: TColumnVector; Z: TMatrix; EVREQD: Boolean);
var
  N: Integer;
  I, K, L, M, ITER: Integer;
  DD, G, R, S, C, P, F, B: Double;
begin
  N := Z.RowCount;
  if (N > 1) then begin
    for I := 2 to N do
      E[I-1] := E[I];
    E[N] := 0;
    for L := 1 to N do begin
      ITER := 0;
      repeat
        M := N;
        for I := L to N-1 do begin
          DD := Abs(D[I])+Abs(D[I+1]);
          if Abs(E[I])+DD = DD then begin
            M := I;
            Break;
          end;
        end;
        if M <> L then begin
          if ITER = 30 then
            raise Exception.Create('Tqli: Too many iterations');
          ITER := ITER+1;
          G := (D[L+1]-D[L])/(2.*E[L]);
          R := SQRT(Sqr(G)+1);
          G := D[M]-D[L]+E[L]/(G+Abs(R)*Sign(G));
          S := 1;
          C := 1;
          P := 0;
          for I := M-1 downto L do begin
            F := S*E[I];
            B := C*E[I];
            IF ABS(F) >= ABS(G) then begin
              C := G/F;
              R := SQRT(Sqr(C)+1);
              E[I+1] := F*R;
              S := 1/R;
              C := C*S;
            end else begin
              S := F/G;
              R := SQRT(Sqr(S)+1);
              E[I+1] := G*R;
              C := 1/R;
              S := S*C;
            end;
            G := D[I+1]-P;
            R := (D[I]-G)*S+2*C*B;
            P := S*R;
            D[I+1] := G+P;
            G := C*R-B;
            if (EVREQD) then
              for K := 1 to N do begin
                F := Z[K,I+1];
                Z[K,I+1] := S*Z[K,I]+C*F;
                Z[K,I] := C*Z[K,I]-S*F;
              end;
          end;
          D[L] := D[L]-P;
          E[L] := G;
          E[M] := 0;
        end;
      until M = L;
    end;
  end;
end;

var
 EigenValuesTime: string = 'Eigen';

procedure Eigen(Eigenvals: TColumnVector; Matrix: TMatrix; EvReqd: Boolean);
var
  Work: TColumnVector;
  Offset, I, J: Integer;
  IsInOrder: Boolean;
  T: Double;
begin
  TStart(EigenValuesTime);
  if Matrix.RowCount <> Matrix.ColCount then
    raise Exception.Create('Eigenvals: Matrix must be square');
  if  Matrix.RowCount <> Eigenvals.Size then
    raise Exception.Create('Eigenvals: Matrix and vector must be same size');
  Work := TColumnVector.Create(Eigenvals.Size);
  try
    Tred2(Matrix, Eigenvals, Work, EvReqd);
    TQLI(Eigenvals, Work, Matrix, EvReqd);
    { Shell sort }
    Offset := Matrix.RowCount;
    while Offset > 1 do begin
      Offset := Offset shr 1;
      repeat
        IsInOrder := True;
        for J := 1 to Matrix.RowCount - Offset do begin
          I := J + Offset;
          if Eigenvals[I] < Eigenvals[J] then begin
            IsInOrder := False;
            T := Eigenvals[I];
            Eigenvals[I] := Eigenvals[J];
            Eigenvals[J] := T;
            if EvReqd then
              Matrix.SwapColumns(I,J);
          end;
        end;
      until IsInOrder;
    end;
  finally
    Work.Free;
  end;
  TEnd(EigenValuesTime);
end;
{$ENDIF}
end.
