unit LUDecomp;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses MatrixVec;

function dgetrf(
  m: Integer; { The number of rows of the matrix A.  M >= 0. }
  n: Integer; { The number of columns of the matrix A.  N >= 0. }
  var a: Double;  { (input/output) array, dimension (LDA,N)
                    On entry, the M-by-N matrix to be factored.
                    On exit, the factors L and U from the factorization
                    A = P*L*U; the unit diagonal elements of L are not stored. }
  lda: Integer; { The leading dimension of the array A.  LDA >= max(1,M) }
  out ipiv: Integer { INTEGER array, dimension (min(M,N))
                      The pivot indices; for 1 <= i <= min(M,N), row i of the
                      matrix was interchanged with row IPIV(i). }
): Integer;
{ Result:
  = 0:  successful exit
  < 0:  if INFO = -i, the i-th argument had an illegal value
  > 0:  if INFO = i, U(i,i) is exactly zero. The factorization
        has been completed, but the factor U is exactly
        singular, and division by zero will occur if it is used
        to solve a system of equations. }

function dgetri(
  n: Integer; { The order of the matrix A.  N >= 0. }
  var a: Double; { (input/output) array, dimension (LDA,N)
                   On entry, the factors L and U from the factorization
                   A = P*L*U as computed by DGETRF.
                   On exit, if INFO = 0, the inverse of the original matrix A. }
  lda: Integer; { The leading dimension of the array A.  LDA >= max(1,N)}
  var ipiv: Integer { (input) INTEGER array, dimension (N)
                      The pivot indices from DGETRF; for 1<=i<=N, row i of the
                      matrix was interchanged with row IPIV(i). }
): Integer;

type
  TEquationType = (eqNormal, eqTranspose, eqConjugate);

{ DGETRS solves a system of linear equations
    A * X = B  or  A' * X = B
  with a general N-by-N matrix A using the LU factorization computed
  by DGETRF. }
function dgetrs(
  EquationType: TEquationType;
  n: Integer;    { The order of the matrix A.  N >= 0. }
  nrhs: Integer; { The number of right hand sides, i.e., the number of columns
                   of the matrix B.  NRHS >= 0. }
  var a: Double; { (input) DOUBLE PRECISION array, dimension (LDA,N)
                   The factors L and U from the factorization A = P*L*U
                   as computed by DGETRF. }
  lda: Integer;  { The leading dimension of the array A.  LDA >= max(1,N). }
  var ipiv: Integer; { (input) The pivot indices from DGETRF; for 1<=i<=N, row i of the
                       matrix was interchanged with row IPIV(i). }
  var b: Double; { (input/output) array, dimension (LDB,NRHS)
                   On entry, the right hand side matrix B.
                   On exit, the solution matrix X. }
  ldb: Integer   { The leading dimension of the array B.  LDB >= max(1,N). }
): Integer;
{ Result:
  =0:  successful exit
  <0:  if INFO = -i, the i-th argument had an illegal value }

function Inverse(Matrix: TMatrix):TMatrix;

implementation

uses SysUtils, f2c, {$IFNDEF DESIGNPACKAGE} Lapack {$ENDIF};

function dgetrf(
  m: Integer; { The number of rows of the matrix A.  M >= 0. }
  n: Integer; { The number of columns of the matrix A.  N >= 0. }
  var a: Double;  { (input/output) array, dimension (LDA,N)
                    On entry, the M-by-N matrix to be factored.
                    On exit, the factors L and U from the factorization
                    A = P*L*U; the unit diagonal elements of L are not stored. }
  lda: Integer; { (input) The leading dimension of the array A.  LDA >= max(1,M) }
  out ipiv: Integer { (output) INTEGER array, dimension (min(M,N))
                       The pivot indices; for 1 <= i <= min(M,N), row i of the
                       matrix was interchanged with row IPIV(i). }
): Integer;
{ Result:
  = 0:  successful exit
  < 0:  if INFO = -i, the i-th argument had an illegal value
  > 0:  if INFO = i, U(i,i) is exactly zero. The factorization
        has been completed, but the factor U is exactly
        singular, and division by zero will occur if it is used
        to solve a system of equations. }
var
  Info: Integer;
begin
  dgetrf_(m, n, a, lda, ipiv, info);
  Result := info;
end;

function dgetri(
  n: Integer; { The order of the matrix A.  N >= 0. }
  var a: Double; { (input/output) array, dimension (LDA,N)
                   On entry, the factors L and U from the factorization
                   A = P*L*U as computed by DGETRF.
                   On exit, if INFO = 0, the inverse of the original matrix A. }
  lda: Integer; { The leading dimension of the array A.  LDA >= max(1,N)}
  var ipiv: Integer { (input) INTEGER array, dimension (N)
                      The pivot indices from DGETRF; for 1<=i<=N, row i of the
                      matrix was interchanged with row IPIV(i). }
): Integer;
{ Result:
  = 0:  successful exit
  < 0:  if INFO = -i, the i-th argument had an illegal value
  > 0:  if INFO = i, U(i,i) is exactly zero; the matrix is
        singular and its inverse could not be computed.
}
var
  Info, LWork: Integer;
  AWork: Double;
  Work: array of Double;
begin
  LWork := -1;
  dgetri_(n, a, lda, ipiv, AWork, LWork, info);
  if info = 0 then begin
    LWork := Round(AWork);
    SetLength(Work, LWork);
    dgetri_(n, a, lda, ipiv, Work[0], LWork, info);
  end;
  Result := Info;
end;

function dgetrs(
  EquationType: TEquationType;
  n: Integer;    { The order of the matrix A.  N >= 0. }
  nrhs: Integer; { The number of right hand sides, i.e., the number of columns
                   of the matrix B.  NRHS >= 0. }
  var a: Double; { (input) DOUBLE PRECISION array, dimension (LDA,N)
                   The factors L and U from the factorization A = P*L*U
                   as computed by DGETRF. }
  lda: Integer;  { The leading dimension of the array A.  LDA >= max(1,N). }
  var ipiv: Integer; { (input) The pivot indices from DGETRF; for 1<=i<=N, row i of the
                       matrix was interchanged with row IPIV(i). }
  var b: Double; { (input/output) array, dimension (LDB,NRHS)
                   On entry, the right hand side matrix B.
                   On exit, the solution matrix X. }
  ldb: Integer   { The leading dimension of the array B.  LDB >= max(1,N). }
): Integer;
{ Result:
  =0:  successful exit
  <0:  if INFO = -i, the i-th argument had an illegal value }
var
  info: Integer;
  TransCh: Char;
const
  Trans: array [TEquationType] of Char = ('N', 'T', 'C');
begin
  TransCh := Trans[EquationType];
  dgetrs_(TransCh,n,nrhs,a,lda,ipiv,b,ldb,info,1);
  Result := info;
end;

function Inverse(Matrix: TMatrix):TMatrix;
var
  Info: Integer;
  ipiv: array of Integer;
begin
  Result := TMatrix.Create(1,1);
  try
    Result.Assign(Matrix);
    SetLength(ipiv,Matrix.RowCount*Matrix.ColCount);
    Info := dgetrf(Matrix.RowCount,Matrix.ColCount,Matrix.FirstItem^,Matrix.RowCount,ipiv[0]);
    if Info = 0 then
      Info := dgetri(Matrix.RowCount,Matrix.FirstItem^,Matrix.RowCount,ipiv[0]);
    if Info <> 0 then
      raise Exception.Create('Matrix.Inverse fails:'+IntToStr(Info));
  except
    on E: Exception do begin
      Result.Free;
      raise;
    end;
  end;
end;

end.
