unit SmoothSpline;
{$I DFS.inc}{ Copyright (C) 2007 C M Western; see UCompat.pas for license}
interface

uses MathSup, Vector, ShellSort, PiecewisePoly;

function Smooth (
            const x: array of Double;
            const y: array of Double;
            const dy: array of Double;
            const s: Double;
            out a: TDoubleArrayArray): Double;

implementation

procedure setupq (
            const x: array of Double;
            const dx: array of Double;
            const y: array of Double;
            const npoint: Integer;
            out v: TDoubleArrayArray;
            out qty: TDoubleArray);
label
  L14, L16;
var
  i: Integer;
  npm1: Integer;
  diff: Double;
  prev: Double;

//   from  * a practical guide to splines *  by c. de boor
//   to be called in  s m o o t h
//   put  delx = x(.+1) - x(.)  into  v(.,4),
//   put  the three bands of  q-transp*d  into  v(.,1-3), and
//   put the three bands of  (d*q)-transp*(d*q)  at and above the diagonal
//      into  v(.,5-7) .
//      here,  q is  the tridiagonal matrix of order (npoint-2,npoint)
//   with general row  1/delx(i) , -1/delx(i) - 1/delx(i+1) , 1/delx(i+1)
//   and   d  is the diagonal matrix  with general row  dx(i) .


begin
  npm1 := npoint - 1;
  SetLength(v,8,npoint+1);
  SetLength(qty,npoint+1);
  v[4,1] := x[2] - x[1];
  for i := 2 to npm1 do begin
    v[4,i] := x[i + 1] - x[i];
    v[1,i] := dx[i - 1] / v[4, i - 1];
    v[2,i] := -dx[i] / v[4, i] - dx[i] / v[4, i - 1];
    v[3,i] := dx[i + 1] / v[4, i];
  end;
  v[1,npoint] := 0.0;
  for i := 2 to npm1 do begin
    v[5,i] := Sqr(v[1, i]) + Sqr(v[2, i]) + Sqr(v[3, i]);
  end;
  if npm1 < 3 then
    goto L14;
  for i := 3 to npm1 do begin
    v[6,i-1] := v[2, i - 1] * v[1, i] + v[3, i - 1] * v[2, i];
  end;
L14:
  v[6,npm1] := 0.0;
  if npm1 < 4 then
    goto L16;
  for i := 4 to npm1 do begin
    v[7,i-2] := v[3, i - 2] * v[1, i];
  end;
L16:
  v[7,npm1-1] := 0.0;
  v[7,npm1] := 0.0;
// construct  q-transp. * y  in  qty.
  prev := (y[2] - y[1]) / v[4, 1];
  for i := 2 to npm1 do begin
    diff := (y[i + 1] - y[i]) / v[4, i];
    qty[i] := diff - prev;
    prev := diff;
  end;

end;

procedure chol1d (
            const p: Double;
            var v: TDoubleArrayArray;
            const qty: array of Double;
            const npoint: Integer;
            var u: TDoubleArray;
            var qu: TDoubleArray);
label
  L10, L41, L40;
var
  i: Integer;
  npm1: Integer;
  npm2: Integer;
  prev: Double;
  ratio: Double;
  six1mp: Double;
  twop: Double;

//   from  * a practical guide to splines *  by c. de boor
//  to be called in  s m o o t h
//   constructs the upper three diags. in v(i,j), i=2,,npoint-1, j=1,3, of
//   the matrix  6*(1-p)*q-transp.*(d**2)*q + p*r, then computes its
//   l*l-transp. decomposition and stores it also in v, then applies
//   forward and backsubstitution to the right side q-transp.*y in  qty
//   to obtain the solution in  u .


begin
  npm1 := npoint - 1;
//      construct 6*(1-p)*q-transp.*(d**2)*q  +  p*r
  six1mp := 6 * (1 - p);
  twop := 2 * p;
  for i := 2 to npm1 do begin
    v[1,i] := six1mp * v[5, i] + twop * (v[4, i - 1] + v[4, i]);
    v[2,i] := six1mp * v[6, i] + p * v[4, i];
    v[3,i] := six1mp * v[7, i];
  end;
  npm2 := npoint - 2;
  if npm2 >= 2 then
    goto L10;
  u[1] := 0;
  u[2] := qty[2] / v[1, 2];
  u[3] := 0;
  goto L41;
//   factorization
L10:
  for i := 2 to npm2 do begin
    ratio := v[2, i] / v[1, i];
    v[1,i+1] := v[1, i + 1] - ratio * v[2, i];
    v[2,i+1] := v[2, i + 1] - ratio * v[3, i];
    v[2,i] := ratio;
    ratio := v[3, i] / v[1, i];
    v[1,i+2] := v[1, i + 2] - ratio * v[3, i];
    v[3,i] := ratio;
  end;
//
//   forward substitution
  u[1] := 0.0;
  v[3,1] := 0.0;
  u[2] := qty[2];
  for i := 2 to npm2 do begin
    u[i+1] := qty[i + 1] - v[2, i] * u[i] - v[3, i - 1] * u[i - 1];
  end;
//   back substitution
  u[npoint] := 0.0;
  u[npm1] := u[npm1] / v[1, npm1];
  i := npm2;
L40:
  u[i] := u[i] / v[1, i] - u[i + 1] * v[2, i] - u[i + 2] * v[3, i];
  i := i - 1;
  if i > 1 then
    goto L40;
//   construct q*u
L41:
  prev := 0.0;
  for i := 2 to npoint do begin
    qu[i] := (u[i] - u[i - 1]) / v[4, i - 1];
    qu[i-1] := qu[i] - prev;
    prev := qu[i];
  end;
  qu[npoint] := -qu[npoint];

end;

function smooth (
            const x: array of Double;
            const y: array of Double;
            const dy: array of Double;
            const s: Double;
            out a: TDoubleArrayArray): Double;
label
  L20, L10, L60, L30, L59;
var
  i: Integer;
  npoint: Integer;
  npm1: Integer;
  change: Double;
  ooss: Double;
  oosf: Double;
  p: Double;
  prevsf: Double;
  prevq: Double;
  q: Double;
  sfq: Double;
  sixp: Double;
  six1mp: Double;
  utru: Double;
  v: TDoubleArrayArray;

//   from  * a practical guide to splines *  by c. de boor
// calls  setupq, chol1d
//
//   constructs the cubic smoothing spline  f  to given data  (x(i),y(i)),
//   i=1,...,npoint, which has as small a second derivative as possible
//   while
//   s(f) = sum( ((y(i)-f(x(i)))/dy(i))**2 , i=1,...,npoint ) .le. s .
//
// ******  i n p u t  ******
//   x(1),...,x(npoint)   data abscissae,  a s s u m e d  to be strictly
//         increasing .
//   y(1),...,y(npoint)     corresponding data ordinates .
//   dy(1),...,dy(npoint)     estimate of uncertainty in data,  a s s u m-
//         e d  to be positive .
//   npoint.....number of data points,  a s s u m e d  .gt. 1
//   s.....upper bound on the discrete weighted mean square distance of
//         the approximation  f  from the data .
//
// ******  w o r k  a r r a y s  *****
//   v.....of size (npoint,7)
//   a.....of size (npoint,4)
//
// *****  o u t p u t  *****
//   a(.,1).....contains the sequence of smoothed ordinates .
//   a(i,j) = f^(j-1)(x(i)), j=2,3,4, i=1,...,npoint-1 ,  i.e., the
//         first three derivatives of the smoothing spline  f  at the
//         left end of each of the data intervals .
//      w a r n i n g . . .   a  would have to be transposed before it
//         could be used in  ppvalu .
//
// ******  m e t h o d  ******
//      The matrices  Q-transp*d  and  Q-transp*D**2*Q  are constructed in
//    s e t u p q  from  x  and  dy , as is the vector  qty = Q-transp*y .
//   Then, for given  p , the vector  u  is determined in  c h o l 1 d  as
//   the solution of the linear system
//                (6(1-p)Q-transp*D**2*Q + p*Q)u  = qty  .
//   From  u , the smoothing spline  f  (for this choice of smoothing par-
//   ameter  p ) is obtained in the sense that
//                         f(x(.))  =  y - 6(1-p)D**2*Q*u        and
//                       f''(x(.))  =  6*p*u                      .
//      The smoothing parameter  p  is found (if possible) so that
//                 sf(p)  =  s ,
//   with  sf(p) = s(f) , where  f  is the smoothing spline as it depends
//   on  p .  if  s = 0, then p = 1 . if  sf(0) .le. s , then p = 0 .
//   Otherwise, the secant method is used to locate an appropriate  p  in
//   the open interval  (0,1) . However, straightforward application of
//   the secant method, as done in the original version of this program,
//   can be very slow and is influenced by the units in which  x  and  y
//   are measured, as C. Reinsch has pointed out. Instead, on recommend-
//   ation from C. Reinsch, the secant method is applied to the function
//            g:q |--> 1/sqrt{sfq(q)} - 1/sqrt{s} ,
//   with  sfq(q) := sf(q/(1+q)), since  1/sqrt{sfq}  is monotone increasing
//   and close to linear for larger  q . One starts at  q = 0  with a
//   Newton step, i.e.,
//                 q_0 = 0,  q_1 = -g(0)/g'(0)
//   with  g'(0) = -(1/2) sfq(0)^{-3/2} dsfq, where dsfq = -12*u-transp*r*u ,
//   and  u  as obtained for  p = 0 . Iteration terminates as soon as
//    abs(sf - s) .le. .01*s .
//
//      logical test
//      parameter (test = .true.)
//      integer itercnt


begin
  npoint := High(x);
  SetLength(a,5,npoint+1);
  setupq(x,dy,y,npoint,v,a[4]);
  if s > 0.0 then
    goto L20;
L10:
  p := 1.0;
  six1mp := 0.0;
  chol1d(p,v,a[4],npoint,a[3],a[1]);
  sfq := 0.0;
  goto L60;
L20:
  p := 0.0;
  six1mp := 6.0;
  chol1d(p,v,a[4],npoint,a[3],a[1]);
  sfq := 0.0;
  for i := 1 to npoint do begin
    sfq := sfq + Sqr( (a[1, i] * dy[i]));
  end;
  sfq := sfq * 36.0;
  if sfq <= s then
    goto L60;
  utru := 0.0;
  for i := 2 to npoint do begin
    utru := utru + v[4, i - 1] * (a[3, i - 1] * (a[3, i - 1] + a[3, i]) + Sqr(a[3, i]));
  end;
  ooss := 1.0 / sqrt(s);
  oosf := 1.0 / sqrt(sfq);
  q := -(oosf - ooss) * sfq / (6.0 * utru * oosf);
//   secant iteration for the determination of p starts here.
//      itercnt = 0
  prevq := 0.0;
  prevsf := oosf;
L30:
  chol1d(q/(1.0+q),v,a[4],npoint,a[3],a[1]);
  sfq := 0.0;
  for i := 1 to npoint do begin
    sfq := sfq + Sqr( (a[1, i] * dy[i]));
  end;
  sfq := sfq * 36.0 / Sqr( (1.0 + q));
  if abs(sfq - s) <= 0.01 * s then
    goto L59;
  oosf := 1.0 / sqrt(sfq);
  change := (q - prevq) / (oosf - prevsf) * (oosf - ooss);
  prevq := q;
  q := q - change;
  prevsf := oosf;
//      itercnt = itercnt + 1
  goto L30;
L59:
  p := q / (1.0 + q);
  six1mp := 6.0 / (1.0 + q);
// correct value of p has been found.
// compute pol.coefficients from  Q*u (in a(.,1)).
L60:
  smooth := sfq;
//      if (test) then
//         print *, 'number of iterations = ', itercnt
//      end if
  for i := 1 to npoint do begin
    a[1,i] := y[i] - six1mp * Sqr(dy[i]) * a[1, i];
  end;
  sixp := 6.0 * p;
  for i := 1 to npoint do begin
    a[3,i] := a[3, i] * sixp;
  end;
  npm1 := npoint - 1;
  for i := 1 to npm1 do begin
    a[4,i] := (a[3, i + 1] - a[3, i]) / v[4, i];
    a[2,i] := (a[1, i + 1] - a[1, i]) / v[4, i] - (a[3, i] + a[4, i] / 3.0 * v[4, i]) / 2.0 * v[4, i];
  end;

end;

end.
