unit TempStrm;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes;

{ This stream provides a safe way of saving a file, leaving a back up copy
  and not loosing the original on disk full or other error. To use, create
  with the desired filename, write to the stream (which writes to a temporary file
  on disk) and then call Save (which does some file renaming). The destructor will
  discard the temporary file if Save has not benn called.
  Typical usage:
    Steam := TBackupStream.Create(FileName);
    try
      Text.SaveToStream(Steam);
      Steam.Save;
    finally
      Steam.Free;
      Raise;
    end;
}

type
  TBackupStream = class (THandleStream)
  private
    FTmpName: string;
    FFileName: string;
    {$IFNDEF DFS_COMPILER_6_UP}
    HandleClosed: Boolean;
    {$ENDIF}
  public
    constructor Create(const AFileName: string);
    destructor Destroy; override;
    procedure Save;
    property TmpName: string read FTmpName;
    property FileName: string read FFileName;
  end;

function TempStreamFor(Name: string; var TmpName: string): THandleStream;
function TmpFileHandleFor(Name: string; var TmpName: string): Integer;

implementation

{$IFDEF UNIX}
uses BaseUnix, SysUtils;

{$LINKLIB libc}
function mkstemp(Template: PChar):Integer; cdecl; external;

function TmpFileHandleFor(Name: string; var TmpName: string): Integer;
begin
  TmpName := Concat(ExpandFilename(Name), 'XXXXXX');
  UniqueString(TmpName);
  Result := mkstemp(PChar(TmpName));
  if Result = -1 then
     RaiseLastOSError;
end;
{$ENDIF}
{$IFDEF WIN32}
uses Windows, SysUtils;
function TmpFileHandleFor(Name: string; var TmpName: string): Integer;
var
  Buffer: array [0..MAX_PATH+1] of char;
begin
  {$IFNDEF offLCL_}
  TmpName := ExtractFilePath(ExpandFilename(Name));
  if Windows.GetTempFileName(PChar(TmpName), 'txt', 0, Buffer) = 0 then
    raise Exception.CreateFmt('Can''t create temporary file in %s', [TmpName]);
  TmpName := Buffer;
  {$ELSE}
  TmpName := GetTempFileName(ExtractFilePath(ExpandFilename(Name)), 'txt');
  {$ENDIF}
  Result := FileCreate(TmpName);
  if Result < 0 then
    RaiseLastOSError;
end;
{$ENDIF}
{$IFDEF VER80}
uses WinProcs, WinTypes, SysUtils, Compat;
function TmpFileHandleFor(Name: string; var TmpName: string): Integer;
var
  BackFile, Buffer, CurDir: string;
begin
  TmpName := ExtractFilePath(Name);
  try
    GetDir(0, CurDir);
    if TmpName[Length(Tmpname)] = '\' then
      SetLength(TmpName, Length(Tmpname)-1);
    ChDir(TmpName);
    BackFile := 'log'#0;
    GetTempFileName(Char(TF_FORCEDRIVE), AddressOfFirstChar(BackFile), 0, AddressOfFirstChar(Buffer));
    SetLength(Buffer, StrLen(AddressOfFirstChar(Buffer)));
  finally
    ChDir(CurDir);
  end;
  TmpName := Concat(TmpName, '\', ExtractFileName(Buffer));
  Result := FileCreate(TmpName);
  if Result < 0 then
    raise Exception.Create('Can''t create temporary file '+TmpName);
end;
{$ENDIF}

function TempStreamFor(Name: string; var TmpName: string): THandleStream;
begin
  Result := THandleStream.Create(TmpFileHandleFor(Name,TmpName));
end;

{ TBackupStream }

procedure TBackupStream.Save;
var
  BackFile: string;
begin
  if (Handle >= 0) {$IFNDEF DFS_COMPILER_6_UP} and not HandleCLosed {$ENDIF} then begin
    FileClose(Handle);
    {$IFDEF DFS_COMPILER_6_UP}
    FHandle := -1;
    {$ELSE}
    HandleClosed := True;
    {$ENDIF}
    {$IFNDEF VER80}
    BackFile := Concat(FileName, '.bak');
    {$ELSE}
    BackFile := ChangeFileExt(FileName, '.bak');
    if CompareText(BackFile, FileName) = 0 then
      BackFile := ChangeFileExt(FileName, '.bk');
    {$ENDIF}
    if FileExists(BackFile) then
      if not DeleteFile(BackFile) then
        raise Exception.CreateFmt('Can''t delete old backup file %s', [BackFile]);
    if FileExists(FileName) and not RenameFile(FileName, BackFile) then
        raise Exception.CreateFmt('Can''t rename old file (%s) to %s', [FileName, BackFile]);
    if not RenameFile(TmpName, FileName) then
      raise Exception.CreateFmt('Can''t rename new file (%s) to %s', [TmpName, FileName]);
  end;
end;

{$IFDEF FPC}
  {$IF defined(ver1) or (defined(ver2_0) and (fpc_patch<=4))}
    {$IFNDEF DARWIN}
      {$DEFINE OLDMODE}
    {$ENDIF}
  {$IFEND}
{$ENDIF}

constructor TBackupStream.Create(const AFileName: string);
var
  {$IFDEF MSWINDOWS}
  Attr: DWORD;
  {$ENDIF}
  {$IFDEF UNIX}
  Info: BaseUnix.Stat;
  {$ENDIF}
begin
  {$IFDEF DFS_COMPILER_6_UP}
  FHandle := -1;
  {$ELSE}
  HandleClosed := False;
  {$ENDIF}
  {$IFDEF UNIX}
  Info.{$IFNDEF OLDMODE} st_mode {$ELSE} mode {$ENDIF} := 0;
  if (FpStat(AFileName, Info) >= 0) and { (Libc.Access(PChar(AFileName),W_OK)<> 0) } FileIsReadOnly(AFileName) then
      raise Exception.Create(AFileName + ' is read only');
  {$ENDIF}
  {$IFDEF MSWINDOWS}
  Attr := GetFileAttributes(PChar(AFileName));
  if (Integer(Attr) <> -1) and ((Attr and FILE_ATTRIBUTE_READONLY) <> 0) then
    raise Exception.Create(AFileName + ' is read only');
  {$ENDIF}
  inherited Create(TmpFileHandleFor(AFileName, FTmpName));
  {$IFDEF UNIX}
  if (Info.{$IFNDEF OLDMODE} st_mode {$ELSE} mode {$ENDIF} <> 0) and (FpChmod(FTmpName, Info.{$IFNDEF OLDMODE} st_mode {$ELSE} mode {$ENDIF}) < 0) then
    RaiseLastOSError;
  {$ENDIF}
  FFileName := AFileName;
end;

destructor TBackupStream.Destroy;
begin
  if (Handle >= 0) {$IFNDEF DFS_COMPILER_6_UP} and not HandleClosed {$ENDIF} then begin
    FileClose(Handle);
    {$IFDEF DFS_COMPILER_6_UP}
    FHandle := -1;
    {$ELSE}
    HandleClosed := True;
    {$ENDIF}
    DeleteFile(TmpName);
  end;
  inherited Destroy;
end;

end.


