unit Vector;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes;

type

  { TVector }

  TVector = class
  private
    FSize: Integer;
    FFirstItem: PDouble;
    function GetPointer(n: Longint): PDouble;
    function GetValid(i:Longint):Boolean;
    procedure SetValid(i:Longint; IsValid:Boolean);
    function GetValue(i:Longint):Double;
    procedure SetValue(i:Longint; v:Double);
    procedure SetSize(const n:Longint);
  public
    Offset: Integer;
    constructor Create(ASize: Longint); overload;
    constructor Create(AVector: TVector); overload;
    destructor Destroy; override;
    procedure SetSizeAndClear(n:Longint);
    procedure Clear; { to NAN }
    function GetCell(i:Longint; var v: Double): Boolean;
    procedure Assign(Source: TVector); virtual;
    procedure AssignAndNil(var Source: TVector); virtual;
    procedure ClearRange(First, Count:Longint);
    function LastIndex:Integer;

    function DotProduct(v:TVector):Double;
    procedure Sqr; virtual;
    procedure Add(v: TVector);
    procedure Zero;
    procedure Scale(x: Double);
    procedure Write(var F: Text; Title: string; Width: Integer = 10; Fixed: Integer = -1);
    function SameValues(AVector: TVector; Epsilon: Double = 0): Boolean;

    property Size:Integer read FSize write SetSize;
    property Item[n:Longint]:PDouble read GetPointer;
    property Valid[i:Longint]:Boolean read GetValid write SetValid;
    property Value[i:Longint]:Double read GetValue write SetValue; default;
    property FirstItem: PDouble read FFirstItem;
  end;

implementation

uses SysUtils, Math, MathSup;

constructor TVector.Create(ASize: Longint);
begin
  inherited Create;
  SetSize(ASize);
end;

procedure TVector.Scale(x: Double);
var
  I: Integer;
  p: PDouble;
begin
  p := FirstItem;
  for I := 1 to Size do begin
    p^ := p^ * x;
    Inc(p);
  end;
end;

function TVector.SameValues(AVector: TVector; Epsilon: Double = 0): Boolean;
var
  I: Integer;
  p, q: PDouble;
begin
  Result := Size = AVector.Size;
  if Result then begin
    p := FirstItem;
    q := AVector.FirstItem;
    for I := 1 to Size do begin
      if not SameValue(p^,q^,Epsilon) then begin
        Result := False;
        Exit;
      end;
      Inc(p);
      Inc(q);
    end;
  end;
end;

procedure TVector.SetSizeAndClear(n: Integer);
var
  Old: Longint;
begin
  Old := Size;
  Size := n;
  if Size > Old then
    ClearRange(Old+Offset, Size-Old);
end;

procedure TVector.SetSize(const n:Longint);
begin
  FSize := n;
  ReallocMem(FFirstItem,n*SizeOf(Double));
end;

procedure TVector.Assign(Source: TVector);
begin
  Size := TVector(Source).Size;
  Offset := TVector(Source).Offset;
  Move(TVector(Source).FirstItem^, FirstItem^, Size*SizeOf(Double));
end;

procedure TVector.AssignAndNil(var Source: TVector);
begin
  FreeMem(FFirstItem);
  FSize := TVector(Source).Size;
  Offset := TVector(Source).Offset;
  FFirstItem := TVector(Source).FFirstItem;
  TVector(Source).FFirstItem := nil;
  FreeAndNil(Source);
end;

procedure TVector.Clear;
begin
  ClearRange(Offset, Size);
end;

procedure TVector.ClearRange(First, Count:Longint);
var
  p:PDouble;
begin
  if Count = 0 then
    Exit; { Avoid range check on next line }
  p := Item[First];
  while Count > 0 do begin
    SetNan(p^);
    Inc(p,1);
    Dec(Count);
  end;
end;

function TVector.LastIndex: Integer;
begin
   Result := Offset + Size - 1;
end;

constructor TVector.Create(AVector: TVector);
begin
  Create(0);
  Assign(AVector);
end;

function TVector.GetValid(i:Longint):Boolean;
var
  p:PDouble;
begin
  p := Item[i];
  Result := not IsNanOrInf(p^);
end;

procedure TVector.SetValid(i:Longint; IsValid:Boolean);
var
  p:PDouble;
begin
  p := Item[i];
  if IsValid then
    p^ := 0
  else
    SetNan(p^);
end;

function TVector.GetValue(i:Longint):Double;
begin
  Result := Item[i]^;
end;

procedure TVector.SetValue(i:Longint; v:Double);
var
  p:PDouble;
begin
  p := Item[i];
  p^ := v;
end;

destructor TVector.Destroy;
begin
  SetSize(0);
  inherited;
end;

function TVector.DotProduct(v:TVector):Double;
var
  p, p1: PDouble;
  i: Longint;
  Sum: Extended;
begin
  if v.Size <> Size then
    raise Exception.CreateFmt('TVector.DotProduct(%d,%d)', [Size, v.Size]);
  p := FirstItem;
  p1 := v.FirstItem;
  Sum := 0;
  for i := 1 to Size do begin
    Sum := Sum + p^*p1^;
    Inc(p);
    Inc(p1);
  end;
  Result := Sum;
end;

function TVector.GetCell(i: Integer; var v: Double): Boolean;
var
  p: PDouble;
begin
  p := Item[i];
  Result := not IsNanOrInf(p^);
  System.Move(p^, v, SizeOf(Double));  { So that specific Invalid values can be seen }
end;

function TVector.GetPointer(n: Integer): PDouble;
begin
  Dec(n,Offset);
  {$IFOPT R+}
  if (n < 0) or (n >= FSize) then
    raise Exception.CreateFmt('TVector.GetPointer(%d) Size=%d', [n,Size])
  else
  {$ENDIF}
  Result := FirstItem;
  Inc(Result,n);
end;

procedure TVector.Sqr;
var
  p: PDouble;
  i: Longint;
begin
  p := FirstItem;
  for i := 1 to Size do begin
    p^ := System.Sqr(p^);
    Inc(p);
  end;
end;

procedure TVector.Write(var F: Text; Title: string; Width, Fixed: Integer);
var
  I: Integer;
begin
  WriteLn(F, Title, ' ', Offset, '..', Size+Offset-1);
  for I := Offset to Size+Offset - 1 do
    System.Write(F, DtosMA(Value[I],Width,Fixed));
  WriteLn(F);
end;

procedure TVector.Add(v: TVector);
var
  p, p1: PDouble;
  i: Longint;
begin
  if v.Size <> Size then
    raise Exception.CreateFmt('TVector.Add(%d,%d)', [Size, v.Size]);
  p := FirstItem;
  p1 := v.FirstItem;
  for i := 1 to Size do begin
    p^ := p^ + p1^;
    Inc(p);
    Inc(p1);
  end;
end;

procedure TVector.Zero;
begin
  FillChar(FirstItem^, Size*SizeOf(Double), #0);
end;

end.
