unit Wofz;

interface

{ Double precision Voigt profile }
function Voigt(dw:     Double; { Offset from line centre }
               gamma:  Double; { Lorenztian FWHM }
               gwidth: Double  { Gaussian FWHM }
               ): Double;

procedure Faddeeva(
  XI: Double;    {REAL PART OF Z}
  YI: Double;    {IMAGINARY PART OF Z}
  var U: Double; {REAL PART OF W(Z)}
  var V: Double  {IMAGINARY PART OF W(Z)}
);

implementation

uses Math, MathSup;

var
  FACTOR:    Double; { = 2/Sqrt(Pi) }
  SqrtLn2:   Double; { = Sqrt(ln(2)) }
  SqrtLn2Pi: Double; { = Sqrt(ln(2)/Pi) }

function Voigt(dw:     Double; { Offset from line centre }
               gamma:  Double; { Lorenztian FWHM }
               gwidth: Double  { Gaussian FWHM }
               ): Double;
var
  x, y, u, v: Double;
begin
  if gwidth > 0 then begin
    x := 2*SqrtLn2*dw/gwidth;
    if gamma > 0 then begin
      y := SqrtLn2*gamma/gwidth;
      Faddeeva(x, y, u, v);
    end else
      u := Exp(-Sqr(x));
    Voigt := 2*SqrtLn2Pi*U/gwidth;
  end else
    Voigt := (gamma/(2.*Pi))/(Sqr(dw)+Sqr(gamma/2));
end;

{    ALGORITHM 680, COLLECTED ALGORITHMS FROM ACM.
     THIS WORK PUBLISHED IN TRANSACTIONS ON MATHEMATICAL SOFTWARE,
     VOL. 16, NO. 1, PP. 47.}
procedure Faddeeva(
  XI: Double;    {REAL PART OF Z}
  YI: Double;    {IMAGINARY PART OF Z}
  var U: Double; {REAL PART OF W(Z)}
  var V: Double  {IMAGINARY PART OF W(Z)}
);
VAR
  A,B: boolean;
  C,DAUX,H,H2,QLAMBDA,QRHO,RX,RY,SX,SY,TX,
  TY,U1,U2,V1,V2,W1,X,XABS,XABSQ,XAUX,XQUAD,XSUM,Y,YABS,YQUAD,YSUM: Double;
  I,J,KAPN,N,NP1,NU: Integer;
BEGIN
{ GIVEN A COMPLEX NUMBER Z = (XI,YI), THIS SUBROUTINE COMPUTES
  THE VALUE OF THE FADDEEVA-FUNCTION W(Z) = EXP(-Z**2)*ERFC(-I*Z),
  WHERE ERFC IS THE COMPLEX COMPLEMENTARY ERROR-FUNCTION AND I
  MEANS SQRT(-1).
  THE ACCURACY OF THE ALGORITHM FOR Z IN THE 1ST AND 2ND QUADRANT
  IS 14 SIGNIFICANT DIGITS; IN THE 3RD AND 4TH IT IS 13 SIGNIFICANT
  DIGITS OUTSIDE A CIRCULAR REGION WITH RADIUS 0.126 AROUND A ZERO
  OF THE FUNCTION.
  THE ROUTINE IS NOT UNDERFLOW-PROTECTED BUT ANY VARIABLE CAN BE
  PUT TO 0 UPON UNDERFLOW
  REFERENCE - GPM POPPE, CMJ WIJERS; MORE EFFICIENT COMPUTATION OF
  THE COMPLEX ERROR-FUNCTION, ACM TRANS. MATH. SOFTWARE.}

        XABS:=ABS(XI);
        YABS:=ABS(YI);
        X:=XABS/6.3;
        Y:=YABS/4.4;
        QRHO:=sqr(X)+sqr(Y);
{*}
        XABSQ:=sqr(XABS);
        XQUAD:=XABSQ-sqr(YABS);
        YQUAD:=2*XABS*YABS;
{*}
        A:=QRHO < 0.085264E0;
{*}
        IF (A) THEN
        BEGIN
{C}
{C IF (QRHO.LT.0.085264D0) THEN THE FADDEEVA-FUNCTION IS EVALUATED}
{C USING A POWER-SERIES (ABRAMOWITZ/STEGUN, EQUATION (7.1.5), P.297)}
{C N IS THE MINIMUM NUMBER OF TERMS NEEDED TO OBTAIN THE REQUIRED}
{C ACCURACY}
{C}
          QRHO:=(1-0.85*Y)*SQRT(QRHO);
          N:=Round(6+72*QRHO);
          J:=2*N+1;
          XSUM:=1.0/J;
          YSUM:=0.0E0;
          for I := N downto 1 do begin
            J:=J-2;
            XAUX:=(XSUM*XQUAD-YSUM*YQUAD)/I;
            YSUM:=(XSUM*YQUAD+YSUM*XQUAD)/I;
            XSUM:=XAUX+1.0/J;
          end;
          U1:=-FACTOR*(XSUM*YABS+YSUM*XABS)+1.0;
          V1:=FACTOR*(XSUM*XABS-YSUM*YABS);
          DAUX:=EXP(-XQUAD);
          U2:=DAUX*COS(YQUAD);
          V2:=-DAUX*SIN(YQUAD);
{*}
          U:=U1*U2-V1*V2;
          V:=U1*V2+V1*U2;
        END ELSE BEGIN
{C IF (QRHO.GT.1.O) THEN W(Z) IS EVALUATED USING THE LAPLACE}
{C CONTINUED FRACTION}
{C NU IS THE MINIMUM NUMBER OF TERMS NEEDED TO OBTAIN THE REQUIRED}
{C ACCURACY}
{C}
{C IF ((QRHO.GT.0.085264D0).AND.(QRHO.LT.1.0)) THEN W(Z) IS EVALUATED}
{C BY A TRUNCATED TAYLOR EXPANSION, WHERE THE LAPLACE CONTINUED FRACTION}
{C IS USED TO CALCULATE THE DERIVATIVES OF W(Z)}
{C KAPN IS THE MINIMUM NUMBER OF TERMS IN THE TAYLOR EXPANSION NEEDED}
{C TO OBTAIN THE REQUIRED ACCURACY}
{C NU IS THE MINIMUM NUMBER OF TERMS OF THE CONTINUED FRACTION NEEDED}
{C TO CALCULATE THE DERIVATIVES WITH THE REQUIRED ACCURACY}
{C}
{*}
          IF (QRHO > 1.0) THEN
          BEGIN
            H:=0.0E0;
            KAPN:=0;
            QRHO:=SQRT(QRHO);
            NU:=Round(3+(1442/(26*QRHO+77)));
          END
          ELSE
          BEGIN
            QRHO:=(1-Y)*SQRT(1-QRHO);
            H:=1.88*QRHO;
            H2:=2*H;
            KAPN:=Round(7+34*QRHO);
            NU:=Round(16+26*QRHO);
          END;
{*}
          B:=(H > 0.0);
{*}
          IF (B) THEN QLAMBDA:=IntPower(H2,KAPN);
{*}
          RX:=0.0;
          RY:=0.0;
          SX:=0.0;
          SY:=0.0;
{*}
          for N := NU downto 0 do begin
            NP1:=N+1;
            TX:=YABS+H+NP1*RX;
            TY:=XABS-NP1*RY;
            C:=0.5/(sqr(TX)+sqr(TY));
            RX:=C*TX;
            RY:=C*TY;
            IF B AND (N <= KAPN) THEN
            BEGIN
              TX:=QLAMBDA+SX;
              SX:=RX*TX-RY*SY;
              SY:=RY*TX+RX*SY;
              QLAMBDA:=QLAMBDA/H2;
            END;
          end;
{*}
          IF (H = 0.0) THEN
          BEGIN
            U:=FACTOR*RX;
            V:=FACTOR*RY;
          END
          ELSE
          BEGIN
            U:=FACTOR*SX;
            V:=FACTOR*SY;
          END;
{*}
          IF (YABS = 0.0) THEN U:=EXP(-sqr(XABS));
{*}
        END;
{*}
{*}
{C}
{C EVALUATION OF W(Z) IN THE OTHER QUADRANTS}
{C}
{*}
        IF (YI < 0.0) THEN
        BEGIN
{*}
          IF (A) THEN
          BEGIN
            U2:=2*U2;
            V2:=2*V2;
          END
          ELSE
          BEGIN
            XQUAD:=-XQUAD;
            W1:=2*EXP(XQUAD);
            U2:=W1*COS(YQUAD);
            V2:=-W1*SIN(YQUAD);
          END;
{*}
          U:=U2-U;
          V:=V2-V;
          IF (XI > 0.0) THEN V:=-V;
        END ELSE IF (XI < 0.0) THEN V:=-V;
END;

begin
  FACTOR := 2/Sqrt(Pi);
  SqrtLn2 := Sqrt(ln(Two));
  SqrtLn2Pi := Sqrt(ln(Two)/Pi);
end.
