unit FixForm;
{$I DFS.inc}
interface

{ Derive forms from this to get more conventional MS Windows behaviour,
  especially under Vista; See:
  http://www.installationexcellence.com/articles/VistaWithDelphi/Index.html
  http://www.installationexcellence.com/articles/VistaWithDelphi/Original/Index.html
  Note that this requires Forms.pas in the VCL to be patched to avoid problems with ShowModal.

  Windows derived from this will have an icon on the TaskBar, and will minimize to the Taskbar.
  Forms without this will Minimize to the desktop (like MDI child windows)
}

uses {$IFDEF MSWINDOWS} Windows, Messages, {$ENDIF}
     Classes, Controls, Forms, Menus;

type
  TFixForm = class(TForm)
  private
    FFixThisForm: Boolean;
  protected
    {$IFDEF VCL_}
    FOriginalStyle: Longint;
    procedure WMSysCommand(var Message: TWMSysCommand); message WM_SYSCOMMAND;
    procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure UpdateActions; override;
    {$ENDIF}
    property FixThisForm: Boolean read FFixThisForm;
  public
    { Used to display a shortcut key in a menu, but not handle it via the standard
      shortcut mechanism. Appropriate for basic printing keys like a..z 0..9, which
      would otherwise be unusable in, say, text fields. Currently requires manual
      handling of character in KeyPress or related events }
    procedure AddShortCut(Item: TMenuItem; ShortCut: string);
  end;

var
  FixAllForms: Boolean = True;

type
  TWindowManagerMode = (wmmStandard, wmmOnTop, wmmIndependent);

function GetCurrentWindowManagerMode: TWindowManagerMode;
procedure SetCurrentWindowManagerMode(AMode: TWindowManagerMode);

implementation

var
  WindowManagerMode: TWindowManagerMode = wmmStandard;
  PendingWindowManagerMode: TWindowManagerMode = wmmStandard;

function GetCurrentWindowManagerMode: TWindowManagerMode;
begin
  Result := WindowManagerMode;
end;

procedure SetCurrentWindowManagerMode(AMode: TWindowManagerMode);
begin
  PendingWindowManagerMode := AMode;
end;

{ TFixForm }

{$IFDEF VCL_}
var
  OriginalStyle: LongInt;

procedure AdjustWindow(Handle: HWND; NewStyle: Longint);
var
  OldStyle: Longint;
begin
  OldStyle := GetWindowLong(Handle, GWL_EXSTYLE);
  if OldStyle <> NewStyle then begin
    ShowWindow(Handle, SW_HIDE);
    SetWindowLong(Handle, GWL_EXSTYLE, NewStyle);
    ShowWindow(Handle, SW_SHOW);
  end;
end;

procedure AdjustHiddenWindow;
begin
  if OriginalStyle = 0 then
    OriginalStyle := GetWindowLong(Application.Handle,GWL_EXSTYLE);
  if WindowManagerMode <> wmmStandard then
    AdjustWindow(Application.Handle, OriginalStyle and not WS_EX_APPWINDOW or WS_EX_TOOLWINDOW)
  else
    AdjustWindow(Application.Handle, OriginalStyle)
end;

procedure TFixForm.CreateParams(var Params: TCreateParams);
begin
  FFixThisForm := FixAllForms;
  inherited;
  if FOriginalStyle = 0 then
    FOriginalStyle := Params.ExStyle;
  if FixThisForm then begin
    AdjustHiddenWindow;
    if WindowManagerMode <> wmmStandard then
      Params.ExStyle := FOriginalStyle and not WS_EX_TOOLWINDOW or WS_EX_APPWINDOW;
  end;
  { Must switch ownership of every form to MainForm, so that the
    application minimizes corrrectly }
  if Assigned(Application.MainForm) and (Application.MainForm <> Self) then
    case WindowManagerMode of
      wmmOnTop:       Params.WndParent := Application.MainFormHandle;
      wmmIndependent: Params.WndParent := 0;
      wmmStandard:    Params.WndParent := Application.Handle;
    end;
end;

type TOpenWindow = class(TWinControl);

procedure TFixForm.UpdateActions;
var
  i: Integer;
begin
  inherited;
  { Update is deferred to here as VCL seems to be confused about recreated window handles }
  if PendingWindowManagerMode <> WindowManagerMode then begin
    WindowManagerMode := PendingWindowManagerMode;
    if Assigned(Application.MainForm) and Application.MainForm.HandleAllocated then begin
      AdjustHiddenWindow;
      if WindowManagerMode <> wmmStandard then
        AdjustWindow(Application.MainForm.Handle, FOriginalStyle and not WS_EX_TOOLWINDOW or WS_EX_APPWINDOW)
      else
        AdjustWindow(Application.MainForm.Handle, FOriginalStyle);
      for i := 0 to Screen.CustomFormCount - 1 do
        if (Screen.CustomForms[i] <> Application.MainForm)
           and Screen.CustomForms[i].HandleAllocated
           and (Screen.CustomForms[i] is TFixForm) then
          TOpenWindow(Screen.CustomForms[i]).RecreateWnd;
    end;
  end;
end;

procedure TFixForm.WMActivate(var Message: TWMActivate);
begin
 if (Message.Active = WA_ACTIVE) and not IsWindowEnabled(Handle) then begin
   SetActiveWindow(Application.Handle);
   Message.Result := 0;
 end else
   inherited;
end;

procedure TFixForm.WMSysCommand(var Message: TWMSysCommand);
begin
  if FixThisForm and (WindowManagerMode <> wmmStandard) then
    case (Message.CmdType and $FFF0) of
      SC_MINIMIZE:
      begin
        ShowWindow(Handle, SW_MINIMIZE);
        Message.Result := 0;
      end;
      SC_RESTORE:
      begin
        ShowWindow(Handle, SW_RESTORE);
        Message.Result := 0;
      end;
    else
      inherited;
    end
  else
    inherited;
end;
{$ENDIF}

{ TFixForm }

procedure TFixForm.AddShortCut(Item: TMenuItem; ShortCut: string);
begin
  Item.Caption := Item.Caption + #9 + Shortcut;
end;

end.

