unit Log;

interface

{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  {$IFNDEF CLX_} Messages, Graphics, Controls, Forms, Menus, StdCtrls, Dialogs, ExtCtrls,
  {$ELSE} QControls, Types, QGraphics, QForms, QStdCtrls, QExtCtrls, QDialogs, Qt, {$ENDIF}
  ScrlCtrl, FixFnt, Classes, SysUtils, CharType, SrchStrs, LogFind;

type

  TLogChangeType = (loNone, loRange, loAll);

  TLogChangedEvent = procedure (Sender: TObject; LogChangeType: TLogChangeType; First, Last: Integer) of object;

  TUndoFlag = (unModified, unEndUpdate);

  TUndoFlags = set of TUndoFlag;

  TUndoActions = (aPut, aDelete, aInsert, aExchange, aPoint);

  TMyPoint = record {To stop C++ Builder complaining }
    x: Integer;
    y: Integer;
  end;

  TUndoAction = record
    Flags: TUndoFlags;
    {$IFNDEF VER80} Line: string; {$ENDIF} { To avoid "String needs finalization error" }
    case Action:TUndoActions of
      aPut, aDelete, aInsert: (Number: Integer; {$IFDEF VER80} Line:string; {$ENDIF});
      aExchange: (Line1, Line2: Integer);
      aPoint:    (Point:TMyPoint);
  end;

  PUndoAction = ^TUndoAction;

  TLogStringList = class(TStringList)
  private
    FOnLogChanged: TLogChangedEvent;
    FUpdateType: TLogChangeType;
    FFirst, FLast: Integer;
    FUpdating: Boolean;
    FModified: Boolean;
    FUndoList: TList;
    FNextUndoItem: Integer;
    FRecordUndo: Boolean;
    FUndoAborted: Boolean;      { Set during an update if too many actions recorded }
    procedure UpdateRange(Line: Integer);
    procedure DoLogChange;
    function GetCanUndo: Boolean;
    function GetCanRedo: Boolean;
    function GetRecordUndo: Boolean;
    procedure SetRecordUndo(Value: Boolean);
  protected
    procedure Put(Index: Integer; const S: string); override;
    procedure SetUpdateState(IsUpdating: Boolean); override;
    procedure Changed; override;
    procedure ClearUndoListFrom(Index: Integer);
    function NextUndoAction: PUndoAction;
  public
    GetPointVar: function :TPoint of object;
    SetPointVar: procedure (Point:TPoint) of object;
    constructor Create;
    destructor Destroy; override;
    function Add(const S: string): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure DiscardFirstThird;
    procedure Exchange(Index1, Index2: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure Undo; virtual;
    procedure Redo; virtual;
    procedure ClearUndo;
    procedure SaveToStreamAlien(Stream: TStream);
    procedure SaveToFileAlien(const Filename: string);
    procedure SaveToFileEnh(const FileName: string; MakeBackup, AlienEol: Boolean);
    property CanUndo: Boolean read GetCanUndo;
    property CanRedo: Boolean read GetCanRedo;
    property Modified: Boolean read FModified write FModified;
    property OnLogChanged: TLogChangedEvent read FOnLogChanged write FOnLogChanged;
    property RecordUndo:Boolean read GetRecordUndo write SetRecordUndo;
    property Updating: Boolean read FUpdating;
  end;

  TNewLineEvent = procedure (Sender: TObject; var Line:string; var Echo:Boolean) of object;

  TLog = class(TScrollableControl)
  private
    FLines: TLogStringList;
    MidLine: Boolean;
    FSelRect: TRect;
    FCharSize: TPoint;
    nVextra: Integer; { Extra lines for keyboard input }
    FMaxWidth: Integer;
    FAutoScroll: Boolean;
    FScrollToEndPending: Boolean;
    FSelecting, FLineSelect, FWordSelect, FWasColumnSelecting: Boolean;
    FAnchor: TPoint;
    FCaretPos: TPoint;
    FCaretTarget: Integer;
    FWordRect: TRect;

    { To implement scrolling by dragging mouse outside control area }
    FLastMouseX, FLastMouseY: Integer;
    FRepeatTimer: TTimer;
    FDragScrollX, FDragScrollY: Integer;

    FFixedFont: TFixedFont;
    FComboBox: TComboBox;
    FForcePending: Boolean;
    FWantTabs: Boolean;
    FOnNewLine: TNewLineEvent;
    FEcho: Boolean;
    FTypeAhead: Boolean;
    FTypeAheadBuffer: string[255];
    FnReturns: Integer;
    FAlwaysShowComboBox: Boolean;
    FAutoHistory: Boolean;
    FOnChange: TNotifyEvent;
    FHideSelection: Boolean;
    FTabSpacing: Integer;
    FMaxLines: Integer;
    FAutoDiscard: Boolean;
    procedure SetLines(const Value: TLogStringList);
    procedure SetSelRect(NewRect: TRect);
    procedure UpdateCharSize;
    procedure LinesChanged(Sender: TObject; LogChangeType: TLogChangeType; First, Last: Integer);
    procedure SetAutoScroll(const Value: Boolean);
    procedure SetAreaSelected(const Value: Boolean);
    function GetAreaSelected: Boolean;
    procedure CheckLines(First, Last: Integer);
    procedure TimerExpired(Sender: TObject);
    procedure InvalidateCharRect(r: TRect);
    procedure UpdateSelectedArea(X, Y: Integer; ForceVisible: Boolean);
    function FindWordAt(m: TPoint): TRect;
    {$IFDEF MSWINDOWS}
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMSetFocus); message WM_KILLFOCUS;
    procedure WMEraseBkgnd(var Message: TWmEraseBkgnd); message WM_ERASEBKGND;
    {$ENDIF}
    procedure SetFixedFont(const Value: TFixedFont);
    procedure FixedFontChanged(Sender: TObject);
    procedure SetAcceptingInput(const Value: Boolean);
    function GetAcceptingInput: Boolean;
    procedure PositionEditWindow;
    procedure SetTypeAhead(const Value: Boolean);
    procedure SetAlwaysShowComboBox(const Value: Boolean);
    function GetHasSelection: Boolean;
    function GetCanDeleteKey: Boolean;
    function GetRecordUndo: Boolean;
    procedure SetRecordUndo(Value: Boolean);
    procedure SetHideSelection(AHide: Boolean);
    function GetModified: Boolean;
    procedure SetModified(Value: Boolean);
    function GetSelLength: {$IFDEF WIN32} Integer; {$ELSE} Longint; {$ENDIF}
    function GetSelText: string;
    procedure SetTabSpacing(Value: Integer);
    procedure SetWantTabs(AWantTabs: Boolean);
    procedure SetMaxLines(const Value: Integer);
  protected
    FLastCaretSize: Integer; { Non zero if a caret has been set up }
    FColumnSelect: Boolean;
    {$IFDEF LCL_}
    function CanTab: Boolean; override;
    {$ENDIF}
    function ConvertMouse(x, y: Integer): TPoint;
    function IsInWord(c: Char):Boolean;
    procedure InitWidget; {$IFDEF CLX_} override; {$ENDIF}
    procedure UpdateScrollBars; override;
    procedure Resize; override;
    {$IFDEF CLX_}
    function NeedKey(Key: Integer; Shift: TShiftState; const KeyText: WideString): Boolean; override;
    {$ENDIF}
    procedure Paint; override;
    procedure MakeSane(var p: TPoint);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    {$IFDEF DFS_DELPHI_4_UP}
    function DoMouseWheelDown(Shift: TShiftState; {$IFDEF CLX_} const {$ENDIF} MousePos: TPoint): Boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; {$IFDEF CLX_} const {$ENDIF} MousePos: TPoint): Boolean; override;
    {$ENDIF}
    procedure DoEnter; override;
    procedure DoNewLine(Line:string); virtual;
    function GetPopupMenu: TPopupMenu; override;
    procedure FindWithDialog(Sender: TObject);
    function FindDialogComponent: TLogFindDialog;
    function ReplaceDialogComponent: TLogReplaceDialog;
    procedure DoReplace(Sender: TObject);
    procedure SetCaretPos(Pos: TPoint); virtual;
    procedure CaretTargetIsHere; { May need to provide other functions, but this is the only current one required }
    property Anchor: TPoint read FAnchor write FAnchor;
    property CaretTarget: Integer read FCaretTarget;
  public
    {$IFDEF CLX_}
    procedure ScrollBy(DeltaX, DeltaY: Integer); override;
    {$ENDIF}
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function AboveLeft(A, B:TPoint):Boolean;

    procedure ScrollToEnd;
    procedure ExtendSelectionTo(p: TPoint; ForceVisible: Boolean; Shift: TShiftState);
    procedure ExtendLineSelectionTo(Line: Integer; ForceVisible: Boolean);

    procedure Clear;
    procedure CutToClipboard;
    procedure DeleteKey;
    procedure ClearUndo;
    function Search(const Target: string; Options:TFindOptions): Boolean; virtual;
    function CanSortOnSelectedColumns: Boolean;
    procedure SortOnSelectedColumns(Ascending: Boolean);

    procedure CheckDiscardLines;
    procedure ChangeCaretPos(x, y: Integer; ExtendSelection: Boolean);
    procedure ForceInputWindowVisible;
    procedure ShowGotoLineDialog;
    procedure ShowFindDialog;
    procedure ShowReplaceDialog;
    procedure FindNext;
    procedure FindThisWord;
    procedure ClearTypeAhead;
    procedure Add(const s: string);
    procedure MakeVisible(CharPoint: TPoint);
    procedure SaveAreaToStream(Area:TRect; Stream: TStream; OnlyColumn: Boolean);
    function GetSelTextBuf(Buffer: PChar; BufSize: Integer): Integer;
    procedure BeginUpdate; virtual;
    procedure EndUpdate; virtual;

    { Characters in selected area - TopLeft is first, Bottom right is one beyond }
    property SelRect: TRect read FSelRect write SetSelRect;
    property AreaSelected:Boolean read GetAreaSelected write SetAreaSelected;
    function WordAtCursor(WantedChar: CharFilter): string;

    function TabCharPos(x, y:Integer):Integer; { x shited right to account for tabs on line y }
    function UnTabCharPos(x, y: Integer): Integer;

    property Canvas;
    property MaxWidth:Integer read FMaxWidth;
    property Modified: Boolean read GetModified write SetModified;
    property HasSelection:Boolean read GetHasSelection;
    property CanDeleteKey: Boolean read GetCanDeleteKey;
    property CharSize: TPoint read FCharSize;
    property ComboBox: TComboBox read FComboBox;
    property CaretPos: TPoint read FCaretPos write SetCaretPos;
    property SelLength: {$IFDEF WIN32} Integer {$ELSE} Longint {$ENDIF} read GetSelLength;
    property SelText: string read GetSelText;

  published
    procedure CopyToClipboard; virtual;
    function CanCopyToClipboard: Boolean; virtual;
    procedure ClearSelection; virtual;
    procedure PasteFromClipboard; virtual;
    function CanPasteFromClipboard: Boolean; virtual;
    function CanUndo: Boolean;
    function CanRedo: Boolean;
    procedure Undo;
    procedure Redo;
    procedure SelectAll; virtual;

    property Lines: TLogStringList read FLines write SetLines;
    property AutoScroll:Boolean read FAutoScroll write SetAutoScroll default True;
    property TopMargin default 2;
    property LeftMargin default 4;
    property FixedFont:TFixedFont read FFixedFont write SetFixedFont;
    property WantTabs:Boolean read FWantTabs write SetWantTabs default False;
    property AcceptingInput:Boolean read GetAcceptingInput write SetAcceptingInput default False;
    property OnNewLine:TNewLineEvent read FOnNewLine write FOnNewLine ;
    property Echo: Boolean read FEcho write FEcho default False;
    property TypeAhead: Boolean read FTypeAhead write SetTypeAhead default False;
    property AlwaysShowComboBox: Boolean read FAlwaysShowComboBox write SetAlwaysShowComboBox default True;
    property AutoHistory:Boolean read FAutoHistory write FAutoHistory default True;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property HideSelection: Boolean read FHideSelection write SetHideSelection default True;
    property RecordUndo:Boolean read GetRecordUndo write SetRecordUndo default True;
    property TabSpacing: Integer read FTabSpacing write SetTabSpacing default 8;
    property MaxLines: Integer read FMaxLines write SetMaxLines default 70000;
    property AutoDiscard: Boolean read FAutoDiscard write FAutoDiscard default False;

    property Align;  { publish properties }
    {$IFDEF DFS_COMPILER_4_UP}
    property Anchors;
    property Constraints;
    {$ENDIF}
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderWidth;
    property BorderStyle;
    property Color;
    {$IFDEF MSWINDOWS}
    property Ctl3D;
    property ParentCtl3D;
    {$ENDIF}
    property ParentColor;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;  { publish events }
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

{$IFDEF SINGLE}
procedure Register;
{$ENDIF}

implementation

uses {$IFNDEF CLX_} ClipBrd, {$ELSE} QClipBrd, Libc, QCaret, {$ENDIF}
     {$IFDEF LCL_} LCLType, LCLIntf, {$ENDIF}
     HugeClip, TempStrm, ComboFix, IntVec, Huge, ResActn, EditLog, ExActn,
     UCompat, WCompat, MathSup;

{$IFDEF SINGLE}
{$IFDEF VER80}
{$R Log.d16}
{$ELSE}
{$R Log.d32}
{$ENDIF}
procedure Register;
begin
  RegisterComponents('CMW', [TLog]);
end;
{$ENDIF}

type
  TLogComboBox = class(TComboBox)
  private
    Log: TLog;
    FAnyFocused: Boolean;
    {$IFNDEF DFS_COMPILER_2_UP}
    { FEditHandle: HWnd; }
    {$ENDIF}
  protected
    {$IFDEF MSWINDOWS}
    {$IFNDEF LCL}
    procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;
    {$ENDIF}
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    {$ENDIF}
    {$IFDEF UNIX}
    procedure DoEnter; override;
    procedure DoExit; override;
    {$ENDIF}
    {$IFNDEF CLX_}
    procedure CreateWnd; override;
    {$ELSE}
    procedure CreateWidget; override;
    {$ENDIF}
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
  public
    constructor CreateWithLog(AOwner:TLog);
    property AnyFocused:Boolean read FAnyFocused; { As Focused, but includes child controls }
    {$IFDEF MSWINDOWS}
    {$IFNDEF DFS_COMPILER_2_UP}
    property EditHandle: HWnd read FEditHandle;
    {$ELSE}
    {$IFNDEF LCL}
    property EditHandle;
    {$ENDIF}
    {$ENDIF}
    {$ENDIF}
  published
    procedure CopyToClipboard; virtual;
    function CanCopyToClipboard: Boolean; virtual;
  end;

function Max(a, b: Integer):Integer;
begin
  if a > b then
    Result := a
  else
    Result := b;
end;

function Min(a, b: Integer):Integer;
begin
  if a < b then
    Result := a
  else
    Result := b;
end;

{ TLog }

procedure TLog.InitWidget;
begin
  {$IFDEF CLX_}
  inherited InitWidget;
  {$ENDIF}
  Color := clWindow;
  FCharSize.x := 0;
end;

constructor TLog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csFramed, csCaptureMouse]
                  {$IFDEF CLX_} - [csNoFocus] {$ENDIF};
  {$IFDEF LCL_}
  // FCompStyle:= csFixed; { Force control type that can get the focus, and have a caret }
  {$ENDIF}
  ParentColor := False;
  ParentFont := False;
  FFixedFont := TFixedFont.Create;
  FFixedFont.OnChange := {$IFDEF FPC}@{$ENDIF}FixedFontChanged;
  FLines := TLogStringList.Create;
  FLines.OnLogChanged := {$IFDEF FPC}@{$ENDIF}LinesChanged;
  FLines.RecordUndo := True;
  Height := 100;
  Width := 100;
  FMaxLines := 70000;
{  EditWindowActive := False; }
  MidLine := False;
  FSelRect := Rect(-1,-1,-1,-1);
  nVextra := 0;
  FAutoScroll := True;
  FTopMargin := 2;
  FLeftMargin := 4;
  TabStop := True;
  FMaxWidth := -1;
  FAlwaysShowComboBox := True;
  FAutoHistory := True;
  FHideSelection := True;
  FTabSpacing := 8;
  WantTabs := False;
  {$IFNDEF CLX_}
  InitWidget;
  {$ENDIF}
end;

procedure TLog.SetWantTabs(AWantTabs: Boolean);
begin
  FWantTabs := AWantTabs;
  {$IFDEF CLX_}
  InputKeys := [ikArrows, ikChars, ikReturns, ikEdit, ikNav];
  if FWantTabs then
    InputKeys := InputKeys + [ikTabs];
  {$ENDIF}
  { TODO : LCL_ WantTabs }
end;

{$IFDEF LCL_}
function TLog.CanTab: Boolean;
begin
  Result := True;
end;
{$ENDIF}

procedure TLog.SetTabSpacing(Value: Integer);
begin
  if Value <> TabSpacing then begin
    FTabSpacing := Value;
    Invalidate;
  end;
end;

{ x shifted right to account for tabs on line y }
function TLog.TabCharPos(x, y:Integer):Integer;
var
  p, e, i: Integer;
  Line: string;
begin
  if (x > 0) and (x < MaxInt) and (y >= 0) and (y < Lines.Count) then begin
    p := Pos(#9, Lines[y]);
    if (p > 0) and (p <= x) then begin
      Line := Lines[y];
      e := x;
      if e > Length(Line) then
        e := Length(Line);
      Result := 0;
      for i := p to e do
        if Line[i] = #9 then
          Inc(Result, TabSpacing - (i+Result-1) mod TabSpacing - 1);
      Inc(Result, x);
      Exit;
    end;
  end;
  Result := x;
end;

function TLog.UnTabCharPos(x, y: Integer): Integer;
{ Convert from char cell count to character count, i.e. subtract off tabs }
var
  p, TotalExtra, Extra: Integer;
  Line: string;
begin
  Result := x;
  if (x > 0) and (x < MaxInt) and (y >= 0) and (y < Lines.Count) then begin
    p := Pos(#9, Lines[y]);
    if (p > 0) and (p <= x) then begin
      Line := Lines[y];
      TotalExtra := 0;
      while (p+TotalExtra <= Result) and (p <= Length(Line)) do
        if Line[p] = #9 then begin
          Extra := TabSpacing - (p+TotalExtra-1) mod TabSpacing - 1;
          if p+TotalExtra+Extra > Result then begin
            if p+TotalExtra+Extra div 2 > Result then
              Result := p-1
            else
              Result := p;
            Exit;
          end else begin
            Inc(TotalExtra, Extra);
            Inc(p);
          end;
        end else
          Inc(p);
      Dec(Result, TotalExtra);
    end;
  end;
end;

destructor TLog.Destroy;
begin
  FLines.Free;
  FLines := nil;
  FFixedFont.Free;
  FRepeatTimer.Free;
  inherited Destroy;
end;

function TLog.GetModified: Boolean;
begin
  Result := Lines.Modified;
end;

procedure TLog.SetModified(Value: Boolean);
begin
  Lines.Modified := Value;
end;

procedure TLog.LinesChanged(Sender: TObject; LogChangeType: TLogChangeType;
  First, Last: Integer);
{ var
  r: TRect;
  t: Longint; }
begin
  if not (csLoading in ComponentState) then begin
    if LogChangeType = loAll then begin
      Invalidate;
      VertScrollBar.Position := 0;
      HorzScrollBar.Position := 0;
      FMaxWidth := -1;
      if AutoScroll then ScrollToEnd;
    end else begin
      CheckLines(First, Last);
      if AutoScroll and (Last >= Lines.Count-1) then ScrollToEnd;
      InvalidateCharRect(Rect(0, First, MaxInt, Last));
    end;
    CheckDiscardLines;
  end else
    FMaxWidth := -1;
  SetCaretPos(CaretPos); { Force sensible value }
  AreaSelected := False;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TLog.CheckLines(First, Last: Integer);
var
  i: Integer;
begin
  for i := Max(0, First) to Min(Last, Lines.Count-1) do
    FMaxWidth := Max(TabCharPos(Length(Lines[i]),i), FMaxWidth);
end;

{$IFDEF CLX_}
procedure TLog.ScrollBy(DeltaX, DeltaY: Integer);
begin
  HideCaret;
  inherited ScrollBy(DeltaX, DeltaY);
  ScrollCaretPos(DeltaX, DeltaY);
  ShowCaret;
end;

type
  TOpenCanvas = class(TCanvas);
{$ENDIF}

{$IFDEF LCL_}
type
  TOpenCanvas = class(TCanvas)
  public
    procedure RequiredState(ReqState: TCanvasState);
  end;
  
procedure TOpenCanvas.RequiredState(ReqState: TCanvasState);
begin
  inherited RequiredState(ReqState);
end;
{$ENDIF}
  
procedure TLog.Paint;
var
  r: TRect;
  PaintBeg, PaintEnd, i, lasty: Integer;
  Reversed, ShowSelection: Boolean;
  Fgnd, Bgnd: TColor;
  p: PChar; { Next character to display }
  cx: Integer; { Position to display it (in character cell units) }
  CharWidth: Integer; { Number of character cells visible in a row }
  {$IFDEF VCL_}
  TabStopPos, TabOrigin, MaxChunk: Integer;
  {$ENDIF}
  {$IFNDEF VCL_}
  { Must expand tabs to spaces, as tab output does not erase spaces }
  procedure MyTabbedTextOut(l: Integer);
  var
    PTab: PChar;
    TabStart, TabEnd: Integer;
    n, nSpaces: Integer;
    w: {$IFDEF CLX_}WideString{$ELSE}string{$ENDIF}; { Text that will actually be printed }
    wx: Integer;   { Where it will start }
    {$IFDEF LCL_}
    r: TRect;
    {$ENDIF}
    function FindNextTab: Boolean;
    begin
      PTab := StrScan(p, #9);
      Result := PTab <> nil;
      if Result then begin
        TabStart := PTab-p + cx;
        nSpaces := TabSpacing - (TabStart+HorzScrollBar.Position) mod TabSpacing;
        TabEnd := TabStart + nSpaces;
      end;
    end;
    procedure AddChunk(s: string);
    begin
      if wx < 0 then begin
        w := s;
        wx := cx;
      end else
        w := Concat(w, s);
    end;
  begin
    wx := -1;
    while (cx < 0) and (l > 0) do begin
      { Skip to first visible char }
      n := -cx;
      if n > l then n := l;
      if FindNextTab then begin
        if (n > PTab-p) and (TabStart < 0) then begin
          n := PTab-p+1;
          if TabEnd > 0 then begin
            { Tab at least partly visible }
            cx := 0;
            AddChunk(StringOfChar(' ', TabEnd));
          end;
          cx := TabEnd
        end;
      end else
        Inc(cx, n); { No tabs to left of visible area }
      Dec(l, n);
      Inc(p, n);
    end;
    while (l > 0) and (cx < CharWidth) do begin
      n := l;
      if FindNextTab and (n > PTab - p) then
        n := PTab - p;
      if n > CharWidth - cx then
        n := CharWidth - cx;
      if n > 0 then begin
        AddChunk(Copy(p, 1, n));
        Inc(p, n);
        Inc(cx, n);
        Dec(l, n);
      end;
      if (l > 0) and (p^ = #9) then begin
        AddChunk(StringOfChar(' ', nSpaces));
        Inc(p);
        Dec(l);
        Inc(cx, nSpaces);
      end;
    end;
    if wx >= 0 then begin
      {$IFDEF CLX_}
      QPainter_drawText(Canvas.Handle, wx*CharSize.x+PaintRect.Left, lasty,
                        Length(w)*CharSize.x, CharSize.y, Ord(AlignmentFlags_DontClip),
                        @w, -1, nil, nil);
      {$ELSE}
       TOpenCanvas(Canvas).RequiredState([csHandleValid, csFontValid, csBrushValid]);
       r.Left := wx*CharSize.x+PaintRect.Left;
       r.Top := lasty;
       r.Bottom := Min(lasty + CharSize.y, PaintRect.Bottom);
       r.Right := r.Left + CharSize.x*Length(w);
       ExtUTF8Out(Canvas.Handle, r.left, lasty, ETO_OPAQUE, @r, PChar(w), Length(w), nil);
      {$ENDIF}
    end;
  end;
  {$ENDIF}
  {$IFDEF VCL_}
  procedure MyTabbedTextOut(l: Integer);
  var
    n: Integer;
  begin
    { Split large TabbedTextOut to avoid 16 bit GDI overflow in result
      Also some optimisation of paint for long lines }
    while l > 0 do begin
      if cx >= CharWidth then begin
        Inc(p, l);
        Exit;
      end;
      if l > MaxChunk then
        n := MaxChunk
      else
        n := l;
      cx := cx + LOWORD(TabbedTextOut(Canvas.Handle,
                                      cx*CharSize.x+PaintRect.Left, lasty,
                                      p, n, 1, TabStopPos, TabOrigin))
                 div CharSize.x;
      Inc(p, n);
      Dec(l, n);
    end;
  end;
  {$ENDIF}
var
  l, c0, cl: Integer;
  EffRect: TRect;
{$IFDEF LCL_}
  TextStyle: TTextStyle;
{$ENDIF}
begin
  {$IFNDEF VCL_}
  if FLastCaretSize <> 0 then
    HideCaret{$IFDEF LCL_}(Handle){$ENDIF};
  {$ENDIF}
  if FMaxWidth < 0 then CheckLines(0, Lines.Count-1);
  UpdateScrollBars;
  if FForcePending then
    ForceInputWindowVisible;
  FForcePending := False;
  inherited Paint;
  Fgnd := ColorToRGB(Canvas.Font.Color);
  Bgnd := ColorToRGB(Canvas.Brush.Color);
  IntersectRect(r, Canvas.ClipRect, PaintRect);
  Reversed := False;
  ShowSelection := Focused or not HideSelection;
  if not Focused and HideSelection and AcceptingInput then
      if ComboBox is TLogComboBox then
        ShowSelection := TLogComboBox(ComboBox).AnyFocused
      else
        ShowSelection := ComboBox.Focused;
  PaintBeg := VertScrollBar.Position + Max( 0, (r.top - PaintRect.Top) div CharSize.y );
  PaintEnd := Min( Lines.Count, VertScrollBar.Position + r.bottom div CharSize.y + 1);
  lasty := PaintRect.Top + CharSize.y*(PaintBeg-VertScrollBar.Position);
  CharWidth := PaintWidth div CharSize.x + 1;
  {$IFDEF VCL_}
  TabStopPos := TabSpacing * CharSize.x;
  MaxChunk := High(SmallInt) div TabStopPos;
  { Origin for tab stops }
  TabOrigin := PaintRect.Left - CharSize.x * HorzScrollBar.Position;
  {$ENDIF}
  {$IFDEF CLX_}
  QPainter_setBackgroundMode(Canvas.Handle, BGMode_OpaqueMode);
  QPainter_setBackgroundColor(Canvas.Handle, QColor(Bgnd));
  TOpenCanvas(Canvas).RequiredState([csHandleValid, csFontValid]);
  {$ENDIF}
  {$IFDEF LCL_}
  TextStyle := Canvas.TextStyle;
  TextStyle.Opaque := True;
  Canvas.TextStyle := TextStyle;
  {$ENDIF}
  EffRect := SelRect;
  if FColumnSelect then begin
    InOrder(EffRect.Left,EffRect.Right);
    InOrder(EffRect.Top,EffRect.Bottom);
  end;
  for i := PaintBeg to PaintEnd-1 do begin
    l := Length(Lines[i]);
    p := AddressOfFirstChar(Lines[i]);
    cx := - HorzScrollBar.Position;
    if ShowSelection and (EffRect.Top >= 0) and (i >= EffRect.Top) and (i <= EffRect.Bottom) then begin
      c0 := 0; { first char on line selected }
      cl := l-1; { last char on line selected }
      if FColumnSelect or ((i = EffRect.Top) and (EffRect.Left > 0)) then begin { First line unselected part }
        MyTabbedTextOut(Min(EffRect.Left,l));
        c0 := EffRect.Left;
      end;
      if not Reversed then begin
        Reversed := True;
        {$IFNDEF CLX_}
        SetTextColor(Canvas.Handle, GetSysColor(COLOR_HIGHLIGHTTEXT));
        SetBkColor(Canvas.Handle, GetSysColor(COLOR_HIGHLIGHT));
        {$ELSE}
        Canvas.Font.Color := clHighlightedText;
        TOpenCanvas(Canvas).RequiredState([csHandleValid, csFontValid]);
        QPainter_setBackgroundColor(Canvas.Handle, QColor(clHighlight));
        {$ENDIF}
      end;
      if FColumnSelect or (i = EffRect.Bottom) then
        cl := Min(EffRect.Right,l)-1;
      MyTabbedTextOut(cl-c0+1);
      Dec(l,Max(cl-c0+1,0)+c0); { Chars remaining }
      if FColumnSelect or (i = EffRect.Bottom) then begin { last line selected }
        Reversed := False;
        {$IFNDEF CLX_}
        SetTextColor(Canvas.Handle, Fgnd);
        SetBkColor(Canvas.Handle, Bgnd);
        {$ELSE}
        Canvas.Font.Color := Fgnd;
        TOpenCanvas(Canvas).RequiredState([csHandleValid, csFontValid]);
        QPainter_setBackgroundColor(Canvas.Handle, QColor(Bgnd));
        {$ENDIF}
        { unselected part on last line }
        MyTabbedTextOut(l);
      end;
    end else
      MyTabbedTextOut(l);
    if cx < CharWidth then begin
      if cx < 0 then cx := 0;
      if lasty < PaintRect.Bottom then
        Canvas.FillRect(Rect(PaintRect.Left+cx*CharSize.x, lasty, PaintRect.Right, Min(lasty+CharSize.y, PaintRect.Bottom)));
    end;
    Inc(lasty, CharSize.y);
  end;
  if lasty < PaintRect.Bottom then
    Canvas.FillRect(Rect(PaintRect.Left, lasty, PaintRect.Right, PaintRect.Bottom));
  if Reversed then begin
    {$IFNDEF CLX_}
    SetTextColor(Canvas.Handle, Fgnd);
    SetBkColor(Canvas.Handle, Bgnd);
    {$ELSE}
    Canvas.Font.Color := Fgnd;
    QPainter_setBackgroundColor(Canvas.Handle, QColor(Bgnd));
    {$ENDIF}
  end;
  {$IFNDEF VCL_}
  if FLastCaretSize <> 0 then
    ShowCaret{$IFDEF LCL_}(Handle){$ENDIF};
  {$ENDIF}
end;

procedure TLog.ScrollToEnd;
begin
  if not (csLoading in ComponentState) and HandleAllocated then begin
    FScrollToEndPending := False;
    UpdateScrollBars;
    if Lines.Count >= VertScrollBar.Position + PaintHeight div CharSize.y then
      VertScrollBar.Position := Lines.Count - PaintHeight div CharSize.y + 5;
    if AcceptingInput then
      ForceInputWindowVisible;
  end else
    FScrollToEndPending := True;
end;

procedure TLog.SetAutoScroll(const Value: Boolean);
begin
  FAutoScroll := Value;
  if AutoScroll then
    ScrollToEnd
  else
    FScrollToEndPending := False;
end;

procedure TLog.SetLines(const Value: TLogStringList);
begin
  FLines.Assign(Value);
end;

procedure TLog.SetHideSelection(AHide: Boolean);
begin
  if FHideSelection <> AHide then begin
    FHideSelection := AHide;
    if not (csLoading in ComponentState) then
      InvalidateCharRect(SelRect)
  end;
end;

procedure TLog.SetSelRect(NewRect: TRect);
var
  r1, r2: TRect;
  procedure Order(var a, b: TPoint);
  var
    t: TPoint;
  begin
    if not AboveLeft(a, b) then begin
      t := a;
      a := b;
      b := t;
    end;
  end;
  procedure InvalidateLines(r: TRect);
  begin
    Inc(r.Bottom);
    r.Left := 0;
    r.Right := 0;
    InvalidateCharRect(r);
  end;
begin
  with NewRect do begin
    Order(TopLeft, BottomRight);
    if AboveLeft(TopLeft, BottomRight) then begin
      MakeSane(TopLeft);
      MakeSane(BottomRight);
    end else begin
      Top := -1;
      Bottom := -1;
      Left := -1;
      Right := -1;
    end;
  end;
  if not AreaSelected then begin
    { No old selection }
    if NewRect.Top >= 0 then
      InvalidateCharRect(NewRect)
  end else if NewRect.Top < 0 then
    { No new selection }
    if FWasColumnSelecting then
      InvalidateLines(SelRect)
    else
      InvalidateCharRect(SelRect)
  else begin
    { Selection changes }
    if not (FWasColumnSelecting or FColumnSelect) then begin
      { Limit repaint region to newly selected/unselected areas }
      if AboveLeft(NewRect.TopLeft, SelRect.TopLeft) then begin
        r1 := NewRect;
        r2 := SelRect;
      end else begin
        r1 := SelRect;
        r2 := NewRect;
      end;
      Order(r1.BottomRight, r2.TopLeft);
      with r2 do Order(TopLeft, BottomRight);
      InvalidateCharRect(r1);
      InvalidateCharRect(r2);
    end else begin
      { skip opimization for column select mode; just invalidate all lines in both areas }
      r1.Top := Min(SelRect.Top,NewRect.Top);
      r1.Bottom := Max(SelRect.Bottom,NewRect.Bottom);
      InvalidateLines(r1);
    end;
    FWasColumnSelecting := FColumnSelect;
  end;
  FSelRect := NewRect;
end;

procedure TLog.UpdateCharSize;
var
  {$IFNDEF CLX_}
  tm: TTextMetric;
  {$ELSE}
  tm: QFontMetricsH;
  {$ENDIF}
begin
  FixedFont.CopyTo(Canvas.Font);
  if AcceptingInput then
    FixedFont.CopyTo(ComboBox.Font);
  {$IFNDEF CLX_}
  GetTextMetrics(Canvas.Handle, tm);
  FCharSize.x := Max(1, tm.tmAveCharWidth);
  FCharSize.y := Max(1, tm.tmHeight); { TODO: Should add tm.tmExternalLeading, but then blank areas between lines must be painted };
                                      { Of the common MSWINDOWS fonts, only seems to be an issue for Lucida Sans Typewriter }
  {$ELSE}
  tm := QFontMetrics_create(Canvas.Font.Handle);
  FCharSize.x := Max(1, QFontMetrics_width(tm, 'A'));
  FCharSize.y := Max(1, QFontMetrics_lineSpacing(tm));
  QFontMetrics_destroy(tm);
  {$ENDIF}
  VertScrollBar.ScaleFactor := FCharSize.y;
  HorzScrollBar.ScaleFactor := FCharSize.x;
  UpdateScrollBars;
  PositionEditWindow;
  SetCaretPos(CaretPos); { Force reposition of Caret }
end;

procedure TLog.UpdateScrollBars;
begin
  if not (Assigned(VertScrollBar) and Assigned(HorzScrollBar) and HandleAllocated) then
    Exit; { During creation }
  if CharSize.x = 0 then
    UpdateCharSize;
  { Note: some possibility of random turning on/off of scroll bars close to threshold }
  VertScrollBar.Range := Lines.Count - PaintHeight div CharSize.y + nVextra + 2;
  HorzScrollBar.Range := 5 + MaxWidth - PaintWidth div CharSize.x;
  HorzScrollBar.PageSize := PaintWidth div CharSize.x;
  VertScrollBar.PageSize := PaintHeight div CharSize.y;
  inherited UpdateScrollBars;
  if FScrollToEndPending then ScrollToEnd;
end;

procedure TLog.Resize;
begin
  inherited Resize;
  { Invalidate; // Required on CLX_/VCL_? }
  PositionEditWindow;
end;

{$IFDEF CLX_}
function TLog.NeedKey(Key: Integer; Shift: TShiftState; const KeyText: WideString): Boolean;
begin
  Result := inherited NeedKey(Key, Shift, KeyText);
  if Result and ((Key = Key_Enter) or (Key = Key_Return)) and (Shift <> []) then
    Result := False;
end;
{$ENDIF}

{$IFDEF MSWINDOWS}
procedure TLog.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := Message.Result or DLGC_WANTARROWS;
  if WantTabs then
    Message.Result := Message.Result or DLGC_WANTTAB;
end;

procedure TLog.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;
  if HideSelection and (not AcceptingInput or (ComboBox.Handle <> Message.FocusedWnd) ) then
    InvalidateCharRect(SelRect);
end;

procedure TLog.WMKillFocus(var Message: TWMSetFocus);
begin
  inherited;
  if HideSelection and (not AcceptingInput or (ComboBox.Handle <> Message.FocusedWnd) ) then
    InvalidateCharRect(SelRect);
end;

procedure TLog.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;
{$ENDIF}

procedure TLog.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  m: TPoint;
begin
  SetFocus;
  {$IFDEF UNIX}
  if not Focused then { MDI bug fix? - parent form has focus but control does not }
    GetParentForm(Self).SetFocusedControl(Self);
  {$ENDIF}
  if (Button = mbLeft) and (Lines.Count > 0) then begin
    FLineSelect := False;
    m := ConvertMouse(X, Y);
    if not FSelecting and not (ssDouble in Shift) then begin
      { Single Click }
      FSelecting := True;
      FWordSelect := False;
      if m.x < 0 then begin
        { Line selction }
        if ssCtrl in Shift then begin
          { entire text }
          SelRect := Rect(0, 0, 0, Lines.Count);
          FSelecting := False;
        end else if ssShift in Shift then begin
          { modified line }
          if SelRect.Left < 0 then
            FSelecting := False
          else begin
            FLineSelect := True;
            ExtendLineSelectionTo(m.y, False);
          end
        end else begin
          { just this line }
          FLineSelect := True;
          m.x := 0;
          MakeSane(m);
          FAnchor := m;
          ExtendLineSelectionTo(m.y, False);
        end;
      end else begin
        { character selection }
        if ssCtrl in Shift then begin
          { entire line }
          FLineSelect := True;
          m.x := 0;
          MakeSane(m);
          if not (ssShift in Shift) then
            FAnchor := m;
          ExtendSelectionTo(Point(32767, m.y), False, []);
        end else begin
          MakeSane(m);
          if not (ssShift in Shift) then
            FAnchor := m;
          FColumnSelect := ssAlt in Shift;
          ExtendSelectionTo(m, False, []);
        end;
      end;
      { *** SetCapture(edit_window); - implied by default handling }
    end else if (ssDouble in Shift) and (m.x >= 0) then begin
      { Double Click }
      FWordSelect := True;
      FSelecting := True;
      FWordRect := FindWordAt(m);
      with FWordRect do
        if ssShift in Shift then begin
          { modified line }
          if AboveLeft(SelRect.TopLeft, TopLeft) then
            ExtendSelectionTo(Point(Right, Top), False, [])
          else
            ExtendSelectionTo(TopLeft, False, []);
        end else begin
          { Start Selection }
          SelRect := FWordRect;
          FAnchor := SelRect.TopLeft;
        end;
    end;
  end;
  CaretTargetIsHere;
  inherited MouseDown(Button, Shift, X,  Y);
end;

procedure TLog.TimerExpired(Sender: TObject);
begin
  FRepeatTimer.Interval := 50;
  if MouseCapture then begin
    if FDragScrollX <> 0 then
      HorzScrollBar.Position := HorzScrollBar.Position + FDragScrollX;
    if FDragScrollY <> 0 then
      VertScrollBar.Position := VertScrollBar.Position + FDragScrollY;
    UpdateSelectedArea(FLastMouseX, FLastMouseY, False);
  end;
end;

procedure TLog.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if FSelecting and (ssLeft in Shift) then begin
    if X < PaintRect.Left then
      FDragScrollX := -Max((PaintRect.Left-X) div CharSize.x, 1)
    else if X > PaintRect.Right then
      FDragScrollX := Max((X-PaintRect.Right) div CharSize.x, 1)
    else
      FDragScrollX := 0;
    if Y < PaintRect.Top then
      FDragScrollY := -Max(-(Y-PaintRect.Top) div CharSize.y, 1)
    else if Y > PaintRect.Bottom then
      FDragScrollY := Max((Y-PaintRect.Bottom) div CharSize.y, 1)
    else
      FDragScrollY := 0;
    if (FDragScrollX or FDragScrollY) <> 0 then begin
      if FRepeatTimer = nil then
        FRepeatTimer := TTimer.Create(Self);
      FRepeatTimer.OnTimer := {$IFDEF FPC}@{$ENDIF}TimerExpired;
      FRepeatTimer.Interval := 50;
      FRepeatTimer.Enabled  := True;
    end;
    FLastMouseX := X;
    FLastMouseY := Y;
    UpdateSelectedArea(X, Y, False);
  end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TLog.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) then begin
    if FRepeatTimer <> nil then
      FRepeatTimer.Enabled  := False;
    FDragScrollX := 0;
    FDragScrollY := 0;
    if FSelecting then begin
      UpdateSelectedArea(X, Y, True);
      if not AreaSelected then
        AreaSelected := False { Force cannonical state - all -1's }
      {$IFNDEF MSWINDOWS}
        {$IFDEF LCL_}
      else
        PrimarySelection.AsText := GetSelText
        {$ENDIF}
      {$ENDIF}
       ;
      FSelecting := False;
      FWordSelect := False;
    end;
  end;
  if AcceptingInput then
    ComboBox.SetFocus;
  inherited MouseUp(Button, Shift, X,  Y);
end;

function TLog.ConvertMouse(x, y: Integer): TPoint;
begin
  if y < PaintRect.Top then
    Result.y := -1
  else
    Result.y := (y-PaintRect.Top) div CharSize.y;
  Inc(Result.y, VertScrollBar.Position);
  if FLineSelect or ((HorzScrollBar.Position = 0) and (x < PaintRect.Left)) then
    Result.x := -1
  else
    Result.x := UnTabCharPos((x-PaintRect.Left) div CharSize.x + HorzScrollBar.Position, Result.y);
end;

function TLog.AboveLeft(A, B: TPoint): Boolean;
begin
  if A.y = B.y then
    Result := A.x < B.x
  else
    Result := A.y < B.y;
end;

procedure TLog.InvalidateCharRect(r: TRect);
var
  p: TRect; { Rectangle in pixel units }
  procedure InvalidateChars(Left, Right: Integer);
  begin
    if Left < HorzScrollBar.Position then
      Left := HorzScrollBar.Position;
    if Right > HorzScrollBar.Position + HorzScrollBar.PageSize + 1then
      Right := HorzScrollBar.Position + HorzScrollBar.PageSize + 1;
    if Left < Right then begin
      p.Left := (Left - HorzScrollBar.Position)*CharSize.x + PaintRect.Left;
      p.Right := (Right - HorzScrollBar.Position)*CharSize.x + PaintRect.Left;
      if p.Right > PaintRect.Right then
        p.Right := PaintRect.Right;
      InvalidateRect({$IFDEF MSWINDOWS}Handle, @{$ENDIF}p, False);
    end;
  end;
begin
  if r.Top > VertScrollBar.Position + VertScrollBar.PageSize then Exit;
  if r.Top < VertScrollBar.Position then begin
    r.Top := VertScrollBar.Position-1;
    r.Left := 0;
  end;
  if r.Bottom > VertScrollBar.Position + VertScrollBar.PageSize then begin
    r.Bottom := VertScrollBar.Position + VertScrollBar.PageSize;
    r.Right := MaxInt;
  end;
  if not AboveLeft(r.TopLeft, r.BottomRight) then Exit;
  p.Top := (r.Top - VertScrollBar.Position) * CharSize.y + PaintRect.Top;
  if (r.Left > 0) and (r.Top < r.Bottom) then begin
    { Do first (part) Line }
    { p.Left :=  TabCharPos(r.Left, r.Top) * CharSize.x; }
    p.Bottom := p.Top + CharSize.y;
    InvalidateChars(TabCharPos(r.Left, r.Top), MaxInt);
    Inc(r.Top);
    p.Top := p.Bottom;
    r.Left := 0;
  end;
  if r.Top < r.Bottom then begin
    { Do Block }
    p.Bottom := (r.Bottom - VertScrollBar.Position) * CharSize.y + PaintRect.Top;
    p.Left := PaintRect.Left;
    p.Right := PaintRect.Right;
    InvalidateRect({$IFDEF MSWINDOWS}Handle, @{$ENDIF}p, False);
    p.Top := p.Bottom;
  end;
  { Last Line }
  p.Bottom := p.Top + CharSize.y;
  InvalidateChars(TabCharPos(r.Left, r.Bottom), TabCharPos(r.Right, r.Bottom));
end;

procedure TLog.UpdateSelectedArea(X, Y: Integer; ForceVisible: Boolean);
{ Private method, called during MouseMove/MouseUp/Autoscroll Timer events only }
var
  m: TPoint;
  w: TRect;
begin
  if FSelecting and (Lines.Count > 0) then begin
    m := ConvertMouse(X, Y);
    if FLineSelect then
      ExtendLineSelectionTo(m.y, ForceVisible)
    else if FWordSelect then begin
      w := FindWordAt(m);
      if AboveLeft(SelRect.TopLeft, w.TopLeft) then
        ExtendSelectionTo(w.BottomRight, ForceVisible, [])
      else if (SelRect.Top <> w.Top) or (SelRect.Left <> w.Left) then
        ExtendSelectionTo(w.TopLeft, ForceVisible, []);
    end else begin
      ExtendSelectionTo(m, ForceVisible, []);
    end;
    CaretTargetIsHere;
  end;
end;

function TLog.IsInWord(c: Char):Boolean;
begin
{  Result := not ((c = ' ') or (c = #9)); }
  Result := c in ['A'..'Z', 'a'..'z', '_', '$', '0'..'9', '.'];
end;

function TLog.FindWordAt(m: TPoint): TRect;
var
  l: Integer;
begin
  with Result do begin
    { Make Sane, but different defaults }
    if m.y < 0 then
      m.y := 0
    else if m.y >= Lines.Count then
      m.y := Lines.Count - 1;
    l := Length(Lines[m.y]);
    Right := Min(m.x, Max(l-1,0));
    Left := Right;
    Bottom := m.y;
    Top := Bottom;
    { Find start of word }
    while (Left > 0) and not IsInWord(Lines[m.y][Left+1]) do Dec(Left);
    while (Left > 0) and IsInWord(Lines[m.y][Left]) do Dec(Left);
    { Find start of next word }
    while (Right < l) and IsInWord(Lines[m.y][Right+1]) do Inc(Right);
    while (Right < l) and not IsInWord(Lines[m.y][Right+1]) do Inc(Right);
  end;
end;

procedure TLog.MakeVisible(CharPoint: TPoint);
var
  xtab: Integer;
begin
  if not (Visible and HandleAllocated
          {$IFDEF CLX_} and QWidget_isVisible(Handle) {$ENDIF}) then
    Exit;
  { Note that any partial repaints outstanding are not shifted
    by ScrollBy, so force a full invalidate if this causes a problem;
    if changing the selected area, do this *after* the call to MakeVisible() }
  UpdateScrollBars;
  if CharPoint.y < VertScrollBar.Position then
    VertScrollBar.Position := CharPoint.y - 1
  else if CharPoint.y >= VertScrollBar.Position + VertScrollBar.PageSize then
    VertScrollBar.Position := CharPoint.y - VertScrollBar.PageSize + 2;
  xtab := TabCharPos(CharPoint.x, CharPoint.y);
  if xtab < HorzScrollBar.Position then
    HorzScrollBar.Position := xtab - 1
  else if xtab >= HorzScrollBar.Position + HorzScrollBar.PageSize then
    HorzScrollBar.Position := xtab - HorzScrollBar.PageSize + 2;
end;

procedure TLog.ExtendSelectionTo(p: TPoint; ForceVisible: Boolean; Shift: TShiftState);
begin
  CaretPos := p;
  if ForceVisible then MakeVisible(CaretPos);
  if not AreaSelected then
    FColumnSelect := FColumnSelect or (ssAlt in Shift);
  SelRect := Rect(FAnchor.x, FAnchor.y, CaretPos.x, CaretPos.y);
end;

procedure TLog.ExtendLineSelectionTo(Line: Integer; ForceVisible: Boolean);
var
  NewSel: TRect;
begin
  if Line < FAnchor.y then begin
    CaretPos := Point(0, Max(0, Line));
    NewSel.TopLeft := CaretPos;
    NewSel.BottomRight := FAnchor;
    if FAnchor.x = 0 then Inc(NewSel.Bottom);
    if ForceVisible then MakeVisible(CaretPos);
    SelRect := NewSel;
  end else
    ExtendSelectionTo(Point(0, Line+1), ForceVisible, []);
end;

procedure TLog.SetAreaSelected(const Value: Boolean);
begin
  if not Value then begin
    FColumnSelect := False;
    SelRect := Rect(-1, -1, -1, -1);
  end;
end;

function TLog.GetAreaSelected: Boolean;
begin
  with SelRect do
    Result := (Top >= 0) and AboveLeft(TopLeft, BottomRight);
end;

function TLog.GetSelLength: {$IFDEF WIN32} Integer {$ELSE} Longint {$ENDIF};
var
  l: Integer;
begin
  Result := 0;
  if AcceptingInput then
    Result := ComboBox.SelLength
  else if AreaSelected then
    with SelRect do
      if Top < Bottom then begin
        Result := Length(Lines[Top])-Left+NewLineLength+Right;
        for l := Top+1 to Bottom-1 do
          Inc(Result, Length(Lines[l])+NewLineLength);
      end else
        Result := Right - Left;
end;

procedure TLog.SelectAll;
begin
  if AcceptingInput then
    ComboBox.SelectAll
  else
    SelRect := Rect(0, 0, 0, Lines.Count);
end;

procedure TLog.MakeSane(var p: TPoint);
begin
  if p.y < 0 then begin
    p.y := 0;
    p.x := 0;
  end else if p.y >= Lines.Count then begin
    p.y := Lines.Count;
    p.x := 0;
  end else
    p.x := Max(0,Min(p.x, Length(Lines[p.y])));
end;

procedure TLog.SetFixedFont(const Value: TFixedFont);
begin
  FFixedFont.Assign(Value);
end;

procedure TLog.FixedFontChanged(Sender: TObject);
begin
  if FCharSize.x <> 0 then begin
    FCharSize.x := 0;
    Invalidate;
  end;
end;

procedure TLog.KeyDown(var Key: Word; Shift: TShiftState);
var
  p: TPoint;
begin
  case Key of
    VK_HOME:  if ssShift in Shift then
                if ssCtrl in Shift then
                  ExtendLineSelectionTo(0, True)
                else
                  ExtendSelectionTo(Point(0, CaretPos.y), True, Shift)
              else
                if ssCtrl in Shift then
                  VertScrollBar.Position := 0
                else
                  HorzScrollBar.Position := 0;
    VK_END:  if ssShift in Shift then
                if ssCtrl in Shift then
                  ExtendLineSelectionTo(Lines.Count, True)
                else
                  ExtendSelectionTo(Point(32767, CaretPos.y), True, Shift)
              else
                if ssCtrl in Shift then
                  VertScrollBar.Position := VertScrollBar.Range
                else
                  HorzScrollBar.Position := HorzScrollBar.Range;
    VK_PRIOR: if ssShift in Shift then begin
                p := CaretPos;
                Dec(p.y, VertScrollBar.PageSize - 2);
                ExtendSelectionTo(p, True, Shift);
              end else
                VertScrollBar.Position := VertScrollBar.Position - VertScrollBar.PageSize + 2;
    VK_NEXT:  if ssShift in Shift then begin
                p := CaretPos;
                Inc(p.y, VertScrollBar.PageSize - 2);
                ExtendSelectionTo(p, True, Shift);
              end else
                VertScrollBar.Position := VertScrollBar.Position + VertScrollBar.PageSize - 2;
    VK_UP:    if ssShift in Shift then begin
                p := CaretPos;
                Dec(p.y, 1);
                ExtendSelectionTo(p, True, Shift);
              end else
                VertScrollBar.Position := VertScrollBar.Position - 1;
    VK_DOWN:  if ssShift in Shift then begin
                p := CaretPos;
                Inc(p.y, 1);
                ExtendSelectionTo(p, True, Shift);
              end else
                VertScrollBar.Position := VertScrollBar.Position + 1;
    VK_LEFT:  if ssShift in Shift then begin
                p := CaretPos;
                Dec(p.x, 1);
                ExtendSelectionTo(p, True, Shift);
              end else
                HorzScrollBar.Position := HorzScrollBar.Position - 1;
    VK_RIGHT: if ssShift in Shift then begin
                p := CaretPos;
                Inc(p.x, 1);
                ExtendSelectionTo(p, True, Shift);
              end else
                HorzScrollBar.Position := HorzScrollBar.Position + 1;
    VK_DELETE:if ssShift in Shift then
                CutToClipboard
              else
                ClearSelection;
              { *** Pass on to edit window also }
    VK_INSERT:if Shift = [ssCtrl] then
                CopyToClipboard
              else if Shift = [ssShift] then
                PasteFromClipboard
              else
                Exit;
    VK_F3:    if Shift = [] then
                FindNext
              else
                Exit;
    VK_F2:    if Shift = [] then
                FindThisWord
              else
                Exit;
    Ord('Z'): if Shift = [ssCtrl, ssShift] then
                Redo
              else
                Exit;
  else
    Exit;
  end;
  Key := 0;
end;

{$IFDEF DFS_DELPHI_4_UP}
function TLog.DoMouseWheelDown(Shift: TShiftState;
  {$IFDEF CLX_} const {$ENDIF} MousePos: TPoint): Boolean;
begin
  VertScrollBar.Position := VertScrollBar.Position + WheelScrollLines;
  inherited DoMouseWheelDown(Shift, MousePos);
  Result := True;
end;

function TLog.DoMouseWheelUp(Shift: TShiftState;
  {$IFDEF CLX_} const {$ENDIF} MousePos: TPoint): Boolean;
begin
  VertScrollBar.Position := VertScrollBar.Position - WheelScrollLines;
  inherited DoMouseWheelUp(Shift, MousePos);
  Result := True;
end;
{$ENDIF}

procedure TLog.Clear;
begin
  SelectAll;
  ClearSelection;
end;

procedure TLog.DeleteKey;
begin
  if AcceptingInput and (ComboBox.SelLength = 0) and (ComboBox.SelStart < Length(ComboBox.Text)) then
    ComboBox.SelLength := 1;
  ClearSelection;
end;

procedure TLog.BeginUpdate;
begin
  Lines.BeginUpdate;
end;

procedure TLog.EndUpdate;
begin
  Lines.EndUpdate;
  if not Lines.Updating then begin
    UpdateScrollBars;
    PositionEditWindow;
  end;
end;

procedure TLog.ClearSelection;
var
  l: Integer;
begin
  if AcceptingInput and (ComboBox.SelLength > 0) then
    ComboBox.SelText := ''
  else if FColumnSelect then
    AreaSelected := False
  else if AreaSelected then
   try
    BeginUpdate;
    with Lines do begin
      { Trim all complete lines }
      for l := SelRect.Bottom - 1 downto SelRect.Top + 1 do begin
        Delete(l);
        Dec(FSelRect.Bottom);
      end;
      if SelRect.Bottom > Lines.Count-1 then
        { Deleted last line }
        if SelRect.Left < 1 then
          Delete(SelRect.Top)
        else
          Strings[SelRect.Top] := Copy(Strings[SelRect.Top], 1, SelRect.Left)
      else begin
        { Merge of two parts; here SelRect.Top = SelRect.Bottom + 0 or 1 }
        Strings[SelRect.Top] := Concat(Copy(Strings[SelRect.Top], 1, SelRect.Left),
                                        Copy(Strings[SelRect.Bottom], SelRect.Right+1, Length(Strings[SelRect.Bottom])));
        if SelRect.Top < SelRect.Bottom then
          Delete(SelRect.Bottom);
      end;
      CaretPos := SelRect.TopLeft;
      Anchor := CaretPos;
      AreaSelected := False; { // NB: Done in EndUpdate also, but must be done here}
      CaretTargetIsHere;
    end;
   finally
    EndUpdate;
   end;
end;

procedure TLog.SaveAreaToStream(Area:TRect; Stream: TStream; OnlyColumn: Boolean);
var
  l: Integer;
  procedure WriteChunk(const s:string; First, Last: Integer);
  begin
    Last := Min(Last,Length(s));
    if Last > First then
      Stream.Write(s[First+1], Last-First);
  end;
begin
  if OnlyColumn then
    for l := Area.Top to Area.Bottom do begin
      WriteChunk(Lines[l], Area.Left, Area.Right);
      Stream.Write(LineEnd[1], NewLineLength);
    end
  else if Area.Top = Area.Bottom then
    WriteChunk(Lines.Strings[Area.Top], Area.Left, Area.Right)
  else begin
    WriteChunk(Lines[Area.Top], Area.Left, Length(Lines[Area.Top]));
    Stream.Write(LineEnd[1], NewLineLength);
    for l := Area.Top + 1 to Area.Bottom - 1 do begin
      WriteChunk(Lines[l], 0, Length(Lines[l]));
      Stream.Write(LineEnd[1], NewLineLength);
    end;
    if SelRect.Right > 0 then
      WriteChunk(Lines[Area.Bottom], 0, Area.Right);
  end;
  l := 0;
  Stream.Write(l, 1);
end;

function TLog.GetSelTextBuf(Buffer: PChar; BufSize: Integer): Integer;
var
  nBytes: {$IFDEF WIN32} Integer {$ELSE} Longint {$ENDIF};
  s: string;
  Stream: THugeMemoryStream;
  p: PChar;
begin
  Result := 1;
  Buffer^ := Chr(0);
  if AcceptingInput and (ComboBox.SelLength > 0) then begin
    s := ComboBox.SelText;
    Result := Length(s)+1;
    if Result > BufSize then
      Result := BufSize;
    StrPLCopy(Buffer, s, BufSize-1)
  end else if AreaSelected then begin
    Stream := nil;
    try
      nBytes := SelLength+1;
      Stream := THugeMemoryStream.Create;
      Stream.SetCapacity(nBytes);
      SaveAreaToStream(SelRect, Stream, FColumnSelect);
      if nBytes > BufSize then begin
        p := Stream.Memory;
        {$IFNDEF VER80}
          Inc(p, BufSize-1);
        {$ELSE}
          HugeInc(Pointer(p), BufSize-1);
        {$ENDIF}
        p^ := Chr(0);
        Result := BufSize;
      end else
        Result := nBytes;
      hmemcpy(Buffer, Stream.Memory, Result);
    finally
      Stream.Free;
    end;
  end;
end;

function TLog.GetSelText: string;
var
  Len: Integer;
begin
{$IFNDEF VER80}
  Len := SelLength;
  SetLength(Result, Len);
  if Len > 0 then
    GetSelTextBuf(PChar(Result), Len+1);
{$ELSE}
  Len := GetSelTextBuf(@Result, 256);
  Move(Result[0], Result[1], Len);
  SetLength(Result, Len);
{$ENDIF}
end;

function TLog.CanCopyToClipboard: Boolean;
begin
  Result := AcceptingInput and (ComboBox.SelLength > 0);
  if not Result then
    Result := AreaSelected;
end;

procedure TLog.CopyToClipboard;
var
  Stream: THugeMemoryStream;
begin
  if AcceptingInput and (ComboBox.SelLength > 0) then
    Clipboard.AsText := ComboBox.SelText
  else if AreaSelected then begin
    Stream := nil;
    try
      Stream := THugeMemoryStream.Create;
      Stream.SetCapacity(SelLength + 1);
      SaveAreaToStream(SelRect, Stream, FColumnSelect);
      Stream.CopyToClipboard(CF_TEXT);
    finally
      Stream.Free;
    end;
  end;
end;

procedure TLog.CutToClipboard;
begin
  CopyToClipboard;
  ClearSelection;
end;

function TLog.CanPasteFromClipboard: Boolean;
begin
  Result := AcceptingInput and ClipboardHasText;
end;

procedure TLog.PasteFromClipboard;
begin
  if CanPasteFromClipboard then
    ComboBox.SelText := Clipboard.AsText;
end;

procedure TLog.SetAcceptingInput(const Value: Boolean);
var
  p: Integer;
  HasFocus: Boolean;
begin
  if AcceptingInput <> Value then
    if Value then begin
      if TypeAhead and (FnReturns > 0) then begin
        Dec(FnReturns);
        p := Pos(#13, FTypeAheadBuffer);
        if p > 0 then begin
          DoNewLine(Copy(FTypeAheadBuffer, 1, p-1));
          FTypeAheadBuffer := Copy(FTypeAheadBuffer, p+1, Length(FTypeAheadBuffer));
        end;
      end else begin
        {$IFDEF DFS_COMPILER_2_UP}
        HasFocus := (Owner is TCustomForm)
                     and (TCustomForm(Owner).ActiveControl = Self);
        {$ELSE}
        HasFocus := (Owner is TForm)
                     and (TForm(Owner).ActiveControl = Self);
        {$ENDIF}
        if not Assigned(FComboBox) then
          FComboBox := TLogComboBox.CreateWithLog(Self);
        ForceInputWindowVisible;
        if TypeAhead and (Length(FTypeAheadBuffer) > 0) then
          FComboBox.Text := FTypeAheadBuffer;
        ClearTypeAhead;
        if HasFocus then FComboBox.SetFocus;
        FComboBox.SelStart := Length(FComboBox.Text);
      end;
    end else begin
      FForcePending := False;
      if Assigned(FComboBox) then begin
        {$IFDEF DFS_DELPHI_2_UP}
        HasFocus := (Owner is TCustomForm)
                     and (TCustomForm(Owner).ActiveControl = ComboBox);
        {$ELSE}
        HasFocus := (Owner is TForm)
                     and (TForm(Owner).ActiveControl = ComboBox);
        {$ENDIF}
        ComboBox.Visible := False;
        if HasFocus then SetFocus;
        if csDesigning in ComponentState then begin
          FComboBox.Free;
          FComboBox := nil;
        end;
      end;
    end;
end;

procedure TLog.ForceInputWindowVisible;
begin
  if not (csLoading in ComponentState) then begin
    UpdateScrollBars;
    if Lines.Count >= VertScrollBar.Position + VertScrollBar.PageSize - 1 then
      VertScrollBar.Position := VertScrollBar.Range;
    PositionEditWindow;
    ComboBox.Visible := True;
  end else
    FForcePending := True;
end;

procedure TLog.ShowGotoLineDialog;
var
  str: string;
begin
  str := '';
  if InputQuery('Go to Line:', 'Line Number:', str) then
    try
      ChangeCaretPos(CaretPos.x, StrToInt(str)-1,False);
    except
    end;
end;

procedure TLog.FindWithDialog(Sender: TObject);
begin
  with Sender as TLogFindDialog do
    if not Search(FindText, Options) then
      SetResultForControl(Self, 'Cannot find "' + FindText + '".')
    else
      SetResultForControl(Self, '');
end;

function TLog.FindDialogComponent: TLogFindDialog;
begin
  if LogFindDialog = nil then begin
    Application.CreateForm(TLogFindDialog, LogFindDialog);
    LogFindDialog.OnFind := {$IFDEF FPC}@{$ENDIF}FindWithDialog;
  end;
  Result := LogFindDialog;
end;

function TLog.ReplaceDialogComponent: TLogReplaceDialog;
begin
  if LogReplaceDialog = nil then begin
    Application.CreateForm(TLogReplaceDialog, LogReplaceDialog);
    LogReplaceDialog.OnFind := {$IFDEF FPC}@{$ENDIF}FindWithDialog;
    LogReplaceDialog.OnReplace := {$IFDEF FPC}@{$ENDIF}DoReplace;
  end;
  Result := LogReplaceDialog;
end;

procedure TLog.DoReplace(Sender: TObject);
var
  n: Integer;
begin
  with Sender as TLogReplaceDialog do begin
    n := (Self as TEditLog).Replace(FindText, ReplaceText, Options);
    if (n <> 1) or (frReplaceAll in Options) then
      SetResultForControl(Self, Concat(IntToStr(n), ' replacment(s) made'))
    else
      SetResultForControl(Self, '');
  end;
end;

procedure TLog.ShowFindDialog;
begin
  FindDialogComponent.Execute;
end;

procedure TLog.FindNext;
begin
  if FindDialogComponent.FindText = '' then
    ShowFindDialog
  else
    FindDialogComponent.DoFind(self);
end;

procedure TLog.FindThisWord;
begin
  FindDialogComponent.FindText := WordAtCursor({$IFDEF FPC}@{$ENDIF}IsCharIdentifier);
  FindNext;
end;

procedure TLog.ShowReplaceDialog;
begin
  ReplaceDialogComponent.Execute;
end;

function TLog.GetAcceptingInput: Boolean;
begin
  Result := Assigned(FComboBox) and ComboBox.Visible;
end;

procedure TLog.DoEnter;
begin
  if AcceptingInput then
    ComboBox.SetFocus;
  inherited DoEnter;
end;

procedure TLog.SetCaretPos(Pos: TPoint);
begin
  MakeSane(Pos);
  FCaretPos := Pos;
end;

procedure TLog.CaretTargetIsHere; { May need to provide other functions, but this is the only current one required }
begin
  FCaretTarget := TabCharPos(CaretPos.x, CaretPos.y);
end;

procedure TLog.PositionEditWindow;
begin
  if Assigned(FComboBox) then
    ComboBox.SetBounds(PaintRect.Left, (Lines.Count - VertScrollBar.Position)*CharSize.y + PaintRect.Top,
                       PaintRect.Right - 2*PaintRect.Left, ComboBox.Height);
end;

procedure TLog.DoNewLine(Line: string);
var
  DoEcho: Boolean;
begin
  DoEcho := FEcho;
  if Assigned(FOnNewLine) then FOnNewLine(Self, Line, DoEcho);
  if DoEcho then
    Lines.Add(Line);
  if AutoHistory and (Length(Line) > 0) and Assigned(FComboBox) then
    with ComboBox do begin
      if (Items.Count > 0) and (Length(Items[Items.Count-1]) = 0) then
        Items[Items.Count-1] := Line
      else
        Items.Add(Line);
      Items.Add('');
      ItemIndex := Items.Count - 1;
    end;
end;

type
  TActionClass = class of TBasicAction;
  
function TLog.GetPopupMenu: TPopupMenu;
  procedure A(ActionClass:TActionClass);
  var
    M: TMenuItem;
  begin
    M := TMenuItem.Create(Owner);
    M.Action := ActionClass.Create(Owner);
    Result.Items.Add(M);
  end;
begin
  Result:=inherited GetPopupMenu;
  if not Assigned(Result) and not (csDesigning in ComponentState) and not (csLoading in ComponentState) and not (csDestroying in ComponentState) then begin
    { Create default popup menu }
    Result := TPopupMenu.Create(Owner);
    PopupMenu := Result;
    // PopupMenu.Parent := Self;
    A(TExUndo);
    A(TExRedo);
    A(TExCut);
    A(TExCopy);
    A(TExPaste);
    A(TExSelectAll);
  end;
end;

procedure TLog.KeyPress(var Key: Char);
var
  p: Integer;
begin
  inherited KeyPress(Key);
  case Key of
    Char(Ord('A') and 31): SelectAll;
    Char(Ord('C') and 31): CopyToClipboard;
    Char(Ord('F') and 31): ShowFindDialog;
    Char(Ord('G') and 31): ShowGotoLineDialog;
    Char(Ord('R') and 31): ShowReplaceDialog;
    Char(Ord('V') and 31): PasteFromClipboard;
    Char(Ord('X') and 31): CutToClipboard;
    Char(Ord('Z') and 31): Undo;
  else
    if TypeAhead and (Key <> #0) then begin
      p := Length(FTypeAheadBuffer);
      if Key = #8 then begin
        if (p > 0) and (FTypeAheadBuffer[p] <> #13) then
          SetLength(FTypeAheadBuffer, p-1);
      end else if p < High(FTypeAheadBuffer) then begin
        SetLength(FTypeAheadBuffer, p+1);
        FTypeAheadBuffer[p+1] := Key;
        if Key = #13 then Inc(FnReturns);
      end;
    end;
    Exit;
  end;
  Key := #0;
end;

procedure TLog.SetTypeAhead(const Value: Boolean);
begin
  if Value <> FTypeAhead then begin
    FTypeAhead := Value;
    ClearTypeAhead;
  end;
end;

procedure TLog.ClearTypeAhead;
begin
  SetLength(FTypeAheadBuffer, 0);
  FnReturns := 0;
end;

procedure TLog.SetAlwaysShowComboBox(const Value: Boolean);
begin
  if not FAlwaysShowComboBox and Value and not AcceptingInput then
    AcceptingInput := True;
  FAlwaysShowComboBox := Value;
end;

procedure TLog.Add(const s: string);
begin
  BeginUpdate;
  Lines.Add(s);
  EndUpdate;
end;

function TLog.GetHasSelection: Boolean;
begin
  Result := AcceptingInput and (ComboBox.SelLength > 0);
  if not Result then
    Result := AreaSelected;
end;

function TLog.GetCanDeleteKey: Boolean;
begin
  Result := HasSelection;
  if not Result then
    Result :=  AcceptingInput and (ComboBox.SelStart < Length(ComboBox.Text));
end;

function TLog.CanUndo: Boolean;
begin
  if AcceptingInput then
    Result := ComboBoxCanUndo(ComboBox)
  else
    Result := Lines.CanUndo;
end;

function TLog.CanRedo: Boolean;
begin
  if AcceptingInput then
    Result := ComboBoxCanRedo(ComboBox)
  else
    Result := Lines.CanRedo;
end;

procedure TLog.ClearUndo;
begin
  Lines.ClearUndo;
end;

procedure TLog.Undo;
begin
  if not AcceptingInput then
    Lines.Undo
  else
    ComboBoxUndo(ComboBox);
end;

procedure TLog.Redo;
begin
  if not AcceptingInput then
    Lines.Redo
  else
    ComboBoxRedo(ComboBox);
end;

function TLog.GetRecordUndo: Boolean;
begin
  Result := Lines.RecordUndo;
end;

procedure TLog.SetRecordUndo(Value: Boolean);
begin
  Lines.RecordUndo := Value;
end;

function TLog.Search(const Target: string; Options:TFindOptions): Boolean;
var
  p: TPoint;
  r: TRect;
begin
  p := CaretPos;
  Result := SearchStrings(Lines, p, Options, Target, r);
  if Result then begin
    CaretPos := p;
    MakeVisible(CaretPos);
    SelRect := r;
    CaretTargetIsHere;
  end else
    AreaSelected := False;
end;

function TLog.CanSortOnSelectedColumns: Boolean;
begin
  Result := AreaSelected and (SelRect.Top = SelRect.Bottom);
end;

procedure TLog.SortOnSelectedColumns(Ascending: Boolean);
var
  Offset, I, J, T: Longint;
  IsInOrder: Boolean;
  OrgOrder: TIntVector;
  c, s, e: Integer;
begin
  if not CanSortOnSelectedColumns then Exit;
  s := SelRect.Left+1;
  e := SelRect.Right-SelRect.Left;
  OrgOrder := nil;
  BeginUpdate;
  try
    OrgOrder := TIntVector.Create(Lines.Count);
    for I := 0 to Lines.Count-1 do
      OrgOrder[I] := I;
    Offset := Lines.Count;
    while Offset > 1 do begin
      Offset := Offset shr 1;
      repeat
        IsInOrder := True;
        for J := 0 to Lines.Count - Offset - 1 do begin
          I := J + Offset;
          c := CompareStr(Copy(Lines[I],s,e), Copy(Lines[J],s,e));
          if not Ascending then c := -c;
          if (c < 0) or ((c = 0) and (OrgOrder[I] < OrgOrder[J])) then begin
            IsInOrder := False;
            T := OrgOrder[I];
            OrgOrder[I] := OrgOrder[J];
            OrgOrder[J] := T;
            Lines.Exchange(I,J);
          end;
        end;
      until IsInOrder;
    end;
  finally
    EndUpdate;
    OrgOrder.Free;
  end;
  AreaSelected := False;
end;


function TLog.WordAtCursor(WantedChar: CharFilter): string;
var
  e, s, l: Integer;
  Line: string;
begin
  Result := '';
  if AreaSelected then begin
    Line := SelText;
    s := 1;
    l := Length(Line);
    while (s <= l) and not WantedChar(Line[s]) do Inc(s);
  end else begin
    if (CaretPos.y < 0) or (CaretPos.y >= Lines.Count) then Exit;
    s := CaretPos.x + 1;
    Line := Lines[CaretPos.y];
    while (s > 1) and WantedChar(Line[s-1]) do Dec(s);
    l := Length(Line);
  end;
  e := s;
  while (e < l) and WantedChar(Line[e+1]) do Inc(e);
  Result := Copy(Line, s, e-s+1);
end;

{ TLogStringList }

{$IFNDEF VER80}
procedure FreeUndoItem(UndoItem:PUndoAction);
begin
  case UndoItem^.Action of
    aPut, aDelete, aInsert: SetLength(UndoItem^.Line, 0);
  end;
end;
{$ENDIF}
{$IFNDEF FPC}
type PtrInt = Longint;
{$ENDIF}
function TLogStringList.NextUndoAction: PUndoAction;
var
  NewN, i: Integer;
  p: PUndoAction;
begin
  if FUndoList = nil then begin
    FUndoList := TList.Create;
    FNextUndoItem := -1;
  end;
  Inc(FNextUndoItem);
  ClearUndoListFrom(FNextUndoItem);
  { Check Undo list not too big }
  if FNextUndoItem >= {$IFNDEF VER80} 65520 {$ELSE} MaxListSize-32 {$ENDIF} then begin
    { Decide how many to delete - at least half the buffer }
    NewN := FUndoList.Count div 2;
    while (NewN < FUndoList.Count) and not (unEndUpdate in PUndoAction(FUndoList[NewN])^.Flags) do
      Inc(NewN);
    if NewN = FUndoList.Count then begin
      { Disaster - deleteing current recorded operation;
        clear and abort undo recording for this operation }
      FRecordUndo := False;
      FUndoAborted := True;
    end;
    for i := NewN to FUndoList.Count-1 do begin
      {$IFNDEF VER80} FreeUndoItem(PUndoAction(FUndoList[i-NewN])); {$ENDIF}
      p := PUndoAction(FUndoList[i-NewN]);
      FreeMem(p); {, SizeOf(TUndoAction));}
      FUndoList[i-NewN] := FUndoList[i];
    end;
    i := FUndoList.Count-NewN;
    FUndoList.Count := NewN;
    ClearUndoListFrom(i);
  end;
  GetMem(Result, SizeOf(TUndoAction));
  {$IFNDEF VER80} PtrInt(Result^.Line) := 0; { Should use New to initialise string, but this is quicker } {$ENDIF}
  if Modified then
    Result^.Flags := [unModified]
  else
    Result^.Flags := [];
  FUndoList.Add(Result);
end;

function TLogStringList.GetRecordUndo: Boolean;
begin
  Result := FRecordUndo or (FUndoList <> nil);
end;

procedure TLogStringList.SetRecordUndo(Value: Boolean);
begin
  if Value <> RecordUndo then begin
    if not Value and (FUndoList <> nil) then begin
      ClearUndo;
      FUndoList.Free;
      FUndoList := nil;
    end;
    FNextUndoItem := -1;
    FRecordUndo := Value;
  end;
end;

procedure TLogStringList.ClearUndoListFrom(Index: Integer);
var
  i: Integer;
  p: PUndoAction;
begin
  If Index < 0 then Index := 0;
  if FUndoList = nil then Exit;
  for i := FUndoList.Count-1 downto Index do begin
    {$IFNDEF VER80} FreeUndoItem(PUndoAction(FUndoList[i])); {$ENDIF}
    p := PUndoAction(FUndoList[i]);
    FreeMem(p); {, SizeOf(TUndoAction));}
    FUndoList.Delete(i);
  end;
  FNextUndoItem := Index;
end;

procedure TLogStringList.ClearUndo;
begin
  ClearUndoListFrom(0);
  FNextUndoItem := -1;
end;

procedure TLogStringList.Undo;
var
  str: string;
  FRecordUndoSave, PointFound: Boolean;
  ThePoint: TMyPoint;
begin
  PointFound := False;
  FRecordUndoSave := FRecordUndo;
  if (FNextUndoItem >= 0) and (FNextUndoItem < FUndoList.Count) then
    try
      FRecordUndo := False;
      BeginUpdate;
      repeat
        with PUndoAction(FUndoList[FNextUndoItem])^ do begin
          case Action of
           aPut:     begin
                       str := Line;
                       Line := Strings[Number];
                       Strings[Number] := str;
                     end;
           aDelete:  begin
                       Insert(Number, Line);
                     end;
           aInsert:  Delete(Number);
           aExchange: Exchange(Line1, Line2);
           aPoint:   begin
                       PointFound := True;
                       ThePoint := Point;
                     end;
          end;
          Modified := unModified in Flags;
        end;
        Dec(FNextUndoItem);
      until (FNextUndoItem < 0) or (unEndUpdate in PUndoAction(FUndoList[FNextUndoItem])^.Flags);
    finally
      EndUpdate;
      FRecordUndo := FRecordUndoSave;
      if PointFound and Assigned(SetPointVar) then SetPointVar(TPoint(ThePoint));
    end;
end;

procedure TLogStringList.Redo;
var
  str: string;
  FRecordUndoSave, PointFound: Boolean;
  ThePoint: TMyPoint;
begin
  FRecordUndoSave := FRecordUndo;
  PointFound := False;
  try
    FRecordUndo := False;
    BeginUpdate;
    while (FNextUndoItem >= -1) and (FNextUndoItem < FUndoList.Count-1) do begin
      Inc(FNextUndoItem);
      with PUndoAction(FUndoList[FNextUndoItem])^ do begin
        case Action of
         aPut:     begin
                     str := Line;
                     Line := Strings[Number];
                     Strings[Number] := str;
                   end;
         aDelete:  Delete(Number);
         aInsert:  Insert(Number, Line);
         aExchange: Exchange(Line1, Line2);
         aPoint:   begin
                     PointFound := True;
                     ThePoint := Point;
                   end;
        end;
        if unEndUpdate in Flags then Break;
      end;
    end;
  finally
    EndUpdate;
    FRecordUndo := FRecordUndoSave;
    if PointFound and Assigned(SetPointVar) then SetPointVar(TPoint(ThePoint));
  end;
end;

function TLogStringList.Add(const S: string): Integer;
begin
  UpdateRange(Count);
  if FRecordUndo then
    with NextUndoAction^ do begin
      Action := aInsert;
      Number := Count;
      Line := S;
    end;
  Result := inherited Add(S);
end;

procedure TLogStringList.Changed;
begin
  inherited Changed;
  if not FUpdating then DoLogChange;
end;

procedure TLogStringList.Clear;
begin
  FUpdateType := loAll;
  inherited Clear;
end;

constructor TLogStringList.Create;
begin
  inherited Create;
  FUpdateType := loNone;
end;

destructor TLogStringList.Destroy;
begin
  RecordUndo := False;
  BeginUpdate;
  inherited Destroy;
end;

procedure TLogStringList.Delete(Index: Integer);
begin
  UpdateRange(Index);
  UpdateRange(Count);
  if FRecordUndo then
    with NextUndoAction^ do begin
      Action := aDelete;
      Line := Strings[Index];
      Number := Index;
    end;
  inherited Delete(Index);
end;

procedure TLogStringList.DoLogChange;
begin
  if not FUpdating then begin
    if Assigned(FOnLogChanged) and (FUpdateType <> loNone) then
      FOnLogChanged(Self, FUpdateType, FFirst, FLast);
    FUpdateType := loNone;
  end;
end;

procedure TLogStringList.Exchange(Index1, Index2: Integer);
begin
  UpdateRange(Index1);
  UpdateRange(Index2);
  if FRecordUndo then
    with NextUndoAction^ do begin
      Action := aExchange;
      Line1 := Index1;
      Line2 := Index2;
    end;
  inherited Exchange(Index1, Index2);
end;

procedure TLogStringList.Insert(Index: Integer; const S: string);
begin
  UpdateRange(Index);
  UpdateRange(Count);
  if FRecordUndo then
    with NextUndoAction^ do begin
      Action := aInsert;
      Number := Index;
      Line := S;
    end;
  inherited Insert(Index, S);
end;

procedure TLogStringList.Put(Index: Integer; const S: string);
begin
  UpdateRange(Index);
  if FRecordUndo then
    with NextUndoAction^ do begin
      Action := aPut;
      Line := Strings[Index];
      Number := Index;
    end;
  inherited Put(Index, S);
end;

procedure TLogStringList.SetUpdateState(IsUpdating: Boolean);
var
  p: TPoint;
begin
  if not FUpdating and IsUpdating and FRecordUndo and Assigned(GetPointVar) then
    with NextUndoAction^ do begin
      Action := aPoint;
      p := GetPointVar();    { Without the variable it fails... }
      Point := TMyPoint(p);
    end;
  inherited SetUpdateState(IsUpdating);
  FUpdating := IsUpdating;
  if not IsUpdating then begin
    DoLogChange;
    if FRecordUndo then
      if Assigned(GetPointVar) then
        with NextUndoAction^ do begin
          Action := aPoint;
          p := GetPointVar();
          Point := TMyPoint(p);
          Flags := Flags + [unEndUpdate];
        end
      else if CanUndo then
        with PUndoAction(FUndoList[FNextUndoItem])^ do
          Flags := Flags + [unEndUpdate];
    if FUndoAborted then begin
      FRecordUndo := True;
      FUndoAborted := False;
      ClearUndo;
    end;
  end;
end;

procedure TLogStringList.UpdateRange(Line: Integer);
begin
  if FUpdateType = loNone then begin
    FUpdateType := loRange;
    FFirst := Line;
    FLast := Line;
    Modified := True;
  end else if FUpdateType = loRange then begin
    FFirst := Min(Line, FFirst);
    FLast := Max(Line, FLast);
    Modified := True;
  end;
end;

function TLogStringList.GetCanUndo: Boolean;
begin
  Result := (FNextUndoItem >= 0) and (FUndoList <> nil)
            and (FNextUndoItem < FUndoList.Count);
end;

function TLogStringList.GetCanRedo: Boolean;
begin
  Result := (FNextUndoItem >= -1) and (FUndoList <> nil)
            and (FNextUndoItem < FUndoList.Count-1);
end;

procedure TLogStringList.SaveToStreamAlien(Stream: TStream);
var
  i: Integer;
{$IFDEF VER80}
  Line: string;
{$ENDIF}
const
{$IFDEF MSWINDOWS}
  NewLine: string = #10;
{$ENDIF}
{$IFDEF UNIX}
  NewLine: string = #13#10;
{$ENDIF}
begin
  for i := 0 to Count-1 do begin
    if Length(Strings[i]) > 0 then
    {$IFNDEF VER80}
      Stream.Write(PChar(Strings[i])^, Length(Strings[i]));
    {$ELSE}
      begin
        Line := Strings[i];
        Stream.Write(Line[1], Length(Line))
      end;
    {$ENDIF}
    Stream.Write(NewLine[1], Length(NewLine));
  end;
end;

procedure TLogStringList.SaveToFileAlien(const Filename: string);
var
  Stream:TStream;
begin
  Stream := nil;
  try
    Stream := TFileStream.Create(FileName, fmCreate or fmShareExclusive);
    SaveToStreamAlien(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TLogStringList.SaveToFileEnh(const FileName: string; MakeBackup, AlienEol: Boolean);
var
  Name: string;
  Stream: THandleStream;
begin
  Name := ExpandFileName(FileName);
  if MakeBackup then
    Stream := TBackupStream.Create(Name)
  else
    Stream := TFileStream.Create(Name, fmCreate or fmShareExclusive);
  try
    if AlienEol then
      SaveToStreamAlien(Stream)
    else
      SaveToStream(Stream);
    if Stream is TBackupStream then
      TBackupStream(Stream).Save;
  finally
    Stream.Free;
  end;
end;

{ procedure TLog.SetMaxLines(const Value: Integer);
begin
  FMaxLines := Value;
end;

TLogComboBox }

constructor TLogComboBox.CreateWithLog(AOwner: TLog);
begin
  inherited Create(AOwner);
  Log := AOwner;
  ParentFont := False;
  Parent := Log;
  TabStop := False;
end;

procedure TLogComboBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  Log.ForceInputWindowVisible;
  case Key of
    VK_RETURN: if DroppedDown then
                 DroppedDown := False
               else begin
                 Log.DoNewLine(Text);
                 Text := '';
                 if Log.AlwaysShowComboBox then
                   Log.ForceInputWindowVisible
                 else
                   Log.AcceptingInput := False;
                 Key := 0;
               end;
    VK_INSERT, Ord('C'):
               if ssCtrl in Shift then begin
                 Log.CopyToClipboard;
                 Key := 0;
               end;
  end;
  inherited KeyDown(Key, Shift);
end;

{$IFDEF UNIX}
procedure TLogComboBox.DoEnter;
begin
  inherited;
  FAnyFocused := True;
  if Log.HideSelection then
    Log.InvalidateCharRect(Log.SelRect);
end;

procedure TLogComboBox.DoExit;
begin
  inherited;
  FAnyFocused := False;
  if Log.HideSelection then
    Log.InvalidateCharRect(Log.SelRect);
end;
{$ENDIF}

{$IFDEF CLX_}
procedure TLogComboBox.CreateWidget;
begin
  inherited;
  Font.Handle := Log.FixedFont.Handle;
end;
{$ENDIF}

{$IFDEF MSWINDOWS}
procedure TLogComboBox.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  if Log.WantTabs then
    Message.Result := Message.Result or DLGC_WANTTAB;
end;

{$IFNDEF LCL} { TODO: LCL equivalent }
procedure TLogComboBox.CNCommand(var Message: TWMCommand);
begin
  inherited;
  case Message.NotifyCode of
    CBN_SETFOCUS, CBN_KILLFOCUS:
      begin
        FAnyFocused := Message.NotifyCode = CBN_SETFOCUS;
        if Log.HideSelection then
          Log.InvalidateCharRect(Log.SelRect);
      end;
  end;
end;
{$ENDIF}
{$ENDIF}

{$IFNDEF CLX_}
procedure TLogComboBox.CreateWnd;
begin
  inherited CreateWnd;
  Font.Handle := Log.Canvas.Font.Handle;
end;
{$ENDIF}

function TLogComboBox.CanCopyToClipboard: Boolean;
begin
  Result := Log.CanCopyToClipboard;
end;

procedure TLogComboBox.CopyToClipboard;
begin
  Log.CopyToClipboard;
end;

procedure TLogComboBox.KeyPress(var Key: Char);
begin
  if Key = #13 then
    Key := #0;
  inherited KeyPress(Key);
end;

procedure TLog.ChangeCaretPos(x, y: Integer; ExtendSelection: Boolean);
begin
  if ExtendSelection then
    ExtendSelectionTo(Point(x,y), True, [])
  else begin
   CaretPos := Point(x, y);
   Anchor := CaretPos;
   AreaSelected := False;
  end;
end;

procedure TLog.CheckDiscardLines;
begin
  if (MaxLines > 0) and (Lines.Count >= MaxLines) then begin
    if not AutoDiscard then
      case MessageDlg('Log is now rather full - discard first third?', mtConfirmation,
                      [mbYes, mbYesToAll, mbNoToAll],0) of
        mrNoToAll:  begin MaxLines := 0; Exit; end;
        mrYesToAll: AutoDiscard := True;
      end;
    Lines.DiscardFirstThird;
  end;
end;

procedure TLog.SetMaxLines(const Value: Integer);
begin
  FMaxLines := Value;
  CheckDiscardLines;
end;

{
  Problem with log - if all of line off end, not erased
  very long lines don't paint right
  Todo:
  Print Preview
}

procedure TLogStringList.DiscardFirstThird;
var
  l, nDiscard: Longint;
begin
  nDiscard := Count div 3 - 1;
  if nDiscard <= 0 then
    Exit;
  BeginUpdate;
  { Deleteing first item is the slowest item to delete, as the
    entire array of strings has to be copied, so do it by exchanging }
  FUpdateType := loAll;
  for l := 0 to Count-1-nDiscard do
    inherited Exchange(l, l+nDiscard);
  for l := Count-1 downto Count-nDiscard do
    inherited Delete(l);
  EndUpdate;
end;

end.


